/******************************************************/
/* adjacents.c                                        */
/* Frank Wallner, Feb. 99                             */
/******************************************************/

#include "qq_header.h"

/***************************************************************************/
void Find_AcptBue(ConfS c, int event)
/* detect latest Buechi- and accepting event in configuration c,
   and set maxBuchi(event) and maxAcpt(event) appropriately */
{
        register int   bit, w, i;
	BitArray       segment;
	char 	       find_buchi = 1;	

        i = event/BA_BASE;
	maxBuchi(event) = maxAcpt(event) = 0;
	while( i >= 0 ) {
	     if ( ((segment=c[i]) & buechi_conf[i]) == 0 ) 
		  { i--; continue; }
             /* c[i] contains some buechi events */
	     w = BA_BASE * (i--) + BA_BASE;
             for ( bit = (BA_BASE-1);  bit >= 0;  bit-- ) {
		   w--;
                   if ( bitmask[bit] != (bitmask[bit] & segment) ) 
			continue;
                   segment -= bitmask[bit];
                   if ( find_buchi && IS(BUECHI, w) ) 
		      { maxBuchi(event) = w; find_buchi = 0; }
                   if ( IS(FINAL,w) ) 
		      { maxAcpt(event) = w; return; }
		   if ( segment == 0 ) break;
                   }
             }
}
/***************************************************************************/
int Find_Corr(ConfS c, int event)
/* detect latest corresponding event in configuration c ;
   'event' is element with max. index in c */
{
        register int   bit, w, i;
	BitArray       segment;

        i = event/BA_BASE;
	while( i >= 0 ) {
	     if ( (segment=(c[i] & corr_conf[i])) == 0 ) 
		  { i--; continue; }
	     w = BA_BASE * (i--) + BA_BASE;
             for ( bit = (BA_BASE-1);  bit >= 0;  bit-- ) {
		   w--;
                   if ( bitmask[bit] != (bitmask[bit] & segment) ) 
			continue;
                   segment -= bitmask[bit];
		   if ( (w <= event) && IS(CORR,w) ) return(w);
		   if ( segment == 0 ) break; }
             }
	return(0);
}
/***************************************************************************/
int Find_Buechi(ConfS c, int event)
/* detect latest Buechi event in configuration c;
   'event' is the element with max. index in c */
{
        register int   bit, w, i;
	BitArray       segment;

        i = event/BA_BASE;
	while( i >= 0 ) {
	     if ( (segment=(c[i] & buechi_conf[i])) == 0) 
		  { i--; continue; }
	     w = BA_BASE * (i--) + BA_BASE;
             for ( bit = (BA_BASE-1);  bit >= 0;  bit-- ) {
		   w--;
                   if ( bitmask[bit] != (bitmask[bit] & segment) ) 
		        continue;
                   segment -= bitmask[bit];
		   if ( (w <= event) && IS(BUECHI,w) ) return(w);
		   if ( segment == 0 ) break; }
             }
	return(0);
}
/***************************************************************************/
int edge_label(int e1, int e2)
/* returns F_11 if [e1]\[e2] contains an accepting event,                    */
/*         F_12 if [e1]\[e2] contains no accepting events but buechi events, */
/*         F_10 if [e1]\[e2] contains no accepting and no buechi event       */
/* !! maxBuchi(e1|e2), maxAcpt(e1|e2), [e1], [e2] are already computed !!  */
{
	int last = 0;

	if ( ( (last=maxAcpt(e1)) > e2 ) ) return(F_11);
	if ( ( last ) && 
	     (! elem_of(config(e2), last)) ) return(F_11);

	if ( ( (last=maxBuchi(e1)) > e2 ) ) return(F_12);
	if ( ( last ) && 
	     (! elem_of(config(e2), last)) ) return(F_12);

	return(F_10);
} 
/***************************************************************************/
int edge_label_gl(ConfS c, int e)
/* returns F_11 if c\[e] contains an accepting event,                     */
/*         F_12 if c\[e] contains no accepting events but buechi events,  */
/*         F_10 if c\[e] contains no accepting and no buechi event        */
{
	int 		result=F_10;
	register int 	i=0;
	BitArray        c1, c2;

	while ( ((c1=c[i]) != BA_ALL) && ((c2=config(e)[i]) != BA_ALL) ) {
		if ( c1 == 0 ) { i++; continue;}
		c1 &= ~(c2);
		if ( c1 == 0 ) { i++; continue; }
                if ( (c1 &  final_conf[i]) != 0 ) return(F_11);
                if ( (c1 & buechi_conf[i]) != 0 ) result=F_12;
                i++; }
	if ( c[i] == BA_ALL ) return(result);
	while ( (c1=c[i]) != BA_ALL ) {
	        if (c1 == 0) {i++; continue;}
                if ( (c1 &  final_conf[i]) != 0 ) return(F_11);
                if ( (c1 & buechi_conf[i]) != 0 ) result=F_12;
                i++; }
	return(result);
} 
/***************************************************************************/
int edge_label_lg(int e, ConfS c)
/* returns F_11 if [e]\c contains an accepting event,                     */
/*         F_12 if [e]\c contains no accepting events but buechi events,  */
/*         F_10 if [e]\c contains no accepting and no buechi event        */
{
	int 		result=F_10;
	register int 	i=0;
	BitArray        c1;

	while ( ((c1=config(e)[i]) != BA_ALL) && (c[i] != BA_ALL) ) {
		if ( c1 == 0 ) { i++; continue; }
		c1 &= ~(c[i]);
		if ( c1 == 0 ) { i++; continue; }
                if ( (c1 &  final_conf[i]) != 0 ) return(F_11);
                if ( (c1 & buechi_conf[i]) != 0 ) result=F_12;
                i++; }
	if ( c1 == BA_ALL ) return(result);
	while ( (c1=config(e)[i]) != BA_ALL ) {
	        if ( c1 == 0 ) {i++; continue;}
                if ( (c1 &  final_conf[i]) != 0 ) return(F_11);
                if ( (c1 & buechi_conf[i]) != 0 ) result=F_12;
                i++; }
	return(result);
} 
/***************************************************************************/
char ADJ_Cutoff(int v)
/* computes the causal predecessors of cutoff 'v' and add
   representative(v) to the adjacent-List of these predecessors
   [[   if (e1 < e2 < v) add REP(e2) to the adj-List of e1   ]]     */
{
	int 		final, w, result=0, border;
	register int 	i=0, bit;
	BitArray        segment;
	ConfS           conf_v;
	Int_List        *new, *il, *prev, *mil, *milp;

        InitConfig(v); 
	InitMaxBuchi(v); 
	conf_v = config(v);
	il = prev = IL_NULL;
  	border = v/BA_BASE;
 	for ( i=0; i<=border; i++ ) {
	    if ( (segment = (conf_v[i] & corr_conf[i])) == 0 ) continue;
	    w = (BA_BASE * i);
	    for ( bit=0;  bit<BA_BASE;  bit++, w++ ) {
      		if ( segment < bitmask[bit] ) continue;
		segment -= bitmask[bit];
		/*** w is a corresponding event contained in [v] ***/
		InitConfig(w); 
		InitMaxBuchi(w);
		INFO(w) |= VISITED;
		new = NEW_INTLIST; new->i = w; new->next = IL_NULL;
		if ( prev ) { prev->next = new; prev = new; }
		else        { il = prev = new; }
                if ( w == correspond(v) ) {
		     final = edge_label(v, w);
                     if ( F__1(final) ) {
			  if ( ! all_counterexamples ) {
                               printf("\n\nNO. (1)\n\n");
                               alreadyNO = 1;
			       printf("Counterexample:\n");
			       PrintTS_Simple(stdout, conf_v, config(w));
			       return(2); }
			  fprintf(fres,"Counterexample:\n");
			  PrintTS_Simple(fres, conf_v, config(w));
			  final_result = 1;
			  result = 2; }
                     if ( ( F__0(final) ) && 
			  ( ! IS( CHCKBC, representative(v)) ) ) 
			  { INFO( representative(v) ) |= CHCKBC;
			    if (aut_acc_ctr <= MAX2_COUNTEREXAMPLES) {
			        Check_Buechi_Circle(maxBuchi(w), conf_v, 2);
			        if ( accepting_component) {
                                    if ( ! good_counterexample )  {
				      printf("\nNV-Counterexample:\n");
				      PrintTS_Simple(stdout, conf_v, config(w));
				      printf("\n");
				      return(1); }
			            fprintf(fres, _LINE_);
			            fprintf(fres,"NV-Counterexample:\n");
				    PrintTS_Simple(fres,conf_v,config(w));
				    result = 1; }
                                }
                          }
		     }
                  if ( segment == 0 ) break; }
	    } /* for (... i <= border ...) */
        /*** il contains all corr events in [v] ***/
	if (prev) maxCorr(v) = prev->i;
	else return(result);
	while ( il ) {
	      w = il->i; new = il; il = il->next; free(new); new = il;
	      prev = milp = mil = IL_NULL;
	      while ( new ) {
		    if ( ! elem_of(config(new->i), w)) { 
			 new = new->next; continue; }
		    milp = mil; final = 0;
		    while ( milp ) {
			 if ( elem_of(config(new->i), milp->i)) 
			      { final = 1; break; }
                         milp = milp->next; }
                    if ( ! final ) {
			 milp = NEW_INTLIST; milp->i = new->i;
		         if ( prev ) { prev->next = milp; prev = milp; }
		         else        { mil = prev = milp; }
			 }
                    new = new->next;
		    }
              if ( ! mil )  { Update_Adjacents(v, w); continue; }
	      while ( mil ) { milp = mil->next;
			      Update_Adjacents(mil->i, w);
			      free(mil); mil = milp; }
	      }
	return(result);
}
/***************************************************************************/
void ADJ_Corr(int v)
/* computes the causal predecessors of corresponding event 'v' and add
   representative(v) to the adjacent-List of these predecessors
   [[   if (e1 < e2 < v) add REP(e2) to the adj-List of e1   ]]     */
{
	int             w, final, border;
	register int    i=0, bit;
	BitArray	segment;
	Int_List        *il, *ilp, *lilp, *mil, *milp;

	InitConfig(v); 
	InitMaxBuchi(v);
	border = v/BA_BASE;
	il = ilp = lilp = IL_NULL;
	for ( i=0; i<=border; i++) {
 	    if ( (segment=(config(v)[i] & corr_conf[i])) == 0 ) continue;
	    w = ( BA_BASE * i );
	    for ( bit=0;  bit<BA_BASE;  bit++, w++ ) {
		if ( segment < bitmask[bit] ) continue;
		segment -= bitmask[bit];
		/*** w is a corresponding event contained in [v] ***/	
		if ( v != w ) {
		     INFO(w) |= VISITED;
		     ilp = NEW_INTLIST; ilp->i = w; ilp->next=IL_NULL;
		     if (lilp) { lilp->next = ilp; lilp = ilp; }
		     else { il = lilp = ilp; }
		     }
                if ( segment == 0 ) break;
	        } /* for (... bit<BA_BASE ...) */
            } /* for (i=0; i<border; ) */

        /*** il contains all corr events in [v] ***/
	while ( il ) {
	      w = il->i; ilp = il; il = il->next; free(ilp); ilp = il;
	      lilp = milp = mil = IL_NULL;
	      while ( ilp ) {
		    if ( ! elem_of(config(ilp->i), w)) { 
			 ilp = ilp->next; continue; }
		    milp = mil; final = 0;
		    while ( milp ) {
			   if ( elem_of(config(ilp->i), milp->i)) {
				final = 1; break; }
                           milp = milp->next; }
                    if ( ! final ) {
			   milp = NEW_INTLIST; milp->i = ilp->i;
		           if ( lilp ) { lilp->next = milp; lilp = milp; }
		           else { mil = lilp = milp; }
			   }
                    ilp = ilp->next;
		    }
              if ( ! mil )  { Update_Adjacents(v, w); continue; }
	      while ( mil ) { milp = mil->next;
			      Update_Adjacents(mil->i, w);
			      free(mil); mil = milp; }
	      }
	return;
}
/***************************************************************************/
void ADJ_EN_loc(int v)
/* find the causal successors (cutoff, corr) of 'simple' event v,
   and add them to the adjacent-List of v */
{
	int             *a, as, x;
       	Edge            *nedge, *lastedge;
	Int_List 	*il, *el, *lastel, *del;
	TE_List         *tel;

	maxCorr(v) = Find_Corr(config(v), v);
	nedge = lastedge = (Edge *) NULL;
	il = el = del = IL_NULL;
	el = NEW_INTLIST; el->i = v; el->next = lastel = IL_NULL;	
	while ( el ) {
	      a = postB(el->i); as = post_size(el->i);
	      while ( as-- ) {
		    if ( (*a) > B_Fin ) break;
		    tel = b_post[(*a)]; 
		    while ( tel ) {
			  il = tel->e; tel = tel->next;
			  while ( il ) {
				if ( ( ! IS (CUTOFF, il->i) ) &&
				     ( ! IS (CORR, il->i) )   ) 
				     { Insert_IL(&el, il->i, &lastel);
				       il = il->next;
				       continue; }
                                /* il->i is cutoff or corr */
				x = ( (IS(CUTOFF,il->i)) ?
				      representative(il->i) :
				      cutoff_num(il->i, 1 )   );
				nedge = S_Grab_Edge(v, x, &lastedge);
	                        if ( ! F__1(nedge->final) ) { 
	                             x = edge_label(il->i, v);
				     if ( x > nedge->final) {
	                                  nedge->final = (2 * x); 
					  nedge->n = il->i; }
                                     }
                                il = il->next; }
                          }
                    a++; 
		    }
	      del = el; el = el->next; 
	      if ( lastel == del ) lastel = el;
	      free(del);
              }
}
/***************************************************************************/
void ADJ_EN_glob(ENCorr *encorr, ConfS conf)
/* find the causal successors of the global config conf and add
   them to the adj-List of encorr */
{
	int             nc, w, cv, final, v;
	int             valid, vdiv, vmod, i=0;
       	Edge            *lce, *nedge, *lastedge;
	Int_List        *il, *ilp;


	lastedge = (Edge *) NULL;
	ilp = il = encorr->max;
	while ( ilp ) { if ( IS(CORR, ilp->i) ) break; 
			ilp = ilp->next; }

	/*** case1: (ilp != NULL) --> il contains a corr event ***/
	/*** case2: (ilp == NULL) --> il contains no corr event ***/

	/*** case1 ***/
	if ( ilp ) { 
	     encorr->last_corr = v = ilp->i; 
	     lce = __adjacents(v); 
	     while ( lce ) {
	           if ( F_1_(lce->final) ) { lce = lce->next; continue; }
		   il = encorr->max;
		   valid = 1;
		   while ( il ) {
			 if ( (il->i > lce->n) ||
			      ! elem_of(config(lce->n), il->i) ) 
			      { valid = 0; break; }
                         il = il->next; }
                   if ( valid ) {
		        nedge = EN_Grab_Edge(encorr, lce->adj, &lastedge);
		        if ( ! F__1(nedge->final) ) { 
		             final = edge_label_lg(lce->n, conf);
		             if ( final > nedge->final ) {
			        nedge->final = ( 2 * final );
			        nedge->n = lce->n; }
                             }
                        lce = lce->next;
			continue;
			}
                   if ( ! IS(CUTOFF, lce->adj) ) { lce = lce->next; continue; }
                   /*** perhaps another representative of lce->adj ***/
		   nc = correspond(lce->adj);
	           w = cutoff_quantity(nc);
		   nedge = (Edge *) NULL;
	           for ( i=1; i <= w; i++ ) {
			 if ( nedge ) break;
		         cv = cutoff_num(nc, i);
		         il = encorr->max; 
			 valid = 1;
		         while ( il ) {
		                  if ( (cv < il->i) || 
		                       ( ! elem_of(config(cv), il->i))) 
				       { valid = 0; break; }
                                  il = il->next; }
                         if ( ! valid ) continue;
		         nedge = EN_Grab_Edge(encorr, lce->adj, &lastedge);
		         if (F__1(nedge->final)) break;
		         final = edge_label_lg(cv, conf);
		         if ( final > nedge->final ) {
			      nedge->final = ( 2 * final );
			      nedge->n = cv; }
		         }
                   lce = lce->next; 
		   }
	     return; 
	     }

        /*** case2 ***/
	v = encorr->max->i;
	encorr->last_corr = ( - Find_Corr(conf, v) );
	vdiv = v/BA_BASE; vmod = v%BA_BASE; 
        /* corrs to include ? */
	i = 0;
	while ( (corrlist[i] < v) && (i < E_corr) ) { i++; }
	while ( i < E_corr ) { 
	      cv = corrlist[i++];
	      if ( ! (config(cv)[vdiv] & vmod) ) continue;
	      il = encorr->max; valid = 1; 
	      while ( il ) {
		    if ( ! elem_of(config(cv), il->i) ) {
			   valid = 0; break; }
                    il = il->next; }
              if ( ! valid ) continue;
	      nedge = EN_Grab_Edge(encorr, cutoff_num(cv, 1), &lastedge);
	      if ( F__1(nedge->final) ) continue; 
	      final = edge_label_lg(cv, conf);
	      if ( final > nedge->final ) {
	           nedge->final = ( 2 * final); nedge->n = cv; }
	      }
	/* cutoffs to include ? */
	i = 0; 
	while ( (cutofflist[i] < v) && (i < E_off) ) { i++; }
	while ( i < E_off ) {
	      cv = cutofflist[i++];
	      if ( ! (config(cv)[vdiv] & vmod) ) continue;
	      il = encorr->max; valid = 1;
	      while ( il ) { 
		    if ( ! elem_of(config(cv), il->i) ) { valid = 0; break; }
                    il = il->next; }
              if ( ! valid ) continue;
	      nedge = EN_Grab_Edge(encorr, representative(cv), &lastedge);
	      if ( ! F__1(nedge->final) ) {
	           final = edge_label_lg(cv, conf);
	           if ( final > nedge->final ) {
	                nedge->final = ( 2 * final); nedge->n = cv; }
                   }
	      }
}
/***************************************************************************/
void ADJ_co(int v)
{
     	int      rep_v, j, v0, w0, lc;
     	int      tarray[MAXTRANSLIST];
	ConfS    conflict_lc, config_v;
     	Edge     *nedge;
	ENode    *ci;
     	Int_List *tl;
	UC_List  *ucl;
	char     cut2_done=0, final;

        v0 = correspond(v);
	rep_v = representative(v);
	if ( (lc=maxCorr(v)) > 0 ) conflict_lc = (ConfS) confls(lc); 
	config_v = config(v);
	j = ( ( corrlist[0] == 0 ) ? 1 : 0 );
	while ( (w0=corrlist[j]) < v0 ) {
	     j++;
	     if ( IS(VISITED,w0) ) { 
		  INFO(w0) &= ~(VISITED);
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }
	     if ( elem_of(config_v, w0) ) continue;
	     if ( elem_of(config(v0), w0) ) continue;
	     final = ( (lc > w0) ? \
		  CIn_Conflict(config(w0), conflict_lc, (w0/BA_BASE)+1) :
                  CIn_Conflict(config_v, (ConfS) confls(w0), (w0/BA_BASE)+1));
             if ( final ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; } 

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }

	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /*(w0 co v) is already done*/
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = ( 2 * edge_label(v, w0) );
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
        j++; /** skip the case (w0 == v0) **/
	while ( ( (w0=corrlist[j]) < v) && (j < E_corr) ) {
	     j++;
	     if ( IS(VISITED, w0) ) { 
		  INFO(w0) &= ~(VISITED);
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }
	     if ( elem_of(config_v, w0) ) continue;
	     final = ( (lc > w0)? \
		  CIn_Conflict(config(w0), conflict_lc, (w0/BA_BASE)+1):
                  CIn_Conflict(config_v, (ConfS) confls(w0), (w0/BA_BASE)+1));
             if ( final ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; } 

	     if ( (nedge=(Edge *)__lastEdge(w0)) && (nedge->adj < rep_v) ) {
		  while ( nedge && (nedge->adj < rep_v) ) nedge=nedge->next;
		  __lastEdge(w0) = (int) nedge; }
             if ( nedge && (nedge->adj == rep_v) &&
		  ((final=edge_label(v, w0)) <= nedge->final) ) continue;

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }
	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is already done */
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = 2 * final;
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
        lc = (v/BA_BASE) + 1;
	while ( j < E_corr ) {
	     w0 = corrlist[j++];
	     if ( IS(VISITED, w0) ) { 
		  INFO(w0) &= ~(VISITED);
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }
             if ( elem_of((BitArray *) confls(w0), v) || 
		  CIn_Conflict(config_v, (ConfS) confls(w0), lc) ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; } 

	     if ( (nedge=(Edge *)__lastEdge(w0)) && (nedge->adj < rep_v) ) {
		  while ( nedge && (nedge->adj < rep_v) ) nedge=nedge->next;
		  __lastEdge(w0) = (int) nedge; }
             if ( nedge && (nedge->adj == rep_v) &&
		  ((final=edge_label(v, w0)) <= nedge->final) ) continue;

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) { 
		  ucl = (UC_List *) nextcorrs(w0);
		  while (ucl) { *(ucl->c) |= VISITED; ucl = ucl->next; }
		  continue; }
	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is already done */
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = 2 * final;
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
}
/***************************************************************************/
char ADJ_co2(int v, Int_List *il, Int_List *gl)
{
     	int      rep_v, j, v0, w0, mlc, mlm, ctr, vlc;
     	int      tarray[MAXTRANSLIST];
     	Edge     *nedge;
	ENode    *ci, *enode;
     	Int_List *tl, *ilp, *ml;
	char     cut2_done=0, final;
	ENCorr   *encorr;
	UC_List  *ucl;
	ConfS    vlc_confl;

        v0 = correspond(v);
	rep_v = representative(v);
	if ( (vlc = maxCorr(v)) ) vlc_confl = (ConfS) confls(vlc);
	/* ---- the global ENCorrs without corresponding event --- */
	ilp = gl; 
	while ( ilp ) {
	      encorr = (ENCorr *) ilp->i; ilp = ilp->next;
	      ml = encorr->max;
	      while ( ml ) { 
		    if ( (ml->i <= vlc) && elem_of(vlc_confl, ml->i) ) {
			 encorr = (ENCorr *) NULL; break; }
		    ml=ml->next; }
              if ( ! encorr ) { continue; }
	      mlm = encorr->max->i;
	      IL2CONF(encorr->conf, conf1, ba_idx(mlm));
              if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	      if ( ! GConcurrent(conf1, mlm, v, tarray, cut2) ) continue; 
	      if ( tarray[0] == 0 ) continue;
	      tl = array2listTA(tarray);
	      ci = ExtGrab_ENode(encorr->enode, rep_v);
	      if ( EN_IS(SHIFTED, ci) ) continue; /* ci is already done */
	      Shift(v, &tl, &ci);
	      if (tl) { Free_Int_List(tl); }
	      ci->final = 2 * edge_label_lg(v, conf1);
	      if ( F__0(ci->final) ) 
		   EN_maxBuchi(ci) = MAX(maxBuchi(v),Find_Buechi(conf1, mlm));
              }
	/* ---- the local ENCorrs --- */
	ilp = il;
	while ( ilp && ((w0=ilp->i) < v0 ) ) {
	     ilp = ilp->next; 
	     if ( elem_of(config(v ), w0) ) continue;
	     if ( elem_of(config(v0), w0) ) continue;
             if ( NC_In_Conflict_1gt2(v, w0) ) continue; 
	     if ( (enode=(ENode *) S_lastENode(w0)) ) {
	          while (enode && (enode->rep < rep_v) ) { enode=enode->next; }
	          S_lastENode(w0) = (int) enode; 
	          if (  enode && 
		       (enode->rep != rep_v) &&
		       (! elem_of(config(v), maxCorr(w0))) ) continue; }

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) continue; 
	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is already done */
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = 2 * edge_label(v, w0);
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
	while ( ilp && ((w0=ilp->i) < v) ) {
	     ilp = ilp->next;
	     if ( elem_of(config(v ), w0) ) continue;
             if ( NC_In_Conflict_1gt2(v, w0) ) continue;
	     if ( (enode=(ENode *) S_lastENode(w0)) ) {
	          while (enode && (enode->rep < rep_v) ) { enode=enode->next; }
	          S_lastENode(w0) = (int) enode; 
	          if (  enode && 
		       (enode->rep != rep_v) &&
		       (! elem_of(config(v), maxCorr(w0))) ) continue; }

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) continue; 
	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is already done */
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = 2 * edge_label(v, w0);
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
	while ( ilp ) {
	     w0 = ilp->i; ilp = ilp->next;
             if ( NC_In_Conflict_2gt1(v, w0) ) continue;
	     if ( (enode = (ENode *) S_lastENode(w0)) ) {
	          while (enode && (enode->rep < rep_v) ) { enode=enode->next; }
	          S_lastENode(w0) = (int) enode; 
	          if (  enode && 
		       (enode->rep != rep_v) &&
		       (! elem_of(config(v), maxCorr(w0))) ) continue; }

             if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	     if ( ! Concurrent(w0, v, tarray, cut2) ) { continue; }
	     ci = Grab_ENode(w0, rep_v);
	     if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is already done */
	     tl = array2listTA(tarray);
	     Shift(v, &tl, &ci);
	     if (tl) { Free_Int_List(tl); }
	     ci->final = 2 * edge_label(v, w0);
	     if ( F__0(ci->final) ) 
		  EN_maxBuchi(ci) = MAX( maxBuchi(v), maxBuchi(w0) );
	     }
	/* ---- the global ENCorrs with corresponding event --- */
	ilp = (Int_List *) ENCorr_co_cands(v);
	while ( ilp ) {
	     encorr = (ENCorr *) ilp->i;
	     tl = ilp->next; free(ilp); ilp = tl;
	     enode = (ENode *) ilp->i;
	     tl = ilp->next; free(ilp); ilp = tl;
	     mlm = encorr->max->i; 
	     if ( (mlc=encorr->last_corr) > 0 ) {
	     /* encorr contains a corresponding event in max */
	        IL2CONF(encorr->conf, conf1, ba_idx(mlm));
                if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	        if ( ! GConcurrent(conf1, mlm, v, tarray, cut2) ) continue; 
	        tl = array2listTA_minus_c(tarray, mlc);
	        if (  ! tl ) continue;
	        ci = ExtGrab_ENode(encorr->enode, rep_v);
	        if ( EN_IS(SHIFTED, ci) ) continue; /* (w0 co v) is done */
	        if ( EN_IS(LOCAL,enode) ) 
		     Extend_LC((int) enode->corr, v, &tl, &ci);
	        else Extend_GC(enode->corr, v, &tl, &ci);
	        if ( tl ) Free_Int_List(tl); 
		ci->final = 2 * edge_label_lg(v, conf1);
		if ( F__0(ci->final) ) 
		     EN_maxBuchi(ci) = MAX(maxBuchi(v),Find_Buechi(conf1,mlm));
		} /* mlc > 0 */
             else { /* no maximal event of enco is corresponding, but conf 
		       contains a corresponding event */
	        ml = encorr->max;
	        while ( ml ) { 
		        if ( (ml->i <= vlc) && elem_of(vlc_confl,ml->i)) 
			     { encorr = (ENCorr *) NULL; break; }
		        ml=ml->next; }
                if ( ! encorr ) { continue; }
	        IL2CONF(encorr->conf, conf1, ba_idx(mlm));
                if ( ! cut2_done ) { Cut(cut2, config(v), v); cut2_done = 1; }
	        if ( ! GConcurrent(conf1, mlm, v, tarray, cut2) ) continue; 
	        if ( tarray[0] == 0 ) continue;
	        tl = array2listTA(tarray);
	        ci = ExtGrab_ENode(encorr->enode, rep_v);
	        if ( EN_IS(SHIFTED, ci) ) continue; /*(w0 co v) is done*/
	        Shift(v, &tl, &ci);
	        if ( tl ) Free_Int_List(tl);
		ci->final = 2 * edge_label_lg(v, conf1);
	        if ( F__0(ci->final) ) 
		     EN_maxBuchi(ci) = MAX(maxBuchi(v),Find_Buechi(conf1,mlm));
		} /* else */
	     }
        ENCorr_co_cands(v) = (int) 0;
}
/***************************************************************************/
void Update_Adjacents(int v, int w)
/* includes or updates the edge rep_w --> rep_v */
{
	int rep_v, rep_w, final;
	Edge *nedge;

	if ( IS(CUTOFF,v) )      rep_v = representative(v);
	else { if ( IS(CORR,v) ) rep_v = cutoff_num(v,1);
	       else              rep_v = v; }
	rep_w = cutoff_num(w, 1);
	nedge = (Edge *) __lastEdge(w); 
	if ( (nedge == NULL) || (nedge->adj != rep_v) )  
	     nedge = Grab_Edge(rep_w, rep_v); 
        if ( F__1(nedge->final) ) return; 
	if ( (final=edge_label(v, w)) > nedge->final) {
	      nedge->final = (2 * final); nedge->n = v; }
}
/***************************************************************************/
char Check_Corr_Cycles(int v)
{
	int cv, final, i, offs_v, result=0;
	int tarray[MAXTRANSLIST];

        Cut(cut2, config(v), v); 
	offs_v = cutoff_quantity(v);
	for (i=1; i <= offs_v; i++) {
	     cv = cutoff_num(v,i);
	     if ( elem_of(config(cv), v) ) continue; 
             if ( Concurrent(cv, v, tarray, cut2) ) {
		  final = edge_label(cv, v);
                  if ( F__1(final) ) {
                       if (! all_counterexamples) {
			     printf("\n\nNO. (3)\n\nCounterexample\n");
			     PrintTS_Simple(stdout, config(cv), config(v)); 
			     return(2); }
		       fprintf(fres,"Counterexample:\n");
		       PrintTS_Simple(fres, config(cv), config(v));
 		       result = 2; }
                  if ( ( F__0(final) ) && 
		       ( ! IS(CHCKBC, representative(cv)) ) &&
		       ( aut_acc_ctr <= MAX2_COUNTEREXAMPLES) ) {
			 INFO(representative(cv)) |= CHCKBC;
			 Check_Buechi_Circle(maxBuchi(v), config(cv), 4);
			 if ( accepting_component ) { 
                              if (! good_counterexample ) {
				         printf("\nNV-Counterexample:\n");
				 	 PrintTS_Simple(stdout, 
							config(cv),config(v));
					 printf("\n");
				         return(1); }
			      fprintf(fres,"- - - - - - - - - - - -\n");
			      fprintf(fres,"NV-Counterexample:\n");
			      PrintTS_Simple(fres,config(cv),config(v));
			      fprintf(fres,"\n--------------------");
			      fprintf(fres,"--------------------\n");
			      result = 1; } 
			 }
                  }
	     } 
	return(result);
}
/***************************************************************************/
UC_List *NextCorrList(int v)
/* computes the causal successors of v that are corresponding events
   and save their INFO-component into a List (cf. ADJ_co)      */
{
	Edge *nedge; 
	UC_List *il, *ilp;

	nedge = __adjacents(v);
	il = ilp = UC_NULL;
	while ( nedge ) {
		if (IS(CORR, nedge->n)) {
		    ilp = NEW_UCLIST; ilp->c = &(INFO(nedge->n)); 
		    ilp->next = il; il = ilp; }
                nedge = nedge->next;
		}
	return(il);
}
/***************************************************************************/
/* adjacents.c */
/***************/
