/*****************************************************************************/
/* convertPEP.c                       Copyright (c) Stefan Roemer 1993-1998  */
/*                                              Version h-1.5.2 (04.03.1998) */
/*****************************************************************************/
/* Schnittstellen-Bibliothek zum Lesen (Schreiben) des PEP-Netzdateiformats  */
/* in die (aus den) grundlegenden Datenstrukturen fuer Petri-Netze.          */
/*****************************************************************************/
/*                                   Stefan Roemer, 13.12.1993 - 01.08.1994  */
/* Neues PEP-Netzdateiformat:        Stefan Roemer, 19.12.1994 - 24.01.1995  */
/* STG-Dateiformat lesen:            Stefan Roemer, 18.08.1995 - 28.02.1996  */
/* FSA-Dateiformat lesen:            Stefan Roemer, 04.09.1995 - 28.02.1996  */
/* Erweiterungen FORMAT_N2(Phantom): Stefan Roemer, 31.05.1996 - 11.06.1996  */
/* Erweiterungen FORMAT_N2(Bloecke): Stefan Roemer, 07.03.1997 - 07.03.1997  */
/*                                   Stefan Roemer, 04.03.1998 - 04.03.1998  */
/*****************************************************************************/
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/* ACHTUNG: Diese Datei enthaelt elementare Routinen, die von vielen ver-    */
/* schiedenen Programmen benoetigt werden.                                   */
/* Jede beabsichtigte Aenderung am Programmcode sowie das Kopieren von       */
/* Programmteilen deshalb bitte vorher mit mir absprechen. Danke.            */
/* E-mail: roemer@informatik.tu-muenchen.de                                  */
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "neunetze.h"
#include "convertPEP.h"


/*****************************************************************************/

#define NAMESIZE 4096
#define MEANINGSIZE 4096
#define GUARDSIZE 8192
#define REFSIZE 8192
#define OMEGA 0x7FFFFFFF
#define NAMES_START 2000
#define NAMES_OFFSET 1000


/*****************************************************************************/

char       CM_nettype[ 31 ];
bool       CM_storeMeaning      = False,
           CM_storeCoordinates  = False,
           CM_storeReferences   = False,
           CM_use_imarks        = True;
M_LIST     CM_placeMList        = NULL,
           CM_transMList        = NULL;
Cardinal32 CM_countMPlaces      = 0,
           CM_countPMeanings    = 0,
           CM_countMTransitions = 0,
           CM_countTMeanings    = 0;

#define V (void)

/*****************************************************************************/
/*                                Funktionen                                 */
/*****************************************************************************/

/*****************************************************************************/
/* Einlesen eines Strings ( von '...' eingeschlossen ) aus einer Textdatei.  */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static unsigned char ReadString( FILE * file, char * string, int maxlen ) 

#else

static unsigned char ReadString( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register char ch;
    register int  len = 0;

    if ( ( ch = ReadWhiteSpace( file ) ) != '\'' )
      {
        NNError( "ReadString: \"'\" erwartet" );
      } /* if */
    len++;
    while( ( ( *string = getc( file ) ) != '\'' ) && ( len < maxlen ) )
      {
        string++;
        len++;
      } /* while */
    if ( len == maxlen )
      {
        while ( ( ( ch = getc( file ) ) != EOF ) && ( ch != '\'' ) )
          skip;
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadString() */


/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet.                        */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static unsigned char ReadAlphaName( FILE * file, char * string, int maxlen )

#else

static unsigned char ReadAlphaName( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register char ch;
    register int  len = 0;

    if ( ( *string = ReadWhiteSpace( file ) ) == '^' )
      {
        len++;
        *++string = ReadWhiteSpace( file );
      } /* if */
    if ( ! isalnum( *( string++ ) ) )
      {
        NNError( "ReadAlphaName: alphanumerische Zeichenkette erwartet" );
      } /* if */
    len++;
    while( ( isalnum( *string = getc( file ) ) || ( *string == '_' )
            || ( *string == '?' ) || ( *string == '!' ) ) && ( len < maxlen ) )
      {
        string++;
        len++;
      } /* while */
    V ungetc( *string, file );
    if ( len == maxlen )
      {
        while ( isalnum( ch = getc( file ) ) )
          skip;
        V ungetc( ch, file );
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadAlphaName() */


/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet; Kommentarzeilen werden */
/* ueberlesen. Funktion fuer das neue Dateiformat.                           */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static unsigned char ReadCmdToken( FILE * file, char * string, int maxlen )

#else

static unsigned char ReadCmdToken( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register char ch;
    register int  len = 0;

    if ( ! isalnum( *( string++ ) = ReadCharComment( file ) ) )
      {
        NNError( "ReadCmdToken(): alphanumerical string expected" );
      } /* if */
    len++;
    while (    ( isalnum( *string = getc( file ) ) || ( *string == '_' ) )
            && ( len < maxlen ) )
      {
        string++;
        len++;
      } /* while */
    V ungetc( *string, file );
    if ( len == maxlen )
      {
        while ( isalnum( ch = getc( file ) ) || ( ch == '_' ) )
          skip;
        V ungetc( ch, file );
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadCmdToken() */


/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet; Kommentarzeilen werden */
/* ueberlesen. Funktion fuer das neue Dateiformat.                           */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static void ReadEnclString( FILE * file, char * string, int maxlen )

#else

static void ReadEnclString( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register char ch    = '\0',
                  * str = string;
    register int  len    = 0;
    char          delimiter;

    if (    ( ( delimiter = ReadCharComment( file ) ) != '\'' )
         && ( delimiter != '"' ) )
      {
        NNError( "ReadEnclString(): string leading ' or \" expected" );
      } /* if */
    while ( ! feof( file ) && ( ( *str = getc( file ) ) != delimiter )
            && ( len < maxlen ) )
      {
        str++;
        len++;
      } /* while */
    if ( ( len == maxlen ) && ( *str != delimiter ) )
      {
        while ( ! feof( file ) && ( ( ch = getc( file ) ) != delimiter ) )
          skip;
      } /* if */
    else
      {
        ch = *str;
      } /* else */
    if ( ch != delimiter )
      {
        V sprintf( string, "ReadEnclString(): closing %c is missing",
                                                                   delimiter );
        NNError( string );
      } /* if */
    *str = '\0';
    
  } /* ReadEnclString() */


/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet.                        */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static void ReadCommand( FILE * file, char * string, int maxlen )

#else

static void ReadCommand( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    if ( ReadAlphaName( file, string, maxlen ) != 3 )
      {
        NNError( "ReadCommand(): unknown command found" );
      } /* if */
    ReadSymbol( file, ':' );

  } /* ReadCommand() */


/*****************************************************************************/
/* Lesen eines Koordinatenpaares `x@y` aus einer Textdatei.                  */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static int ReadCoordinates( FILE * file, int * x )

#else

static int ReadCoordinates( file, x )

  FILE * file;
  int  * x;

#endif

  {

    *x = ReadNumber( file );
    ReadSymbol( file, '@' );
    return ( ReadNumber( file ) );
    
  } /* ReadCoordinates() */


/*****************************************************************************/
/*********************** insert_place_meaning() ******************************/
/*****************************************************************************/
/*                                                  08.08.1994 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static M_LIST insert_place_meaning( char * meaning, Place pl )

#else

static M_LIST insert_place_meaning( meaning, pl )

  char  * meaning;
  Place pl;

#endif

  {

    register M_LIST mptr,
                    newmptr;
    register P_LIST pptr;
    register char   * rptr,
                    * wptr;
    register int    comp = -1;

    rptr = wptr = meaning;
    while ( *rptr )
      {
        if ( ! isspace( *rptr ) )
          {
            *( wptr++ ) = *rptr;
          } /* if */
        rptr++;
      } /* while */
    *wptr = *rptr;
    if (    ! ( mptr = CM_placeMList )
         || ( strcmp( meaning, mptr->meaning ) < 0 ) )
      {
        ( newmptr = (M_LIST) Malloc( sizeof( M_LIST_ELEM ) ) )
                                                        ->next = CM_placeMList;
        ( CM_placeMList = newmptr )->number = 1;
        newmptr->meaning = Strdup( meaning );
        CM_countPMeanings++;
        ( pptr = newmptr->list = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                                 ->next = NULL;
        pptr->item = pl;
      } /* if */
    else
      {
        while (    ( newmptr = mptr->next )
                && ( ( comp = strcmp( newmptr->meaning, meaning ) ) < 0 ) )
          {
            mptr = newmptr;
          } /* while */
        if ( newmptr && ! comp )
          {
            ( pptr = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                        ->next = newmptr->list;
            ( newmptr->list = pptr )->item = pl;
            newmptr->number++;
          } /* if */
        else
          {
            ( newmptr = (M_LIST) Malloc( sizeof( M_LIST_ELEM ) ) )
                                                           ->next = mptr->next;
            mptr->next = newmptr;
            newmptr->meaning = Strdup( meaning );
            newmptr->number = 1;
            CM_countPMeanings++;
            ( pptr = newmptr->list = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                                 ->next = NULL;
            pptr->item = pl;
          } /* else */
      } /* else */

    CM_countMPlaces++;

    return ( CM_placeMList );

  } /* insert_place_meaning() */


/*****************************************************************************/
/*********************** insert_trans_meaning() ******************************/
/*****************************************************************************/
/*                                                  08.08.1994 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static M_LIST insert_trans_meaning( char * meaning, Transition tr )

#else

static M_LIST insert_trans_meaning( meaning, tr )

  char       * meaning;
  Transition tr;

#endif

  {

    register M_LIST mptr,
                    newmptr;
    register P_LIST pptr;
    register char   * rptr,
                    * wptr;
    register int    comp = -1;

    rptr = wptr = meaning;
    while ( *rptr )
      {
        if ( ! isspace( *rptr ) )
          {
            *( wptr++ ) = *rptr;
          } /* if */
        rptr++;
      } /* while */
    *wptr = *rptr;
    if (    ! ( mptr = CM_transMList )
         || ( strcmp( meaning, mptr->meaning ) < 0 ) )
      {
        ( newmptr = CM_transMList = (M_LIST) Malloc( sizeof( M_LIST_ELEM ) ) )
                                                                 ->next = NULL;
        newmptr->meaning = Strdup( meaning );
        newmptr->number = 1;
        CM_countTMeanings++;
        ( pptr = newmptr->list = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                                 ->next = NULL;
        pptr->item = (Place) tr;
      } /* if */
    else
      {
        while (    ( newmptr = mptr->next )
                && ( ( comp = strcmp( newmptr->meaning, meaning ) ) < 0 ) )
          {
            mptr = newmptr;
          } /* while */
        if ( newmptr && ! comp )
          {
            ( pptr = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                        ->next = newmptr->list;
            ( newmptr->list = pptr )->item = (Place) tr;
            newmptr->number++;
          } /* if */
        else
          {
            ( newmptr = (M_LIST) Malloc( sizeof( M_LIST_ELEM ) ) )
                                                           ->next = mptr->next;
            mptr->next = newmptr;
            newmptr->meaning = Strdup( meaning );
            newmptr->number = 1;
            CM_countTMeanings++;
            ( pptr = newmptr->list = (P_LIST) Malloc( sizeof( P_LIST_ELEM ) ) )
                                                                 ->next = NULL;
            pptr->item = (Place) tr;
          } /* else */
      } /* else */

    CM_countMTransitions++;

    return ( CM_transMList );

  } /* insert_trans_meaning() */


/*****************************************************************************/
/************************ delete_meaning_list() ******************************/
/*****************************************************************************/
/*                                                  08.08.1994 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void delete_meaning_list( M_LIST list )

#else

void delete_meaning_list( list )

  M_LIST list;

#endif

  {

    register M_LIST mdel;
    register P_LIST pptr,
                    pdel;

    mdel = list;
    while ( mdel )
      {
        Free( mdel->meaning );
        pdel = mdel->list;
        while ( pdel )
          {
            pdel = ( pptr = pdel )->next;
            Free( pptr );
          } /* while */
        mdel = ( list = mdel )->next;
        Free( list );
      } /* while */

  } /* delete_meaning_list() */


/*****************************************************************************/
/******************************* read_PEP ************************************/
/*****************************************************************************/
/*                                                  13.12.1993 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet read_PEP( char * PEPfilename )

#else

LMPNet read_PEP( PEPfilename )

  char * PEPfilename;

#endif

  {

    int                 x, y = 0,
                        r1, r2,
                        tx, ty,
                        cmarks, imarks, width, capacity,
                        num, num2, weight;
    register char       ch,
                        ** Namen,
                        * S_T_Flags;
    char                cmd[513],
                        name[4097],
                        meaning[4097],
                        name2[NAMELENGTH];
    register FILE       * infile;
    register LMPNet     net;
    register Cardinal32 AnzNamen,
                        MaxNamen,
                        count;
    Place               pl_last = NULL;
    Transition          tr_last = NULL;
    STEdge              st_last = NULL;
    TSEdge              ts_last = NULL;

    NNinput_file = PEPfilename;
    NNinput_line = 1;

    /* Oeffne PEP-Datei */
    if ( ! ( infile = fopen( PEPfilename, "r" ) ) )
      {
        V fprintf( stderr,
              "ERROR: CM.read_PEP(): could not open file '%s' for reading.\n",
                 PEPfilename );
        exit ( 4 );
      } /* if */

    ReadAlphaName( infile, cmd, 512 );
    if ( strcmp( cmd, "PEP" ) )
      {
        NNError( "keyword 'PEP' expected" );
      } /* if */
    ReadAlphaName( infile, cmd, 512 );
    if (    strcmp( cmd, "PetriBox" ) && strcmp( cmd, "Net" )
         && strcmp( cmd, "PTNet" )    && strcmp( cmd, "CENet" )
         && strcmp( cmd, "HLNet" ) )
      {
        NNError( "file identifier ( 'PTNet', 'PetriBox', ... ) expected" );
      } /* if */
    strncpy( CM_nettype, cmd, 30 );
    ReadAlphaName( infile, cmd, 512 );
    if ( ! ( strncmp( cmd, "FORMAT_N", 8 ) ) )
      {
        return ( read_new_PEP( PEPfilename ) );
      } /* if */
    else if ( strcmp( cmd, "FORMAT_A" ) )
      {
        NNError( "keyword 'FORMAT_A' expected" );
        return ( NULL );
      } /* if */

    /* Initialisiere net */
    Namen = (char **) Malloc( ( count = MaxNamen = NAMES_START )
                                                          * sizeof( char * ) );
    S_T_Flags = (char *) Malloc( count * sizeof( char ) );
    AnzNamen = 0;
    while ( --count )
      {
        Namen[ count ] = (char *) NULL;
        S_T_Flags[ count ] = '\0';
      } /* while */
    net = empty_net();
    net->NumComponents = 1;

    ReadCommand( infile, cmd, 512 );
    while ( strcmp( cmd, "END" ) )
      {
        if ( ! strcmp( cmd, "PLA" ) )
          {
            ReadCommand( infile, cmd, 512 );
            num = 0;
            imarks = cmarks = 0;
            *name = *meaning = '\0';
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "idt" ) )
                  {
                    num = ReadNumber( infile );
                  } /* if */
                else if ( ! strcmp( cmd, "nam" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "mea" ) )
                  {
                    ReadString( infile, meaning, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dnp" ) )
                  {
                    ty = ReadCoordinates( infile, &tx );
                  } /* else if */
                else if ( ! strcmp( cmd, "dsz" ) )
                  {
                    r2 = ReadCoordinates( infile, &r1 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                else if ( ! strcmp( cmd, "dwd" ) )
                  {
                    width = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cap" ) )
                  {
                    while ( ( ch = getc( infile ) ) == ' ' )
                      skip;
		    if ( isdigit( ch ) )
                      {
                        V ungetc( ch, infile );
                        capacity = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        V ungetc( ch, infile );
                        ReadAlphaName( infile, name2, NAMELENGTH-1 );
                        if ( strcmp( name2, "Omega" ) )
                          {
                            NNError( "incorrect capacity" );
                          } /* if */
                      } /* else */
                  } /* else if */
                else if ( ! strcmp( cmd, "ima" ) )
                  {
                    imarks = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cma" ) )
                  {
                    cmarks = ReadNumber( infile );
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
            if ( ! num )
              {
                NNError( "missing place identifier" );
              } /* if */
            if ( num > AnzNamen )
              {
                AnzNamen = num;
              } /* if */
            else if ( Namen[ num ] )
              {
                NNError( "multiply defined place identificator" );
              } /* else if */
            if ( AnzNamen >= MaxNamen )
              {
                Namen = (char **) Realloc( (char *) Namen,
                             ( MaxNamen += NAMES_OFFSET ) * sizeof( char * ) );
                S_T_Flags = (char *) Realloc( (char *) S_T_Flags,
                                                   MaxNamen * sizeof( char ) );
                for ( count = MaxNamen - NAMES_OFFSET ;
                                                   count < MaxNamen ; count++ )
                  {
                    Namen[ count ] = (char *) NULL;
                    S_T_Flags[ count ] = '\0';
                  } /* for */
              } /* if */
            if ( ! *name )
              {
                V sprintf( name, "S%d", num );
              } /* if */
            Namen[ num ] = fast_insert_place( net, name,
                           CM_use_imarks ? imarks : cmarks, &pl_last ) -> name;
            if ( CM_storeMeaning && *meaning )
              {
                insert_place_meaning( meaning, pl_last );
              } /* if */
            pl_last->label = EMPTY;
            pl_last->component = 1;
            S_T_Flags[ num ] = 'S';
            if ( CM_storeCoordinates )
              {
                Coords co;
                pl_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
                co->next = NULL;
                co->x = x;
                co->y = y;
              } /* if */
          } /* if */
        else if (    ! strcmp( cmd, "TRA" )
                  || ! strcmp( cmd, "RTR" ) )
          {
            ReadCommand( infile, cmd, 512 );
            num = 0;
            *name = *meaning =  '\0';
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "idt" ) )
                  {
                    num = ReadNumber( infile );
                  } /* if */
                else if ( ! strcmp( cmd, "nam" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "mea" ) )
                  {
                    ReadString( infile, meaning, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dnp" ) )
                  {
                    ty = ReadCoordinates( infile, &tx );
                  } /* else if */
                else if ( ! strcmp( cmd, "dsz" ) )
                  {
                    r2 = ReadCoordinates( infile, &r1 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                else if ( ! strcmp( cmd, "dwd" ) )
                  {
                    width = ReadNumber( infile );
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
            if ( ! num )
              {
                NNError( "missing transition identifier" );
              } /* if */
            if ( num > AnzNamen )
              {
                AnzNamen = num;
              } /* if */
            else if ( Namen[ num ] )
              {
                NNError( "multiply defined transition identifier" );
              } /* else if */
            if ( AnzNamen >= MaxNamen )
              {
                Namen = (char **) Realloc( (char *) Namen,
                                     ( MaxNamen += NAMES_OFFSET ) * sizeof( char * ) );
                S_T_Flags = (char *) Realloc( (char *) S_T_Flags,
                                                   MaxNamen * sizeof( char ) );
                for ( count = MaxNamen - NAMES_OFFSET ; count < MaxNamen ; count++ )
                  {
                    Namen[ count ] = (char *) NULL;
                    S_T_Flags[ count ] = '\0';
                  } /* for */
              } /* if */
            if ( ! *name )
              {
                V sprintf( name, "T%d", num );
              } /* if */
            Namen[ num ] = fast_insert_transition( net, name, &tr_last )
                                                                       -> name;
            if ( CM_storeMeaning && *meaning )
              {
                insert_trans_meaning( meaning, tr_last );
              } /* if */
            tr_last->label = NULL;
            S_T_Flags[ num ] = 'T';
            if ( CM_storeCoordinates )
              {
                Coords co;
                tr_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
                co->next = NULL;
                co->x = x;
                co->y = y;
              } /* if */
          } /* else if */
        else if ( ! strcmp( cmd, "ARC" ) )
          {
            Coords lastco  = NULL,
                   firstco = NULL;
            ReadCommand( infile, cmd, 512 );
            num = num2 = weight = 0;
            *name = *meaning = '\0';
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "lpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                else if ( ! strcmp( cmd, "dsz" ) )
                  {
                    width = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "wgt" ) )
                  {
                    weight = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "sou" ) )
                  {
                    num = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "des" ) )
                  {
                    num2 = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "nam" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "mea" ) )
                  {
                    ReadString( infile, meaning, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "vsf" ) )
                  {
                    char flag[2];
                    ReadAlphaName( infile, flag, 1 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                    if ( CM_storeCoordinates )
                      {
                        if ( lastco )
                          {
                            lastco = lastco->next
                                   = (Coords) Malloc( sizeof( CoordEntry ) );
                          } /* if */
                        else
                          {
                            firstco = lastco
                                    = (Coords) Malloc( sizeof( CoordEntry ) );
                          } /* else */
                        lastco->next = NULL;
                        lastco->x = x;
                        lastco->y = y;
                      } /* if */
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
            if ( ! num || ( num > AnzNamen ) || ! Namen[ num ] )
              {
                NNError( "arc: missing source identifier" );
              } /* if */
            if ( ! num2 || ( num2 > AnzNamen ) || ! Namen[ num2 ]  )
              {
                NNError( "arc: missing destination identifier" );
              } /* if */
            if ( ( S_T_Flags[ num ] == 'S' ) && ( S_T_Flags[ num2 ] == 'T' ) )
              {
                fast_insert_PT_con( net, Namen[ num ], Namen[ num2 ],
                                        weight, &pl_last, &tr_last, &st_last );
                if ( CM_storeCoordinates )
                  {
                    lastco = (Coords) Malloc( sizeof( CoordEntry ) );
                    lastco->x = weight;
                    lastco->y = 0;
                    lastco->next = firstco;
                    st_last->weight = (Cardinal32) lastco;
                  } /* if */
              } /* if */
            else if (    ( S_T_Flags[ num ] == 'T' )
                      && ( S_T_Flags[ num2 ] == 'S' ) )
              {
                fast_insert_TP_con( net, Namen[ num2 ], Namen[ num ],
                                        weight, &pl_last, &tr_last, &ts_last );
                if ( CM_storeCoordinates )
                  {
                    lastco = (Coords) Malloc( sizeof( CoordEntry ) );
                    lastco->x = weight;
                    lastco->y = 0;
                    lastco->next = firstco;
                    ts_last->weight = (Cardinal32) lastco;
                  } /* if */
              } /* if */
            else
              {
                NNError( "faulty PEP net file: S->S resp. T->T arc found" );
              } /* else */
          } /* else if */
        else if ( ! strcmp( cmd, "IPL" ) )
          {
            ReadCommand( infile, cmd, 512 );
            num = 0;
            imarks = 1;
            cmarks = 0;
            *name = *meaning = '\0';
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "idt" ) )
                  {
                    num = ReadNumber( infile );
                  } /* if */
                else if ( ! strcmp( cmd, "nam" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "mea" ) )
                  {
                    ReadString( infile, meaning, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dnp" ) )
                  {
                    ty = ReadCoordinates( infile, &tx );
                  } /* else if */
                else if ( ! strcmp( cmd, "dsz" ) )
                  {
                    r2 = ReadCoordinates( infile, &r1 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                else if ( ! strcmp( cmd, "dwd" ) )
                  {
                    width = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cap" ) )
                  {
                    while ( ( ch = getc( infile ) ) == ' ' )
                      skip;
		    if ( isdigit( ch ) )
                      {
                        V ungetc( ch, infile );
                        capacity = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        V ungetc( ch, infile );
                        ReadAlphaName( infile, name2, NAMELENGTH-1 );
                        if ( strcmp( name2, "Omega" ) )
                          {
                            V fprintf( stderr,
                                            "ERROR: CM.read_PEP(): capacity" );
                            V fprintf( stderr, "'Omega' expected.\n" );
                          } /* if */
                      } /* else */
                  } /* else if */
                else if ( ! strcmp( cmd, "ima" ) )
                  {
                    imarks = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cma" ) )
                  {
                    cmarks = ReadNumber( infile );
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
            if ( ! num )
              {
                NNError( "missing place identifier" );
              } /* if */
            if ( num > AnzNamen )
              {
                AnzNamen = num;
              } /* if */
            else if ( Namen[ num ] )
              {
                NNError( "multiply defined place identifer" );
              } /* else if */
            if ( AnzNamen >= MaxNamen )
              {
                Namen = (char **) Realloc( (char *) Namen,
                              ( MaxNamen += NAMES_OFFSET ) * sizeof( char * ) );
                S_T_Flags = (char *) Realloc( (char *) S_T_Flags,
                                                    MaxNamen * sizeof( char ) );
                for ( count = MaxNamen - NAMES_OFFSET ;
                                                    count < MaxNamen ; count++ )
                  {
                    Namen[ count ] = (char *) NULL;
                    S_T_Flags[ count ] = '\0';
                  } /* for */
              } /* if */
            if ( ! *name )
              {
                V sprintf( name, "S%d", num );
              } /* if */
            Namen[ num ] = fast_insert_place( net, name,
                                 CM_use_imarks ? 1 : cmarks, &pl_last ) -> name;
            if ( CM_storeMeaning && *meaning )
              {
                insert_place_meaning( meaning, pl_last );
              } /* if */
            pl_last->label = ENTRY;
            pl_last->component = 0;
            S_T_Flags[ num ] = 'S';
            if ( CM_storeCoordinates )
              {
                Coords co;
                pl_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
                co->next = NULL;
                co->x = x;
                co->y = y;
              } /* if */
          } /* else if */
        else if ( ! strcmp( cmd, "OPL" ) )
          {
            ReadCommand( infile, cmd, 512 );
            num = imarks = cmarks = 0;
            *name = *meaning = '\0';
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "idt" ) )
                  {
                    num = ReadNumber( infile );
                  } /* if */
                else if ( ! strcmp( cmd, "nam" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "mea" ) )
                  {
                    ReadString( infile, meaning, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dnp" ) )
                  {
                    ty = ReadCoordinates( infile, &tx );
                  } /* else if */
                else if ( ! strcmp( cmd, "dsz" ) )
                  {
                    r2 = ReadCoordinates( infile, &r1 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                else if ( ! strcmp( cmd, "dwd" ) )
                  {
                    width = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cap" ) )
                  {
                    while ( ( ch = getc( infile ) ) == ' ' )
                      skip;
		    if ( isdigit( ch ) )
                      {
                        V ungetc( ch, infile );
                        capacity = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        V ungetc( ch, infile );
                        ReadAlphaName( infile, name2, NAMELENGTH-1 );
                        if ( strcmp( name2, "Omega" ) )
                          {
                            V fprintf( stderr,
                                            "ERROR: CM.read_PEP(): capacity" );
                            V fprintf( stderr, "'Omega' expected.\n" );
                          } /* if */
                      } /* else */
                  } /* else if */
                else if ( ! strcmp( cmd, "ima" ) )
                  {
                    imarks = ReadNumber( infile );
                  } /* else if */
                else if ( ! strcmp( cmd, "cma" ) )
                  {
                    cmarks = ReadNumber( infile );
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
            if ( ! num )
              {
                NNError( "missing place identifier" );
              } /* if */
            if ( num > AnzNamen )
              {
                AnzNamen = num;
              } /* if */
            else if ( Namen[ num ] )
              {
                NNError( "multiply defined place identifier" );
              } /* else if */
            if ( AnzNamen >= MaxNamen )
              {
                Namen = (char **) Realloc( (char *) Namen,
                              ( MaxNamen += NAMES_OFFSET ) * sizeof( char * ) );
                S_T_Flags = (char *) Realloc( (char *) S_T_Flags,
                                                    MaxNamen * sizeof( char ) );
                for ( count = MaxNamen - NAMES_OFFSET ;
                                                    count < MaxNamen ; count++ )
                  {
                    Namen[ count ] = (char *) NULL;
                    S_T_Flags[ count ] = '\0';
                  } /* for */
              } /* if */
            if ( ! *name )
              {
                V sprintf( name, "S%d", num );
              } /* if */
            Namen[ num ] = fast_insert_place( net, name,
                           CM_use_imarks ? imarks : cmarks, &pl_last ) -> name;
            if ( CM_storeMeaning && *meaning )
              {
                insert_place_meaning( meaning, pl_last );
              } /* if */
            pl_last->label = EXIT;
            pl_last->component = 0;
            S_T_Flags[ num ] = 'S';
            if ( CM_storeCoordinates )
              {
                Coords co;
                pl_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
                co->next = NULL;
                co->x = x;
                co->y = y;
              } /* if */
          } /* else if */
        else if ( ! strcmp( cmd, "TXT" ) )
          {
            ReadCommand( infile, cmd, 512 );
            num = 0;
            while ( strcmp( cmd, "end" ) )
              {
                if ( ! strcmp( cmd, "str" ) )
                  {
                    ReadString( infile, name, 512 );
                  } /* else if */
                else if ( ! strcmp( cmd, "dpt" ) )
                  {
                    y = ReadCoordinates( infile, &x );
                  } /* else if */
                ReadCommand( infile, cmd, 512 );
              } /* while */
          } /* else if */
        else if ( ! strcmp( cmd, "FML" ) )
          {
            ReadNewline( infile, 1 );
            V ungetc( '\n', infile );
          } /* else if */
        else if ( ! strcmp( cmd, "haf" ) )
          {
            ReadNewline( infile, 1 );
          } /* else if */
        else if ( ! strcmp( cmd, "set" ) )
          {
            ReadNewline( infile, 1 );
          } /* else if */
        else if ( ! strcmp( cmd, "pap" ) )
          {
            ReadNewline( infile, 1 );
          } /* else if */
        else if ( ! strcmp( cmd, "com" ) )
          {
            ReadString( infile, name, 4096 );
          } /* else if */
        else if ( ! strcmp( cmd, "end" ) )
          {
          } /* else if */
        else
          {
            V fprintf( stderr,
            "ERROR: CM.read_PEP(): unknown PEP command found: '%s'.\n", cmd );
          } /* else */
        ReadNewline( infile, 1 );
        ReadCommand( infile, cmd, 512 );
      } /* while */

    V fclose( infile );

    Free( Namen );

    return( net );

  } /* read_PEP() */


/*****************************************************************************/
/****************************** write_PEP ***********************************/
/*****************************************************************************/
/*                                                  13.12.1993 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_PEP( LMPNet net, char * PEPfilename )

#else

void write_PEP( net, PEPfilename )
  /* Gibt die mit net bezeichnete Matrix als Inzidenzmatrix nach 'file' aus  */

  LMPNet net;
  char   * PEPfilename;

#endif

  {

    /* NOCH NICHT IMPLEMENTIERT !!! */

  } /* write_PEP() */


/*****************************************************************************/
/***************************** read_new_PEP **********************************/
/*****************************************************************************/
/* Neu eingefuegt (neues Format)       19.12.1994 - 09.01.1995 Stefan Roemer */
/*                                     Bloecke      07.03.1997 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet read_new_PEP( char * PEPfilename )

#else

LMPNet read_new_PEP( PEPfilename )

  char * PEPfilename;

#endif

  {

    int                 cx, cy,
                        sx, sy,
                        px, py,
                        prx, pry,
                        gx, gy,
                        grx, gry,
                        wrx, wry,
                        arx, ary,
                        mx, my,
                        mrx, mry,
                        color, thickness,
                        cmarks, imarks, capacity,
                        num, num2, weight,
                        def_pl_radius = 10,
                        def_pl_prx = 10,
                        def_pl_pry = 0,
                        def_pl_mrx = 10,
                        def_pl_mry = 10,
                        def_pl_wrx = 10,
                        def_pl_wry = 10,
                        def_pl_imarks = 0,
                        def_pl_cmarks = 0,
                        def_pl_capacity = OMEGA,
                        def_pl_thickness = 2,
                        def_pl_color = 1,
                        def_tr_size_x = 10,
                        def_tr_size_y = 10,
                        def_tr_prx = 10,
                        def_tr_pry = 0,
                        def_tr_mrx = 10,
                        def_tr_mry = 10,
                        def_tr_thickness = 2,
                        def_tr_color = 1;
    register char       ch;
    Place               * PlArray;
    Transition          * TrArray,
                        * PhTrArray;
    char                cmd[513],
                        name[NAMESIZE+1],
                        meaning[MEANINGSIZE+1],
                        marking[MEANINGSIZE+1],
                        guard[GUARDSIZE+1],
                        reference[REFSIZE+1],
                        treference[REFSIZE+1],
                        blocklist[REFSIZE+1],
                        synclist[REFSIZE+1],
                        name2[MEANINGSIZE+1];
    register FILE       * infile;
    register LMPNet     net;
    register Cardinal32 AnzPlNamen,
                        MaxPlNamen,
                        AnzTrNamen,
                        MaxTrNamen,
                        AnzPhTrNamen,
                        MaxPhTrNamen,
                        autocounter,
                        ident,
                        count;
    Cardinal32          max_ident = 0,
                        invisible;
    Place               pl_last = NULL;
    Transition          tr_last = NULL;
    STEdge              st_last = NULL;
    TSEdge              ts_last = NULL;
    bool                entry_place,
                        exit_place,
                        refined,
                        is_sync_trans,
                        autonumbering,
                        abs_meaning,
                        abs_name,
                        num_marking,
                        specified;

    NNinput_file = PEPfilename;
    NNinput_line = 1;
    PlArray = (Place *) Malloc( ( count = MaxPlNamen = MaxTrNamen
                            = MaxPhTrNamen = NAMES_START ) * sizeof( Place ) );
    TrArray = (Transition *) Malloc( count * sizeof( Transition ) );
    PhTrArray = (Transition *) Malloc( count * sizeof( Transition ) );
    AnzPlNamen = AnzTrNamen = AnzPhTrNamen = 0;
    while ( --count )
      {
        PlArray[ count ] = (Place) NULL;
        TrArray[ count ] = (Transition) NULL;
        PhTrArray[ count ] = (Transition) NULL;
      } /* while */

    if ( ! ( infile = fopen( PEPfilename, "r" ) ) )
      {
        V fprintf( stderr,
               "ERROR: CM.read_PEP(): could not open file '%s' for reading.\n",
                 PEPfilename );
        exit ( 4 );
      } /* if */

    ReadCmdToken( infile, cmd, 512 );
    if ( strcmp( cmd, "PEP" ) )
      {
        NNError( "keyword 'PEP' expected" );
        return ( NULL );
      } /* if */
    ReadNewline( infile, 1 );
    ReadCmdToken( infile, cmd, 512 );
    if (    strcmp( cmd, "PetriBox" ) && strcmp( cmd, "HLNet" )
         && strcmp( cmd, "PTNet" ) )
      {
      NNError( "format identifier ( 'PTNet', 'PetriBox', 'HLNet' ) expected" );
      } /* if */
    strncpy( CM_nettype, cmd, 30 );
    ReadNewline( infile, 1 );
    ReadCmdToken( infile, cmd, 512 );
    if ( strncmp( cmd, "FORMAT_N", 8 ) )
      {
        NNError( "keyword 'FORMAT_N' or 'FORMAT_N2' expected" );
        return ( NULL );
      } /* if */

    net = empty_net();
    net->NumComponents = 1;

    ReadNewline( infile, 1 );
    ReadCmdToken( infile, cmd, 512 );
    while ( *cmd == 'D' )
      {
        if ( ! strcmp( cmd, "DPL" ) )
          {
            /* To be done */
            ReadNewline( infile, 1 );
          } /* if */
        else if ( ! strcmp( cmd, "DTR" ) )
          {
            /* To be done */
            ReadNewline( infile, 1 );
          } /* else if */
        else if ( ! strcmp( cmd, "DPT" ) )
          {
            /* To be done */
            ReadNewline( infile, 1 );
          } /* else if */
        else
          {
            V sprintf( name, "read_new_PEP(): unknown keyword '%s'", cmd );
            NNError( name );
            return ( NULL );
          } /* else */
        ReadCmdToken( infile, cmd, 512 );
      } /* while */

    /* Read in block information *********************************************/
    if ( ! strcmp( cmd, "BL" ) )
      {
        ReadNewline( infile, 1 );
 
        while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'P' ) )
          {
            while ( ch != '\n' )
              {
                if ( isdigit( ch ) || ( ch == '-' ) )
                  {
                    V ungetc( ch, infile );
                    num = ReadNumber( infile );
                    if ( TestSymbol( infile, '@' ) )
                      {
                        cx = num;
                        cy = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        if ( ( ident = num ) != autocounter )
                          {
                            autonumbering = False;
                          } /* if */
                        specified = True;
                      } /* else */
                  } /* if */
                else
                  {
                    switch ( ch )
                      {
                        case '\'':
                        case '"' : V ungetc( ch, infile );
                                   ReadEnclString( infile, name, NAMESIZE );
                                   break;
                        case 'n' : pry = ReadCoordinates( infile, &prx );
                                   break;
                        case 'N' : py = ReadCoordinates( infile, &px );
                                   abs_name = True;
                                   break;
                        case 'b' : ReadEnclString( infile, meaning,
                                                                 MEANINGSIZE );
                                   break;
                        case 'a' : mry = ReadCoordinates( infile, &mrx );
                                   break;
                        case 'A' : my = ReadCoordinates( infile, &mx );
                                   abs_meaning = True;
                                   break;
                        case 'R' : ReadEnclString( infile, reference, REFSIZE );
                                   break;
                        case 'T' : ReadEnclString( infile, treference,REFSIZE );
                                   break;
                        case 'u' : ReadEnclString( infile, blocklist, REFSIZE );
                                   break;
                        default  : V sprintf( name, "unknown token '%c'", ch );
                                   NNError( name );
                      } /* switch */
                  } /* else */
                ch = ReadCharComment( infile );
              } /* while */
 
            NNinput_line++;
          } /* while */

        V ungetc( ch, infile );
        ReadCmdToken( infile, cmd, 512 );
      } /* if */
 
    /* Read in place nodes ***************************************************/
    if ( strcmp( cmd, "PL" ) )
      {
        NNError( "keyword 'PL' expected" );
        return ( NULL );
      } /* if */
    ReadNewline( infile, 1 );

    autocounter = 0;
    autonumbering = True;
    while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'T' ) )
      {
        autocounter++; 
        ident = 0;
        cx = cy =  mx = my = px = py = 0;
        prx = def_pl_prx;
        pry = def_pl_pry;
        wrx = def_pl_wrx;
        wry = def_pl_wry;
        mrx = def_pl_mrx;
        mry = def_pl_mry;
        cmarks = def_pl_cmarks;
        imarks = def_pl_imarks;
        num_marking = True;
        invisible = 0;
        abs_name = abs_meaning = specified = False;
        *marking = *guard = *name = *name2 = *meaning = *reference
                                             = *treference = *blocklist = '\0';
        capacity = def_pl_capacity;
        color = def_pl_color;
        sx = def_pl_radius;
        thickness = def_pl_thickness;
        entry_place = exit_place = False;
        while ( ch != '\n' )
          {
            if ( isdigit( ch ) || ( ch == '-' ) )
              {
                V ungetc( ch, infile );
                num = ReadNumber( infile );
                if ( TestSymbol( infile, '@' ) )
                  {
                    cx = num;
                    cy = ReadNumber( infile );
                  } /* if */
                else
                  {
                    if ( ( ident = num ) != autocounter )
                      {
                        autonumbering = False;
                      } /* if */
                    if ( ident > max_ident )
                      {
                        max_ident = ident;
                      } /* if */
                    specified = True;
                  } /* else */
              } /* if */ 
            else
              {
                switch ( ch )
                  {
                    case '\'':
                    case '"' : V ungetc( ch, infile );
                               ReadEnclString( infile, name, NAMESIZE );
                               break;
                    case 'n' : pry = ReadCoordinates( infile, &prx );
                               break;
                    case 'N' : py = ReadCoordinates( infile, &px );
                               abs_name = True;
                               break;
                    case 'b' : ReadEnclString( infile, meaning, MEANINGSIZE );
                               break;
                    case 'a' : mry = ReadCoordinates( infile, &mrx );
                               break;
                    case 'A' : my = ReadCoordinates( infile, &mx );
                               abs_meaning = True;
                               break;
                    case 'M' : imarks = ReadNumber( infile );
                               num_marking = True;
                               break;
                    case 'm' : cmarks = ReadNumber( infile );
                               num_marking = True;
                               break;
                    case 'Z' : ReadEnclString( infile, guard, GUARDSIZE );
                               num_marking = False;
                               break;
                    case 'B' : wry = ReadCoordinates( infile, &wrx );
                               break;
                    case 'z' : ReadEnclString( infile, name2, MEANINGSIZE );
                               num_marking = False;
                               break;
                    case 'y' : ReadEnclString( infile, marking, MEANINGSIZE );
                               num_marking = False;
                               break;
                    case 'R' : ReadEnclString( infile, reference, REFSIZE );
                               break;
                    case 'T' : ReadEnclString( infile, treference, REFSIZE );
                               break;
                    case 'u' : ReadEnclString( infile, blocklist, REFSIZE );
                               break;
                    case 'k' : capacity = ReadNumber( infile );
/* POSITION KAPAZITAET */
                               break;
                    case 'e' : if ( exit_place )
                                 {
                                   NNError( "both exit and entry flag found" );
                                 } /* else */
                               else
                                 {
                                   entry_place = True;
                                 } /* else */
                               break;
                    case 'x' : if ( entry_place )
                                 {
                                   NNError( "both entry and exit flag found" );
                                 } /* else */
                               else
                                 {
                                   exit_place = True;
                                 } /* else */
                               break;
                    case 'v' : invisible = ReadNumber( infile );
                               break;
                    case 's' : sx = ReadNumber( infile );
                               break;
                    case 't' : thickness = ReadNumber( infile );
                               break;
                    case 'c' : color = ReadNumber( infile );
                               break;
                    default  : V sprintf( name, "unknown token '%c'", ch );
                               NNError( name );
                  } /* switch */
              } /* else */
            ch = ReadCharComment( infile );
          } /* while */

        /* insert place node in the net */
        if ( autonumbering )
          {
            ident = autocounter;
          } /* if */
        else if ( ! specified )
          {
            NNError( "missing place identifier" );
          } /* if */
        if ( ident > AnzPlNamen )
          {
            AnzPlNamen = ident;
          } /* if */
        else if ( PlArray[ ident ] )
          {
            V sprintf( guard, "place identifier %d multiply used", ident );
            NNError( guard );
          } /* else if */
        if ( AnzPlNamen >= MaxPlNamen )
          {
            PlArray = (Place *) Realloc( (char *) PlArray,
                            ( MaxPlNamen += NAMES_OFFSET ) * sizeof( Place ) );
            for ( count = MaxPlNamen - NAMES_OFFSET ;
                                                 count < MaxPlNamen ; count++ )
              {
                PlArray[ count ] = (Place) NULL;
              } /* for */
          } /* if */
        if ( ! *name )
          {
            V sprintf( name, "PL%d", ident );
          } /* if */
        PlArray[ ident ] = fast_insert_place( net, name,
                                   CM_use_imarks ? imarks : cmarks, &pl_last );
        if ( CM_storeMeaning && *meaning )
          {
            insert_place_meaning( meaning, pl_last );
          } /* if */
        if ( ! strcmp( CM_nettype, "PetriBox" ) )
          {
            if ( CM_use_imarks )
              {
                pl_last->marks = entry_place ? 1 : num_marking ? imarks : 0;
              } /* if */
            pl_last->label = entry_place ? ENTRY : exit_place ? EXIT : EMPTY;
          }
        pl_last->component = 1;
        if ( ! abs_name )
          {
            px = cx + prx;
            py = cy + pry;
          } /* if */
        if ( ! abs_meaning )
          {
            mx = cx + mrx;
            my = cy + mry;
          } /* if */
        if ( CM_storeCoordinates || CM_storeReferences )
          {
            Coords co;
            pl_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
            co->x = cx;
            co->y = cy;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = mx;
            co->y = my;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( meaning );
            co->y = (int) Strdup( reference );
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( guard );
            co->y = (int) Strdup( name2 );
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) cmarks;
            co->next = NULL;
          } /* if */

        NNinput_line++;
      } /* while */

    /* Read in transition nodes **********************************************/
    V ungetc( ch, infile );
    ReadCmdToken( infile, cmd, 512 );
    if ( strcmp( cmd, "TR" ) )
      {
        NNError( "keyword 'TR' expected" );
        return ( NULL );
      } /* if */
    ReadNewline( infile, 1 );

    autocounter = 0;
    autonumbering = True;
    while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'T' )
                                               && ( ch != 'P' ) )
      {
        autocounter++; 
        ident = 0;
        cx = cy =  mx = my = px = py = 0;
        arx = ary = 0;
        prx = def_tr_prx;
        pry = def_tr_pry;
        mrx = def_tr_mrx;
        mry = def_tr_mry;
        num_marking = True;
        invisible = 0;
        abs_name = abs_meaning = specified = is_sync_trans = False;
        *guard = *name = *name2 = *meaning = *reference = *treference
                                                           = *blocklist = '\0';
        color = def_tr_color;
        sx = def_tr_size_x;
        sy = def_tr_size_y;
        thickness = def_tr_thickness;
        while ( ch != '\n' )
          {
            refined = False;
            if ( isdigit( ch ) || ( ch == '-' ) )
              {
                V ungetc( ch, infile );
                num = ReadNumber( infile );
                if ( TestSymbol( infile, '@' ) )
                  {
                    cx = num;
                    cy = ReadNumber( infile );
                  } /* if */
                else
                  {
                    if ( ( ident = num ) != autocounter )
                      {
                        autonumbering = False;
                      } /* if */
                    if ( ident > max_ident )
                      {
                        max_ident = ident;
                      } /* if */
                    specified = True;
                  } /* else */
              } /* if */ 
            else
              {
                switch ( ch )
                  {
                    case '\'':
                    case '"' : V ungetc( ch, infile );
                               ReadEnclString( infile, name, NAMESIZE );
                               break;
                    case 'n' : pry = ReadCoordinates( infile, &prx );
                               break;
                    case 'N' : py = ReadCoordinates( infile, &px );
                               break;
                    case 'b' : ReadEnclString( infile, meaning, MEANINGSIZE );
                               break;
                    case 'a' : mry = ReadCoordinates( infile, &mrx );
                               break;
                    case 'A' : my = ReadCoordinates( infile, &mx );
                               break;
                    case 'v' : invisible = ReadNumber( infile );
                               break;
                    case 'g' : ReadEnclString( infile, guard, GUARDSIZE );
                               break;
                    case 'h' : gry = ReadCoordinates( infile, &grx );
                               break;
                    case 'i' : ReadEnclString( infile, name2, MEANINGSIZE );
                               break;
                    case 'j' : ary = ReadCoordinates( infile, &arx );
                               break;
                    case 'H' : gy = ReadCoordinates( infile, &gx );
                               break;
                    case 'R' : ReadEnclString( infile, reference, REFSIZE );
                               break;
                    case 'T' : ReadEnclString( infile, treference, REFSIZE );
                               break;
                    case 'u' : ReadEnclString( infile, blocklist, REFSIZE );
                               break;
                    case 'P' : ReadEnclString( infile, synclist, REFSIZE );
                               break;
                    case 's' : sy = sx = ReadNumber( infile );
                               break;
#undef _OLD_VERSION_
#ifdef _OLD_VERSION_
                    case 'S' : sx = ReadNumber( infile );
                               if ( TestSymbol( infile, '@' ) )
                                 {
                                   sy = ReadNumber( infile );
                                 } /* if */
                               else
                                 {
                                   NNError( "missing '@'" );
                                 } /* else */
                               break;
#else  /* _OLD_VERSION_ */
                    case 'S' : is_sync_trans = True;
                               break;
#endif /* _OLD_VERSION_ */
                    case 't' : thickness = ReadNumber( infile );
                               break;
                    case 'r' : if ( refined )
                                 {
                                   NNError( "refined flag doubly defined" );
                                 } /* else */
                               else
                                 {
                                   refined = True;
                                 } /* else */
                               break;
                    case 'c' : color = ReadNumber( infile );
                               break;
                    default  : V sprintf( name, "unknown token '%c'", ch );
                               NNError( name );
                  } /* switch */
              } /* else */
            ch = ReadCharComment( infile );
          } /* while */

        /* insert transition node in the net */
        if ( autonumbering )
          {
            ident = autocounter;
          } /* if */
        else if ( ! specified )
          {
            NNError( "missing transition identifier" );
          } /* if */
        if ( ident > AnzTrNamen )
          {
            AnzTrNamen = ident;
          } /* if */
        else if ( TrArray[ ident ] )
          {
            V sprintf( guard, "transition identifier %d multiply used", ident );
            NNError( guard );
          } /* else if */
        if ( AnzTrNamen >= MaxTrNamen )
          {
            TrArray = (Transition *) Realloc( (char *) TrArray,
                       ( MaxTrNamen += NAMES_OFFSET ) * sizeof( Transition ) );
            for ( count = MaxTrNamen - NAMES_OFFSET ;
                                                 count < MaxTrNamen ; count++ )
              {
                TrArray[ count ] = (Transition) NULL;
              } /* for */
          } /* if */
        if ( ! *name )
          {
            V sprintf( name, "TR%d", ident );
          } /* if */
        TrArray[ ident ] = fast_insert_transition( net, name, &tr_last );
        if ( CM_storeMeaning && *meaning )
          {
            insert_trans_meaning( meaning, tr_last );
          } /* if */
        tr_last->label = NULL;

        if ( ! abs_name )
          {
            px = cx + prx;
            py = cy + pry;
          } /* if */
        if ( ! abs_meaning )
          {
            mx = cx + mrx;
            my = cy + mry;
          } /* if */
        if ( CM_storeCoordinates || CM_storeReferences )
          {
            Coords co;
            tr_last->inscript = (char *)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
            co->x = cx;
            co->y = cy;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = mx;
            co->y = my;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( name2 );
            co->y = (int) Strdup( meaning );
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( guard );
            co->y = (int) Strdup( reference );
            co->next = NULL;
          } /* if */

        NNinput_line++;
      } /* while */

    /* Read in phantom transitions *******************************************/
    if ( ch == 'P' )
      {
        V ungetc( ch, infile );
        ReadCmdToken( infile, cmd, 512 );
        if ( strcmp( cmd, "PTR" ) )
          {
            NNError( "keyword 'PTR' expected" );
            return ( NULL );
          } /* if */
        ReadNewline( infile, 1 );

        autocounter = 0;
        autonumbering = True;
        while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'T' ) )
          {
            autocounter++; 
            ident = 0;
            cx = cy =  mx = my = px = py = 0;
            arx = ary = 0;
            prx = def_tr_prx;
            pry = def_tr_pry;
            mrx = def_tr_mrx;
            mry = def_tr_mry;
            num_marking = True;
            invisible = 0;
            abs_name = abs_meaning = specified = False;
            *guard = *name = *name2 = *meaning = *reference
                                            = *treference = *blocklist = '\0';
            color = def_tr_color;
            sx = def_tr_size_x;
            sy = def_tr_size_y;
            thickness = def_tr_thickness;
            while ( ch != '\n' )
              {
                refined = False;
                if ( isdigit( ch ) || ( ch == '-' ) )
                  {
                    V ungetc( ch, infile );
                    num = ReadNumber( infile );
                    if ( TestSymbol( infile, '@' ) )
                      {
                        cx = num;
                        cy = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        if ( ( ident = num ) != autocounter )
                          {
                            autonumbering = False;
                          } /* if */
                        if ( ident > max_ident )
                          {
                            max_ident = ident;
                          } /* if */
                        specified = True;
                      } /* else */
                  } /* if */ 
                else
                  {
                    switch ( ch )
                      {
                        case '\'':
                        case '"' : V ungetc( ch, infile );
                                   ReadEnclString( infile, name, NAMESIZE );
                                   break;
                        case 'n' : pry = ReadCoordinates( infile, &prx );
                                   break;
                        case 'N' : py = ReadCoordinates( infile, &px );
                                   break;
                        case 'b' : ReadEnclString( infile, meaning,
                                                                  MEANINGSIZE );
                                   break;
                        case 'a' : mry = ReadCoordinates( infile, &mrx );
                                   break;
                        case 'A' : my = ReadCoordinates( infile, &mx );
                                   break;
                        case 'v' : invisible = ReadNumber( infile );
                                   break;
                        case 'g' : ReadEnclString( infile, guard, GUARDSIZE );
                                   break;
                        case 'h' : gry = ReadCoordinates( infile, &grx );
                                   break;
                        case 'i' : ReadEnclString( infile, name2, MEANINGSIZE );
                                   break;
                        case 'j' : ary = ReadCoordinates( infile, &arx );
                                   break;
                        case 'H' : gy = ReadCoordinates( infile, &gx );
                                   break;
                        case 'R' : ReadEnclString( infile, reference, REFSIZE );
                                   break;
                        case 'T' : ReadEnclString( infile, treference,REFSIZE );                                   break;
                        case 'u' : ReadEnclString( infile, blocklist, REFSIZE );                                   break;
                        case 'P' : ReadEnclString( infile, synclist, REFSIZE );
                                   break;
                        case 's' : sy = sx = ReadNumber( infile );
                                   break;
                        case 'S' : sx = ReadNumber( infile );
                                   if ( TestSymbol( infile, '@' ) )
                                     {
                                       sy = ReadNumber( infile );
                                     } /* if */
                                   else
                                     {
                                       NNError( "missing '@'" );
                                     } /* else */
                                   break;
                        case 't' : thickness = ReadNumber( infile );
                                   break;
                        case 'r' : if ( refined )
                                     {
                                       NNError( "refined flag doubly defined" );
                                     } /* else */
                                   else
                                     {
                                       refined = True;
                                     } /* else */
                                   break;
                        case 'c' : color = ReadNumber( infile );
                                   break;
                        default  : V sprintf( name, "unknown token '%c'", ch );
                                   NNError( name );
                      } /* switch */
                  } /* else */
                ch = ReadCharComment( infile );
              } /* while */

            /* insert phantom transition node in the net */
            if ( autonumbering )
              {
                ident = autocounter;
              } /* if */
            else if ( ! specified )
              {
                NNError( "missing phantom transition identifier" );
              } /* if */
            if ( ident > AnzPhTrNamen )
              {
                AnzPhTrNamen = ident;
              } /* if */
            else if ( PhTrArray[ ident ] )
              {
                V sprintf( guard,
                     "phantom transition identifier %d multiply used", ident );
                NNError( guard );
              } /* else if */
            if ( AnzPhTrNamen >= MaxPhTrNamen )
              {
                PhTrArray = (Transition *) Realloc( (char *) PhTrArray,
                     ( MaxPhTrNamen += NAMES_OFFSET ) * sizeof( Transition ) );
                for ( count = MaxPhTrNamen - NAMES_OFFSET ;
                                               count < MaxPhTrNamen ; count++ )
                  {
                    PhTrArray[ count ] = (Transition) NULL;
                  } /* for */
              } /* if */
            if ( ! *name )
              {
                V sprintf( name, "TR%d", ident );
              } /* if */
            PhTrArray[ ident ] = (Transition) 1;
            NNinput_line++;
          } /* while */
      } /* if */

    /* Read in T->P arc nodes ************************************************/
    V ungetc( ch, infile );
    ReadCmdToken( infile, cmd, 512 );
    if ( strcmp( cmd, "TP" ) )
      {
        NNError( "keyword 'TP' expected" );
        return ( NULL );
      } /* if */
    ReadNewline( infile, 1 );

    while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'P' ) )
      {
        weight = 1;
        gx = gy = 0;
        invisible = 0;
        *guard = *name = *name2 = *meaning = '\0';
        num = num2 = 0;
        while ( ch != '\n' )
          {
            if ( isdigit( ch ) )
              {
                V ungetc( ch, infile );
                num = ReadNumber( infile );
                if ( TestSymbol( infile, '<' ) )
                  {
                    num2 = ReadNumber( infile );
                  } /* if */
                else
                  {
                    NNError( "'<' expected" );
                  } /* else */
              } /* if */ 
            else
              {
                switch ( ch )
                  {
                    case 'J' :
                    case ',' : my = ReadCoordinates( infile, &mx );
                               break;
                    case 'w' : weight = ReadNumber( infile );
                               break;
                    case 'n' : pry = ReadCoordinates( infile, &prx );
                               break;
                    case 'N' : py = ReadCoordinates( infile, &px );
                               break;
                    case 'p' : ReadEnclString( infile, guard, GUARDSIZE );
                               break;
                    case 'q' : gry = ReadCoordinates( infile, &grx );
                               break;
                    case 'Q' : gy = ReadCoordinates( infile, &gx );
                               break;
                    case 'v' : invisible = ReadNumber( infile );
                               break;
                    case 't' : thickness = ReadNumber( infile );
                               break;
                    case 'c' : color = ReadNumber( infile );
                               break;
                    default  : V sprintf( name, "unknown token '%c'", ch );
                               NNError( name );
                  } /* switch */
              } /* else */
            ch = ReadCharComment( infile );
          } /* while */
        if ( ! num || ( num > AnzTrNamen ) || ! TrArray[ num ] )
          {
            NNError( "arc: incorrect transition identifier" );
          } /* if */
        if ( ! num2 || ( num2 > AnzPlNamen ) || ! PlArray[ num2 ]  )
          {
            NNError( "arc: incorrect place identifier" );
          } /* if */
        tr_last = TrArray[ num ];
        pl_last = PlArray[ num2 ];
        fast_insert_TP_con( net, pl_last->name, tr_last->name, weight,
                                                &pl_last, &tr_last, &ts_last );
        if ( CM_storeCoordinates )
          {
            Coords co;
            ts_last->weight = (Cardinal32)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
            co->x = gx;
            co->y = gy;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( guard );
            co->y = (int) invisible;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) weight;
            co->next = NULL;
          } /* if */

        NNinput_line++;
      } /* while */

    /* Read in P->T arc nodes ************************************************/
    V ungetc( ch, infile );
    ReadCmdToken( infile, cmd, 512 );
    if ( strcmp( cmd, "PT" ) )
      {
        NNError( "keyword 'PT' expected" );
        return ( NULL );
      } /* if */
    ReadNewline( infile, 1 );

    while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'T' )
            && ( ch != 'P' ) && ! feof( infile ) )
      {
        weight = 1;
        gx = gy = 0;
        invisible = 0;
        *guard = *name = *name2 = *meaning = '\0';
        num = num2 = 0;
        while ( ch != '\n' )
          {
            if ( isdigit( ch ) )
              {
                V ungetc( ch, infile );
                num = ReadNumber( infile );
                if ( TestSymbol( infile, '>' ) )
                  {
                    num2 = ReadNumber( infile );
                  } /* if */
                else
                  {
                    NNError( "'>' expected" );
                  } /* else */
              } /* if */ 
            else
              {
                switch ( ch )
                  {
                    case 'J' :
                    case ',' : my = ReadCoordinates( infile, &mx );
                               break;
                    case 'w' : weight = ReadNumber( infile );
                               break;
                    case 'n' : pry = ReadCoordinates( infile, &prx );
                               break;
                    case 'N' : py = ReadCoordinates( infile, &px );
                               break;
                    case 'p' : ReadEnclString( infile, guard, GUARDSIZE );
                               break;
                    case 'q' : gry = ReadCoordinates( infile, &grx );
                               break;
                    case 'Q' : gy = ReadCoordinates( infile, &gx );
                               break;
                    case 'v' : invisible = ReadNumber( infile );
                               break;
                    case 't' : thickness = ReadNumber( infile );
                               break;
                    case 'c' : color = ReadNumber( infile );
                               break;
                    default  : V sprintf( name, "unknown token '%c'", ch );
                               NNError( name );
                  } /* switch */
              } /* else */
            ch = ReadCharComment( infile );
          } /* while */
        if ( ! num || ( num > AnzPlNamen ) || ! PlArray[ num ] )
          {
            NNError( "arc: incorrect place identifier" );
          } /* if */
        if ( ! num2 || ( num2 > AnzTrNamen ) || ! TrArray[ num2 ]  )
          {
            NNError( "arc: incorrect transition identifier" );
          } /* if */
        pl_last = PlArray[ num ];
        tr_last = TrArray[ num2 ];
        fast_insert_PT_con( net, pl_last->name, tr_last->name, weight,
                                                &pl_last, &tr_last, &st_last );
        if ( CM_storeCoordinates )
          {
            Coords co;
            st_last->weight = (Cardinal32)
                              ( co = (Coords) Malloc( sizeof( CoordEntry ) ) );
            co->x = gx;
            co->y = gy;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) Strdup( guard );
            co->y = (int) invisible;
            co = co->next = (Coords) Malloc( sizeof( CoordEntry ) );
            co->x = (int) weight;
            co->next = NULL;
          } /* if */

        NNinput_line++;
      } /* while */

    /* Read in phantom T->P arc nodes ****************************************/
    if ( ch == 'P' )
      {
        V ungetc( ch, infile );
        ReadCmdToken( infile, cmd, 512 );
        if ( strcmp( cmd, "PTP" ) )
          {
            NNError( "keyword 'PTP' expected" );
            return ( NULL );
          } /* if */
        ReadNewline( infile, 1 );

        while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'P' ) )
          {
            weight = 1;
            gx = gy = 0;
            invisible = 0;
            *guard = *name = *name2 = *meaning = '\0';
            num = num2 = 0;
            while ( ch != '\n' )
              {
                if ( isdigit( ch ) )
                  {
                    V ungetc( ch, infile );
                    num = ReadNumber( infile );
                    if ( TestSymbol( infile, '<' ) )
                      {
                        num2 = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        NNError( "'<' expected" );
                      } /* else */
                  } /* if */ 
                else
                  {
                    switch ( ch )
                      {
                        case 'J' :
                        case ',' : my = ReadCoordinates( infile, &mx );
                                   break;
                        case 'w' : weight = ReadNumber( infile );
                                   break;
                        case 'n' : pry = ReadCoordinates( infile, &prx );
                                   break;
                        case 'N' : py = ReadCoordinates( infile, &px );
                                   break;
                        case 'p' : ReadEnclString( infile, guard, GUARDSIZE );
                                   break;
                        case 'q' : gry = ReadCoordinates( infile, &grx );
                                   break;
                        case 'Q' : gy = ReadCoordinates( infile, &gx );
                                   break;
                        case 'v' : invisible = ReadNumber( infile );
                                   break;
                        case 't' : thickness = ReadNumber( infile );
                                   break;
                        case 'c' : color = ReadNumber( infile );
                                   break;
                        default  : V sprintf( name, "unknown token '%c'", ch );
                                   NNError( name );
                      } /* switch */
                  } /* else */
                ch = ReadCharComment( infile );
              } /* while */
            if ( ! num || ( num > AnzPhTrNamen ) || ! PhTrArray[ num ] )
              {
                NNError( "arc: incorrect phantom transition identifier" );
              } /* if */
            if ( ! num2 || ( num2 > AnzPlNamen ) || ! PlArray[ num2 ]  )
              {
                NNError( "arc: incorrect place identifier" );
              } /* if */

            NNinput_line++;
          } /* while */
      } /* if */

    /* Read in phantom P->T arc nodes ****************************************/
    if ( ch == 'P' )
      {
        V ungetc( ch, infile );
        ReadCmdToken( infile, cmd, 512 );
        if ( strcmp( cmd, "PPT" ) )
          {
            NNError( "keyword 'PPT' expected" );
            return ( NULL );
          } /* if */
        ReadNewline( infile, 1 );

        while ( ( ch = ReadCharComment( infile ) ) && ( ch != 'T' )
                && ! feof( infile ) )
          {
            weight = 1;
            gx = gy = 0;
            invisible = 0;
            *guard = *name = *name2 = *meaning = '\0';
            num = num2 = 0;
            while ( ch != '\n' )
              {
                if ( isdigit( ch ) )
                  {
                    V ungetc( ch, infile );
                    num = ReadNumber( infile );
                    if ( TestSymbol( infile, '>' ) )
                      {
                        num2 = ReadNumber( infile );
                      } /* if */
                    else
                      {
                        NNError( "'>' expected" );
                      } /* else */
                  } /* if */ 
                else
                  {
                    switch ( ch )
                      {
                        case 'J' :
                        case ',' : my = ReadCoordinates( infile, &mx );
                                   break;
                        case 'w' : weight = ReadNumber( infile );
                                   break;
                        case 'n' : pry = ReadCoordinates( infile, &prx );
                                   break;
                        case 'N' : py = ReadCoordinates( infile, &px );
                                   break;
                        case 'p' : ReadEnclString( infile, guard, GUARDSIZE );
                                   break;
                        case 'q' : gry = ReadCoordinates( infile, &grx );
                                   break;
                        case 'Q' : gy = ReadCoordinates( infile, &gx );
                                   break;
                        case 'v' : invisible = ReadNumber( infile );
                                   break;
                        case 't' : thickness = ReadNumber( infile );
                                   break;
                        case 'c' : color = ReadNumber( infile );
                               break;
                        default  : V sprintf( name, "unknown token '%c'", ch );
                                   NNError( name );
                      } /* switch */
                  } /* else */
                ch = ReadCharComment( infile );
              } /* while */
            if ( ! num || ( num > AnzPlNamen ) || ! PlArray[ num ] )
              {
                NNError( "arc: incorrect place identifier" );
              } /* if */
            if ( ! num2 || ( num2 > AnzPhTrNamen ) || ! PhTrArray[ num2 ]  )
              {
                NNError( "arc: incorrect phantom transition identifier" );
              } /* if */

            NNinput_line++;
          } /* while */
      } /* if */

    /* Read in additional text strings ***************************************/
    if ( ! feof( infile ) )
      {
        V ungetc( ch, infile );
        ReadCmdToken( infile, cmd, 512 );

        if ( ! strcmp( cmd, "TX" ) )
          {
            ReadNewline( infile, 1 );
            while ( ( ch = ReadCharComment( infile ) ) && ! feof( infile ) )
              {
                while ( ch != '\n' )
                  {
                    switch ( ch )
                      {
                        case '\'':
                        case '"' : V ungetc( ch, infile );
                                   ReadEnclString( infile, guard, NAMESIZE );
                                   break;
                        case 'N' : py = ReadCoordinates( infile, &px );
                                   break;
                        default  : V sprintf( name, "unknown token '%c'", ch );
                                   NNError( name );
                      } /* switch */
                    ch = ReadCharComment( infile );
                  } /* while */
                NNinput_line++;
              } /* while */
          } /* if */
      } /* if */

    Free( PlArray );
    Free( TrArray );
    Free( PhTrArray );

    fclose( infile );

    return( net );

  } /* read_new_PEP() */


/*****************************************************************************/
/***************************** write_new_PEP *********************************/
/*****************************************************************************/
/*                                                  09.01.1995 Stefan Roemer */
/*                              Letzte Aenderungen: 13.06.1996 Stefan Roemer */
/*****************************************************************************/

#   define f( STRING ) V fprintf( file, STRING )
#   define fx( STRING, X ) fprintf( file, STRING, X )
#   define fxy( STRING, X, Y ) fprintf( file, STRING, X, Y )
#   define fxyz( STRING, X, Y, Z ) V fprintf( file, STRING, X, Y, Z )
#   define fxyxy( STRING, X1, Y1, X2, Y2 ) V fprintf( file, STRING, X1,Y1,X2,Y2 )
#   define fxyxyz( STRING, X1,Y1,X2,Y2,Z ) V fprintf( file, STRING,X1,Y1,X2,Y2,Z )

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_new_PEP( LMPNet net, char * PEPfilename )

#else

void write_new_PEP( net, PEPfilename )

  LMPNet net;
  char   * PEPfilename;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st;
    register TSEdge     ts;
    Cardinal32          count,
                        x1,
                        nx;     /* Koordinaten Kopfknoten */
    Coords              co;
    char                * ptr;
    FILE                * file;

    if ( ! ( file = fopen( PEPfilename, "w" ) ) )
      {
        NNError( "Could not open file for writing" );
      } /* if */
    if ( ! CM_storeCoordinates )
      {
        NNError( "Internal Error: 'CM_storeCoordinates' not set" );
      } /* if */

    f( "PEP\n" );
    V fx( "%s\n", CM_nettype );
    f( "FORMAT_N\n" );
    f( "DPL s7n10@-9t2\n" );
    f( "DTR s7n10@-9t2\n" );
    f( "DPT w1t1\n" );

                      /* Liste der Stellen ausgeben */
    f( "PL\n" );
    for ( count = 0, pl = net->Places ; pl ; pl = pl->nextplace )
      {
        pl->numname = ++count;
        V fx( "\"%s\"", pl->name );
        co = (Coords) pl->inscript;
	if (! co) {f("\n"); continue;}
	if (co->x) fx("%u@", co->x); else f("0@");
	if (co->y) fx("%u", co->y); else f("0");
        if ( pl->label && ( ! strcmp( CM_nettype, "PetriBox" )
                                         || ! strcmp( CM_nettype, "HLNet" ) ) )
          {
            V fx( "%s", ( pl->label == ENTRY ? "e"
                                             : pl->label == EXIT ? "x" : "" ) );
          } /* if */
        else if ( pl->marks )
          {
            V fx( "M%u", pl->marks );

          } /* else if */
        if ( ( co = co->next ) && ( co = co->next ) )
          {
	    if ( (ptr=(char *) co->x) && *ptr ) V fx("b\"%s\"", ptr);
	    if ( (ptr=(char *) co->y)  && *ptr) V fx("R\"%s\"", ptr);
            if ( co = co->next )
              {
	        if ( (ptr=(char *) co->x)  && *ptr) V fx("Z\"%s\"", ptr);
	        if ( (ptr=(char *) co->y)  && *ptr) V fx("z\"%s\"", ptr);
                if ( co = co->next )
                  {
		    if (co->x) {
		       x1 = (Cardinal32) co->x;
		       V fxy("M%um%u", x1, x1);
		       }
                  } /* if */
              } /* if */
          } /* if */
        f( "\n" );
      } /* for ( pl ... ) */

		     /* Liste der Transitionen ausgeben */
    f( "TR\n" );
    for ( count = 0, tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        tr->numname = ++count;
        V fx("\"%s\"",tr->name);
        co = (Coords) tr->inscript;
	if (! co) {f("\n"); continue;}
	if (co->x) fx("%u@", co->x); else f("0@");
	if (co->y) fx("%u", co->y); else f("0");
        if ( tr->label && ( ! strcmp( CM_nettype, "PetriBox" )
                                         || ! strcmp( CM_nettype, "HLNet" ) ) )
          {
            V fx( "%s", "r" + ! ( tr->label && ( *tr->label == 1 ) ) );
          } /* if */
        if ( ( co = co->next ) && ( co = co->next ) )
          {
	    if ( (ptr=(char *) co->x) && *ptr) V fx("i\"%s\"", ptr);
	    if ( (ptr=(char *) co->y) && *ptr) V fx("b\"%s\"", ptr);
            if ( co = co->next )
              {
	        if ( (ptr=(char *) co->x) && *ptr) V fx("g\"%s\"", ptr);
	        if ( (ptr=(char *) co->y) && *ptr) V fx("R\"%s\"", ptr);
              } /* if */
          } /* if */
        f( "\n" );
      } /* for ( tr ... ) */

    f( "TP\n" );
    for ( tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        for ( ts = tr->postset ; ts ; ts = ts->nextplace )
          {
            pl = ts->place;
            fxy( "%u<%u", tr->numname, pl->numname );
            co = (Coords) ts->weight;
            if ( co && ( co = co->next ) )
              {
	        if ( (ptr=(char *) co->x) && *ptr) V fx("p\"%s\"", ptr);
	        if ( (nx=(Cardinal32) co->y) ) V fx("v%u", nx);
                if ( co = co->next )
                  {
		    if ( (nx=(Cardinal32)co->x) > 1) V fx("w%u",nx);
                  } /* if */
              } /* if */
            f( "\n" );
          } /* for */
      } /* for */

    f( "PT\n" );
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        for ( st = pl->postset ; st ; st = st->nexttrans )
          {
            tr = st->trans;
            fxy( "%u>%u", pl->numname, tr->numname );
            co = (Coords) st->weight;
            if ( co && ( co = co->next ) )
              {
	        if ( (ptr=(char *) co->x) && *ptr) V fx("p\"%s\"", ptr);
	        if ( (nx=(Cardinal32) co->y) ) V fx("v%u", nx);
                if ( co = co->next )
                  {
		    if ( (nx=(Cardinal32)co->x) > 1) V fx("w%u",nx);
                  } /* if */
              } /* if */
            f( "\n" );
          } /* for */
      } /* for */

    f( "TX\n" );
    f( "\n" );

    fclose( file );

  } /* write_new_PEP() */


/*****************************************************************************/
/***************************** write_new_Tobi ********************************/
/*****************************************************************************/
/*                                              Do.,19.01.1995 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_new_Tobi( LMPNet net, char * TobiFilename )

#else

void write_new_Tobi( net, TobiFilename )

  LMPNet net;
  char   * TobiFilename;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st;
    register TSEdge     ts;
    register Cardinal32 count;
    Coords              co;
    int                 x,
                        y;
    FILE                * file;

    if ( ! CM_storeCoordinates )
      {
        NNError( "Internal Error: 'CM_storeCoordinates' not set" );
      } /* if */

    if ( ! ( file = fopen( TobiFilename, "w" ) ) )
      {
        NNError( "Could not open file for writing" );
      } /* if */

    f( "Place\n" );
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        V fx( "Name %s ", pl->name );
        co = (Coords) pl->inscript;
        fxy( "Position %d %d\n", co->x, co->y );
        co = co->next;
        fxy( " %d %d", co->x, co->y );
        co = co->next;
        if ( *( (char *) co->x ) )
          {
            V fx( "\"%s\"", (char *) co->x );
          } /* if */
        else if ( ! strcmp( CM_nettype, "HLNet" ) )
          {
            V fx( "\"%c\"", "0ex"[ (int) pl->label ] );
          } /* else if */
        else
          {
            f( "\"\"" );
          } /* else */
        co = co->next;
        if ( *( (char *) co->x ) )
          {
            V fx( " \"%s\"\n", (char *) co->x );
          } /* if */
        else if ( ! strcmp( CM_nettype, "HLNet" ) )
          {
            f( " \"dot\"\n" );
          } /* else if */
        else
          {
            f( " \"\"\n" );
          } /* else */
      } /* for */

    f( "\nTransition\n" );
    for ( tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        V fx( "Name %s ", tr->name );
        co = (Coords) tr->inscript;
        fxy( "Position %d %d\n", co->x, co->y );
        co = co->next;
        fxy( " %d %d", co->x, co->y );
        co = co->next;
        V fx( "\"%s\"", (char *) co->x );
        co = co->next;
        if ( *( (char *) co->x ) )
          {
            V fx( " \"%s\"\n", (char *) co->x );
          } /* if */
        else if ( ! strcmp( CM_nettype, "HLNet" ) )
          {
            f( " \"true\"\n" );
          } /* else if */
        else
          {
            f( " \"\"\n" );
          } /* else */
      } /* for */

    f( "\nEdge\n" );
    count = 0;
    for ( tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        for ( ts = tr->postset ; ts ; ts = ts->nextplace )
          {
            V fx( "Name HLEdge%u ", ++count );
            V fx( "From %s ", ts->trans->name );
            V fx( "To %s\n", ts->place->name );
            co = (Coords) ts->weight;
            x = ( (Coords) ts->trans->inscript )->x;
            y = ( (Coords) ts->place->inscript )->x;
            x += ( y - x ) >> 1;
            V fx( " %d", x );
            x = ( (Coords) ts->trans->inscript )->y;
            y = ( (Coords) ts->place->inscript )->y;
            x += ( y - x ) >> 1;
            V fx( " %d", x );
            co = co->next;
            V fx( "\"%s\"\n", (char *) co->x );
          } /* for */
      } /* for */
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        for ( st = pl->postset ; st ; st = st->nexttrans )
          {
            V fx( "Name HLEdge%u ", ++count );
            V fx( "From %s ", st->place->name );
            V fx( "To %s\n", st->trans->name );
            co = (Coords) st->weight;
            x = ( (Coords) st->place->inscript )->x;
            y = ( (Coords) st->trans->inscript )->x;
            x += ( y - x ) >> 1;
            V fx( " %d", x );
            x = ( (Coords) st->place->inscript )->y;
            y = ( (Coords) st->trans->inscript )->y;
            x += ( y - x ) >> 1;
            V fx( " %d", x );
            co = co->next;
            V fx( "\"%s\"\n", (char *) co->x );
          } /* for */
      } /* for */

    f( "\nToken\n" );
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        co = ( (Coords) pl->inscript )->next->next;
        if ( co->y && *( (char *) co->y ) )
          {
            V fx( "Name %s ", pl->name );
            V fx( " \"%s\"\n", (char *) co->y );
          } /* if */
      } /* for */

    f( "\n" );

    fclose( file );

  } /* write_new_Tobi() */


/*****************************************************************************/
/* Erzeugen von Namen fuer die Elemente eines Netzes:                        */
/* Stellenknoten werden mit "Sn" bezeichnet, wobei 'n' eine laufende Nummer  */
/* ( 1..net->NumPlaces ) ist; die Transitionen werden entsprechend mit       */
/* "Tm", m = 1..net->NumTransitions, benannt.                                */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet generate_S_T_names( LMPNet net )

#else

LMPNet generate_S_T_names( net )

  LMPNet net;

#endif

  {

    register Cardinal32 counter;
    register Place      pl;
    register Transition tr;
    char                str[30];

    counter = 0;
    pl = net->Places;
    while ( pl )
      {
        if ( pl->name )
          {
            Free( pl->name );
          } /* if */
        V sprintf( str, "S%u", ++counter );
        pl->name = Strdup( str );
        pl = pl->nextplace;
      } /* while */

    counter = 0;
    tr = net->Transitions;
    while ( tr )
      {
        if ( tr->name )
          {
            Free( tr->name );
          } /* if */
        V sprintf( str, "T%u", ++counter );
        tr->name = Strdup( str );
        tr = tr->nexttrans;
      } /* while */

    return ( net );

  } /* generate_S_T_names() */


/*****************************************************************************/

static Cardinal32 CP_AnzTrNamen,
                  CP_AnzTrExtra,
                  CP_MaxTrExtra;
static char       ** CP_TrNamen,
                  ** CP_TrExtra;
static Place      CP_pllast;
static Transition CP_trlast,
                  * CP_TransArray = NULL,
                  * CP_TransExtraArray = NULL;


/*****************************************************************************/
/* Einlesen einer Zeichenkette aus einer Textdatei; fuehrende Leerzeichen    */
/* werden ueberlesen und nicht dem String zugeordnet.                        */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static char * Read_STG_Name( FILE * file, char * string, int maxlen,
                             bool * endofline )

#else

static char * Read_STG_Name( file, string, maxlen, endofline )

  FILE * file;
  char * string;
  int  maxlen;
  bool * endofline;

#endif

  {

    register char ch;
    register int  len = 0;
    char          * pos = NULL;

    while( ( ( ch = getc( file ) ) == ' ' ) || ( ch == '\n' ) || ( ch == '#' ) )
      {
        if ( ch == '\n' )
          {
            NNinput_line++;
          } /* if */
        else if ( ch == '#' )
          {
            while ( ( ch = getc( file ) ) != '\n' )
              skip;
            NNinput_line++;
          } /* else */
      } /* while */
    V ungetc( ch, file );
    while(    ( ( *string = ch = getc( file ) ) != ' ' ) && ( ch != '\n' )
           && ( ch != ',' ) && ( ch != '>' ) && ( ch != '}' ) && ( ch != '=' )
           && ( ch != EOF ) && ( len < maxlen ) )
      {
        if ( ! pos && ( ( ch == '+' ) || ( ch == '-' ) || ( ch == '~' ) ) )
          {
            pos = string;
          } /* if */
        string++;
        len++;
      } /* while */
    V( ( ( ch == '>' ) || ( ch == ',' ) || ( ch == '}' ) || ( ch == '=' ) )
                                                      && ungetc( ch, file ) );
    if ( len == maxlen )
      {
        NNError( "read_STG(): length of identifier too long" );
        return ( NULL );
      } /* if */
    *endofline = (bool) ( *string != ' ' );
    V ( ( *string == '\n' ) && NNinput_line++ );
    V ( pos || ( pos = string ) );
    *string = '\0';
    
    return ( pos );

  } /* Read_STG_Name() */


/*****************************************************************************/
/* Lesen eines Stellenknotens.                                               */
/*                                                             Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static bool Read_STG_Node( FILE   * file,
                           char   * string,
                           LMPNet net,
                           bool   * endofline )

#else

static bool Read_STG_Node( file, string, net, endofline )

  FILE   * file;
  char   * string;
  LMPNet net;
  bool   * endofline;

#endif

  {

    register char       ch,
                        ** nptr;
    register Cardinal32 count;
    char                * end;
    unsigned short int  index;
    Transition          * adr;

    ch = *( end = Read_STG_Name( file, string, NAMESIZE, endofline ) );
    *end = '\0';
    for ( nptr = CP_TrNamen, count = 0 ;
                 ( ++count <= CP_AnzTrNamen ) && strcmp( string, *++nptr ) ; )
      /* skip */;
    *end = ch;
    if ( count <= CP_AnzTrNamen )
      {
        if ( ( index = ch == '+'? 0 : ch == '-'? 1 : ch == '~'? 2 : 3 ) == 3 )
          {
            char msg[200];
            V sprintf( msg,
            "CP_read_STG(): transition name without trailing +/-/~ found (%s)",
                                                                      string );
printf( "%s\n", msg );
/*
            NNError( msg );
*/
          } /* if */
        if ( *( end + 1 ) == '/' )
          {
            for ( nptr = CP_TrExtra, count = 0 ;
                 ( ++count <= CP_AnzTrExtra ) && strcmp( string, *++nptr ) ; )
              /* skip */;
            if ( count <= CP_AnzTrExtra )
              {
                CP_trlast = CP_TransExtraArray[ count ];
              } /* if */
            else if ( ( CP_AnzTrExtra = count ) >= CP_MaxTrExtra )
              {
                CP_TrExtra = (char **) Realloc( (char *) CP_TrExtra,
                        ( CP_MaxTrExtra += NAMES_OFFSET ) * sizeof( char * ) );
                CP_TransExtraArray = (Transition *) Realloc( (char *)
                    CP_TransExtraArray, CP_MaxTrExtra * sizeof( Transition ) );
                CP_TrExtra[ CP_AnzTrExtra ] = Strdup( string );
                CP_TransExtraArray[ CP_AnzTrExtra ] =
                             fast_insert_transition( net, string, &CP_trlast );
              } /* else if */
            else
              {
                CP_TrExtra[ CP_AnzTrExtra ] = Strdup( string );
                CP_TransExtraArray[ CP_AnzTrExtra ] =
                             fast_insert_transition( net, string, &CP_trlast );
              } /* else */
          } /* if */
        else if ( ! ( CP_trlast = *( adr = CP_TransArray + count*3 + index ) ) )
          {
            *adr = fast_insert_transition( net, string, &CP_trlast );
          } /* else if */
        return ( True );
      } /* if */
    else
      {
        if ( ! ( net->Places && ( CP_pllast = find_P( net, string ) ) ) )
          {
            V fast_insert_place( net, string, 0, &CP_pllast );
          } /* if */
        return ( False );
      } /* else */

  } /* Read_STG_Node() */


/*****************************************************************************/
/***************************** read_STG **************************************/
/*****************************************************************************/
/* Neu eingefuegt                      18.08.1995 - 31.01.1996 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet read_STG( char * STGfilename )

#else

LMPNet read_STG( STGfilename )

  char * STGfilename;

#endif

  {

    Place               * PlArray,
                        pl;
    char                ** nptr;
    char                cmd[513],
                        name[NAMESIZE+1],
                        plname[NAMESIZE+1],
                        trname[NAMESIZE+1],
                        guard[GUARDSIZE+1];
    register FILE       * STGfile;
    register LMPNet     net;
    register Cardinal32 AnzPlNamen,
                        MaxPlNamen,
                        MaxTrNamen,
                        count;
    Transition          tr1,
                        tr2;
    STEdge              stlast = NULL;
    TSEdge              tslast = NULL;
    bool                endofline;
    Cardinal32          placecounter = 0;

    CP_trlast = NULL;
    CP_pllast = NULL;
    NNinput_file = STGfilename;
    NNinput_line = 1;
    PlArray = (Place *) Malloc( ( count = MaxPlNamen = MaxTrNamen
                           = CP_MaxTrExtra = NAMES_START ) * sizeof( Place ) );
    CP_TrNamen = (char **) Malloc( count * sizeof( char * ) );
    CP_TrExtra = (char **) Malloc( count * sizeof( char * ) );
    CP_TransExtraArray = (Transition *) Malloc( count * sizeof( Transition ) );
    AnzPlNamen = CP_AnzTrNamen = CP_AnzTrExtra = 0;
    while ( --count )
      {
        PlArray[ count ] = (Place) NULL;
      } /* while */

    /* Oeffne STG-Datei */
    if ( ! ( STGfile = fopen( STGfilename, "r" ) ) )
      {
        V fprintf( stderr,
               "ERROR: CM.read_STG(): could not open file '%s' for reading.\n",
                 STGfilename );
        exit ( 4 );
      } /* if */

    net = empty_net();
    net->NumComponents = 1;

    while ( True )
      {
        Read_STG_Name( STGfile, cmd, 512, &endofline );
        if ( *cmd != '.' )
          {
            V sprintf( guard,
                "read_STG(): command with prefix '.' expected, but '%s' found",
                                                                         cmd );
            NNError( guard );
            return ( NULL );
          } /* if */
        if ( ! strcmp( cmd, ".end" ) )
          {
            break;
          } /* if */
        else if ( ! strcmp( cmd, ".name" ) )
          {
            Read_STG_Name( STGfile, name, NAMESIZE, &endofline );
          } /* else if */
        else if ( ! strcmp( cmd, ".model" ) )
          {
            Read_STG_Name( STGfile, name, NAMESIZE, &endofline );
          } /* else if */
        else if ( ! (    strcmp( cmd, ".dummy" )
                      && strcmp( cmd, ".internal" )
                      && strcmp( cmd, ".inputs" )
                      && strcmp( cmd, ".outputs" ) ) )
          {
            do {
                 Read_STG_Name( STGfile, name, NAMESIZE, &endofline );
                 for ( nptr = CP_TrNamen, count = 0 ;
                    ( ++count <= CP_AnzTrNamen ) && strcmp( name, *++nptr ) ; )
                   /* skip */;
                 if ( count <= CP_AnzTrNamen )
                   {
                     V sprintf( guard, "transition '%s' multiply used", name );
                     NNError( guard );
                   } /* if */
                 if ( ( CP_AnzTrNamen = count ) >= MaxTrNamen )
                   {
                     CP_TrNamen = (char **) Realloc( (char *) CP_TrNamen,
                           ( MaxTrNamen += NAMES_OFFSET ) * sizeof( char * ) );
                   } /* if */
                 CP_TrNamen[ CP_AnzTrNamen ] = Strdup( name );
               } while ( ! endofline );
          } /* else if */
        else if ( ! strcmp( cmd, ".graph" ) )
          {
            CP_TransArray = (Transition *) Malloc( 3 * ( CP_AnzTrNamen + 1 )
                                                      * sizeof( Transition ) );
            while ( ! ( feof( STGfile ) || TestSymbol( STGfile, '.' ) ) )
              {
                if ( TestSymbol( STGfile, '\n' ) )
                  {
                    continue;
                  } /* if */
                if ( Read_STG_Node( STGfile, name, net, &endofline ) )
                  {
                    tr1 = CP_trlast;
                    strcpy( trname, name );
                    if ( endofline )
                      {
                        NNError( "CP_read_STG(): missing destination name" );
                      } /* if */
                    while ( ! endofline )
                      {
                        if ( Read_STG_Node( STGfile, name, net, &endofline ) )
                          {
                            tr2 = CP_trlast;
                            V sprintf( plname, "zzzplace%u",
                                                              ++placecounter );
                            CP_pllast = net->LastPlace;
                            V fast_insert_place( net, plname, 0,
                                                                  &CP_pllast );
                            CP_trlast = tr1;
                            V fast_insert_TP_con( net, plname, trname, 1, 
                                             &CP_pllast, &CP_trlast, &tslast );
                            CP_trlast = tr2;
                            V fast_insert_PT_con( net, plname, name, 1,
                                             &CP_pllast, &CP_trlast, &stlast );
                          } /* if */
                        else
                          {
                            CP_trlast = tr1;
                            V fast_insert_TP_con( net, name, trname, 1, 
                                             &CP_pllast, &CP_trlast, &tslast );
                          } /* else */
                      } /* while */;
                  } /* if */
                else
                  {
                    /* Place name found */
                    strcpy( plname, name );
                    if ( endofline )
                      {
                        NNError( "CP_read_STG(): missing destination name" );
                      } /* if */
                    while ( ! endofline )
                      {
                        if ( ! Read_STG_Node( STGfile, name, net, &endofline ) )
                          {
                            NNError( "CP_read_STG(): transition name expected" );
                          } /* if */
                        V fast_insert_PT_con( net, plname, name, 1,
                                             &CP_pllast, &CP_trlast, &stlast );
                      } /* while */;
                  } /* else */
              } /* while */
            V ( feof( STGfile ) || ungetc( '.', STGfile ) );
            Free( CP_TransArray );
          } /* else if */
        else if ( ! strcmp( cmd, ".marking" ) )
          {
            if ( ! TestSymbol( STGfile, '{' ) )
              {
                NNError( "read_STG(): '{' expected" );
              } /* if */
            while ( ! TestSymbol( STGfile, '}' ) )
              {
                if ( TestSymbol( STGfile, '<' ) )
                  {
                    if ( ! Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): transition name expected" );
                      } /* if */
                    tr1 = CP_trlast;
                    ReadSymbol( STGfile, ',' );
                    if ( ! Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): transition name expected" );
                      } /* if */
                    ReadSymbol( STGfile, '>' );
                    for ( tslast = tr1->postset ; tslast ;
                                                   tslast = tslast->nextplace )
                      {
                        if (    tslast->place
                             && ! tslast->place->preset->nexttrans
                             && tslast->place->postset
                             && ! tslast->place->postset->nexttrans
                             && tslast->place->postset->trans == CP_trlast )
                          {
                            tslast->place->marks++;
                            break;
                          } /* if */
                      } /* for */
                  } /* if */
                else
                  {
                    if ( Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): place name expected" );
                      } /* if */
                    CP_pllast->marks++;
                  } /* else */
              } /* while */
          } /* else if */
        else if ( ! strcmp( cmd, ".capacity" ) )
          {
            while ( ! TestSymbol( STGfile, '.' ) )
              {
                pl = NULL;
                if ( TestSymbol( STGfile, '<' ) )
                  {
                    if ( ! Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): transition name expected" );
                      } /* if */
                    tr1 = CP_trlast;
                    ReadSymbol( STGfile, ',' );
                    if ( ! Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): transition name expected" );
                      } /* if */
                    ReadSymbol( STGfile, '>' );
                    for ( tslast = tr1->postset ; tslast ;
                                                   tslast = tslast->nextplace )
                      {
                        if (    tslast->place
                             && ! tslast->place->preset->nexttrans
                             && tslast->place->postset
                             && ! tslast->place->postset->nexttrans
                             && tslast->place->postset->trans == CP_trlast )
                          {
                            pl = tslast->place;
                            break;
                          } /* if */
                      } /* for */
                  } /* if */
                else
                  {
                    if ( Read_STG_Node( STGfile, name, net, &endofline ) )
                      {
                        NNError( "CP_read_STG(): place name expected" );
                      } /* if */
                    pl = CP_pllast;
                  } /* else */
                ReadSymbol( STGfile, '=' );
                count = ReadNumber( STGfile );
                if ( pl )
                  {
/*
                    pl->capacity = count;
*/
                  } /* if */
              } /* while */
          } /* else if */
        else
          {
            V sprintf( name, "read_STG(): unknown keyword '%s'", cmd );
            NNError( name );
            return ( NULL );
          } /* else */
      } /* while */

    Free( PlArray );
    for ( nptr = CP_TrNamen, count = 0 ; ++count <= CP_AnzTrNamen ;
                                                              Free( *++nptr ) )
      /* skip */;
    Free( CP_TrNamen );
    for ( nptr = CP_TrExtra, count = 0 ; ++count <= CP_AnzTrExtra ;
                                                              Free( *++nptr ) )
      /* skip */;
    Free( CP_TrExtra );
    Free( CP_TransArray );
    Free( CP_TransExtraArray );

    fclose( STGfile );

    return( net );

  } /* read_STG() */


/*****************************************************************************/
/***************************** write_STG *************************************/
/*****************************************************************************/
/*                                                  18.08.1995 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_STG( LMPNet net, char * STGfilename )

#else

void write_STG( net, STGfilename )
  /* Gibt die mit net bezeichnete Matrix als Inzidenzmatrix nach 'file' aus  */

  LMPNet net;
  char   * STGfilename;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st;
    register TSEdge     ts;
    Cardinal32          count,
                        x1,
                        nx;
    Coords              co;
    char                * ptr;
    FILE                * file;

    if ( ! ( file = fopen( STGfilename, "w" ) ) )
      {
        NNError( "Could not open file for writing" );
      } /* if */
    if ( ! CM_storeCoordinates )
      {
        NNError( "Internal Error: 'CM_storeCoordinates' not set" );
      } /* if */

    f( "STG\n" );
    V fx( "%s\n", CM_nettype );
    f( "FORMAT_N\n" );
    f( "DPL s7n10@-9t2\n" );
    f( "DTR s7n10@-9t2\n" );
    f( "DPT w1t1\n" );

    f( "PL\n" );
    for ( count = 0, pl = net->Places ; pl ; pl = pl->nextplace )
      {
        pl->numname = ++count;
        V fx( "\"%s\"", pl->name );
        co = (Coords) pl->inscript;
        fxy( "%u@%u", co->x, co->y );
        if ( pl->label && ( ! strcmp( CM_nettype, "PetriBox" )
                                         || ! strcmp( CM_nettype, "HLNet" ) ) )
          {
            V fx( "%c", ( pl->label == ENTRY ? 'e'
                                           : pl->label == EXIT ? 'x' : ' ' ) );
          } /* if */
        else if ( pl->marks )
          {
            V fx( "M%u", pl->marks );
          } /* else if */
        if ( ( co = co->next ) && ( co = co->next ) )
          {
            V ( *( ptr = (char *) co->x ) && fx( "b\"%s\"", ptr ) );
            V ( *( ptr = (char *) co->y ) && fx( "R\"%s\"", ptr ) );
            if ( co = co->next )
              {
                V ( *( ptr = (char *) co->x ) && fx( "Z\"%s\"", ptr ) );
                V ( *( ptr = (char *) co->y ) && fx( "z\"%s\"", ptr ) );
                if ( co = co->next )
                  {
                    V ( ( x1 = (Cardinal32) co->x ) && fx( "m%u", x1 ) );
                  } /* if */
              } /* if */
          } /* if */
        f( "\n" );
      } /* for ( pl ... ) */

    f( "TR\n" );
    for ( count = 0, tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        tr->numname = ++count;
        V fx("\"%s\"",tr->name);
        co = (Coords) tr->inscript;
        fxy( "%u@%u", co->x, co->y );
        if ( tr->label && ( ! strcmp( CM_nettype, "PetriBox" )
                                         || ! strcmp( CM_nettype, "HLNet" ) ) )
          {
            V fx( "%s", "r" + ! ( tr->label && ( *tr->label == 1 ) ) );
          } /* if */
        if ( ( co = co->next ) && ( co = co->next ) )
          {
            V ( *( ptr = (char *) co->x ) && fx( "i\"%s\"", ptr ) );
            V ( *( ptr = (char *) co->y ) && fx( "b\"%s\"", ptr ) );
            if ( co = co->next )
              {
                V ( *( ptr = (char *) co->x ) && fx( "g\"%s\"", ptr ) );
                V ( *( ptr = (char *) co->y ) && fx( "R\"%s\"", ptr ) );
              } /* if */
          } /* if */
        f( "\n" );
      } /* for ( tr ... ) */
                                             /* Liste der TS-Kanten ausgeben */
    f( "TP\n" );
    for ( tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        for ( ts = tr->postset ; ts ; ts = ts->nextplace )
          {
            pl = ts->place;
            fxy( "%u<%u", tr->numname, pl->numname );
            co = (Coords) ts->weight;
            if ( co && ( co = co->next ) )
              {
                V ( *( ptr = (char *) co->x ) && fx( "p\"%s\"", ptr ) );
                V ( ( nx = (Cardinal32) co->y ) && fx( "v%u", nx ) );
                if ( co = co->next )
                  {
                    V ( ( ( nx = (Cardinal32) co->x )>1 ) && fx( "w%u", nx ) );
                  } /* if */
              } /* if */
            f( "\n" );
          } /* for */
      } /* for */

    f( "PT\n" );
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        for ( st = pl->postset ; st ; st = st->nexttrans )
          {
            tr = st->trans;
            fxy( "%u>%u", pl->numname, tr->numname );
            co = (Coords) st->weight;
            if ( co && ( co = co->next ) )
              {
                V ( *( ptr = (char *) co->x ) && fx( "p\"%s\"", ptr ) );
                V ( ( nx = (Cardinal32) co->y ) && fx( "v%u", nx ) );
                if ( co = co->next )
                  {
                    V ( ( ( nx = (Cardinal32) co->x )>1 ) && fx( "w%u", nx ) );
                  } /* if */
              } /* if */
            f( "\n" );
          } /* for */
      } /* for */

    f( "TX\n" );
    f( "\n" );

    fclose( file );

  } /* write_STG() */


/*****************************************************************************/
/* Einlesen eines Namens aus einer Textdatei.                                */
/*                                                             Stefan Roemer */
/*****************************************************************************/
 
#if defined(__STDC__) || defined(__cplusplus)
 
static unsigned char ReadFSAName( FILE * file, char * string, int maxlen )
 
#else
 
static unsigned char ReadFSAName( file, string, maxlen )
 
  FILE * file;
  char * string;
  int  maxlen;
 
#endif
 
  {
 
    register unsigned char len = 1;
 
    if ( isalnum( *string = ReadWhiteSpace( file ) ) )
      {
        while ( ( isalnum( *++string = getc( file ) ) || ( *string == '-' )
               || ( *string == '_' ) ) && ( (int) ++len < maxlen ) )
          skip;
        if ( len == maxlen )
          {
            NNError( "ReadFSAName(): identifier string size exceeded" );
          } /* if */
      } /* if */
    else if ( ! feof( file ) )
      {
        NNError( "ReadFSAName(): identifier expected" );
      } /* else if */
    V ungetc( *string, file );
    *string = '\0';
 
    return ( len );
 
  } /* ReadFSAName() */
 

/*****************************************************************************/
/* Lesen einer Zahl ( im Dezimalsystem ) aus einer Textdatei.                */
/*                                                             Stefan Roemer */
/*****************************************************************************/
 
#if defined(__STDC__) || defined(__cplusplus)
 
static Cardinal32 ReadFSANumber( FILE * file )
 
#else
 
static Cardinal32 ReadFSANumber( file )
 
  FILE * file;
 
#endif
 
  {
 
    register Cardinal32 number;
    register char       digit = '\0';
 
    while (    ! feof( file )
            && ( ( ( digit = getc( file ) ) == ' ' ) || ( digit == '\t' ) ) )
      skip;
    if ( isdigit( digit ) )
      {
        number = digit - '0';
        while ( isdigit( digit = getc( file ) ) )
          {
            number = number * 10 + digit - '0';
          } /* while */
        if ( digit != ' ' )
          {
            V ungetc( digit, file );
          } /* if */
        return ( number );
      } /* if */
    else if ( digit == '\n' )
      {
        NNinput_line++;
        return ( 0 );
      } /* else if */
    else
      {
        NNError( "ReadFSANumber(): digit expected" );
      } /* else */

    return ( 0 );
 
  } /* ReadFSANumber() */
 

/*****************************************************************************/
/***************************** read_FSA **************************************/
/*****************************************************************************/
/* Neu eingefuegt                      04.09.1995 - 04.09.1995 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet read_FSA( char * FSAfilename )

#else

LMPNet read_FSA( FSAfilename )

  char * FSAfilename;

#endif

  {

    Place               pl,
                        * PlArray;
    register Transition tr,
                        tr2;
    Transition          lasttr,
                        * trarray,
                        * trarrayptr;
    char                cmd[513],
                        name[NAMESIZE+1],
                        fsaname[NAMESIZE+1],
                        numericname[50];
    register FILE       * FSAfile;
    register LMPNet     net;
    LMPNet              * netarray,
                        * netarrayptr;
    char                ** netname,
                        ** identifier;
    register char       ** netnameptr,
                        ** idptr;
    register Cardinal32 AnzPlNamen,
                        MaxPlNamen,
                        MaxTrNamen,
                        NumNets,
                        MaxNets,
                        NumIdents,
                        MaxIdents,
                        id,
                        ident,
                        count,
                        count2,
                        comtype = 0,
                        * comlabel,
                        * comlabel2;
    Cardinal32          startstate;
    STEdge              stlast = NULL;
    TSEdge              tslast = NULL;

    CP_trlast = NULL;
    CP_pllast = NULL;
    NNinput_file = FSAfilename;
    NNinput_line = 1;
    PlArray = (Place *) Malloc( ( count = MaxPlNamen = MaxTrNamen
                                           = NAMES_START ) * sizeof( Place ) );
    CP_TrNamen = (char **) Malloc( count * sizeof( char * ) );
    AnzPlNamen = CP_AnzTrNamen = 0;
    while ( --count )
      {
        PlArray[ count ] = (Place) NULL;
      } /* while */

    netarrayptr = netarray = (LMPNet *) Malloc( ( MaxNets = 100 )
                                                           * sizeof( LMPNet ) );
    netname = netnameptr = (char **) Malloc( MaxNets * sizeof( char * ) );
    identifier = idptr = (char **) Malloc( ( MaxIdents = NAMES_START )
                                                           * sizeof( char * ) );
    NumNets = NumIdents = 0;

    if ( ! ( FSAfile = fopen( FSAfilename, "r" ) ) )
      {
        V fprintf( stderr,
                "ERROR: CM.read_FSA(): could not open file '%s' for reading.\n",
                 FSAfilename );
        exit ( 4 );
      } /* if */

    ReadFSAName( FSAfile, cmd, 512 );
    while ( ! feof( FSAfile ) )
      {
        if ( strcmp( cmd, "FSA" )  )
          {
            NNError( "CM.read_FSA(): keyword 'FSA' expected" );
            return ( NULL );
          } /* if */
        /* Initialisiere net */
        if ( ++NumNets >= MaxNets )
          {
            netarrayptr = &( ( netarray = (LMPNet *) Realloc( (char *) netarray,
                     ( MaxNets += 100 ) * sizeof( LMPNet ) ) )[ NumNets - 1 ] );
            netnameptr = &( ( netname = (char **) Realloc( (char *) netname,
                                MaxNets * sizeof( char * ) ) )[ NumNets - 1 ] );
          } /* if */
        *( netarrayptr++ ) = net = empty_net();
        net->NumComponents = 1;
        ReadFSAName( FSAfile, fsaname, NAMESIZE );
        *( netnameptr++ ) = Strdup( fsaname );
        ReadSymbol( FSAfile, ':' );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "STATES" )  )
          {
            NNError( "CM.read_FSA(): keyword 'STATES' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "TRANSITIONS" )  )
          {
            NNError( "CM.read_FSA(): keyword 'TRANSITIONS' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "START" )  )
          {
            NNError( "CM.read_FSA(): keyword 'START' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        startstate = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "ACCEPT" )  )
          {
            NNError( "CM.read_FSA(): keyword 'ACCEPT' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );
        ident = 0;

        ReadFSAName( FSAfile, cmd, 512 );
        while ( ! strcmp( cmd, "STATE" ) )
          {
            count = ReadFSANumber( FSAfile );
            V sprintf( numericname, "%05u%016u", NumNets, count );
            fast_insert_place( net, numericname, startstate == count,
                                                  &CP_pllast )->numname = count;
            ReadSymbol( FSAfile, ':' );
            ReadFSAName( FSAfile, cmd, 512 );
            while ( ! strcmp( cmd, "TRANS" )  )
              {
                ReadFSAName( FSAfile, cmd, 512 );
                ReadSymbol( FSAfile, '(' );
                ReadFSAName( FSAfile, name, NAMESIZE );
                count = 1;
                if ( ! strcmp( cmd, "call" ) )
                  {
                    if ( strcmp( name, fsaname ) )
                      {
                        NNError( "CM.read_FSA(): name of FSA expected" );
                      } /* if */
                    comtype = 1;
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        count++;
                        ReadSymbol( FSAfile, '.' );
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        count++;
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        if ( id == NumIdents )
                          {
                            if ( ++NumIdents >= MaxIdents )
                              {
                                idptr = &( ( identifier = (char **) Realloc(
                                               (char *) identifier, MaxIdents
                                     * sizeof( char * ) ) ) [ NumIdents - 1 ] );
                              } /* if */
                            *idptr = Strdup( name );
                          } /* if */
                        while ( TestSymbol( FSAfile, ';' ) )
                          {
                            ReadFSAName( FSAfile, name, NAMESIZE );
                            count++;
                            for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                              {
                                if ( ! strcmp( *idptr, name ) )
                                  {
                                    break;
                                  } /* if */
                              } /* for */
                            if ( id == NumIdents )
                              {
                                if ( ++NumIdents >= MaxIdents )
                                  {
                                    idptr = &( ( identifier = (char **) Realloc(
                                               (char *) identifier, MaxIdents
                                     * sizeof( char * ) ) ) [ NumIdents - 1 ] );
                                  } /* if */
                                *idptr = Strdup( name );
                              } /* if */
                          } /* while */
                      } /* if */
                  } /* if */
                else if ( ! strcmp( cmd, "accept" ) )
                  {
                    comtype = 2;
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        if ( strcmp( name, fsaname ) )
                          {
                            NNError( "CM.read_FSA(): name of FSA expected" );
                          } /* if */
                        count++;
                        ReadSymbol( FSAfile, '.' );
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        count++;
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        if ( id == NumIdents )
                          {
                            if ( ++NumIdents >= MaxIdents )
                              {
                                idptr = &( ( identifier = (char **) Realloc(
                                               (char *) identifier, MaxIdents
                                     * sizeof( char * ) ) ) [ NumIdents - 1 ] );
                              } /* if */
                            *idptr = Strdup( name );
                          } /* if */
                        while ( TestSymbol( FSAfile, ';' ) )
                          {
                            ReadFSAName( FSAfile, name, NAMESIZE );
                            count++;
                            for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                              {
                                if ( ! strcmp( *idptr, name ) )
                                  {
                                    break;
                                  } /* if */
                              } /* for */
                            if ( id == NumIdents )
                              {
                                if ( ++NumIdents >= MaxIdents )
                                  {
                                    idptr = &( ( identifier = (char **) Realloc(
                                               (char *) identifier, MaxIdents
                                     * sizeof( char * ) ) ) [ NumIdents - 1 ] );
                                  } /* if */
                                *idptr = Strdup( name );
                              } /* if */
                          } /* while */
                      } /* if */
                  } /* else if */
                else if ( ! (    strcmp( cmd, "internal" )
                              && strcmp( cmd, "exception" ) ) )
                  {
                    if ( strcmp( name, fsaname ) )
                      {
                        NNError( "CM.read_FSA(): name of FSA expected" );
                      } /* if */
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        count++;
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        if ( id == NumIdents )
                          {
                            if ( ++NumIdents >= MaxIdents )
                              {
                                idptr = &( ( identifier = (char **) Realloc(
                                               (char *) identifier, MaxIdents
                                     * sizeof( char * ) ) ) [ NumIdents - 1 ] );
                              } /* if */
                            *idptr = Strdup( name );
                          } /* if */
                      } /* if */
                    comtype = 0;
                  } /* else if */
                else if ( ! strcmp( cmd, "terminate" ) )
                  {
                    if ( strcmp( name, fsaname ) )
                      {
                        NNError( "CM.read_FSA(): name of FSA expected" );
                      } /* if */
                    comtype = 0;
                  } /* else if */
                else
                  {
                    V sprintf( name,
                                "CM.read_FSA(): unexpected keyword '%s' found",
                                                                          cmd );
                    NNError( name );
                  } /* else */
                V sprintf( numericname, "%05u%016u", NumNets, ++ident );
                fast_insert_transition( net, numericname, &CP_trlast )
                                                              ->numname = count;
                CP_trlast->label = (MultiSet) NumNets;
                comlabel = (Cardinal32 *) ( CP_trlast->inscript
                      = (char *) Malloc( (count + 2) * sizeof( Cardinal32 ) ) );
                *( comlabel + 1 ) = count;
                if ( *comlabel = comtype )
                  {
                    *( comlabel + comtype + 1 ) = NumNets - 1;
                  } /* if */
                else
                  {
                    *( comlabel + 2 ) = NumNets - 1;
                  } /* else */
                V fast_insert_PT_con( net, CP_pllast->name,
                          CP_trlast->name, 1, &CP_pllast, &CP_trlast, &stlast );
                ReadSymbol( FSAfile, ')' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '>' );
                count = ReadFSANumber( FSAfile );

                ReadFSAName( FSAfile, cmd, 512 );
              } /* while ( ! strcmp( cmd, "TRANS" )  ) */
          } /* while ( ! strcmp( cmd, "STATE" ) ) */

      } /* while ( ! feof( FSAfile ) ) */
    fclose( FSAfile );

    NNinput_line = 1;
    comtype = 0;
    /* Open FSA file */
    if ( ! ( FSAfile = fopen( FSAfilename, "r" ) ) )
      {
        V fprintf( stderr,
               "ERROR: CM.read_FSA(): could not open file '%s' for reading.\n",
                 FSAfilename );
        exit ( 4 );
      } /* if */

    netarrayptr = netarray;
    ReadFSAName( FSAfile, cmd, 512 );
    while ( ! feof( FSAfile ) )
      {
        /* Initialisiere net */
        tr = ( net = *( netarrayptr++ ) )->Transitions;
        comtype++;
        if ( strcmp( cmd, "FSA" )  )
          {
            NNError( "CM.read_FSA(): keyword 'FSA' expected" );
            return ( NULL );
          } /* if */
        ReadFSAName( FSAfile, name, NAMESIZE );
        ReadSymbol( FSAfile, ':' );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "STATES" )  )
          {
            NNError( "CM.read_FSA(): keyword 'STATES' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "TRANSITIONS" )  )
          {
            NNError( "CM.read_FSA(): keyword 'TRANSITIONS' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "START" )  )
          {
            NNError( "CM.read_FSA(): keyword 'START' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        if ( strcmp( cmd, "ACCEPT" )  )
          {
            NNError( "CM.read_FSA(): keyword 'ACCEPT' expected" );
            return ( NULL );
          } /* if */
        ReadSymbol( FSAfile, '=' );
        count = ReadFSANumber( FSAfile );

        ReadFSAName( FSAfile, cmd, 512 );
        while ( ! strcmp( cmd, "STATE" ) )
          {
            count = ReadFSANumber( FSAfile );
            ReadSymbol( FSAfile, ':' );

            ReadFSAName( FSAfile, cmd, 512 );
            while ( ! strcmp( cmd, "TRANS" )  )
              {
                ReadFSAName( FSAfile, cmd, 512 );
                ReadSymbol( FSAfile, '(' );
                ReadFSAName( FSAfile, name, NAMESIZE );
                if ( ! strcmp( cmd, "call" ) )
                  {
                    comlabel = (Cardinal32 *) tr->inscript;
                    comlabel++;
                    comlabel++;
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        for ( netnameptr = netname, count = 0 ;
                                       count < NumNets ; netnameptr++, count++ )
                          {
                            if ( ! strcmp( *netnameptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        *++comlabel = count;
                        ReadSymbol( FSAfile, '.' );
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        *++comlabel = id;
                        while ( TestSymbol( FSAfile, ';' ) )
                          {
                            ReadFSAName( FSAfile, name, NAMESIZE );
                            for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                              {
                                if ( ! strcmp( *idptr, name ) )
                                  {
                                    break;
                                  } /* if */
                              } /* for */
                            *++comlabel = id;
                          } /* while */
                      } /* if */
                  } /* if */
                else if ( ! strcmp( cmd, "accept" ) )
                  {
                    comlabel = (Cardinal32 *) tr->inscript;
                    comlabel++;
                    for ( netnameptr = netname, count = 0 ; count < NumNets ;
                                                         netnameptr++, count++ )
                      {
                        if ( ! strcmp( *netnameptr, name ) )
                          {
                            break;
                          } /* if */
                      } /* for */
                    *++comlabel = count;
                    comlabel++;
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        ReadSymbol( FSAfile, '.' );
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        *++comlabel = id;
                        while ( TestSymbol( FSAfile, ';' ) )
                          {
                            ReadFSAName( FSAfile, name, NAMESIZE );
                            for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                              {
                                if ( ! strcmp( *idptr, name ) )
                                  {
                                    break;
                                  } /* if */
                              } /* for */
                            *++comlabel = id;
                          } /* while */
                      } /* if */
                  } /* else if */
                else if ( ! (    strcmp( cmd, "internal" )
                              && strcmp( cmd, "exception" ) ) )
                  {
                    if ( TestSymbol( FSAfile, ';' ) )
                      {
                        comlabel = (Cardinal32 *) tr->inscript;
                        comlabel++;
                        comlabel++;
                        ReadFSAName( FSAfile, name, NAMESIZE );
                        for ( idptr = identifier, id = 0 ; id < NumIdents ;
                                                                 idptr++, id++ )
                          {
                            if ( ! strcmp( *idptr, name ) )
                              {
                                break;
                              } /* if */
                          } /* for */
                        *++comlabel = id;
                      } /* if */
                  } /* else if */
                else if ( ! strcmp( cmd, "terminate" ) )
                  {
                  } /* else if */
                else
                  {
                    V sprintf( name,
                                "CM.read_FSA(): unexpected keyword '%s' found",
                                                                          cmd );
                    NNError( name );
                  } /* else */
                ReadSymbol( FSAfile, ')' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '-' );
                ReadSymbol( FSAfile, '>' );
                count = ReadFSANumber( FSAfile );
                V sprintf( numericname, "%05u%016u", comtype, count );
                CP_pllast = NULL;
                V fast_insert_TP_con( net, numericname, tr->name, 1, 
                                              &CP_pllast, &CP_trlast, &tslast );
                tr = tr->nexttrans;
                ReadFSAName( FSAfile, cmd, 512 );
              } /* while ( ! strcmp( cmd, "TRANS" )  ) */
          } /* while ( ! strcmp( cmd, "STATE" ) ) */
      } /* while ( ! feof( FSAfile ) ) */
    fclose( FSAfile );

    ident = 0;
    *( trarrayptr = trarray = (Transition *) Malloc( ( NumNets + 1 )
                          * sizeof( Transition ) ) ) = (*netarray)->Transitions;
    for ( netarrayptr = netarray + 1, count = NumNets - 1 ; count-- ;
                                                                 netarrayptr++ )
      {
        *++trarrayptr = ( net = *netarrayptr )->Transitions;
        (*netarray)->NumPlaces += net->NumPlaces;
        (*netarray)->NumTransitions += net->NumTransitions;
        if ( ! (*netarray)->Places )
          {
            (*netarray)->Places = net->Places;
            (*netarray)->LastPlace = net->LastPlace;
          } /* if */
        else if ( (*netarray)->LastPlace->nextplace = net->Places )
          {
            net->Places->prevplace = (*netarray)->LastPlace;
            (*netarray)->LastPlace = net->LastPlace;
          } /* else if */
        if ( ! (*netarray)->Transitions )
          {
            (*netarray)->Transitions = net->Transitions;
            (*netarray)->LastTransition = net->LastTransition;
          } /* if */
        else if ( (*netarray)->LastTransition->nexttrans = net->Transitions )
          {
            net->Transitions->prevtrans = (*netarray)->LastTransition;
            (*netarray)->LastTransition = net->LastTransition;
          } /* else if */
        net->NumPlaces = net->NumTransitions = 0;
        net->Places = net->LastPlace = NULL;
        net->Transitions = net->LastTransition = NULL;
      } /* for */

    net = *netarray;
    lasttr = net->LastTransition;
    for ( tr = net->Transitions ; tr != lasttr ; tr = tr->nexttrans )
      {
        if (    *( comlabel = (Cardinal32 *) tr->inscript )
             && (    ( ( *comlabel == 1 ) && ( comlabel[2] < comlabel[3] ) )
                  || ( ( *comlabel == 2 ) && ( comlabel[2] > comlabel[3] ) ) ) )
          {
            count = comlabel[ 4 - *comlabel ];
            for ( tr2 = trarray[ count ] ; tr2 && ( tr2 != trarray[ count+1 ] ) ;
                                                          tr2 = tr2->nexttrans )
              {
                if (    *( comlabel2 = (Cardinal32 *) tr2->inscript )
                     && ( *comlabel != *comlabel2 ) )
                  {
                    if ( ( count2 = comlabel[1] ) == comlabel2[1] )
                      {
                        while ( count2 && ( comlabel[ count2 + 1 ]
                                                   == comlabel2[ count2 + 1 ] ) )
                          {
                            count2--;
                          } /* while */
                        if ( ! count2 )
                          {
                            V sprintf( numericname, "%05u%016u", NumNets + 1,
                                                                      ++ident );
                            CP_trlast = NULL;
                            fast_insert_transition( net, numericname,
                                                                   &CP_trlast );
                            if ( ident == 1 )
                              {
                                trarray[ NumNets ] = CP_trlast;
                              } /* if */
                            CP_pllast = NULL;
                            CP_trlast = NULL;
                            V fast_insert_PT_con( net,
                                      tr->preset->place->name, numericname, 1,
                                              &CP_pllast, &CP_trlast, &stlast );
                            CP_pllast = NULL;
                            CP_trlast = NULL;
                            V fast_insert_PT_con( net,
                                     tr2->preset->place->name, numericname, 1,
                                              &CP_pllast, &CP_trlast, &stlast );
                            CP_pllast = NULL;
                            CP_trlast = NULL;
                            V fast_insert_TP_con( net, 
                                     tr->postset->place->name, numericname, 1,
                                              &CP_pllast, &CP_trlast, &tslast );
                            CP_pllast = NULL;
                            CP_trlast = NULL;
                            V fast_insert_TP_con( net, 
                                    tr2->postset->place->name, numericname, 1,
                                              &CP_pllast, &CP_trlast, &tslast );
                          } /* if */
                      }
                  } /* if */
              } /* for ( tr2 ... ) */
          } /* if */
      } /* for ( tr ... ) */

    lasttr = trarray[ NumNets ];
    for ( tr = net->Transitions ; tr != lasttr ; skip )
      {
        if ( *( (Cardinal32 *) tr->inscript ) )
          {
            Free( tr->inscript );
            tr->inscript = NULL;
            tr = delete_transition_direct( net, tr );
          } /* if */
        else
          {
            tr = tr->nexttrans;
          } /* else */
      } /* for ( tr ... ) */
    for ( pl = net->Places, count = 0 ; pl ; pl = pl->nextplace )
      {
        pl->numname = ++count;
      } /* for */
    for ( tr = net->Transitions, count = 0 ; tr ; tr = tr->nexttrans )
      {
        tr->numname = ++count;
      } /* for */

    Free( *netname );
    for ( netarrayptr = netarray + 1, netnameptr = netname, count = NumNets ;
                      --count ; Free( *netnameptr ), remove_net( netarrayptr ) )
      skip;
    Free( netarray );
    Free( netname );
    for ( idptr = identifier, count = NumIdents ; count-- ; Free( *idptr ) )
      skip;
    Free( identifier );
    Free( trarray );

    return( net );

  } /* read_FSA() */


/*****************************************************************************/
/***************************** write_FSA *************************************/
/*****************************************************************************/
/*                                                  18.08.1995 Stefan Roemer */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_FSA( LMPNet net, char * FSAfilename )

#else

void write_FSA( net, FSAfilename )

  LMPNet net;
  char   * FSAfilename;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st;
    register TSEdge     ts;
    Cardinal32          count;
    FILE                * file;

    if ( ! ( file = fopen( FSAfilename, "w" ) ) )
      {
        NNError( "Could not open file for writing" );
      } /* if */

    f( "PEP\n" );
    f( "PTNet\n" );
    f( "FORMAT_N\n" );
    f( "DPL s7n10@-9t2\n" );
    f( "DTR s7n10@-9t2\n" );
    f( "DPT w1t1\n" );

    f( "PL\n" );
    for ( count = 0, pl = net->Places ; pl ; pl = pl->nextplace )
      {
        pl->numname = ++count;
        V fx( "\"%s\"", pl->name );
        fxy( "%u@%u", 20 + count * 10, 20 + count * 10 );
        if ( pl->marks )
          {
            V fx( "M%u", pl->marks );
          } /* else if */
        f( "\n" );
      } /* for ( pl ... ) */
    f( "TR\n" );

    for ( count = 0, tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        tr->numname = ++count;
        V fx("\"%s\"",tr->name);
        fxy( "%u@%u", 160 + count * 10, 20 + count * 10 );
        f( "\n" );
      } /* for ( tr ... ) */

    f( "TP\n" );
    for ( tr = net->Transitions ; tr ; tr = tr->nexttrans )
      {
        for ( ts = tr->postset ; ts ; ts = ts->nextplace )
          {
            pl = ts->place;
            fxy( "%u<%u\n", tr->numname, pl->numname );
          } /* for */
      } /* for */

    f( "PT\n" );
    for ( pl = net->Places ; pl ; pl = pl->nextplace )
      {
        for ( st = pl->postset ; st ; st = st->nexttrans )
          {
            tr = st->trans;
            fxy( "%u>%u\n", pl->numname, tr->numname );
          } /* for */
      } /* for */

    f( "TX\n" );
    f( "\n" );

    fclose( file );

  } /* write_FSA() */


/*****************************************************************************/
/* END OF FILE convertPEP.c                                                  */
/*****************************************************************************/

