/**********************************/
/* function.c                     */
/* Burkhard Graves, 06.10.93      */
/* Lars Jenner, Juni 1994         */
/**********************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if defined (sun) || defined (linux)
#include <malloc.h>
#else
#include <alloc.h>
#endif
#ifdef __cplusplus
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include "function.h"
#include "main.h"


/**********************************/
/* Datenstrukturen                */
/**********************************/

const char *MsgType[]={"Info","Warnung","Fehler","Fatal"};


/*****************************************************************************/
/* Diese Funktion wird bei Auftreten eines Fehlers aufgerufen: sie gibt die  */
/* Fehlermeldung `msg` auf dem Bildschirm aus und bricht das Programm ab.    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void Error( char * msg )

#else

void Error( msg )

  char * msg;

#endif

  {

    fprintf( stderr, "\n%s\nProgram aborted.\n", msg );
    exit ( 1 );

  } /* Error */


/**********************************/
/* Implementierung                */
/**********************************/

#ifdef __cplusplus

void Msg( char Type, ... )
{
   va_list argptr;
   char *Fmt;

   va_start(argptr,Type);
   Fmt = va_arg(argptr,char*);
   vfprintf(stderr,Fmt, argptr);
   va_end(argptr);
   fprintf(stderr,"\n");
}

#else
#if defined (sun) || defined (linux)

void Msg(va_alist)
va_dcl
{
   va_list argptr;
   char Type;
   char *Fmt;

   va_start(argptr);
   Type = va_arg(argptr,char);
   Fmt = va_arg(argptr,char*);
   vfprintf(stderr,Fmt, argptr);
   va_end(argptr);
   fprintf(stderr,"\n");
}

#else

void Msg( ... )
{
   va_list argptr;
   char Type;
   char *Fmt;

   va_start(argptr);
   Type = va_arg(argptr,char); 
   Fmt  = va_arg(argptr,char*);
   vfprintf(stderr,Fmt, argptr);
   va_end(argptr);
   fprintf(stderr,"\n");
}

#endif
#endif


#if defined(__STDC__) || defined(__cplusplus)

void * MYmalloc( size_t size )

#else

void * MYmalloc( size )

  size_t size;

#endif

  {

    void * m;

    if ( ! ( m = calloc( size, 1 ) ) )
      {
	Msg(_F,"Speicheranforderung abgewiesen !");
      }

    return ( m );

  } /* MYmalloc() */


#if defined(__STDC__) || defined(__cplusplus)

void * MYrealloc( void * ptr, size_t size )

#else

void * MYrealloc( ptr, size )

  void   * ptr;
  size_t size;

#endif

  {

    void * r;

    if ( ! ( r = realloc( (char *) ptr, size) ) )
      {
        Msg(_F,"Speicheranforderung abgewiesen !");
      }

    return ( r );

  } /* MYrealloc() */


#if defined(__STDC__) || defined(__cplusplus)

char * MYstrdup( char * string )

#else

char * MYstrdup( string )

  char * string;

#endif

  {

    char * str;

    if ( ! ( str = strdup( string ) ) )
      {
	Msg(_F,"Speicheranforderung abgewiesen !");
      }

    return ( str );

  } /* MYstrdup() */


#if defined(__STDC__) || defined(__cplusplus)

int MYfree( void * ptr )

#else

int MYfree( ptr )

  void * ptr;

#endif

  {

    free( ptr );
    return ( 1 );

  } /* MYstrdup() */


/**********************************/
/* END OF FILE function.c         */
/**********************************/
