/******************************************************/
/* graphalgorithms.c                                  */
/* Frank Wallner, Feb. 98                             */
/******************************************************/

#include "qq_header.h"

/*************/
/* FUNCTIONS */
/*************/

/***************************************************************************/
void b_search(int v, char mode) 
{
        int     ii, x, w, start, stop;
	Edge    *nedge; 
	char    tabulator=0, reachable_cycl=0, onst;

	b_push(v); INFO(v) |= ( VISITED | ONSTACK );
	B_lowlink(v) = B_dfn(v) = b_count++; 
	nedge = B_adjacents(v); 
	while ( nedge ) { 
	      w = nedge->adj; 
	      if ( ! IS(VISITED, w) ) { 
		     b_search(w, mode); 
		     B_lowlink(v) = MIN(B_lowlink(v), B_lowlink(w)); } 
	      else { if ( (B_dfn(w) < B_dfn(v)) && IS_ON_STACK(w) ) 
			   B_lowlink(v) = MIN(B_dfn(w),B_lowlink(v)); } 
	      nedge = nedge->next; } 
	if ( B_lowlink(v) != B_dfn(v) ) { return; }
	stop = b_stackpointer; 
	onst = IS_ON_STACK(actual_buechi);
	reachable_cycl = 0;
	do { x = b_pop;  INFO(x) &= ~ONSTACK;
	     reachable_cycl = (reachable_cycl || ( IS(FINAL,x) && onst) );
	   } while ( v != x ); 
	if ( ! reachable_cycl ) { return; }
	start = b_stackpointer; 
	if ( (stop-start == 1) && ( ! IS(SELFLOOP,v) ) ) return;
	accepting_component = final_result = 1; 
	if (aut_acc_ctr++ > MAX2_COUNTEREXAMPLES) return;
	if (! good_counterexample) {
	        printf("\n\nNO.\n\nAccepting Cycle (Automaton) :\n");
		alreadyNO = 1;
	        for (ii = 0; ii < start; ii++) 
	             printf("_c_P%d -> ", (b_stack[ii] + 1));
	        printf("(  ");
	        for (ii=start; ii < stop; ii++) {
	             printf("_c_P%d -> ", (b_stack[ii] + 1));
                     tabulator++; }
                printf("\b\b\b )\n");
		}
	fprintf(fres,"Accepting Cycle (Automaton) :\n");
	for (ii = 0; ii < start; ii++) 
	     fprintf(fres,"_c_P%d -> ", (b_stack[ii] + 1));
	fprintf(fres,"( ");
	for (ii=start; ii < stop; ii++) {
	     fprintf(fres,"_c_P%d -> ", (b_stack[ii] + 1));
	     if (tabulator > 8) {fprintf(fres,"\n"); tabulator=0;}
             tabulator++; }
}
/***************************************************************************/
void Check_Buechi_Circle(int lastbuechi, ConfS conf, char mode)
{
	int b, i=0;
  
	if (lastbuechi == 0) b = initial_state;
	else                 b = (int) postplaces(lastbuechi);
	actual_buechi = --b; /* state i is in buechi_nodes[i-1] !! */
	Evaluate(conf);
	ep = (Event *) buechi_nodes;
	InstantiateAutomaton();
	b_stackpointer = accepting_component = 0;
	b_count = 1;
	while ( i < _PBuechi_ ) { INFO(i) &= ~(VISITED);
				  INFO(i) &= ~(ONSTACK);
				  B_dfn(i) = 0;
				  B_lowlink(i) = 0;
				  i++; }
	b_search(b, mode);
	ep = events;
}
/***************************************************************************/
ENode *Grab_ENode(int corr, int rep)
/* returns the ENode corr-->rep if it exists already                  */
/* else creates a new ENode with entry rep and returns this new ENode */
{
	ENode *q, *q2, *r;

	q2 = (ENode *) NULL;

        if ( ( (q = (ENode *) lastENode(corr)) == (ENode *) NULL ) &&
	     ( enodes(corr) == 0 )  ) {
	     r = (ENode *) MyMalloc(sizeof(ENode)); 
	     r->idx = EN_idx++;
	     r->src = (- corr);
	     r->rep = rep;
	     r->cutoff = 0; 
	     r->next = (ENode *) NULL;
	     enodes(corr) = lastENode(corr) = (int) r;
	     return(r); } 
	    
        if ( ( ! q) || ( q->rep > rep ) ) q = (ENode *) enodes(corr); 
	while ( q && (q->rep < rep) ) { q2=q; q=q->next; continue; }
	if (q && (q->rep == rep) ) { lastENode(corr) = (int) q; 
				     return( q ); }

	r = (ENode *) MyMalloc(sizeof(ENode)); 
	r->idx = EN_idx++;
	r->src = (- corr);
	r->rep = rep;
	r->cutoff = 0; 
	r->next = q;
	if (q2 != (ENode *) NULL) q2->next = r;
	else enodes(corr) = (int) r;
	lastENode(corr) = (int) r;
	return(r);
}
/***************************************************************************/
ENode *ExtGrab_ENode(ENode *enode, int rep)
/* returns the ENode enode-->cutoff if it exists already                  */
/* else creates a new ENode with entry rep and returns this new ENode */
/* enode is NOT LOCAL ! */
{
	ENode *q, *q2, *r;

	q2 = (ENode *) NULL;
        if ( (q = (enode->corr->co_en)) == (ENode *) NULL ) {
	     r = (ENode *) MyMalloc(sizeof(ENode)); 
	     r->idx = EN_idx++;
	     r->src = (int) enode;
	     r->rep = rep;
	     r->cutoff = 0; 
	     r->next = (ENode *) NULL;
	     enode->corr->co_en = r; 
	     return(r); } 

	while ( q && (q->rep < rep) ) { q2=q; q=q->next; continue; }
	if (q && (q->rep == rep) ) { return( q ); }

	r = (ENode *) MyMalloc(sizeof(ENode)); 
	r->idx = EN_idx++;
	r->src = (int) enode;
	r->rep = rep;
	r->cutoff = 0; 
	r->next = q;
	if (q2 != (ENode *) NULL) q2->next = r;
	else { enode->corr->co_en = r; }
	return(r);
}
/***************************************************************************/
Edge *Grab_Edge(int v, int e)
/* returns the edge v-->e if it exists already                    */
/* else creates a new edge with entry e and returns this new edge */
/* both v and e are Representative events (e may be simple)       */
{
	Edge *q, *q2, *r;

	q2 = (Edge *) NULL;
        if ( ( (q = (Edge *) lastEdge(v)) == (Edge *) NULL ) &&
	     ( adjacents(v) == (Edge *) NULL)  ) {
	     r = (Edge *) MyMalloc(sizeof(Edge)); 
	     r->adj = e; r->final = r->n = 0; r->next = (Edge *) NULL;
	     adjacents(v) = r;
	     lastEdge(v) = (int) r;
	     return(r); } 
	    
        if ( ( ! q ) || (q->adj > e) ) q = adjacents(v); 
	while ( q != NULL ) {
	       if (q->adj < e) { q2=q; q=q->next; continue; }
	       if (q->adj > e) { break; }
	       /** else: (q->adj == e) **/
               lastEdge(v) = (int) q; return(q); }

	r = (Edge *) MyMalloc(sizeof(Edge)); 
	r->adj = e; r->final = r->n = 0; r->next = q; 
        lastEdge(v) = (int) r;
	if (q2 != (Edge *) NULL) q2->next = r;
	else adjacents(v) = r;
	return(r);
}
/***************************************************************************/
Edge *S_Grab_Edge(int v, int e, Edge **le)
/* returns the edge v-->e if it exists already                    */
/* else creates a new edge with entry e and returns this new edge */
/* v is a "simple" event, and e a Representative/Simple event     */
/* *le is the last included Edge 			          */
{
	Edge *q, *q2, *r;

	q2 = (Edge *) NULL;
        if ( ( (q = (*le)) == (Edge *) NULL ) &&
	     ( __adjacents(v) == (Edge *) NULL )  ) {
	     r = (Edge *) MyMalloc(sizeof(Edge)); 
	     r->adj = e; r->final = r->n = 0; r->next = (Edge *) NULL;
	     __adjacents(v) = r; (*le) = r; return(r); } 
	    
        if ( ( ! q) || ( q->adj > e ) ) q = __adjacents(v); 
	while ( q && (q->adj < e) )  { q2 = q; q = q->next; }
        if ( q && (q->adj == e) ) { (*le) = q; return(q); }

	r = (Edge *) MyMalloc(sizeof(Edge)); 
	r->adj = e; r->final = r->n = 0; r->next = q; (*le) = r;
	if ( q2 ) q2->next = r; 
	else __adjacents(v) = r;
	return(r);
}
/***************************************************************************/
Edge *EN_Grab_Edge(ENCorr *v, int e, Edge **le)
/* returns the edge v-->e if it exists already                    */
/* else creates a new edge with entry e and returns this new edge */
/* v is a "simple" event, and e a Representative/Simple event     */
{
	Edge *q, *q2, *r;

	q2 = (Edge *) NULL;
        if ( ( (q = (* le)) == (Edge *) NULL ) &&
	     ( v->adjs == (Edge *) NULL )   ) {
	     r = (Edge *) MyMalloc(sizeof(Edge)); 
	     r->adj = e; r->final = r->n = 0; r->next = (Edge *) NULL;
	     v->adjs = r; (*le) = r; return(r); } 
	    
	if ( ( ! q) || ( q->adj > e ) ) q = v->adjs;
	while ( q && (q->adj < e) )  { q2 = q; q = q->next; }
        if ( q && (q->adj == e) ) { (*le) = q; return(q); }

	r = (Edge *) MyMalloc(sizeof(Edge)); 
	r->adj = e; r->final = r->n = 0; r->next = q; (*le) = r;
	if ( q2 ) q2->next = r; 
	else v->adjs = r;
	return(r);
}
/***************************************************************************/
void final_search(int v)
{
	int 	i, w, j=0,
		start, stop, from, to;
        char	tabulator=1;
	Edge 	*nedge;
	int     *cost, *path;

	push(v); INFO(v) |= ( VISITED | ONSTACK );
      	lowlink(v)  = count; dfn(v) = count++;
	scc(v) = 0;
	nedge = adjacents(v);
	while ( nedge ) {
	       w = nedge->adj;
	       /* if ( w == v ) { nedge = nedge->next; continue; } */
	       if ( ! IS(VISITED, w) ) { 
		   if ( F__1(nedge->final) ) acc_succ(v) = w;
		   final_search(w);
		   lowlink(v) = MIN( lowlink(v), lowlink(w) ); }
	       else { if ( (dfn(w) < dfn(v) ) && IS_ON_STACK(w) )
		            lowlink(v) = MIN( dfn(w), lowlink(v) ); }
	       nedge = nedge->next; }
	if ( lowlink(v) != dfn(v) ) return;
	stop = stackpointer;
	do { i = pop; INFO(i) &= ~ONSTACK; 
             scc(i) = comp_ctr;
           } while ( v != i );
        comp_ctr++;
	start = stackpointer;
        if ( (stop-start) == 1 ) return; 
	if ( ! Is_AcceptingSCC(start,stop, &from, &to) ) return;
        if ( ! alreadyNO ) { printf("\n\nNO. (6)\n\n"); alreadyNO=1; }
	accepting_component = final_result = 1; acc_ctr++;
        if ( extended_output ) {
                   for (i=start; i<stop; i++, tabulator++) {
			fprintf(fres,"  %5d(",stack[i]);
			fprintf(fres,"%5d)   ", correspond(stack[i]) );
                        if (tabulator>3) {fprintf(fres,"\n"); tabulator=0; }
			}
                   if (tabulator != 1) fprintf(fres,"\n");
		   fprintf(fres,"--------------------------------------\n");
		   fprintf(fres,"Counterexample:\n"); 
		   }  /*extended_output*/
        else { fprintf(fres,"\n"); }
	cost  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	path  = (int *) MyMalloc( sizeof(int) * (stop-start) );
        j = 0;
	for (i=start; i<stop; i++) { cost[j] = INFTY; scc_idx(stack[i]) = j++; }
        cost[scc_idx(to)] = 0;
        shortest_path(cost, path, from, start, stop);
	PrintTransSeq(fres, path, from, to);
	fprintf(fres,"\n======================================\n");
	if ( ! all_counterexamples ) {
	     printf("Counterexample:\n");
	     PrintTransSeq(stdout, path, from, to);
	     all_time_f = nc_stopclock(all_time); 
	     if ( show_statistics ) { 
	          ffcomp1 = comp_ctr - 1; 
		  ffacc1 = acc_ctr; 
		  ShowStatistics(); }
	     fclose(fres); 
	     if ( no_output) { EXIT_NO_OUTPUT; } 
	     exit(0); }
        free(cost); free(path); 
}
/***************************************************************************/
void non_final_search(int v)
{
	int 	i, w, j = 0,
		start, stop, from, to;
	Edge 	*nedge;
	int     *cost, *path;
	char    tabulator=1;

	push(v); INFO(v) |= ( VISITED | ONSTACK );
      	lowlink(v) = count; dfn(v) = count++;
	nedge = adjacents(v);
	while ( nedge ) {
	   	if ( F__0(nedge->final) ) w = nedge->adj;
		else { nedge=nedge->next; continue; }
		/* if (w==v) { nedge=nedge->next; continue;} */
		if ( ! IS(VISITED,w) ) { 
			non_final_search(w);
			lowlink(v) = MIN(lowlink(v), lowlink(w)); }
		else { if ( ( dfn(w) < dfn(v) ) && IS_ON_STACK(w) )
			      lowlink(v) = MIN( dfn(w) , lowlink(v) ); }
		nedge = nedge->next; }
	if ( lowlink(v) != dfn(v) ) return;
	stop = stackpointer;
	do { i = pop; INFO(i) &= ~ONSTACK; scc(i) = comp_ctr;
	   } while (v != i);
        comp_ctr++;
	start = stackpointer;
	if ( (stop-start) == 1 )  return;
        from = to = (-1);
        for ( i=start; i<stop; i++ ) {
			j = maxBuchi(stack[i]);
			accepting_component = 0;
			Check_Buechi_Circle(j, config(stack[i]), 7);
                        if (accepting_component) { from = stack[i]; break; }
			}
        if (from == (-1) ) { return; }
        nedge = adjacents(from);
        while ( nedge ) {
                if (nedge->adj == from) { nedge=nedge->next; continue; }
		if ( (scc(nedge->adj) == scc(from) ) &&  
		     (F__0(nedge->final))  ) { to = nedge->adj; break; }
                nedge = nedge->next; } 
	fprintf(fres,"\nSCC %d - NV-Counterexample:\n", component);
   	if ( extended_output ) {
		   for (i=start; i< stop; i++, tabulator++){
			fprintf(fres,"  %5d(",stack[i]);
			fprintf(fres,"%5d)   ",correspond(stack[i]));
                        if (tabulator>3) { fprintf(fres,"\n"); tabulator=0; }
			}
                   if (tabulator != 1) fprintf(fres,"\n");
		   fprintf(fres,"-----------------------------------------\n");
		   fprintf(fres,"Circle without Buechi-Transitions:\n"); 
		   } /*extended_output*/
        else { fprintf(fres,"\n"); }
	acc_ctr++;
	cost  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	path  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	j=0;
	for (i=start; i<stop; i++) { cost[j]=INFTY; scc_idx(stack[i]) = j++;}
        cost[scc_idx(to)] = 0;
        NBshortest_path(cost, path, from, start, stop);
	NB_PrintTransSeq(fres, path, from, to);
	fprintf(fres,"\n==========================================\n");
        free(cost); free(path); 
	component++;
}
/***************************************************************************/
int Is_AcceptingSCC(int start, int stop, int *ffrom, int *tto)
/* returns 1 if the strongly connected comp contains a final edge, else 0*/
{
	register int i, v, comp; 
	char fin;
	Edge *vw; 

	for (i=start; i<stop; i++) { 
	     v = stack[i]; 
	     comp = scc(v);
	     if ( acc_succ(v) && (scc(acc_succ(v)) == comp) ) {
		  fprintf(fres,"SCC %d - Counterexample: ", component++);
        	  *ffrom = v; *tto = acc_succ(v); return(1); } 
	     vw = adjacents(v); 
	     while (vw != NULL) { 
		   if (scc(vw->adj) != comp) { vw=vw->next; continue; }
		   if ( F__1(vw->final) )  {
		        fprintf(fres,"SCC %d - Counterexample: ",component++);
        		*ffrom = v; *tto = vw->adj;
                        if ( extended_output) {
			         if (vw->final == (F_01 & vw->final)) fin=1;
			    else if (vw->final == (F_11 & vw->final)) fin=11;
			    fprintf(fres,"  **** %d--acc(%d)", v,fin); 
			    fprintf(fres,"-->%d ****\n", vw->adj); }
			return(1); } 
                   vw=vw->next; } 
	     } 
	return(0); 
}
/***************************************************************************/
int X_Is_AcceptingSCC(int start, int stop, int *ffrom, int *tto)
/* returns 1 if the strongly connected comp contains a final edge, else 0*/
{
	register int i, v, comp, as; 
	char fin;
	Edge *vw; 
	ENode *vn, *vwen;

	if ( stack[start] < 0 ) comp = EN_scc(search_enod(stack[start]));
	else comp = scc(stack[start]);
	for (i=start; i<stop; i++) { 
	     fin = 0;
	     if ( (v=stack[i]) < 0 ) {
		    vn = search_enod( v ); 
		    if ( (as=EN_acc_succ(vn)) != 0 ) {
		         if ( as < 0 ) 
			      fin = (comp == EN_scc(search_enod(as)));
			 else fin = (comp == scc(as));
			 }
                    }
	     else { vn = (ENode *) NULL; 
		    if ( (as=acc_succ(v)) != 0 ) {
			 if ( as < 0 )
                              fin = (comp == EN_scc(search_enod(as)));
			 else fin = (comp == scc(as));
			 }
		  }
	     if ( fin ) {
		  fprintf(fres,"SCC %d - Counterexample: \n", component++);
        	  *ffrom = v; *tto = as; return(1); } 
	     if ( vn ) {
		  if ( EN_IS(LOCAL, vn) ) {
			 vw = __adjacents((int) vn->corr);
			 vwen = (ENode *) enodes((int) vn->corr);
			 }
		  else { vw = vn->corr->adjs;
			 vwen = vn->corr->co_en; 
                       }
		  }
	     else { if ( IS(CUTOFF,v) ) {
			 vw = adjacents(v); 
			 vwen = (ENode *) enodes(correspond(v));
			 }
                    else { vw = __adjacents(v);
			   vwen = (ENode *) enodes(v);
                         }
                  }
	     while ( vw ) { 
		   if (scc(vw->adj) != comp) { vw=vw->next; continue; }
		   if ( F__1(vw->final) )  {
		        fprintf(fres,"SCC %d - Counterexample:\n",component++);
        		*ffrom = v; *tto = vw->adj;
			return(1); } 
                   vw=vw->next; } 
	     while ( vwen ) { 
		   if ( EN_scc(vwen) != comp ) { 
			vwen=vwen->next; continue; }
		   if ( F__1(vwen->final) ) {
		        fprintf(fres,"XSCC %d- Counterexample:\n",component++);
        		*ffrom = v; *tto = search_arg( vwen );
			return(1); } 
                   vwen=vwen->next; } 
	     } 
	return(0); 
}
/***************************************************************************/
void shortest_path(int *costs, int *path, int goal, int s1, int s2)
/* finds a shortest path from stack[s1] to goal */
{
	int i=0, j, border, v, w;
        Edge *nedge;

	border = s2 - s1;	
        while (i < border) {
              for (j=0; j < border; j++) {
    		   if (costs[j] > i) { continue; }
                   v = stack[s1+j];	
                   nedge = adjacents(v);
                   while (nedge) {
                          w = nedge->adj;
                          if ( w!=v && scc(w) == scc(v) ) {
			     if (costs[scc_idx(w)] > i+1) {
                                 costs[scc_idx(w)] = i+1;
			         path[scc_idx(w)] = v; }
 			     if (w == goal) { return; }
                             }
                          nedge = nedge->next;
			  }
                   }
               i++;
               }
}
/***************************************************************************/
void X_shortest_path(int *costs, int *path, int goal, int s1, int s2)
/* finds a shortest path from stack[s1] to goal */
{
	int i=0, j, border, v, w, comp;
        Edge *nedge;
	ENode *vn, *enode;

	border = s2 - s1;	
        while (i < border) {
              for (j=0; j < border; j++) {
    		   if (costs[j] > i) { continue; }
                   if ( (v = stack[s1+j]) < 0 ) {
			vn = search_enod(v);	
			comp = EN_scc(vn);
			if ( EN_IS(LOCAL, vn) ) {
			       nedge = __adjacents((int) vn->corr);
			       enode = (ENode *) enodes((int) vn->corr); }
                        else { nedge = vn->corr->adjs;
			       enode = vn->corr->co_en; }
			}
		   else { vn = (ENode *) NULL;
			  comp = scc(v);
			  if ( IS(CUTOFF, v) ) {
			         nedge = adjacents(v);
				 enode = (ENode *) enodes(correspond(v)); }
			  else { nedge = __adjacents(v);
				 enode = (ENode *) enodes(v); }
			}
                   while ( nedge ) {
                          w = nedge->adj;
                          if ( w!=v && scc(w) == comp ) {
			     if ( costs[scc_idx(w)] > i+1) {
                                  costs[scc_idx(w)] = i+1;
			          path[scc_idx(w)] = v; }
 			     if (w == goal) { return; }
                             }
                          nedge = nedge->next;
			  }
                   while ( enode ) {
                          if ( v!=search_arg(enode) && 
			       EN_scc(enode) == comp ) {
			     if ( costs[EN_scc_idx(enode)] > i+1) {
                                  costs[EN_scc_idx(enode)] = i+1;
			          path[EN_scc_idx(enode)] = v; }
 			     if ( search_arg(enode) == goal ) { return; }
                             }
                          enode = enode->next;
			  }
                   }
               i++;
               }
}
/***************************************************************************/
void NBshortest_path(int *costs, int *path, int goal, int s1, int s2)
/* finds a shortest path (without any Buechi Transitions, NB=no_buechi)
   from stack[s1] to goal */
{
	int i=0, j, border, v, w;
        Edge *nedge;

	border = s2 - s1;	
        while (i < border) {
              for (j=0; j < border; j++) {
    		   if (costs[j] > i) continue;
                   v = stack[s1+j];	
                   nedge = adjacents(v);
                   while (nedge) {
                          if ( ! F__0(nedge->final) ) {
                               nedge = nedge->next; continue; }
                          w = nedge->adj;
                          if ((scc(w)==scc(v)) && w!=v) {
			     if (costs[scc_idx(w)] > i+1) {
                                 costs[scc_idx(w)] = i+1;
			         path[scc_idx(w)] = v; }
 			     if (w == goal) return;
                             }
                          nedge = nedge->next;
			  }
                   }
               i++;
               }
}
/***************************************************************************/
void X_NBshortest_path(int *costs, int *path, int goal, int s1, int s2)
/* finds a shortest path (without any Buechi Transitions, NB=no_buechi)
   from stack[s1] to goal */
{
	int i=0, j, border, v, w, comp;
        Edge *nedge;
	ENode *vn, *enode;

	border = s2 - s1;	
        while (i < border) {
              for (j=0; j < border; j++) {
    		   if (costs[j] > i) { continue; }
                   if ( (v = stack[s1+j]) < 0 ) {
			vn = search_enod(v);	
			comp = EN_scc(vn);
			if ( EN_IS(LOCAL, vn) ) {
			       nedge = __adjacents((int) vn->corr);
			       enode = (ENode *) enodes((int) vn->corr); }
                        else { nedge = vn->corr->adjs;
			       enode = vn->corr->co_en; }
			}
		   else { vn = (ENode *) NULL;
			  comp = scc(v);
			  if ( IS(CUTOFF, v) ) {
			         nedge = adjacents(v);
				 enode = (ENode *) enodes(correspond(v)); }
			  else { nedge = __adjacents(v);
				 enode = (ENode *) enodes(v); }
			}
                   while ( nedge ) {
			  if ( ! F__0(nedge->final) ) {
			       nedge=nedge->next; continue; }
                          w = nedge->adj;
                          if ( w!=v && scc(w) == comp ) {
			     if ( costs[scc_idx(w)] > i+1) {
                                  costs[scc_idx(w)] = i+1;
			          path[scc_idx(w)] = v; }
 			     if (w == goal) { return; }
                             }
                          nedge = nedge->next;
			  }
                   while ( enode ) {
			  if ( ! F__0(enode->final) ) {
			       enode = enode->next; continue; }
                          if ( v!=search_arg(enode) && 
			       EN_scc(enode) == comp ) {
			     if ( costs[EN_scc_idx(enode)] > i+1) {
                                  costs[EN_scc_idx(enode)] = i+1;
			          path[EN_scc_idx(enode)] = v; }
 			     if ( search_arg(enode) == goal ) { return; }
                             }
                          enode = enode->next;
			  }
                   }
               i++;
	       }
}
/***************************************************************************/
void *Insert_IL(Int_List **il, int e, Int_List **lastil)
/* returns the list il (sorted upwards, i.e. il->i < il->next->i),
   with e inserted at the appropriate position */
{
	Int_List *q, *q2, *r;

	q2 = IL_NULL;
        if ( (q = (*lastil)) == IL_NULL ) {
	     (*lastil) = (*il) = NEW_INTLIST; 
	     (*il)->i = e; (*il)->next = IL_NULL;
	     return; } 
	    
        if ( q->i > e) q = (*il); 
	while ( q ) {
	       if (q->i < e) { q2=q; q=q->next; continue; }
	       if (q->i > e) { break; }
	       /** else: (q->i == e) **/
               (*lastil) = q; return; }

	r = NEW_INTLIST; 
	r->i = e; r->next = q; 
        (*lastil) = r;
	if (q2 != IL_NULL) q2->next = r;
	else (*il) = r;
}
/***********************************************************************/
ENCorr *Grab_ENCorr(Int_List *mtl, ENode *en)
{
	Int_List 	*mtlp, *il;
	TE_List  	*tel, *telp, *newtel;
	int 		i, idx;
	ENCorr   	*encorr;
	char 		c;

	mtlp = mtl; i = 0; 
	while ( mtlp ) { i += mtlp->i; mtlp = mtlp->next; }
	if ( i < 0 ) i = ( - i );
	idx = ENC_table_idx(i);
	tel = enCorrs[idx]; telp = TEL_NULL;
	while ( tel && (c=IL_less(tel->e, mtl)) ) { 
		if (c == 2) return( (ENCorr *) tel->t); 
		telp=tel; tel=tel->next; }
        /* entry exists already 
	if ( c == 2 ) {
	     if ( telp ) tel = telp; else tel = enCorrs[idx];
	     encorr = (ENCorr *) tel->t;
	     return(encorr);
	     }
        */

	newtel = NEW_TELIST; 
	if ( telp ) { newtel->next = telp->next; telp->next = newtel; }
	else { newtel->next = enCorrs[idx]; enCorrs[idx] = newtel; }
	newtel->t = (int) MyMalloc(sizeof(ENCorr));
	encorr = (ENCorr *) newtel->t;
	encorr->max = newtel->e = mtl;
	encorr->enode = en;
	return(encorr);
}
/***********************************************************************/
void  Create_Cand_Lists(Int_List **ncl)
{
	ENCorr *encorr;
	ENode *enode;
	Edge *nedge;
	TE_List *tel;
	Int_List *il;
	int i=0, border, lc;

	border = E_Fin/__HT__;
	while ( i < border ) {
	      tel = enCorrs[i++];
	      while ( tel ) {
		    encorr = (ENCorr *) tel->t; tel = tel->next;
		    if (encorr->enode->info & CHECKED) continue;
		    encorr->enode->info |= CHECKED;
		    lc = encorr->last_corr;
		    if ( lc == 0 ) { globEN_nc++;
			 il = NEW_INTLIST; il->i = (int) encorr;
                         il->next = (*ncl); (*ncl) = il; continue; }
		    /** last_corr is set **/
		    globEN++;
		    if ( lc < 0 ) enode = (ENode *) enodes( - lc );
		    else enode = (ENode *) enodes(lc);
		    while ( enode ) {
			  il = NEW_INTLIST; il->i = (int) encorr;
			  il->next = NEW_INTLIST; 
			  il->next->i = (int) enode;
			  il->next->next = 
				    (Int_List *) ENCorr_co_cands(enode->cutoff);
			  ENCorr_co_cands(enode->cutoff) = (int) il;
			  enode = enode->next; }
		    if ( lc < 0 ) {
		       nedge = __adjacents( - lc );
		       while ( nedge ) {
			     if (IS(CUTOFF,nedge->n) ) {
				il = NEW_INTLIST; il->i = (int) encorr;
				il->next = NEW_INTLIST; il->next->i = 0;
				il->next->next =
				    (Int_List *) ENCorr_co_cands(nedge->n);
			        ENCorr_co_cands(nedge->n) = (int) il;
				}
                             nedge = nedge->next;
			     }
                       }
		    }
              }
}
/***************************************************************************/
void X_final_search(int v)
{
	int 	 i, x, w, j=0,
		 start, stop, from, to;
	Edge 	 *nedge;
	ENode    *vn, *wn;
	ENCorr   *encorr;
	int      *cost, *path;

	if ( v < 0 ) vn = search_enod(v);  
	else         vn = (ENode *) NULL;
	if ( vn && EN_IS(LOCAL, vn) ) {
	     if ( IS(VISITED, (int) vn->corr ) ) { 
		  vn->info |= VISITED; return; }
             INFO( (int) vn->corr ) |= VISITED;
	     }
        if ( ! vn ) { INFO(v) |= (ONSTACK | VISITED);
      	              lowlink(v) = count; dfn(v) = count++;
	              scc(v) = 0; push(v); }
        else        { vn->info |= (ONSTACK | VISITED);
		      vn->info &= ~(INITIALIZED);
		      EN_lowlink(vn) = count; EN_dfn(vn) = count++;
		      EN_scc(vn) = 0; push(v); }

	if ( (! vn) ) { /* v is an event */
	     if (IS(CUTOFF, v)) { 
		    nedge = adjacents(v); 
                    wn = (ENode *) enodes(correspond(v)); }
	     else { nedge = __adjacents(v);
                    wn = (ENode *) enodes(v); }
	     while ( nedge ) { w = nedge->adj;
	            if ( ! IS(VISITED, w) ) { 
		         if ( F__1(nedge->final) ) acc_succ(v) = w;
		         X_final_search(w);
		         lowlink(v) = MIN( lowlink(v), lowlink(w) ); }
	            else { if ((dfn(w)<dfn(v)) && IS_ON_STACK(w))
		              lowlink(v) = MIN( dfn(w), lowlink(v) ); }
	            nedge = nedge->next; }
             while ( wn ) { 
		    if ( ! EN_IS(VISITED, wn) ) {
			 if ( F__1(wn->final) && (acc_succ(v) == 0) ) 
			      acc_succ(v) = search_arg(wn);
		         X_final_search( search_arg(wn) );
		         lowlink(v) = MIN(lowlink(v), EN_lowlink(wn) ); }
                    else { if ((EN_dfn(wn)<dfn(v))&&
			       (EN_IS(ONSTACK,wn) ) )
			       lowlink(v) = MIN( lowlink(v), EN_dfn(wn));
                         }
                    wn = wn->next; }
             } /* v is an event */

	if ( vn ) { /* vn is an ENode */
	     if (EN_IS(LOCAL, vn)) { 
		    encorr = (ENCorr *) NULL;
		    nedge = __adjacents( (int) vn->corr ); 
                    wn = (ENode *) enodes( (int) vn->corr ); }
	     else { encorr = vn->corr;
		    nedge = encorr->adjs;
                    wn = encorr->co_en; }
	     while ( nedge ) { w = nedge->adj;
	           if ( ! IS(VISITED, w) ) { 
		          if ( F__1(nedge->final) ) EN_acc_succ(vn) = w;
		          X_final_search(w);
		          EN_lowlink(vn) = MIN( EN_lowlink(vn), lowlink(w) ); }
	           else { if ((dfn(w)<EN_dfn(vn)) && IS_ON_STACK(w))
		              EN_lowlink(vn)=MIN(dfn(w), EN_lowlink(vn)); 
			}
	           nedge = nedge->next; }
             while ( wn ) { 
		   if ( ! EN_IS(VISITED, wn) ) {
			  if ( F__1(wn->final) && (EN_acc_succ(vn) == 0 ) )
			       EN_acc_succ(vn) = search_arg(wn);
			  X_final_search( search_arg (wn) );
			  EN_lowlink(vn) = MIN(EN_lowlink(vn), EN_lowlink(wn));
			  }
                   else { if ((EN_dfn(wn)<EN_dfn(vn))&&
			      (EN_IS(ONSTACK,wn) ) )
			    EN_lowlink(vn)=MIN(EN_lowlink(vn), EN_dfn(wn));
                        }
                   wn = wn->next; }
	     } /* else (v is an ENOde) */

	stop = ((vn)?(EN_lowlink(vn) != EN_dfn(vn)):(lowlink(v) != dfn(v)));
	if ( stop ) return; 
	stop = stackpointer;
	do { i = pop; 
	     if (i < 0) { wn = search_enod(i) ;
	                  wn->info &= ~ONSTACK; EN_scc(wn) = comp_ctr; }
	     else { INFO(i) &= ~ONSTACK; scc(i) = comp_ctr; }
           } while ( v != i );
        comp_ctr++;
	start = stackpointer;
        if ( (stop-start) == 1) return;
	if ( ! X_Is_AcceptingSCC(start,stop, &from, &to) ) return;
	/*
	fprintf(fres,"\n\n %d --> %d \n\n", from, to);
	*/
        if ( ! alreadyNO ) { printf("\n\nNO. (9)\n\n"); alreadyNO=1; }
	accepting_component = final_result = 1; acc_ctr++;
        fprintf(fres,"\n"); 
	cost  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	path  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	for (i=start, j=0; i<stop; i++, j++ ) { 
	     cost[j] = INFTY; 
	     if (stack[i] < 0) 
	          EN_scc_idx( search_enod(stack[i]) ) = j; 
	     else scc_idx(stack[i]) = j; }
        if ( to < 0 ) 
	     cost[EN_scc_idx( search_enod(to) )] = 0;
	else cost[scc_idx(to)] = 0;
        X_shortest_path(cost, path, from, start, stop);
	X_PrintTransSeq(fres, path, from, to);
	fprintf(fres,"\n======================================\n");
	if ( ! all_counterexamples ) { 
	     printf("Counterexample:\n"); 
	     PrintTransSeq(stdout, path, from, to); 
	     all_time_f = nc_stopclock(all_time); 
	     if ( show_statistics) { ffcomp1 = comp_ctr - 1; 
				     ffacc1 = acc_ctr; 
				     ShowStatistics(); } 
	     fclose(fres); 
	     if ( no_output) { EXIT_NO_OUTPUT; } 
	     exit(0); }
        free(cost); free(path); 
}
/***************************************************************************/
void X_non_final_search(int v)
{
	int 	 i, x, w, j=0,
		 start, stop, from, to;
	Edge 	 *nedge;
	ENode    *vn, *wn;
	ENCorr   *encorr;
	int      *cost, *path;
	ConfS    conf;

	if ( v < 0 ) vn = search_enod(v);  
	else         vn = (ENode *) NULL;
	if ( vn && EN_IS(LOCAL, vn) ) {
	     if ( IS(VISITED, (int) vn->corr ) ) { 
		  vn->info |= VISITED; return; }
             INFO( (int) vn->corr ) |= VISITED;
	     }
        if ( ! vn ) { INFO(v) |= (ONSTACK | VISITED);
      	              lowlink(v) = count; dfn(v) = count++;
	              scc(v) = 0; push(v); }
        else        { vn->info |= (ONSTACK | VISITED);
		      EN_lowlink(vn) = count; EN_dfn(vn) = count++;
		      EN_scc(vn) = 0; push(v); }

	if ( (! vn) ) { /* v is an event */
	     if (IS(CUTOFF, v)) { 
		    nedge = adjacents(v); 
                    wn = (ENode *) enodes(correspond(v)); }
	     else { nedge = __adjacents(v);
                    wn = (ENode *) enodes(v); }
	     while ( nedge ) { 
		    if ( F__0(nedge->final) ) w = nedge->adj;
		    else { nedge=nedge->next; continue; }
	            if ( ! IS(VISITED, w) ) { 
		         X_non_final_search(w);
		         lowlink(v) = MIN( lowlink(v), lowlink(w) ); }
	            else { if ((dfn(w)<dfn(v)) && IS_ON_STACK(w))
		              lowlink(v) = MIN( dfn(w), lowlink(v) ); }
	            nedge = nedge->next; }
             while ( wn ) { 
		    if ( ! F__0(wn->final) ) { wn=wn->next; continue; }
		    if ( ! EN_IS(VISITED, wn) ) {
		         X_non_final_search( search_arg(wn) );
		         lowlink(v) = MIN(lowlink(v), EN_lowlink(wn) ); }
                    else { if ((EN_dfn(wn)<dfn(v))&&
			       (EN_IS(ONSTACK,wn) ) )
			       lowlink(v) = MIN( lowlink(v), EN_dfn(wn));
                         }
                    wn = wn->next; }
             } /* v is an event */

	if ( vn ) { /* vn is an ENode */
	     if (EN_IS(LOCAL, vn)) { 
		    encorr = (ENCorr *) NULL;
		    nedge = __adjacents( (int) vn->corr ); 
                    wn = (ENode *) enodes( (int) vn->corr ); }
	     else { encorr = vn->corr;
		    nedge = encorr->adjs;
                    wn = encorr->co_en; }
	     while ( nedge ) { 
		   if ( F__0(nedge->final) ) w = nedge->adj;
		   else { nedge=nedge->next; continue; }
	           if ( ! IS(VISITED, w) ) { 
		          if ( F__1(nedge->final) ) EN_acc_succ(vn) = w;
		          X_non_final_search(w);
		          EN_lowlink(vn) = MIN( EN_lowlink(vn), lowlink(w) ); }
	           else { if ((dfn(w)<EN_dfn(vn)) && IS_ON_STACK(w))
		              EN_lowlink(vn)=MIN(dfn(w), EN_lowlink(vn)); 
			}
	           nedge = nedge->next; }
             while ( wn ) { 
		   if ( ! F__0(wn->final) ) { wn = wn->next; continue; }
		   if ( ! EN_IS(VISITED, wn) ) {
			  if ( F__1(wn->final) && (EN_acc_succ(vn) == 0 ) )
			       EN_acc_succ(vn) = search_arg(wn);
			  X_non_final_search( search_arg (wn) );
			  EN_lowlink(vn) = MIN(EN_lowlink(vn), EN_lowlink(wn));
			  }
                   else { if ((EN_dfn(wn)<EN_dfn(vn))&&
			      (EN_IS(ONSTACK,wn) ) )
			    EN_lowlink(vn)=MIN(EN_lowlink(vn), EN_dfn(wn));
                        }
                   wn = wn->next; }
	     } /* else (v is an ENOde) */

	stop = ( (vn) ? ( EN_lowlink(vn) != EN_dfn(vn) ) :
	                ( lowlink(v) != dfn(v) ) );
	if ( stop ) return; 
	stop = stackpointer;
	do { i = pop; 
	     if (i < 0) { wn = search_enod(i) ;
	            wn->info &= ~ONSTACK; EN_scc(wn) = comp_ctr; }
	     else { INFO(i) &= ~ONSTACK; scc(i) = comp_ctr; }
           } while ( v != i );
        comp_ctr++;
	start = stackpointer;

        if ( (stop-start) == 1 ) return;

	from = to = INFTY;
	for ( i = start; i < stop; i++ ) {
	    if (stack[i] < 0) 
		 { /* an ENode */
		   vn = search_enod(stack[i]); 
		   j = EN_maxBuchi(vn);
		   /* set conf appropriately */
	           if ( EN_IS(LOCAL, vn) )  
			  conf = config( (int) vn->corr );
	           else { encorr = vn->corr;
			  IL2CONF(encorr->conf, conf3, ba_idx(encorr->max->i));
			  conf = conf3; }
                 }
	    else { /* not an ENode */
		   j = maxBuchi(stack[i]);
		   conf = config(stack[i]); }
	    accepting_component = 0;
	    Check_Buechi_Circle(j, conf, 8);
	    if (accepting_component) { from = stack[i]; break; }
	    }

	if ( from == INFTY ) return;

	if ( from < 0 ) {
		vn = search_enod(from);
		if ( EN_IS(LOCAL, vn) ) {
		       nedge = __adjacents( (int) vn->corr );
		       wn = (ENode *) enodes( (int) vn->corr ); }
		else { nedge = vn->corr->adjs;
		       wn = vn->corr->co_en; }
	      }
	else  { if ( IS(CUTOFF, from) ) { 
		       nedge = adjacents(from);
		       wn = (ENode *) enodes(correspond(from)); }
                else { nedge = __adjacents(from);
		       wn = (ENode *) enodes(from); }
	      }

	while ( ( to == INFTY) && nedge ) {
		if ( ( F__0(nedge->final) )  && 
		     ( scc(nedge->adj) == (comp_ctr - 1) ) ) {
		     to = nedge->adj; break; }
                nedge = nedge->next; 
		}
        while ( ( to == INFTY) && wn ) {
		if ( ( F__0(wn->final) )  &&
		     ( EN_scc(wn) == (comp_ctr - 1) ) ) {
		     to = search_arg(wn); break; }
                wn = wn->next;
		}

	if ( to == INFTY ) {
	     printf("Oops... no non-buechi-Transition found.\n");
	     }

	acc_ctr++;

 	fprintf(fres,"\nSCC %d - NV-Counterexample:\n\n",component);
	cost  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	path  = (int *) MyMalloc( sizeof(int) * (stop-start) );
	for (i=start, j=0; i<stop; i++, j++) { 
	     cost[j] = INFTY; 
	     if (stack[i] < 0) 
	          EN_scc_idx( search_enod(stack[i]) ) = j; 
	     else scc_idx(stack[i]) = j; }
        if ( to < 0 ) 
	     cost[EN_scc_idx( search_enod(to) )] = 0;
	else cost[scc_idx(to)] = 0;
        X_NBshortest_path(cost, path, from, start, stop);
	X_NB_PrintTransSeq(fres, path, from, to);
	fprintf(fres,"\n======================================\n");
        free(cost); free(path); 
	component++;
}
/***************************************************************************/
void EN_final_search(ENode *enod)
{
	ENode *enen;

	enod->info &= ~(INITIALIZED);
	if ( ! EN_IS(VISITED, enod) ) X_final_search( search_arg(enod) );
	else return;
	if ( EN_IS(LOCAL, enod) ) 
	     enen = (ENode *) enodes( (int) enod->corr ); 
        else enen = enod->corr->co_en; 
	while (enen) { EN_final_search(enen); enen=enen->next; }
}
/***************************************************************************/
void EN_non_final_search(ENode *enod)
{
	ENode *enen;

	if ( ! EN_IS(VISITED, enod) ) X_non_final_search( search_arg(enod) );
	else return;
	if ( EN_IS(LOCAL, enod) ) 
	     enen = (ENode *) enodes( (int) enod->corr ); 
        else enen = enod->corr->co_en; 
	while (enen) { EN_non_final_search(enen); enen=enen->next; }
}
/***************************************************************************/
void Init_ENodes(ENode *enod)
{
	ENode *enen;

	if ( EN_IS(INITIALIZED, enod) ) return;
	enod->info |= INITIALIZED;
	enod->info &= ~VISITED;
	EN_lowlink(enod) = EN_dfn(enod) = 0;
	EN_scc(enod) = EN_scc_idx(enod) = 0;
	if ( EN_IS(LOCAL, enod) ) {
	     enod->info |= VISITED;
	     enen = (ENode *) enodes( (int) enod->corr ); 
	     }
        else enen = enod->corr->co_en; 
	while (enen) { Init_ENodes(enen); enen=enen->next; }
}
/***************************************************************************/
void UnInit_ENodes(ENode *enod)
{
	ENode *enen;

	if ( ! EN_IS(INITIALIZED, enod) ) return;
	enod->info &= ~(INITIALIZED);
	if ( EN_IS(LOCAL, enod) ) 
	     enen = (ENode *) enodes( (int) enod->corr ); 
        else enen = enod->corr->co_en; 
	while (enen) { UnInit_ENodes(enen); enen=enen->next; }
}
/***************************************************************************/
/* graphalgorithms.c */
/*********************/
