/*****************************************************************************/
/* neunetze.c   - VERSION m-1.8.1 - 08.02.1998 - (c) Stefan Roemer 1993-1998 */
/*****************************************************************************/
/*              Schnittstellen Bibliothek zwischen GreatSPN und              */
/*                 Praktikum Werkzeuge zur Petri Netz Analyse                */
/*                    Author: Robert Riemann                                 */
/*                    Date  : April 1992                                     */
/*****************************************************************************/
/* Version m-1.8.1                     Stefan Roemer,  04.05.1993-09.01.1995 */
/*                                     Stefan Roemer,            -09.03.1997 */
/*                                     Stefan Roemer,            -30.01.1998 */
/*                                     Stefan Roemer,            -08.02.1998 */
/*****************************************************************************/
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/* ACHTUNG: Diese Datei enthaelt elementare Routinen, die von vielen ver-    */
/* schiedenen Programmen benoetigt werden.                                   */
/* Jede beabsichtigte Aenderung am Programmcode sowie das Kopieren von       */
/* Programmteilen deshalb bitte vorher mit mir absprechen. Danke.            */
/* E-mail: roemer@informatik.tu-muenchen.de                                  */
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/*****************************************************************************/


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "neunetze.h"

#if defined(__STDC__) || defined(__cplusplus)

void delete_P_to_T_con_node( LMPNet, STEdge );
void delete_T_to_P_con_node( LMPNet, TSEdge );

#else

void delete_P_to_T_con_node();
void delete_T_to_P_con_node();

#endif

/*****************************************************************************/
/*                              Deklarationen                                */
/*****************************************************************************/

LMPNet        net;
LIST_POINTER  list;
MARKING       mark_list;
int NN_ambigious_nodes = 2; /* 0 = ok, 1 = error, 2 = warning */


/*****************************************************************************/
/*                                Funktionen                                 */
/*****************************************************************************/

/******************/
/* Fehler-Makros: */
/******************/

#define ERROR_MEM_OVERFLOW( FUNCT ) \
   fprintf( stderr, "ERROR: NN.%s(): memory overflow.\n", FUNCT ), \
   exit ( 3 )

#define ERROR_FILE_READ( FUNCT, FILENAME ) \
 fprintf( stderr, "WARNING: NN.%s(): could not open file '%s' for reading.\n",\
            FUNCT, FILENAME ), \
   exit ( 4 )

#define ERROR_FILE_WRITE( FUNCT, FILENAME ) \
 fprintf( stderr, "WARNING: NN.%s(): could not open file '%s' for writing.\n",\
            FUNCT, FILENAME ), \
   exit ( 5 )

#define WARNING_NO_NET( FUNCT ) \
  fprintf( stderr, "WARNING: NN.%s(): not a net.\n", FUNCT )

#define WARNING_NO_NET_OR_P( FUNCT ) \
  fprintf( stderr, "WARNING: NN.%s(): not a net or no place.\n", FUNCT )

#define WARNING_NO_NET_OR_T( FUNCT ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): not a net or no transition.\n", FUNCT )

#define WARNING_P( FUNCT, P_NAME ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): place '%s' does not exist.\n", FUNCT, P_NAME )

#define WARNING_T( FUNCT, T_NAME ) \
   fprintf( stderr, \
   "WARNING: NN.%s(): transition '%s' does not exist.\n", FUNCT, T_NAME )

#define WARNING_P_N( FUNCT, P_NUMNAME ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): place identifier '%u' does not exist.\n", \
           FUNCT, P_NUMNAME )

#define WARNING_T_N( FUNCT, T_NUMNAME ) \
   fprintf( stderr, \
   "WARNING: NN.%s(): transition identifier '%u' does not exist.\n", \
            FUNCT, T_NUMNAME )

#define WARNING_NO_P( FUNCT ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): net without places.\n", FUNCT )

#define WARNING_NO_T( FUNCT ) \
   fprintf( stderr, \
   "WARNING: NN.%s(): net without transitions.\n", FUNCT )

#define WARNING_P_NUM( FUNCT, P_NUMNAME ) \
  fprintf( stderr, \
  "WARNING: NN.%s(): a place numbered %u does not exist\n", FUNCT, P_NUMNAME )

#define WARNING_T_NUM( FUNCT, T_NUMNAME ) \
   fprintf( stderr, \
   "WARNING: NN.%s(): a transition numbered %u does not exist\n", T_NUMNAME )

/*
ERROR_MEM_OVERFLOW(  );
WARNING_NO_NET(  );
WARNING_NO_NET_OR_P(  );
WARNING_NO_NET_OR_T(  );
WARNING_P( , P_name );
WARNING_T( , T_name );
WARNING_NO_P(  );
WARNING_NO_T(  );
*/

/*****************************************************************************/
/* ERGAENZUNG: 04.05.1993 Stefan Roemer                                      */
/* Allokationsfunktion                                                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void * Malloc( size_t size )

#else

void * Malloc( size )

  size_t size;

#endif

  {

    register char * m;

    if ( ! ( m = calloc( size, 1 ) ) )
      {
        ERROR_MEM_OVERFLOW( "Malloc" );
      } /* if */

/*
printf("%p MALLOC\n", m);
*/

    return ( m );

  } /* Malloc() */


/*****************************************************************************/
/* ERGAENZUNG: 04.05.1993 Stefan Roemer                                      */
/* Allokationsfunktion                                                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void * Calloc( size_t nmemb, size_t size )

#else

void * Calloc( nmemb, size )

  size_t nmemb,
         size;

#endif

  {

    register char * m;

    if ( ! ( m = calloc( nmemb, size ) ) )
      {
        ERROR_MEM_OVERFLOW( "Calloc" );
      } /* if */

/*
printf("%p CALLOC\n", m);
*/

    return ( m );

  } /* Calloc() */


/*****************************************************************************/
/* ERGAENZUNG: 04.05.1993 Stefan Roemer                                      */
/* Allokationsfunktion                                                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void * Realloc( void * ptr, size_t size )

#else

void * Realloc( ptr, size )

  void   * ptr;
  size_t size;

#endif

  {

    register void * r;

    if ( ! ( r = realloc( ptr, size ) ) )
      {
        ERROR_MEM_OVERFLOW( "Realloc" );
      } /* if */

/*
printf("%p REALLOC\n", r);
*/

    return ( r );

  } /* Realloc() */


/*****************************************************************************/
/* ERGAENZUNG: 04.05.1993 Stefan Roemer                                      */
/* Allokationsfunktion                                                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

char * Strdup( char * str )

#else

char * Strdup( str )

  char * str;

#endif

  {

    register char * s,
                  * p1,
                  * p2;
    register int  i;

    for ( i = 1, p1 = str ; *( p1++ ) ; i++ )
      skip;
    if ( ! ( s = p2 = (char *) calloc( i, 1 ) ) )
      {
        ERROR_MEM_OVERFLOW( "Strdup" );
      } /* if */
    p1 = str;
    while ( *( p2++ ) = *( p1++ ) )
      skip;

/*
printf("%p STRDUP\n", s);
*/

    return ( s );

  } /* Strdup() */


/*****************************************************************************/
/* ERGAENZUNG: 04.05.1993 Stefan Roemer                                      */
/* Vergleichsfunktionen fuer Strings                                         */
/* In der vorliegenden Version: Sortierung nach lexikalischer Ordnung!!!     */
/*****************************************************************************/

#define Strcpy( _STR1_, _STR2_ ) strncpy( _STR1_, _STR2_, NAMELENGTH )

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static bool LG_Kleiner( char * name1, char * name2 )

#else

static bool LG_Kleiner( name1, name2 )

  char * name1, * name2;

#endif

  {

    register int l1, l2;

    if (    ( ( l1 = (int) strlen( name1 ) ) < ( l2 = (int) strlen( name2 ) ) )
         || ( ( l1 == l2 ) && ( strncmp( name1, name2, NAMELENGTH ) < 0 ) ) )
      {
        return ( True );
      } /* if */
    else
      {
        return ( False );
      } /* else */

  } /* LG_Kleiner() */


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static bool LG_Gleich( char * name1, char * name2 )

#else

static bool LG_Gleich( name1, name2 )

  char * name1, * name2;

#endif

  { 

    if ( strncmp( name1, name2, NAMELENGTH ) == 0 )
      {
        return ( True );
      } /* if */
    else
      {
        return ( False );
      } /* else */

  } /* LG_Gleich() */


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static bool SC_Kleiner( char * name1, char * name2 )

#else

static bool SC_Kleiner( name1, name2 )

  char * name1, * name2;

#endif

  { 

    return ( (bool) ( strncmp( name1, name2, NAMELENGTH ) < 0 ) );

  } /* SC_Kleiner() */


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

static bool SC_Gleich( char * name1, char * name2 )

#else

static bool SC_Gleich( name1, name2 )

  char * name1, * name2;

#endif

  { 

    return ( (bool) ! strncmp( name1, name2, NAMELENGTH ) );

  } /* SC_Gleich() */


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
bool (* Kleiner)( char *, char * ) = LG_Kleiner;
bool (* Gleich)( char *, char * ) = LG_Gleich;
#else
bool (* Kleiner)() = LG_Kleiner;
bool (* Gleich)() = LG_Gleich;
#endif


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void strcmp_lexgraph( void )

#else

void strcmp_lexgraph()

#endif

  { 

    Kleiner = LG_Kleiner;
    Gleich = LG_Gleich;

  } /* strcmp_lexgraph() */


/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void strcmp_strcmp( void )

#else

void strcmp_strcmp()

#endif

  { 

    Kleiner = SC_Kleiner;
    Gleich = SC_Gleich;

  } /* strcmp_strcmp() */


/*****************************************************************************/
/*                             Loeschfunktionen                              */
/*****************************************************************************/

/*****************************************************************************/
/***************************** remove_list() *********************************/
/*****************************************************************************/
/* NEUE FUNKTION: 10.05.1993 Stefan Roemer                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void remove_list( LIST_POINTER * list )

#else

void remove_list( list )
  /* Entfernen einer Liste vom Typ 'LIST_POINTER' aus dem Speicher.          */

  LIST_POINTER * list; /* call-by-reference ! */

#endif

  {

    register LIST_POINTER del = *list;

    while ( del )
      {
        del = ( *list = del ) -> next;
        Free( *list );
      } /* while */
    *list = NULL;

  } /* remove_list() */


/*****************************************************************************/
/****************************** remove_marklist() ****************************/
/*****************************************************************************/
/* NEUE FUNKTION: 10.05.1993 Stefan Roemer                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void remove_marklist( MARKING * list )

#else

void remove_marklist( list )
  /* Entfernen einer Markierungsliste aus dem Speicher.                      */

  MARKING * list; /* call-by-reference ! */

#endif

  {

    register MARKING del = *list;

    while ( del )
      {
        del = ( *list = del ) -> next;
        Free( *list );
      } /* while */
    *list = NULL;

  } /* remove_marklist() */


/*****************************************************************************/
/****************************** remove_net() *********************************/
/*****************************************************************************/
/* NEUE FUNKTION: 10.05.1993 Stefan Roemer                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void remove_net( LMPNet * net )

#else

void remove_net( net )
  /* Entfernen eines Netzes aus dem Speicher.                                */

  LMPNet * net; /* call-by-reference ! */

#endif

  {

    if ( *net )
      {
        while ( (*net)->Transitions )
          {
            (void) delete_transition_direct( *net, (*net)->Transitions );
          } /* while */
        while ( (*net)->Places )
          {
            (void) delete_place_direct( *net, (*net)->Places );
          } /* while */
        if ( (*net)->Flags & 1 )
          {
            Free( (*net)->Places );
            Free( (*net)->Transitions );
          } /* if */
        Free( *net );
        *net = NULL;
      } /* if */

  } /* remove_net() */


/*###########################################################################*/
/*  Matrixinterne Funktionen, die auf die 'numname'-Eintraege zurueckgreifen */
/*###########################################################################*/
/* NEU EINGEFUEGT: 15.02.1994 Stefan Roemer                                  */
/*###########################################################################*/

/*****************************************************************************/
/**************************** num_net_nodes() ********************************/
/*****************************************************************************/
/* NEU EINGEFUEGT: 15.02.1994 Stefan Roemer                                  */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void num_net_nodes( LMPNet net )

#else

void num_net_nodes( net )
  /* Diese Funktion numeriert die Stellenknoten des uebergebenen Netzes in   */
  /* der 'numname'-Komponente von 1..net->NumPlaces durch. Analog werden die */
  /* 'numname'-Eintraege der Transitionsknoten des Netzes durchnumeriert mit */
  /* 1..net->NumTransitions.                                                 */

  LMPNet     net;

#endif

  {

    register Cardinal32 count;
    register Place      pl;
    register Transition tr;

    count = 1;
    pl = net->Places;
    while ( pl )
      {
        pl->numname = count++;
        pl = pl->nextplace;
      } /* while */

    count = 1;
    tr = net->Transitions;
    while ( tr )
      {
        tr->numname = count++;
        tr = tr->nexttrans;
      } /* while */

  } /* num_net_nodes() */


/*****************************************************************************/
/*********************************** find_T_num ******************************/
/*****************************************************************************/
/* KONVERSION: 15.02.1994 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition find_T_num( LMPNet net, Cardinal32 T_numname )

#else

Transition find_T_num( net, T_numname )
  /* liefert Pointer auf Knoten mit T_name in der ersten Zeile.              */
  /* Ergebnis ist NULL-Pointer, wenn: Keine Transitionenliste in der Matrix  */
  /*                                                              existiert  */
  /*                                  Kein Knoten mit T_name in der ersten   */
  /*                                                        Zeile existiert  */

  LMPNet     net;
  Cardinal32 T_numname;    /* String fuer Name der Transition */

#endif

  {

    register Transition tr = NULL;

    if ( net )
      {
        if ( tr = net->Transitions )
          {
            while ( tr && ( tr->numname < T_numname ) )
              {
                tr = tr->nexttrans;
              } /* while */
            if ( ! ( tr && ( tr->numname == T_numname ) ) )
              {
                WARNING_T_NUM( "find_T_num", T_numname );
                tr = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_T( "find_T_num" );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_T_num" );
      } /* else */

    return ( tr );

  } /* find_T_num() */


/*****************************************************************************/
/*********************************** find_P_num ******************************/
/*****************************************************************************/
/* KONVERSION: 15.02.1994 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place find_P_num( LMPNet net, Cardinal32 P_numname )

#else

Place find_P_num( net, P_numname )
  /* liefert Pointer auf Stellenknoten mit P_numname, also Knoten in erster  */
  /* Spalte der Matrix.                                                      */
  /* Ergebnis ist NULL-Pointer, wenn: keine Stellen im Netz vorhanden sind,  */
  /*                                          also die erste Spalte leer ist */
  /*                                  kein Stellenknoten mit Namen P_numname */
  /*                                              in der Liste vorhanden ist */

  LMPNet     net;
  Cardinal32 P_numname;

#endif

  {

    register Place pl = NULL;

    if ( net )
      {
        if ( pl = net->Places )
          {
            while ( pl && ( pl->numname < P_numname ) )
              {
                pl = pl->nextplace;
              } /* while */
            if ( ! ( pl && ( pl->numname == P_numname ) ) )
              {
                WARNING_P_NUM( "find_P_num", P_numname );
                pl = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_P( "find_P_num" );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_P_num" );
      } /* else */

    return ( pl );

  } /* find_P_num() */


/*****************************************************************************/
/***************************** find_PT_connection_num ************************/
/*****************************************************************************/
/* KONVERSION: 15.02.1994 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge find_PT_connection_num( LMPNet net,
                               Cardinal32 P_numname, Cardinal32 T_numname )

#else

STEdge find_PT_connection_num( net, P_numname, T_numname )
  /* liefert Pointer auf Verbindungsknoten fuer die Verbindung von Stelle    */
  /* 'P_name' nach Transition 'T_name'.                                      */
  /* Ergebnis ist NULL-Pointer, wenn: P_name oder T_name Stellen- bzw.       */
  /* Transitionsknoten sind oder kein solcher Verbindungsknoten vorhanden ist*/

  LMPNet     net;
  Cardinal32 P_numname,
             T_numname;

#endif

  {

    register Place  pl;
    register STEdge st;

    if ( net )
      {
        if ( pl = find_P_num( net, P_numname ) )
          {
            if ( st = pl->postset )
              {
                while ( st && ( st->trans->numname < T_numname ) )
                  {
                    st = st->nexttrans;
                  } /* while */
                if ( st && ( T_numname == st->trans->numname ) )
                  {
                    return ( st );
                  } /* if */
              } /* if */
          } /* if */
        else
          {
            WARNING_P_NUM( "find_PT_connection_num", P_numname );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_PT_connection_num" );
      } /* else */

    return ( NULL );

  } /* find_PT_connection_num() */


/*****************************************************************************/
/***************************** find_TP_connection_num ************************/
/*****************************************************************************/
/* KONVERSION: 15.02.1994 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge find_TP_connection_num( LMPNet net,
                               Cardinal32 P_numname, Cardinal32 T_numname )

#else

TSEdge find_TP_connection_num( net, P_numname, T_numname )
  /* liefert Pointer auf Verbindungsknoten fuer die Verbindung von           */
  /* Transition 'T_name' nach Stelle 'P_name'.                               */
  /* Ergebnis ist NULL-Pointer, wenn: P_name oder T_name Stellen- bzw.       */
  /* Transitionsknoten sind oder kein solcher Verbindungsknoten vorhanden ist*/

  LMPNet     net;
  Cardinal32 P_numname,
             T_numname;

#endif

  {

    register Place  pl;
    register TSEdge ts;

    if ( net )
      {
        if ( pl = find_P_num( net, P_numname ) )
          {
            if ( ts = pl->preset )
              {
                while ( ts && ( ts->trans->numname < T_numname ) )
                  {
                    ts = ts->nexttrans;
                  } /* while */
                if ( ts && ( T_numname == ts->trans->numname ) )
                  {
                    return ( ts );
                  } /* if */
              } /* if */
          } /* if */
        else
          {
            WARNING_P_NUM( "find_TP_connection_num", P_numname );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_TP_connection_num" );
      } /* else */

    return ( NULL );

  } /* find_TP_connection_num() */


/*****************************************************************************/
/*                         Matrixinterne Funktionen                          */
/*****************************************************************************/

/*****************************************************************************/
/*********************************** find_T **********************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition find_T( LMPNet net, char * T_name )

#else

Transition find_T( net, T_name )
  /* liefert Pointer auf Knoten mit T_name in der ersten Zeile.              */
  /* Ergebnis ist NULL-Pointer, wenn: Keine Transitionenliste in der Matrix  */
  /*                                                              existiert  */
  /*                                  Kein Knoten mit T_name in der ersten   */
  /*                                                        Zeile existiert  */

  LMPNet net;
  char   * T_name;    /* String fuer Name der Transition */

#endif

  {

    register Transition tr = NULL;

    if ( net && T_name && *T_name)
      {
        if ( tr = net->Transitions )
          {
            while ( tr && Kleiner( tr->name, T_name ) )
              {
                tr = tr->nexttrans;
              } /* while */
            if ( ! ( tr && Gleich( tr->name, T_name ) ) )
              {
/*
                WARNING_T( "find_T", T_name );
*/
                tr = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_T( "find_T" );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_T" );
      } /* else */

    return ( tr );

  } /* find_T() */


/*****************************************************************************/
/*********************************** find_P **********************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place find_P( LMPNet net, char * P_name )

#else

Place find_P( net, P_name )
  /* liefert Pointer auf Stellenknoten mit P_name, also Knoten in erster     */
  /* Spalte der Matrix.                                                      */
  /* Ergebnis ist NULL-Pointer, wenn: keine Stellen im Netz vorhanden sind,  */
  /*                                          also die erste Spalte leer ist */
  /*                                  kein Stellenknoten mit Namen P_name in */
  /*                                                 der Liste vorhanden ist */

  LMPNet net;
  char   * P_name;

#endif

  {

    register Place pl = NULL;

    if ( net && P_name && *P_name )
      {
        if ( pl = net->Places )
          {
            while ( pl && Kleiner( pl->name, P_name ) )
              {
                pl = pl->nextplace;
              } /* while */
            if ( ! ( pl && Gleich( pl->name, P_name ) ) )
              {
/*
                WARNING_P( "find_P", P_name );
*/
                pl = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_P( "find_P" );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_P" );
      } /* else */

    return ( pl );

  } /* find_P() */


/*****************************************************************************/
/***************************** find_PT_connection ****************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge find_PT_connection( LMPNet net, char * P_name, char * T_name )

#else

STEdge find_PT_connection( net, P_name, T_name )
  /* liefert Pointer auf Verbindungsknoten fuer die Verbindung von Stelle    */
  /* 'P_name' nach Transition 'T_name'.                                      */
  /* Ergebnis ist NULL-Pointer, wenn: P_name oder T_name Stellen- bzw.       */
  /* Transitionsknoten sind oder kein solcher Verbindungsknoten vorhanden ist*/

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    register Place  pl;
    register STEdge st;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            if ( st = pl->postset )
              {
                while ( st && Kleiner( st->trans->name, T_name ) )
                  {
                    st = st->nexttrans;
                  } /* while */
                if ( st && Gleich( T_name, st->trans->name ) )
                  {
                    return ( st );
                  } /* if */
              } /* if */
          } /* if */
        else
          {
            WARNING_P( "find_PT_connection", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_PT_connection" );
      } /* else */

    return ( NULL );

  } /* find_PT_connection() */


/*****************************************************************************/
/***************************** find_TP_connection ****************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge find_TP_connection( LMPNet net, char * P_name, char * T_name )

#else

TSEdge find_TP_connection( net, P_name, T_name )
  /* liefert Pointer auf Verbindungsknoten fuer die Verbindung von           */
  /* Transition 'T_name' nach Stelle 'P_name'.                               */
  /* Ergebnis ist NULL-Pointer, wenn: P_name oder T_name Stellen- bzw.       */
  /* Transitionsknoten sind oder kein solcher Verbindungsknoten vorhanden ist*/

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    register Place  pl;
    register TSEdge ts;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            if ( ts = pl->preset )
              {
                while ( ts && Kleiner( ts->trans->name, T_name ) )
                  {
                    ts = ts->nexttrans;
                  } /* while */
                if ( ts && Gleich( T_name, ts->trans->name ) )
                  {
                    return ( ts );
                  } /* if */
              } /* if */
          } /* if */
        else
          {
            WARNING_P( "find_TP_connection", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "find_TP_connection" );
      } /* else */

    return ( NULL );

  } /* find_TP_connection() */


/*****************************************************************************/
/************************************ next_place *****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place next_place( LMPNet net, Place pl )

#else

Place next_place( net, pl )
  /* liefert einen Zeiger auf einen Knoten fuer eine Stelle.                 */
  /* Falls pl ein nil-zeiger ist, so wird ein Zeiger auf das erste     */
  /* Element der Stellenliste zurueckgegeben, andernfalls wird ein Zeiger    */
  /* auf den Nachfolgerknoten von dem Knoten, auf den pl zeigt,        */
  /* geliefert.                                                              */

  LMPNet net;
  Place  pl;

#endif

  {

    if ( net )
      {
        if ( net->Places )
          {
            return ( pl ? pl->nextplace : net->Places );
          } /* if */
        else
          {
            WARNING_NO_P( "next_place" );
            return ( NULL );
          } /* else */
      } /* if */   
    else
      {
        WARNING_NO_NET( "next_place" );
        return( NULL );
      } /* else */

  } /* next_place() */


/*****************************************************************************/
/****************************** next_transition ******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition next_transition( LMPNet net, Transition tr )

#else

Transition next_transition( net, tr )
  /* liefert einen Zeiger auf einen Knoten fuer eine Transition.             */
  /* Falls tr ein nil-zeiger ist, so wird ein Zeiger auf das erste     */
  /* Element der Transitionsliste zurueckgegeben, andernfalls wird ein       */
  /* Zeiger auf den Nachfolgerknoten von dem Knoten, auf den 'tr'      */
  /* zeigt, geliefert.                                                       */

  LMPNet     net;
  Transition tr;

#endif

  {

    if ( net )
      {
        if ( net->Transitions )
          {
            return ( tr ? tr->nexttrans : net->Transitions );
          } /* if */
        else
          {
            WARNING_NO_T( "next_transition" );
            return( NULL );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "next_transition" );
        return( NULL );
      } /* else */

  } /* next_transition() */


/*****************************************************************************/
/***************************** next_PT_connection ****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/
/* Wofuer ist diese Funktion eigentlich gut ???                              */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge next_PT_connection( LMPNet net, STEdge st )

#else

STEdge next_PT_connection( net, st )
  /* liefert einen Pointer auf einen S-T-Verbindungsknoten zurueck.          */
  /* Ist 'st' ein nil-Zeiger, so wird ein Zeiger auf den ersten       */
  /* Verbindungsknoten zurueckgegeben, sonst die Adresse des rechten         */
  /* Nachbarn. Existiert dieser nicht ( Zeile zuende ), so wird die Adresse  */
  /* des ersten Verbindungsknotens in der naechsten Zeile geliefert.         */

  LMPNet net;
  STEdge st;

#endif

  {

    register Place pl;

    if ( net )
      {
        if ( st )
          {
            if ( st->nexttrans )
              {
                return ( st->nexttrans );
              } /* if */
            pl = st->place->nextplace;
            while ( pl && ! pl->postset )
              {
                pl = pl->nextplace;
              } /* while */
            return ( pl ? pl->postset : NULL );
          } /* if */
        else
          {
            if ( pl = net->Places )
              {
                while ( pl && ! pl->postset )
                  {
                    pl = pl->nextplace;
                  } /* while */
                return ( pl ? pl->postset : NULL );
              } /* if */
            else
              {
                WARNING_NO_P( "next_PT_connection" );
              } /* else */
          }
      } /* if */
    else
      {
        WARNING_NO_NET( "next_PT_connection" );
        return ( NULL );
      } /* else */

    return ( NULL );

  } /* next_PT_connection() */

      
/*****************************************************************************/
/***************************** next_TP_connection ****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/
/* Wofuer ist diese Funktion eigentlich gut ???                              */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge next_TP_connection( LMPNet net, TSEdge ts )

#else

TSEdge next_TP_connection( net, ts )
  /* liefert einen Pointer auf einen T-S-Verbindungsknoten zurueck.          */
  /* Ist 'ts' ein nil-Zeiger, so wird ein Zeiger auf den ersten       */
  /* Verbindungsknoten zurueckgegeben, sonst die Adresse des rechten         */
  /* Nachbarn. Existiert dieser nicht ( Zeile zuende ), so wird die Adresse  */
  /* des ersten Verbindungsknotens in der naechsten Zeile geliefert.         */

  LMPNet net;
  TSEdge ts;

#endif

  {

    register Place pl;

    if ( net )
      {
        if ( ts )
          {
            if ( ts->nexttrans )
              {
                return ( ts->nexttrans );
              } /* if */
            pl = ts->place->nextplace;
            while ( pl && ! pl->preset )
              {
                pl = pl->nextplace;
              } /* while */
            return ( pl ? pl->preset : NULL );
          } /* if */
        else
          {
            if ( pl = net->Places )
              {
                while ( pl && ! pl->preset )
                  {
                    pl = pl->nextplace;
                  } /* while */
                return ( pl ? pl->preset : NULL );
              } /* if */
            else
              {
                WARNING_NO_P( "next_TP_connection" );
              } /* else */
          }
      } /* if */
    else
      {
        WARNING_NO_NET( "next_PT_connection" );
        return ( NULL );
      } /* else */

    return ( NULL );

  } /* next_PT_connection() */

      
/*****************************************************************************/
/********************************** row_index ********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 row_index( LMPNet net, char * P_name )

#else

Cardinal32 row_index( net, P_name )
  /* liefert die Zeilennummer der Matrix, in der 'P_name' als Stellenname im */
  /* ersten Knoten (0-te Spalte) steht.                                      */

  LMPNet net;
  char   * P_name;

#endif

  {

    register Place      pl;
    register Cardinal32 index = 1;

    if ( net && P_name && *P_name )
      {
        if ( pl = net->Places )
          {
            while ( pl && Kleiner( pl->name, P_name ) )
              {
                index++;
                pl = pl->nextplace;
              } /* while */
            if ( ! pl || ! Gleich( pl->name, P_name ) )
              {
                WARNING_P( "row_index", P_name );
                index = 0;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_P( "row_index" );
            index = 0;
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "row_index" );
        index = 0;
      } /* else */

    return ( index );

  } /* row_index() */


/*****************************************************************************/
/******************************** column_index *******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 column_index( LMPNet net, char * T_name )

#else

Cardinal32 column_index( net, T_name )
  /* liefert die Spaltennummer, in der 'T_name' als Transitionsname steht.   */
  /* Ist 'T_name' kein Transitionsname, so wird 0 zurueckgeliefert.          */

  LMPNet net;
  char   * T_name;    /* String fuer Name der Transition */

#endif

  {

    register Transition tr;
    register Cardinal32 index = 1;

    if ( net && T_name && *T_name )
      {
        if ( tr = net->Transitions )
          {
            while ( tr && Kleiner( tr->name, T_name ) )
              {
                index++;
                tr = tr->nexttrans;
              } /* while */
            if ( ! tr || ! Gleich( tr->name, T_name ) )
              {
                WARNING_T( "column_index", T_name );
                index = 0;
              } /* if */
          } /* if */
        else
          {
            WARNING_NO_T( "column_index" );
            index = 0;
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "column_index" );
        index = 0;
      } /* else */

    return ( index );

  } /* column_index() */


/*****************************************************************************/
/*                         Initialisierungsfunktionen                        */
/*****************************************************************************/

/*****************************************************************************/
/******************************** empty_net **********************************/
/*****************************************************************************/
/* NEUE FUNKTION: 06.05.1993 Stefan Roemer                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet empty_net( void )

#else

LMPNet empty_net()
  /* Diese Funktion richtet ein neues Netz ( genauer: nur das Kopfelement )  */
  /* ohne Stellen und ohnen Transitionen ein und gibt einen Zeiger darauf    */
  /* als Funktionswert zurueck.                                              */

#endif

  {

    register LMPNet net;

    net = (LMPNet) Malloc( sizeof( LMPNetHead ) );
    net->Places = net->LastPlace = NULL;
    net->Transitions = net->LastTransition = NULL;
    net->NumPlaces = net->NumTransitions = net->NumComponents = 0;

    return ( net );

  } /* empty_net() */


/*###########################################################################*/
/*###          EINIGE HILFSFUNKTIONEN ZUM LESEN VON TEXTDATEIEN           ###*/
/*###########################################################################*/
/* AENDERUNG : 14.02.1994 Stefan Roemer                                      */
/*###########################################################################*/

/*****************************************************************************/
/* Globale Konstante: aktuelle Zeilennummer in der Eingabedatei.             */
/*****************************************************************************/

char * NNinput_file;
int  NNinput_line;


/*****************************************************************************/
/* Programmabbruch wegen fehlerhafter Eingabedatei.                          */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

extern void NNError( char * msg )

#else

extern void NNError( msg )

  char * msg;

#endif

  {

    fprintf( stderr, "ERROR: syntax error in file '%s', line %d: %s.\n",
                     NNinput_file, NNinput_line, msg );
    exit ( 10 );

  } /* NNError() */


/*****************************************************************************/
/* Ueberlesen von 'white spaces'; Rueckgabe ist das erste Zeichen nach dem   */
/* Freiraum.                                                                 */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

char ReadWhiteSpace( FILE * file )

#else

char ReadWhiteSpace( file )

  FILE * file;

#endif

  {

    register char ch;

    while ( ! feof( file ) && isspace( ch = getc( file ) ) )
      {
        if ( ch == '\n' )
          {
            NNinput_line++;
          } /* if */
      } /* while */

    return ( ch );

  } /* ReadWhiteSpace() */


/*****************************************************************************/
/* Ueberlesen von 'white spaces' und Kommentarzeilen.                        */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

char ReadCharComment( FILE * file )

#else

char ReadCharComment( file )

  FILE * file;

#endif

  {

    register char ch;

    do {
         while ( ! feof( file ) && strchr( " \t\b", ch = getc( file ) ) )
           skip;
         if ( ch == '%' )
           {
             while ( ! feof( file ) && ( ( ch = getc( file ) ) != '\n' ) )
               skip;
             if ( ch == '\n' )
               {
                 NNinput_line++;
                 ch = '\0';
               } /* if */
           } /* if */
       } while ( ! ch );

    return ( ch );

  } /* ReadCharComment() */


/*****************************************************************************/
/* Ueberlesen aller Zeichen einer Textdatei bis zum Auftreten eines Zeilen-  */
/* endes ( inklusive ).                                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void ReadNewline( FILE * file, unsigned char count )

#else

void ReadNewline( file, count )

  FILE          * file;
  unsigned char count;

#endif

  {

    while ( count-- )
      {
        while ( ! feof( file ) && ( getc( file ) != '\n' ) )
          skip;
        NNinput_line++;
      } /* while ( count-- ) */

  } /* ReadNewline() */


/*****************************************************************************/
/* Testen, ob das naechste Zeichen einer Textdatei mit `symbol` ueberein-    */
/* stimmt; fuehrende Leerzeichen werden dabei ueberlesen. Bei Nichtueberein- */
/* stimmung erfolgt eine Fehlermeldung ( Programmabbruch ).                  */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool TestSymbol( FILE * file, char symbol )

#else

bool TestSymbol( file, symbol )

  FILE * file;
  char symbol;

#endif

  {

    register char ch;

    if ( ( ch = ReadWhiteSpace( file ) ) == symbol )
      {
        return ( True );
      } /* if */
    else
      {
        ungetc( ch, file );
        return ( False );
      } /* else */

  } /* TestSymbol() */


/*****************************************************************************/
/* Testen, ob das naechste Zeichen einer Textdatei mit `symbol` ueberein-    */
/* stimmt; fuehrende Leerzeichen werden dabei ueberlesen. Bei Nichtueberein- */
/* stimmung erfolgt eine Fehlermeldung ( Programmabbruch ).                  */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void ReadSymbol( FILE * file, char symbol )

#else

void ReadSymbol( file, symbol )

  FILE * file;
  char symbol;

#endif

  {

    register char ch;

    if ( ( ch = ReadWhiteSpace( file ) ) != symbol )
      {
        char msg[100];
        sprintf( msg, "'%c' expected, but '%c' was found", symbol, ch );
        NNError( msg );
      } /* if */

  } /* ReadSymbol() */


/*****************************************************************************/
/* Einlesen eines Namens aus einer Textdatei.                                */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

unsigned char ReadName( FILE * file, char * string, int maxlen )

#else

unsigned char ReadName( file, string, maxlen )

  FILE * file;
  char * string;
  int  maxlen;

#endif

  {

    register unsigned char len = 1;

    *string = ReadWhiteSpace( file );
    while(    ! isspace( *++string = getc( file ) )
           && ( *string != ',' ) && ( (int) ++len < maxlen ) )
      skip;
    if ( len == maxlen )
      {
        while( ! isspace( getc( file ) ) )
          skip;
      } /* if */
    if ( *string == ',' )
      {
        ungetc( *string, file );
      } /* if */
    *string = '\0';
    
    return ( len );

  } /* ReadName() */


/*****************************************************************************/
/* Lesen einer Zahl ( im Dezimalsystem ) aus einer Textdatei; fuehrende      */
/* Leerzeichen und Minuszeichen werden ueberlesen, d.h. es wird nur der      */
/* Absolutbetrag der gelesenen Zahl ermittelt.                               */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

int ReadNumber( FILE * file )

#else

int ReadNumber( file )

  FILE * file;

#endif

  {

    register int  number;
    register char digit;
    register int  vorz = 1;

    if ( ( digit = ReadWhiteSpace( file ) ) == '-' )
      {
        vorz = -1;
        digit = ReadWhiteSpace( file );
      } /* if */

    if ( isdigit( digit ) )
      {
        number = digit - '0';
      } /* if */
    else
      {
        NNError( "digit expected" );
      } /* else */

    while ( isdigit( digit = getc( file ) ) )
      {
        number = number * 10 + digit - '0';
      } /* while */

    if ( digit != ' ' )
      {
        ungetc( digit, file );
      } /* if */

    return ( vorz * number );

  } /* ReadNumber() */


/*###########################################################################*/
/*###                 BESCHLEUNIGTE EINFUEGE-OPERATIONEN                  ###*/
/*###########################################################################*/
/* AENDERUNG : 21.02.1994 Stefan Roemer                                      */
/* AENDERUNG : 31.01.1998 Stefan Roemer                                      */
/*###########################################################################*/

/*****************************************************************************/
/****************************** fast_find_P_num ******************************/
/*****************************************************************************/
/* NEU: 31.01.1998 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place fast_find_P_num( LMPNet net, Cardinal32 P_numname, Place * pl_last )

#else

Place fast_find_P_num( net, P_numnamem pl_last )
  /* liefert Pointer auf Stellenknoten mit P_numname, also Knoten in erster  */
  /* Spalte der Matrix.                                                      */
  /* Ergebnis ist NULL-Pointer, wenn: keine Stellen im Netz vorhanden sind,  */
  /*                                          also die erste Spalte leer ist */
  /*                                  kein Stellenknoten mit Namen P_numname */
  /*                                              in der Liste vorhanden ist */

  LMPNet     net;
  Cardinal32 P_numname;
  Place      * pl_last;

#endif

  {

    register Place pl = NULL;

    if ( net )
      {
        if ( pl = *pl_last )
          {
            while ( pl && ( pl->numname < P_numname ) )
              {
                pl = ( *pl_last = pl )->nextplace;
              } /* while */
            while ( pl && ( pl->numname > P_numname ) )
              {
                *pl_last = pl = pl->prevplace;
              } /* while */
            if ( pl && ( pl->numname == P_numname ) )
              {
                *pl_last = pl;
              } /* if */
            else
              {
                pl = NULL;
              } /* if */
          } /* if */
        else if ( pl = net->Places )
          {
            while ( pl && ( pl->numname < P_numname ) )
              {
                pl = ( *pl_last = pl )->nextplace;
              } /* while */
            if ( ! ( pl && ( pl->numname == P_numname ) ) )
              {
                pl = NULL;
              } /* if */
          } /* else if */
      } /* if */
    else
      {
        WARNING_NO_NET( "fast_find_P_num" );
      } /* else */

    return ( pl );

  } /* fast_find_P_num() */


/*****************************************************************************/
/****************************** fast_find_T_num ******************************/
/*****************************************************************************/
/* NEU: 31.01.1998 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition fast_find_T_num( LMPNet     net,
                            Cardinal32 T_numname,
                            Transition * tr_last )

#else

Transition fast_find_T_num( net, T_numname, tr_last )
  /* liefert Pointer auf Knoten mit T_name in der ersten Zeile.              */
  /* Ergebnis ist NULL-Pointer, wenn: Keine Transitionenliste in der Matrix  */
  /*                                                              existiert  */
  /*                                  Kein Knoten mit T_name in der ersten   */
  /*                                                        Zeile existiert  */

  LMPNet     net;
  Cardinal32 T_numname;
  Transition * tr_last;

#endif

  {

    register Transition tr = NULL;

    if ( net )
      {
        if ( tr = *tr_last )
          {
            while ( tr && ( tr->numname < T_numname ) )
              {
                tr = ( *tr_last = tr )->nexttrans;
              } /* while */
            while ( tr && ( tr->numname > T_numname ) )
              {
                *tr_last = tr = tr->prevtrans;
              } /* while */
            if ( tr && ( tr->numname == T_numname ) )
              {
                *tr_last = tr;
              } /* if */
            else
              {
                tr = NULL;
              } /* if */
          } /* if */
        else if ( tr = net->Transitions )
          {
            while ( tr && ( tr->numname < T_numname ) )
              {
                tr = ( *tr_last = tr )->nexttrans;
              } /* while */
            if ( ! ( tr && ( tr->numname == T_numname ) ) )
              {
                tr = NULL;
              } /* if */
          } /* else if */
      } /* if */
    else
      {
        WARNING_NO_NET( "fast_find_T_num" );
      } /* else */

    return ( tr );

  } /* fast_find_T_num() */


/*****************************************************************************/
/************************** fast_insert_place ********************************/
/*****************************************************************************/
/* NEU EINGEFUEGT : 21.02.1994, Stefan Roemer                                */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place fast_insert_place( LMPNet net, char * P_name, Cardinal32 marks,
                         Place * pl_last )

#else

Place fast_insert_place( net, P_name, marks, pl_last )
  /* setzt einen Knoten fuer die Stelle mit Namen P_name in die Stellenliste */
  /* an lexikographisch richtiger Stelle ein.                                */

  LMPNet      net;
  char        * P_name;
  Cardinal32  marks;
  Place       * pl_last;

#endif

  {

    register Place pl,
                   plnew;

    if ( ! ( pl = *pl_last ) || ! Kleiner( pl->name, P_name ) )
      {
        pl = net->Places;
      } /* if */
    while ( pl && Kleiner( pl->name, P_name ) )
      {
        pl = pl->nextplace;
      } /* while */
    if ( pl && Gleich( P_name, pl->name ) )
      {
        if ( NN_ambigious_nodes == 1 )
          {
            fprintf( stderr, "ERROR: place '%s' already exists.\n",
                     P_name );
            exit ( 1 );
          } /* if */
        else if ( NN_ambigious_nodes == 2 )
          {
            fprintf( stderr, "WARNING: place '%s' already exists.\n",
                     P_name );
          } /* else if */
        return ( *pl_last = pl );
      } /* if */
    plnew = (Place) Malloc( sizeof( PlElem ) );
    if ( pl )
      {
        if ( plnew->prevplace = pl->prevplace )
          {
            plnew->prevplace->nextplace = plnew;
          } /* if */
        else
          {
            net->Places = plnew;
          } /* else */
        ( plnew->nextplace = pl ) -> prevplace = plnew;
      } /* if */
    else
      {
        if ( plnew->prevplace = net->LastPlace )
          {
            net->LastPlace->nextplace = plnew;
          } /* if */
        else
          {
            net->Places = plnew;
          } /* else */
        ( net->LastPlace = plnew ) -> nextplace = NULL;
      } /* else */
    plnew->name = Strdup( P_name );
    plnew->component = plnew->numname = 0;
    plnew->preset = NULL;
    plnew->postset = NULL;
    plnew->label = EMPTY;
    plnew->marks = marks;
    net->NumPlaces++;

    return ( *pl_last = plnew );

  } /* fast_insert_place() */


/*****************************************************************************/
/************************* fast_insert_transition ****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition fast_insert_transition( LMPNet net, char * T_name,
                                   Transition * tr_last )

#else

Transition fast_insert_transition( net, T_name, tr_last )
  /* Setzt einen neuen Knoten fuer eine Transition an der lexikographisch    */
  /* richtigen Stelle in die Matrix ein, also in die erste Zeile.            */

  LMPNet     net;
  char       * T_name;
  Transition * tr_last;

#endif

  {

    register Transition tr,
                        trnew;

    if ( ! ( tr = *tr_last ) || ! Kleiner( tr->name, T_name ) )
      {
        tr = net->Transitions;
      } /* if */
    while ( tr && Kleiner( tr->name, T_name ) )
      {
        tr = tr->nexttrans;
      } /* while */
    if ( tr && Gleich( T_name, tr->name ) )
      {
        if ( NN_ambigious_nodes == 1 )
          {
            fprintf( stderr, "ERROR: transition '%s' already exists.\n",
                     T_name );
            exit ( 1 );
          } /* if */
        else if ( NN_ambigious_nodes == 2 )
          {
            fprintf( stderr, "WARNING: transition '%s' already exists.\n",
                     T_name );
          } /* else if */
        return ( *tr_last = tr );
      } /* if */
    trnew = (Transition) Malloc( sizeof( TrElem ) );
    if ( tr )
      {
        if ( trnew->prevtrans = tr->prevtrans )
          {
            trnew->prevtrans->nexttrans = trnew;
          } /* if */
        else
          {
            net->Transitions = trnew;
          } /* else */
        ( trnew->nexttrans = tr ) -> prevtrans = trnew;
      } /* if */
    else
      {
        if ( trnew->prevtrans = net->LastTransition )
          {
            net->LastTransition->nexttrans = trnew;
          } /* if */
        else
          {
            net->Transitions = trnew;
          } /* else */
        ( net->LastTransition = trnew ) -> nexttrans = NULL;
      } /* else */
    trnew->name = Strdup( T_name );
    trnew->numname = 0;
    trnew->label = NULL;
    trnew->preset = NULL;
    trnew->postset = NULL;
    net->NumTransitions++;

    return ( *tr_last = trnew );

  } /* fast_insert_transition() */


/*****************************************************************************/
/*************************** fast_insert_TP_con ******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge fast_insert_TP_con( LMPNet net, char * P_name, char * T_name,
                           Cardinal32 tsweight, Place * pl_last,
                           Transition * tr_last, TSEdge * ts_last )

#else

TSEdge fast_insert_TP_con( net, P_name, T_name, tsweight,
                           pl_last , tr_last , ts_last )
  /* setzt einen Verbindungsknoten mit dem Gewicht tsweight in die Matrix ein; */
  /* falls ein solcher schon vorhanden, wird das Gewicht entsprechend        */
  /* geaendert.                                                              */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 tsweight;
  Place      * pl_last;
  Transition * tr_last;
  TSEdge     * ts_last;

#endif

  {

    register Place      pl;
    register Transition tr;
    register TSEdge     ts,
                        tsnew;
    register bool       kl1,
                        kl2;

    pl = *pl_last;
    if ( ( ts = *ts_last ) && Gleich( ts->place->name, P_name ) )
      {
        pl = ts->place;
      } /* if */
    else if ( pl && Gleich( pl->name, P_name ) )
      {
        ts = NULL;
      } /* else if */
    else if ( pl && ts && ( ( kl1 = Kleiner( pl->name, P_name ) ),
                            ( kl2 = Kleiner( ts->place->name, P_name ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && Kleiner( pl->name, ts->place->name ) ) )
          {
            pl = ts->place;
          } /* if */
        else
          {
            ts = NULL;
          } /* else */
      } /* else if */
    else
      {
        pl = net->Places;
        ts = NULL;
      } /* else */
    while ( pl && Kleiner( pl->name, P_name ) )
      {
ts = NULL;
        pl = pl->nextplace;
      } /* while */

    if ( ! ( pl && Gleich( pl->name, P_name ) ) )
      {
        WARNING_P( "fast_insert_TP_con", P_name );
        return ( NULL );
      } /* if */

    if ( ( ts && Kleiner( ts->trans->name, T_name ) ) || ( ts = pl->preset ) )
      {
        if ( Kleiner( T_name, ts->trans->name ) )
          {
            ts = NULL;
          } /* if */
        else
          {
            while ( ts->nexttrans
                    && Kleiner( ts->nexttrans->trans->name, T_name ) )
              {
                ts = ts->nexttrans;
              } /* while */
            if ( ( tsnew = ts->nexttrans )
                 && Gleich( tsnew->trans->name, T_name ) )
              {
                *pl_last = tsnew->place;
                *tr_last = tsnew->trans;
                if ( ! ( tsnew->weight = tsweight ) )
                  {
                    delete_T_to_P_con_node( net, tsnew );
                    return ( NULL );
                  } /* if */
                else
                  {
                    return ( *ts_last = tsnew );
                  } /* else */
              } /* if */
          } /* else */
      } /* if */
    tr = ( *tr_last && Gleich( (*tr_last)->name, T_name ) ) ? *tr_last
                                           : ts ? ts->trans : net->Transitions;
    while ( tr && Kleiner( tr->name, T_name ) )
      {
        tr = tr->nexttrans;
      } /* while */
    if ( ! ( tr && Gleich( tr->name, T_name ) ) )
      {
        WARNING_T( "fast_insert_TP_con", T_name );
        return ( NULL );
      } /* if */

    tsnew = (TSEdge) Malloc( sizeof( TSElem ) );
    if ( ts )
      {
        if ( tsnew->nexttrans = ts->nexttrans )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          } /* if */
        ( tsnew->prevtrans = ts ) -> nexttrans = tsnew;
      } /* if */
    else
      {
        if ( tsnew->nexttrans = pl->preset )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          } /* if */
        ( pl->preset = tsnew ) -> prevtrans = NULL;
      } /* else */
    if ( ts = tr->postset )
      {
        if ( Kleiner( P_name, ts->place->name ) )
          {
            ts = NULL;
          } /* if */
        else
          {
            while ( ts->nextplace
                    && Kleiner( ts->nextplace->place->name, P_name ) )
              {
                ts = ts->nextplace;
              } /* while */
          } /* else */
      } /* if */
    if ( ts )
      {
        if ( tsnew->nextplace = ts->nextplace )
          {
            tsnew->nextplace->prevplace = tsnew;
          } /* if */
        ( tsnew->prevplace = ts ) -> nextplace = tsnew;
      } /* if */
    else
      {
        if ( tsnew->nextplace = tr->postset )
          {
            tsnew->nextplace->prevplace = tsnew;
          } /* if */
        ( tr->postset = tsnew ) -> prevplace = NULL;
      } /* else */
    tsnew->weight = tsweight;
    *pl_last = tsnew->place = pl;
    *tr_last = tsnew->trans = tr;

    return ( *ts_last = tsnew );

  } /* fast_insert_TP_con() */


/*****************************************************************************/
/**************************** fast_insert_PT_con *****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge fast_insert_PT_con( LMPNet net, char * P_name, char * T_name,
                           Cardinal32 stweight, Place * pl_last,
                           Transition * tr_last, STEdge * st_last )

#else

STEdge fast_insert_PT_con( net, P_name, T_name, stweight,
                           pl_last, tr_last, st_last )
  /* setzt einen Verbindungsknoten mit dem Gewicht stweight in die Matrix    */
  /* ein; falls ein solcher schon vorhanden, wird das Gewicht entsprechend   */
  /* geaendert.                                                              */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 stweight;                  /* Anzahl der Verbindungen von einer */
                                        /* Stelle zu einer Transition        */
  Place      * pl_last;
  Transition * tr_last;
  STEdge     * st_last;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st,
                        stnew;
    register bool       kl1,
                        kl2;

    pl = *pl_last;
    if ( ( st = *st_last ) && Gleich( st->place->name, P_name ) )
      {
        pl = st->place;
      } /* if */
    else if ( pl && Gleich( pl->name, P_name ) )
      {
        st = NULL;
      } /* else if */
    else if ( pl && st && ( ( kl1 = Kleiner( pl->name, P_name ) ),
                            ( kl2 = Kleiner( st->place->name, P_name ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && Kleiner( pl->name, st->place->name ) ) )
          {
            pl = st->place;
          } /* if */
        else
          {
            st = NULL;
          } /* else */
      } /* else if */
    else
      {
        pl = net->Places;
        st = NULL;
      } /* else */
    while ( pl && Kleiner( pl->name, P_name ) )
      {
st = NULL;
        pl = pl->nextplace;
      } /* while */

    if ( ! ( pl && Gleich( pl->name, P_name ) ) )
      {
        WARNING_P( "fast_insert_PT_con", P_name );
        return ( NULL );
      } /* if */

    if ( ( st && Kleiner( st->trans->name, T_name ) ) || ( st = pl->postset ) )
      {
        if ( Kleiner( T_name, st->trans->name ) )
          {
            st = NULL;
          } /* if */
        else
          {
            while ( st->nexttrans
                    && Kleiner( st->nexttrans->trans->name, T_name ) )
              {
                st = st->nexttrans;
              } /* while */
            if ( ( stnew = st->nexttrans )
                 && Gleich( stnew->trans->name, T_name ) )
              {
                *pl_last = stnew->place;
                *tr_last = stnew->trans;
                if ( ! ( stnew->weight = stweight ) )
                  {
                    delete_P_to_T_con_node( net, stnew );
                    return ( NULL );
                  } /* if */
                else
                  {
                    return ( *st_last = stnew );
                  } /* else */
              } /* if */
          } /* else */
      } /* if */
    tr = ( *tr_last && Gleich( (*tr_last)->name, T_name ) ) ? *tr_last
                                           : st ? st->trans : net->Transitions;
    while ( tr && Kleiner( tr->name, T_name ) )
      {
        tr = tr->nexttrans;
      } /* while */
    if ( ! ( tr && Gleich( tr->name, T_name ) ) )
      {
        WARNING_T( "fast_insert_PT_con", T_name );
        return ( NULL );
      } /* if */

    stnew = (STEdge) Malloc( sizeof( STElem ) );
    if ( st )
      {
        if ( stnew->nexttrans = st->nexttrans )
          {
            stnew->nexttrans->prevtrans = stnew;
          } /* if */
        ( stnew->prevtrans = st ) -> nexttrans = stnew;
      } /* if */
    else
      {
        if ( stnew->nexttrans = pl->postset )
          {
            stnew->nexttrans->prevtrans = stnew;
          } /* if */
        ( pl->postset = stnew ) -> prevtrans = NULL;
      } /* else */
    if ( st = tr->preset )
      {
        if ( Kleiner( P_name, st->place->name ) )
          {
            st = NULL;
          } /* if */
        else
          {
            while ( st->nextplace
                    && Kleiner( st->nextplace->place->name, P_name ) )
              {
                st = st->nextplace;
              } /* while */
          } /* else */
      } /* if */
    if ( st )
      {
        if ( stnew->nextplace = st->nextplace )
          {
            stnew->nextplace->prevplace = stnew;
          } /* if */
        ( stnew->prevplace = st ) -> nextplace = stnew;
      } /* if */
    else
      {
        if ( stnew->nextplace = tr->preset )
          {
            stnew->nextplace->prevplace = stnew;
          } /* if */
        ( tr->preset = stnew ) -> prevplace = NULL;
      } /* else */
    stnew->weight = stweight;
    *pl_last = stnew->place = pl;
    *tr_last = stnew->trans = tr;

    return ( *st_last = stnew );

  } /* fast_insert_PT_con() */

/*****************************************************************************/
/*************************** fast_insert_TP_num ******************************/
/*****************************************************************************/
/* NEU: 08.02.1998 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge fast_insert_TP_num( LMPNet net, Cardinal32 P_numname,
                           Cardinal32 T_numname,
                           Cardinal32 tsweight, Place * pl_last,
                           Transition * tr_last, TSEdge * ts_last )

#else

TSEdge fast_insert_TP_num( net, P_numname, T_numname, tsweight,
                           pl_last , tr_last , ts_last )
  /* setzt einen Verbindungsknoten mit dem Gewicht tsweight in die Matrix ein; */
  /* falls ein solcher schon vorhanden, wird das Gewicht entsprechend        */
  /* geaendert.                                                              */

  LMPNet     net;
  Cardinal32 P_numname,
             T_numname,
             tsweight;
  Place      * pl_last;
  Transition * tr_last;
  TSEdge     * ts_last;

#endif

  {

    register Place      pl;
    register Transition tr;
    register TSEdge     ts,
                        tsnew;
    register bool       kl1,
                        kl2;

    pl = *pl_last;
    if ( ( ts = *ts_last ) && ( ts->place->numname == P_numname ) )
      {
        pl = ts->place;
      } /* if */
    else if ( pl && ( pl->numname == P_numname ) )
      {
        ts = NULL;
      } /* else if */
    else if ( pl && ts && ( ( kl1 = ( pl->numname < P_numname ) ),
                            ( kl2 = ( ts->place->numname < P_numname ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && ( pl->numname < ts->place->numname ) ) )
          {
            pl = ts->place;
          } /* if */
        else
          {
            ts = NULL;
          } /* else */
      } /* else if */
    else
      {
        pl = net->Places;
        ts = NULL;
      } /* else */
    while ( pl && ( pl->numname < P_numname ) )
      {
ts = NULL;
        pl = pl->nextplace;
      } /* while */

    if ( ! ( pl && ( pl->numname == P_numname ) ) )
      {
        WARNING_P_N( "fast_insert_TP_num", P_numname );
        return ( NULL );
      } /* if */

    if ( ( ts && ( ts->trans->numname < T_numname ) ) || ( ts = pl->preset ) )
      {
        if ( T_numname < ts->trans->numname )
          {
            ts = NULL;
          } /* if */
        else
          {
            while (    ts->nexttrans
                    && ( ts->nexttrans->trans->numname < T_numname ) )
              {
                ts = ts->nexttrans;
              } /* while */
            if (    ( tsnew = ts->nexttrans )
                 && ( tsnew->trans->numname == T_numname ) )
              {
                *pl_last = tsnew->place;
                *tr_last = tsnew->trans;
                if ( ! ( tsnew->weight = tsweight ) )
                  {
                    delete_T_to_P_con_node( net, tsnew );
                    return ( NULL );
                  } /* if */
                else
                  {
                    return ( *ts_last = tsnew );
                  } /* else */
              } /* if */
          } /* else */
      } /* if */
    tr = ( *tr_last && ( (*tr_last)->numname == T_numname ) ) ? *tr_last
                                           : ts ? ts->trans : net->Transitions;
    while ( tr && ( tr->numname < T_numname ) )
      {
        tr = tr->nexttrans;
      } /* while */
    if ( ! ( tr && ( tr->numname == T_numname ) ) )
      {
        WARNING_T_N( "fast_insert_TP_con", T_numname );
        return ( NULL );
      } /* if */

    tsnew = (TSEdge) Malloc( sizeof( TSElem ) );
    if ( ts )
      {
        if ( tsnew->nexttrans = ts->nexttrans )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          } /* if */
        ( tsnew->prevtrans = ts ) -> nexttrans = tsnew;
      } /* if */
    else
      {
        if ( tsnew->nexttrans = pl->preset )
          {
            tsnew->nexttrans->prevtrans = tsnew;
          } /* if */
        ( pl->preset = tsnew ) -> prevtrans = NULL;
      } /* else */
    if ( ts = tr->postset )
      {
        if ( P_numname < ts->place->numname )
          {
            ts = NULL;
          } /* if */
        else
          {
            while (    ts->nextplace
                    && ( ts->nextplace->place->numname < P_numname ) )
              {
                ts = ts->nextplace;
              } /* while */
          } /* else */
      } /* if */
    if ( ts )
      {
        if ( tsnew->nextplace = ts->nextplace )
          {
            tsnew->nextplace->prevplace = tsnew;
          } /* if */
        ( tsnew->prevplace = ts ) -> nextplace = tsnew;
      } /* if */
    else
      {
        if ( tsnew->nextplace = tr->postset )
          {
            tsnew->nextplace->prevplace = tsnew;
          } /* if */
        ( tr->postset = tsnew ) -> prevplace = NULL;
      } /* else */
    tsnew->weight = tsweight;
    *pl_last = tsnew->place = pl;
    *tr_last = tsnew->trans = tr;

    return ( *ts_last = tsnew );

  } /* fast_insert_TP_num() */


/*****************************************************************************/
/**************************** fast_insert_PT_num *****************************/
/*****************************************************************************/
/* NEU: 08.02.1998 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge fast_insert_PT_num( LMPNet net, Cardinal32 P_numname,
                           Cardinal32 T_numname,
                           Cardinal32 stweight, Place * pl_last,
                           Transition * tr_last, STEdge * st_last )

#else

STEdge fast_insert_PT_num( net, P_numname, T_numname, stweight,
                           pl_last, tr_last, st_last )
  /* setzt einen Verbindungsknoten mit dem Gewicht stweight in die Matrix    */
  /* ein; falls ein solcher schon vorhanden, wird das Gewicht entsprechend   */
  /* geaendert.                                                              */

  LMPNet     net;
  Cardinal32 P_numname,
             T_numname,
             stweight;                  /* Anzahl der Verbindungen von einer */
                                        /* Stelle zu einer Transition        */
  Place      * pl_last;
  Transition * tr_last;
  STEdge     * st_last;

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st,
                        stnew;
    register bool       kl1,
                        kl2;

    pl = *pl_last;
    if ( ( st = *st_last ) && ( st->place->numname == P_numname ) )
      {
        pl = st->place;
      } /* if */
    else if ( pl && ( pl->numname == P_numname ) )
      {
        st = NULL;
      } /* else if */
    else if ( pl && st && ( ( kl1 = ( pl->numname < P_numname ) ),
                            ( kl2 = ( st->place->numname < P_numname ) ),
                            ( kl1 || kl2 ) ) )
      {
        if ( ( ! kl1 ) || ( kl2 && ( pl->numname < st->place->numname ) ) )
          {
            pl = st->place;
          } /* if */
        else
          {
            st = NULL;
          } /* else */
      } /* else if */
    else
      {
        pl = net->Places;
        st = NULL;
      } /* else */
    while ( pl && ( pl->numname < P_numname ) )
      {
st = NULL;
        pl = pl->nextplace;
      } /* while */

    if ( ! ( pl && ( pl->numname == P_numname ) ) )
      {
        WARNING_P_N( "fast_insert_PT_num", P_numname );
        return ( NULL );
      } /* if */

    if ( ( st && ( st->trans->numname < T_numname ) ) || ( st = pl->postset ) )
      {
        if ( T_numname < st->trans->numname )
          {
            st = NULL;
          } /* if */
        else
          {
            while (    st->nexttrans
                    && ( st->nexttrans->trans->numname < T_numname ) )
              {
                st = st->nexttrans;
              } /* while */
            if ( ( stnew = st->nexttrans )
                 && ( stnew->trans->numname == T_numname ) )
              {
                *pl_last = stnew->place;
                *tr_last = stnew->trans;
                if ( ! ( stnew->weight = stweight ) )
                  {
                    delete_P_to_T_con_node( net, stnew );
                    return ( NULL );
                  } /* if */
                else
                  {
                    return ( *st_last = stnew );
                  } /* else */
              } /* if */
          } /* else */
      } /* if */
    tr = ( *tr_last && ( (*tr_last)->numname == T_numname ) ) ? *tr_last
                                           : st ? st->trans : net->Transitions;
    while ( tr && ( tr->numname < T_numname ) )
      {
        tr = tr->nexttrans;
      } /* while */
    if ( ! ( tr && ( tr->numname == T_numname ) ) )
      {
        WARNING_T_N( "fast_insert_PT_num", T_numname );
        return ( NULL );
      } /* if */

    stnew = (STEdge) Malloc( sizeof( STElem ) );
    if ( st )
      {
        if ( stnew->nexttrans = st->nexttrans )
          {
            stnew->nexttrans->prevtrans = stnew;
          } /* if */
        ( stnew->prevtrans = st ) -> nexttrans = stnew;
      } /* if */
    else
      {
        if ( stnew->nexttrans = pl->postset )
          {
            stnew->nexttrans->prevtrans = stnew;
          } /* if */
        ( pl->postset = stnew ) -> prevtrans = NULL;
      } /* else */
    if ( st = tr->preset )
      {
        if ( P_numname < st->place->numname )
          {
            st = NULL;
          } /* if */
        else
          {
            while (    st->nextplace
                    && ( st->nextplace->place->numname < P_numname ) )
              {
                st = st->nextplace;
              } /* while */
          } /* else */
      } /* if */
    if ( st )
      {
        if ( stnew->nextplace = st->nextplace )
          {
            stnew->nextplace->prevplace = stnew;
          } /* if */
        ( stnew->prevplace = st ) -> nextplace = stnew;
      } /* if */
    else
      {
        if ( stnew->nextplace = tr->preset )
          {
            stnew->nextplace->prevplace = stnew;
          } /* if */
        ( tr->preset = stnew ) -> prevplace = NULL;
      } /* else */
    stnew->weight = stweight;
    *pl_last = stnew->place = pl;
    *tr_last = stnew->trans = tr;

    return ( *st_last = stnew );

  } /* fast_insert_PT_num() */


/*****************************************************************************/
/************************** fast_insert_con **********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void fast_insert_con( LMPNet net, char * P_name, char * T_name,
                      Cardinal32 tsweight, Cardinal32 stweight,
                      Place * pl_last, Transition * tr_last,
                      STEdge * st_last, TSEdge * ts_last )

#else

void fast_insert_con( net, P_name, T_name, tsweight, stweight,
                      pl_last, tr_last, st_last, ts_last )
  /* setzt die beiden Verbindungsknoten mit den Gewichten tsweight und stweight  */
  /* in die Matrix ein. Falls solche schon vorhanden sind, werden die        */
  /* Gewichte entsprechend geaendert.                                        */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 tsweight,                    /* Anzahl der Verbindungen von einer */
                                        /* Transition zu einer Stelle        */
             stweight;                    /* Anzahl der Verbindungen von einer */
                                        /* Stelle zu einer Transition        */
  Place      * pl_last;
  Transition * tr_last;
  STEdge     * st_last;
  TSEdge     * ts_last;

#endif

  {

    (void) fast_insert_PT_con( net, P_name, T_name, stweight,
                               pl_last, tr_last, st_last );
    (void) fast_insert_TP_con( net, P_name, T_name, tsweight,
                               pl_last, tr_last, ts_last );

  } /* fast_insert_con() */


/*###########################################################################*/
/*######                    ENDE DER EINFUEGUNGEN                     #######*/
/*###########################################################################*/

/*****************************************************************************/
/******************************** gspn_ini ***********************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet gspn_ini( char * filename )

#else

LMPNet gspn_ini( filename )
  /* Initialisiert eine Matrix mit dem in der Datei filename abgelegten Netz */
  /* Die Datei filename ist im GSPN-Format.                                  */

  char * filename; /* Name der Datei, in der das Netz abgelegt ist */

#endif

  {

    /* Deklarationen */
    FILE *ifp;               /* Pointer auf Eingabedatei (InputFilePointer)  */
    char       line[LINELENGTH],
               name[NAMELENGTH],
               comment[1024];/* String fuer den Kommentar zum Netz */
    int i, j, k;             /* Laufvariablen */
    int trnum;       /* Anzahl der Transitionen */
    int plnum;            /* Anzahl der Stellen (places)  */
    int group_nr;            /* Anzahl der  Groups */
    int rate_nr;             /* Anzahl der  Rates*/
    int mark_nr;             /* Anzahl der  Marks */
    int marking;             /* Nimmt Anfangsmarkierung einer Stelle auf     */
    float TRATE;             /* */
    int TSERV, TKND;         /* */
    int TINP;                /* Anzahl der Verbindungen ZU einer Transition  */
    int TINH;                /* Anzahl Inhibitor Arcs                        */
    int TROT;                /* */
    int TOUT;                /* Anzahl der Verbindungen VON einer Transition */
    int AMULT;               /* Wertigkeit einer Verbindung                  */
    int APLACE;              /* zur Aufnahme des anderen Endes der Verbindung*/
    int APOINTS;             /* Anzahl der Zeilen mit Koordinaten von        */
                             /* Zwischenpunkten einer Verbindung             */
    int LEVELS;              /* */
    LIST_POINTER place_list,
                 help_list,
                 list_old;

    /* Oeffnen der Datei im GSPN-Format */
    if ( ! ( ifp = fopen( filename,"r" ) ) )
      {
        ERROR_FILE_READ( gspn_ini, filename );
      } /* if */

    /* Initialisiere net */
    net = empty_net();

    /*********************** NONTERMINAL COMMENT *****************************/
    for ( i = 1 ; i <= 2; i++ )
      {
        fgets( line, LINELENGTH, ifp );
      } /* end of for i */
    while ( strncmp( line, "|\n", 2 ) )
      {
        strcat( comment, line );
        fgets( line, LINELENGTH, ifp );
      } /* end of while */
    /* Zeilen mit NONTERMINAL COMMENT gelesen und Kommentar in               */
    /* string comment abgelegt                                               */

    /*********************** NONTERMINAL NOOBJS  *****************************/
    fgets( line, LINELENGTH, ifp );
    sscanf( line, " f   %d %d %d %d %d 0\n",
                  &mark_nr, &plnum, &rate_nr, &trnum, &group_nr );

    /*********************** NONTERMINAL MARKS *******************************/
    for ( i=1 ; i <= mark_nr; i++ )
      {
        fgets( line,LINELENGTH, ifp );
      } /* end of for i */
    /* Zeilen mit Nonterminal MARKS ueberlesen */

    /*********************** NONTERMINAL PLACES ******************************/
    /* Baue Stellenliste in die Matrix ein */
    /* unterhalte in place_list eine Liste der tatsaechlichen Namen der      */
    /* Stellen */
    help_list = place_list = (LIST_POINTER) Malloc( sizeof( LIST_ELEMENT ) );
    for ( i = 1 ; i <= plnum ; i++ )
      {
        fgets( line, LINELENGTH, ifp );
        sscanf( line,"%[A-Za-z0-9]   %d \n", name, &marking );
        Strcpy( help_list->name, name );
        help_list->next = (LIST_POINTER) Malloc( sizeof( LIST_ELEMENT ) );
        list_old = help_list;
        help_list = help_list->next;
        insert_place( net, name, marking );
      } /* end of for i */
    /* schliesse Liste ab */
    list_old->next = NULL;
    /* Stellenliste in die Matrix eingebaut, erste Spalte der Matrix steht.  */

    /*********************** NONTERMINAL GROUPS ******************************/
    for ( i=1; i<=group_nr; i++ )
      {
        fgets( line, LINELENGTH, ifp );
      } /* end of for i */
    /* Zeilen mit Nonterminal GROUPS ueberlesen */

    /*********************** NONTERMINAL TRANS *******************************/
    /* Baue Transitionsliste und Verbindungen in die Matrix ein */
    /* Situation: Bild 1. */
    /* Aufbau von hinten, da GSPN die Transitionen umgekehrt ihrer Benennung */
    /* ablegt */
    for ( i=1; i<= trnum; i++ )
      {
        fgets( line, LINELENGTH, ifp );
        sscanf( line, "%[A-Za-z0-9] %f %d %d %d %d  \n",
                      name, &TRATE, &TSERV, &TKND, &TINP, &TROT );
        insert_transition( net, name );
        /* verarbeite die Zeilen fuer die Verbindungen zu dieser Transition */
        for ( j = 1 ; j<= TINP ; j++ )
          {
            fgets( line, LINELENGTH, ifp );
	    sscanf( line, "    %d  %d  %d %d \n",
                          &AMULT, &APLACE, &APOINTS, &LEVELS );
	    /* Konstruiere den Namen der Stelle,von der die Verbindung kommt */
	    help_list = place_list;
	    for ( k = 1 ; k < APLACE ; k++ )
	      {
                help_list = help_list->next;
              } /* end of for k */
	    insert_PT_con( net, help_list->name, name, AMULT );
	    /* Ueberlese die Koordinatenzeilen mit Zwischenpunkten */
	    for ( k = 1 ; k <= APOINTS; k++ )
	      {
	        fgets( line, LINELENGTH, ifp );
              } /* end of for k */
	  } /* end of for j */
        /* Zu-Verbindungen eingebaut */ 
    
        /* Verarbeite die Zeilen fuer die Verbindungen von dieser Transition */
        fgets( line, LINELENGTH, ifp );
        sscanf( line, "%d \n", &TOUT );
        /* TOUT : Anzahl der Von-Verbindungen ( # Output Arcs) */
        for ( j = 1 ; j <= TOUT; j++ )
          {
            fgets( line, LINELENGTH, ifp );
	    sscanf( line, "   %d  %d  %d %d \n",
                          &AMULT, &APLACE, &APOINTS, &LEVELS );
            /* AMULT : Wertigkeit der Verbindung */
	    /* APLACE :Name der Stelle, von der die Verbindung kommt */
	    /* APOINTS : Anzahl der Zwischenpunkte dieser Verbindung */
	    /* Konstruiere Name, von der die Verbindung kommt */
	    help_list = place_list;
	    for ( k = 1 ; k < APLACE ; k++ )
	      {
	        help_list = help_list->next;
              } /* end of for k */
	    insert_TP_con( net, help_list->name, name, AMULT );
        
	    /* Ueberlese die Koordinatenzeilen mit Zwischenpunkten */
            for ( k = 1 ; k <= APOINTS ; k++ )
	      {
                fgets( line, LINELENGTH, ifp );
              } /* end of for k */
          } /* end of for j */
        /* Von-Verbindungen eingebaut */

        /* verarbeite Zeilen mit Anzahl der Inhibitor-Verbindungen */
        fgets( line, LINELENGTH, ifp );
        sscanf( line, "%d \n", &TINH );
        if ( TINH != 0 )
          {
            /* Fehler: es existieren Inhibitor-Verbindungen */
            fprintf( stderr,
                    "WARNING: NN.gspn_ini(): net contains inhibitor arcs.\n" );
            remove_net( &net );
            fclose( ifp );
            return( NULL );
          } /* end of if */

      } /* end of for i */
    fclose( ifp );

    return( net );

  } /* gspn_ini() */


/*****************************************************************************/
/****************************** praktikum_ini ********************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/* AENDERUNG : 14.02.1994 Stefan Roemer: Keine Beschraenkung mehr bzgl.      */
/*                                       der Zeilenlaenge!                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LMPNet praktikum_ini( char * filename )

#else

LMPNet praktikum_ini( filename )
  /* liefert einen Pointer auf eine Matrix, in der das Netz gespeichert ist. */
  /* Ergebnis ist NULL-Pointer, wenn eine entsprechend benannte .inf Datei   */
  /* nicht existiert.                                                        */

  char * filename;

#endif

  {

    register FILE       * ifp;
    char                str[101],
                        name[NAMELENGTH],
                        P_name[NAMELENGTH];
    register Cardinal32 trnum,
                        plnum,
                        tsweight,
                        stweight,
                        i, j;
    register Place      phelp;
    register Transition thelp;
    register bool       parenth;
    Place               pl_last = NULL;
    Transition          tr_last = NULL;
    STEdge              st_last = NULL;
    TSEdge              ts_last = NULL;

    /* Oeffne .inf-Datei */
    if ( ! ( ifp = fopen( filename, "r" ) ) )
      {
        ERROR_FILE_READ( praktikum_ini, filename );
        exit ( 4 );
      } /* if */
    NNinput_line = 1;
    NNinput_file = filename;

    /* Initialisiere net */
    net = empty_net();

    /* Lese Zeile mit Anzahl-Angaben */
    plnum = ReadNumber( ifp );
    trnum = ReadNumber( ifp );

    /* Lese Zeile mit 'Inzidenzmatrix' */
    ReadName( ifp, str, 100 );
    if ( strcmp( str, "Inzidenzmatrix:" ) )
      {
        fclose( ifp );
        NNError( "keyword 'Inzidenzmatrix:' expected" );
      } /* if */

    /* Lese erste Zeile der Matrix */
    for ( i = 1 ; i <= trnum ; i++ )
      {
        ReadName( ifp, name, NAMELENGTH-1 );
        /* setze neue Transition in die Matrix ein */
        fast_insert_transition( net, name, &tr_last );
      } /* end of for i */
    /* erste Zeile verarbeitet, Transitionsliste in der Matrix initialisiert */
    /* verarbeite weitere Zeilen */
    for ( i = 1; i <= plnum ; i++ )
      {
        ReadName( ifp, P_name, NAMELENGTH-1 );
        fast_insert_place( net, P_name, 0, &pl_last );
        ReadSymbol( ifp, '|' );
        /* lege eine Hilfszeiger in die Transitionsliste, um auf die         */
        /*  Transitionsnamen Zugriff zu haben */
        thelp = net->Transitions;
        for ( j = 1 ; j <= trnum ; j++ )
          {
            ReadSymbol( ifp, '(' );
            tsweight = abs( ReadNumber( ifp ) );
            ReadSymbol( ifp, ',' );
            stweight = abs( ReadNumber( ifp ) );
            ReadSymbol( ifp, ')' );
	    if ( stweight )
	      { fast_insert_PT_con( net, P_name, thelp->name, stweight,
                                    &pl_last, &tr_last, &st_last ); }
	    if ( tsweight )
	      { fast_insert_TP_con( net, P_name, thelp->name, tsweight,
                                    &pl_last, &tr_last, &ts_last ); }
	    thelp = thelp->nexttrans;
          } /* end of for j */
      } /* end of for i */
    /* Inzidenzmatrix verarbeitet */

    /* Lese Anfangsmarkierung ein */
    /* Lese Zeile mit 'Startmarkierung' */
    ReadName( ifp, str, 100 );
    if ( strcmp( str, "Startmarkierung:" ) )
      {
        fclose( ifp );
        NNError( "keyword 'Startmarkierung:' expected" );
      } /* if */

    parenth = False;
    if ( TestSymbol( ifp, '(' ) )
      {
        parenth = True;
      } /* if */
    phelp = net->Places;
    phelp->marks = ReadNumber( ifp );
    for ( i = 2 ; i <= plnum; i++ )
      {
        ReadSymbol( ifp, ',' );
        ( phelp = phelp->nextplace ) -> marks = ReadNumber( ifp );
      } /* for */
    if ( parenth && ! TestSymbol( ifp, ')' ) )
      {
        fclose( ifp );
        NNError( "marking vector: ')' expected" );
      } /* if */

    fclose( ifp );

    return( net );

  } /* praktikum_ini() */


/*****************************************************************************/
/*                             Ausgabefunktionen                             */
/*****************************************************************************/

/*****************************************************************************/
/******************************** write_matrix *******************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void output_matrix( LMPNet net, FILE * ofp )

#else

void output_matrix( net, ofp )
  /* Gibt die mit net bezeichnete Matrix als Inzidenzmatrix nach 'file' aus  */

  LMPNet net;        /* Zeiger auf Matrix */
  FILE   * ofp;

#endif

  {

    register Cardinal32 trnum,
                        plnum;
    register int        i,
                        width1,
                        width2;
    register Place      pl;
    register Transition tr,
                        trhelp;
    register STEdge     st;
    register TSEdge     ts;

    if ( net )
      {
        trnum = net->NumTransitions;
        plnum = net->NumPlaces;
        /* Bestimme Wert fuer Spaltenbreite */
        width1 = max_placename_length( net );
        fprintf( ofp, "%d %d \n", plnum, trnum ); 
        fprintf( ofp, "Inzidenzmatrix: \n" );
        tr = net->Transitions;
        for ( i = 1 ; i <= width1+4 ; i++ )
          {
            fprintf( ofp, " " );
          } /* end of for i */
        while ( tr )
          {
            fprintf( ofp, "%s ", tr->name );
            for ( i = 1 ; i <= (int)( 6 - strlen( tr->name ) ) ; i++ )
	      {
                fprintf( ofp, " " );
              } /* end of for i */
            tr = tr->nexttrans;
          } /* while */
        fprintf( ofp, " \n" );
        pl = net->Places;
        while ( pl )
          {
            trhelp = net->Transitions;
            st = pl->postset;
            ts = pl->preset;
            fprintf( ofp, "%s", pl->name );
	    for ( i = 1 ; i <= (int)( width1 - strlen( pl->name ) ) ; i++ )
	      {
                fprintf( ofp, " " );
              } /* end of for i */
            fprintf( ofp, " |" );
            while ( trhelp )
              {
	        width2 = strlen( trhelp->name );
	        /* falls die Verbindungsliste noch nicht zuende ist */
                if ( ts && ( ts->trans == trhelp ) )
                  {
                    fprintf( ofp, "(-%d,", ts->weight );
                    ts = ts->nexttrans;
                  } /* if */
                else
                  {
                    fprintf( ofp, "( 0," );
                  } /* else */
                if ( st && ( st->trans == trhelp ) )
                  {
                    fprintf( ofp, "%d) ", st->weight );
                    st = st->nexttrans;
                  } /* if */
                else
                  {
                    fprintf( ofp, "0) " );
                  } /* else */
                for ( i = 1; i <= ( width2 - 5 ) ; i++ )
	          {
                    fprintf( ofp, " " );
                  } /* end of for i */
	        trhelp = trhelp->nexttrans;
              } /* end of while */
            fprintf( ofp, " \n" );
            /* eine Zeile geschrieben */
            pl = pl->nextplace;
          } /* end of while */
      } /* end of if net */
    else
      {
        WARNING_NO_NET( "output_matrix" );
      } /* else */

  } /* output_matrix() */


/*---------------------------------------------------------------------------*/

#if defined(__STDC__) || defined(__cplusplus)

void write_matrix( LMPNet net, char * filename )

#else

void write_matrix( net, filename )
  /* Gibt die mit net bezeichnete Matrix als Inzidenzmatrix in die Datei     */
  /* filename aus. */ 

  LMPNet net;        /* Zeiger auf Matrix */
  char   * filename; /* String mit Namen der Ausgabedatei */

#endif

  {

    register FILE * ofp;

    if ( ofp = fopen( filename, "w" ) )
      {
        output_matrix( net, ofp );
        fclose( ofp );
      } /* if */
    else
      {
        ERROR_FILE_WRITE( write_matrix, filename );
      } /* else */

  } /* write_matrix() */


/*****************************************************************************/
/******************************* write_netmarking ****************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void write_netmarking( LMPNet net, char * filename )

#else

void write_netmarking( net, filename )
  /* Gibt die Anfangsmarkierung der Matrix net in die Datei filename aus */

  LMPNet net;
  char   filename[NAMELENGTH];

#endif

  {

    Place pl;
    FILE  * ofp;

    if ( net )
      {
        if ( ofp = fopen( filename, "a" ) )
          {
            fprintf( ofp, "\nStartmarkierung: \n" );
            if ( pl = net->Places )
              {
                fprintf( ofp, "( %d", pl->marks );
                while ( pl = pl->nextplace )
                  {
	            /* mittlere Marke, mit Komma */
                    fprintf( ofp, " ,%d", pl->marks);
                  } /* while */
	        fprintf( ofp, " ) \n" );
              } /* if */
            fclose( ofp );
          } /* if */
        else
          {
            ERROR_FILE_READ( write_netmarking, filename );
          } /* end of else */
      } /* if */
    else
      {
        WARNING_NO_NET( "write_netmarking" );
      } /* else */
   
  } /* write_netmarking() */


/*****************************************************************************/
/****************************** show_matrix **********************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void show_matrix( LMPNet net )

#else

void show_matrix( net )
  /* Gibt die mit net bezeichnete Matrix als Inzidenzmatrix auf dem          */
  /* Bildschirm aus. */ 

  LMPNet net; /* Zeiger auf Matrix */

#endif

  {

    output_matrix( net, stdout );

  } /* show_matrix() */


/*****************************************************************************/
/******************************** show_marking *******************************/
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void show_marking( MARKING marking )

#else

void show_marking(marking)
               /* gibt die Markierungsliste 'marking' auf dem Bildschirm aus */

  MARKING marking;

#endif

  {

    register MARKING mpointer;

    if ( marking )
      {
        mpointer = marking;
        while ( mpointer )
          {
	    if ( mpointer == marking )
	      {
	        printf("(");
              } /* if */
	    if ( ! mpointer->next )
	      {
	        printf( "%d)\n", mpointer->value );
              } /* if */
	    else
	      {
	        printf( "%d, ", mpointer->value );
              } /* else */
	    mpointer = mpointer->next;
          } /* while */
      } /* if */
    else
      {
        fprintf( stderr, "WARNING: NN.show_marking(): not a marking.\n" );
      } /* else */

  } /* show_marking() */


/*****************************************************************************/
/******************************** show_multiset ******************************/
/*****************************************************************************/
/* Gibt die Liste 'list' als Menge auf dem Bildschirm aus */
/* 'list' ist als Datenstruktur fuer Vor- und Nachbereiche konzipiert. */
/* In der Regel wird also eine Menge von Stellen oder Transitionen angezeigt */

#if defined(__STDC__) || defined(__cplusplus)

void show_multiset( LIST_POINTER list )

#else

void show_multiset( list )

LIST_POINTER list;

#endif

  {

    printf( "{" );
    if ( list )
      {
        while ( list )
          {
            printf( "%s", list->name );
            if ( list = list->next )
	     {
	       printf( ", " );
             } /* if */
          } /* while */
        printf("}\n");
      } /* if */
    else
      {
        printf( "}\n" );
      } /* else */

  } /* show_multiset() */


/*****************************************************************************/
/*                          Netzinformationsfunktionen                       */
/*****************************************************************************/

/*****************************************************************************/
/******************************** place_number *******************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 place_number( LMPNet net )

#else

Cardinal32 place_number( net )
  /* Berechnet die Anzahl der Stellen im Netz, also die Anzahl der Elemente  */
  /* in der ersten Spalte der Matrix.                                        */

  LMPNet net;

#endif

  {

    if ( net )
      {
        return ( net->NumPlaces );
      } /* if */
    else
      {
        WARNING_NO_NET( "place_number" );
        return ( 0 );
      } /* else */

  } /* place_number() */


/*****************************************************************************/
/******************************** trans_number *******************************/
/*****************************************************************************/
/* KONVERSION: 04.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 trans_number( LMPNet net )

#else

Cardinal32 trans_number( net )
  /* Berechnet die Anzahl der Transitionen im Netz, also die Anzahl der      */
  /* Elemente in der ersten Zeile der Matrix.                                */

  LMPNet net;

#endif

  {

    if ( net )
      {
        return ( net->NumTransitions );
      } /* if */
    else
      {
        WARNING_NO_NET( "trans_number" );
        return ( 0 );
      } /* else */

  } /* trans_number() */


/*****************************************************************************/
/***************************** component_number ******************************/
/*****************************************************************************/
/* NEUE FUNKTION: 04.05.1993 Stefan Roemer                                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 component_number( LMPNet net )

#else

Cardinal32 component_number( net )
  /* Berechnet die Anzahl der Komponenten im Netz.                           */

  LMPNet net;

#endif

  {

    if ( net )
      {
        return ( net->NumComponents );
      } /* if */
    else
      {
        WARNING_NO_NET( "component_number" );
        return ( 0 );
      } /* else */

  } /* component_number() */


/*****************************************************************************/
/******************************* is_place_successor **************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool is_place_successor( LMPNet net, char * P_name, char * T_name )

#else

bool is_place_successor( net, P_name, T_name )
  /* liefert True, falls Verbindung P_name --> T_name besteht, sonst False.  */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {
    
    return ( (bool) !! ( find_PT_connection( net, P_name, T_name ) ) );

  } /* is_place_successor() */


/*****************************************************************************/
/******************************* is_trans_successor **************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool is_trans_successor( LMPNet net, char * P_name, char * T_name )

#else

bool is_trans_successor( net, P_name, T_name )
  /* liefert True, falls Verbindung T_name --> P_name besteht, sonst False.  */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {
    
    return ( (bool) !! ( find_TP_connection( net, P_name, T_name ) ) );

  } /* is_trans_successor() */


/*****************************************************************************/
/************************ max_placename_length *******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal max_placename_length( LMPNet net )

#else

Cardinal max_placename_length( net )
  /* liefert die Laenge des laengsten Stellennamens.                         */
  /* Wird benoetigt, um die uebersichtlich auszugeben.                       */

  LMPNet net;

#endif

  {

    register Cardinal length,
                      maxlength = 0;
    register Place    pl;

    if ( net )
      {
        pl = net->Places;
        while ( pl )
	  {
            if ( ( length = strlen( pl->name ) ) > maxlength )
	      {
	        maxlength = length;
              } /* if */
	    pl = pl->nextplace;
          } /* while */
      } /* if */
    else
      {
        WARNING_NO_NET( "max_placename_length" );
      } /* else */

    return ( maxlength );

  } /* max_placename_length() */


/*****************************************************************************/
/*********************** max_transitionname_length ***************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal max_transitionname_length( LMPNet net )

#else

Cardinal max_transitionname_length( net )
/* liefert die Laenge des laengsten Transitionsnamens.                       */
/* Wird fuer die Matrixausgabe benoetigt.                                    */

  LMPNet net;

#endif

  {

    register Cardinal   length,
                        maxlength = 0;
    register Transition tr;

    if ( net )
      {
        tr = net->Transitions;
        while ( tr )
	  {
            if ( ( length = strlen( tr->name ) ) > maxlength )
	      {
	        maxlength = length;
              } /* if */
	    tr = tr->nexttrans;
          } /* while */
      } /* if */
    else
      {
        WARNING_NO_NET( "max_transitionname_length" );
      } /* else */

    return ( maxlength );

  } /* max_transitionname_length() */


/*****************************************************************************/
/*                         Expandierungsfunktionen                           */
/*****************************************************************************/

/*****************************************************************************/
/******************************* insert_place ********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place insert_place( LMPNet net, char * P_name, Cardinal32 marks )

#else

Place insert_place( net, P_name, marks )
  /* setzt einen Knoten fuer die Stelle mit Namen P_name in die Stellenliste */
  /* an lexikographisch richtiger Stelle ein.                                */

  LMPNet      net;
  char        * P_name;
  Cardinal32  marks;

#endif

  {

    register Place pl,
                   plnew;

    if ( net && P_name && *P_name )
      {
        pl = net->Places;
        while ( pl && Kleiner( pl->name, P_name ) )
          {
            pl = pl->nextplace;
          } /* while */
        if ( pl && Gleich( P_name, pl->name ) )
          {
            fprintf( stderr,
                    "WARNING: NN.insert_place(): place '%s' already exists.\n",
                     P_name );
            return ( pl );
          } /* if */
        plnew = (Place) Malloc( sizeof( PlElem ) );
        if ( pl )
          {
            if ( plnew->prevplace = pl->prevplace )
              {
                plnew->prevplace->nextplace = plnew;
              } /* if */
            else
              {
                net->Places = plnew;
              } /* else */
            ( plnew->nextplace = pl ) -> prevplace = plnew;
          } /* if */
        else
          {
            if ( plnew->prevplace = net->LastPlace )
              {
                net->LastPlace->nextplace = plnew;
              } /* if */
            else
              {
                net->Places = plnew;
              } /* else */
            ( net->LastPlace = plnew ) -> nextplace = NULL;
          } /* else */
        plnew->name = Strdup( P_name );
        plnew->component = plnew->numname = 0;
        plnew->preset = NULL;
        plnew->postset = NULL;
        plnew->label = EMPTY;
        plnew->marks = marks;
        net->NumPlaces++;
      } /* if */
    else
      {
        WARNING_NO_NET( "insert_place" );
        return ( NULL );
      } /* else */

    return ( plnew );

  } /* insert_place() */


/*****************************************************************************/
/****************************** insert_transition ****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition insert_transition( LMPNet net, char * T_name )

#else

Transition insert_transition( net, T_name )
  /* Setzt einen neuen Knoten fuer eine Transition an der lexikographisch    */
  /* richtigen Stelle in die Matrix ein, also in die erste Zeile.            */

  LMPNet net;
  char   * T_name;

#endif

  {

    register Transition tr,
                        trnew;

    if ( net && T_name && *T_name )
      {
        tr = net->Transitions;
        while ( tr && Kleiner( tr->name, T_name ) )
          {
            tr = tr->nexttrans;
          } /* while */
        if ( tr && Gleich( T_name, tr->name ) )
          {
            fprintf( stderr,
          "WARNING: NN.insert_transition(): transition '%s' already exists.\n",
                     T_name );
            return ( tr );
          } /* if */
        trnew = (Transition) Malloc( sizeof( TrElem ) );
        if ( tr )
          {
            if ( trnew->prevtrans = tr->prevtrans )
              {
                trnew->prevtrans->nexttrans = trnew;
              } /* if */
            else
              {
                net->Transitions = trnew;
              } /* else */
            ( trnew->nexttrans = tr ) -> prevtrans = trnew;
          } /* if */
        else
          {
            if ( trnew->prevtrans = net->LastTransition )
              {
                net->LastTransition->nexttrans = trnew;
              } /* if */
            else
              {
                net->Transitions = trnew;
              } /* else */
            ( net->LastTransition = trnew ) -> nexttrans = NULL;
          } /* else */
        trnew->name = Strdup( T_name );
        trnew->numname = 0;
        trnew->label = NULL;
        trnew->preset = NULL;
        trnew->postset = NULL;
        net->NumTransitions++;
      } /* if */
    else
      {
        WARNING_NO_NET( "insert_transition" );
        return ( NULL );
      } /* else */

    return ( trnew );

  } /* insert_transition() */


/*****************************************************************************/
/******************************* insert_con **********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void insert_con( LMPNet net, char * P_name, char * T_name,
                 Cardinal32 tsweight, Cardinal32 stweight )

#else

void insert_con( net, P_name, T_name, tsweight, stweight )
  /* setzt die beiden Verbindungsknoten mit den Gewichten tsweight und stweight  */
  /* in die Matrix ein. Falls solche schon vorhanden sind, werden die        */
  /* Gewichte entsprechend geaendert.                                        */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 tsweight,                    /* Anzahl der Verbindungen von einer */
                                        /* Transition zu einer Stelle        */
             stweight;                    /* Anzahl der Verbindungen von einer */
                                        /* Stelle zu einer Transition        */

#endif

  {

    (void) insert_PT_con( net, P_name, T_name, stweight );
    (void) insert_TP_con( net, P_name, T_name, tsweight );

  } /* insert_con() */


/*****************************************************************************/
/******************************** insert_TP_con ******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

TSEdge insert_TP_con( LMPNet net, char * P_name, char * T_name,
                      Cardinal32 tsweight )

#else

TSEdge insert_TP_con( net, P_name, T_name, tsweight )
  /* setzt einen Verbindungsknoten mit dem Gewicht tsweight in die Matrix ein; */
  /* falls ein solcher schon vorhanden, wird das Gewicht entsprechend        */
  /* geaendert.                                                              */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 tsweight;                    /* Anzahl der Verbindungen von einer */
                                        /* Transition zu einer Stelle        */

#endif

  {

    register Place      pl;
    register Transition tr;
    register TSEdge     ts,
                        tsnew;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( ! tsweight )
          {
            delete_T_to_P_con_node( net,
                                   find_TP_connection( net, P_name, T_name ) );
            return ( NULL );
          } /* if */
        if ( ( pl = find_P( net, P_name ) ) && ( tr = find_T( net, T_name ) ) )
          {
            /* Position des Knotens suchen */
            if ( ts = pl->preset )
              {
                if ( Kleiner( T_name, ts->trans->name ) )
                  {
                    ts = NULL;
                  } /* if */
                else
                  {
                    while ( ts->nexttrans
                            && Kleiner( ts->nexttrans->trans->name, T_name ) )
                      {
                        ts = ts->nexttrans;
                      } /* while */
                    if ( ( tsnew = ts->nexttrans )
                         && Gleich( tsnew->trans->name, T_name ) )
                      {
                        tsnew->weight = tsweight;
                        return ( tsnew );
                      } /* if */
                  } /* else */
              } /* if */
            tsnew = (TSEdge) Malloc( sizeof( TSElem ) );
            if ( ts )
              {
                if ( tsnew->nexttrans = ts->nexttrans )
                  {
                    tsnew->nexttrans->prevtrans = tsnew;
                  } /* if */
                ( tsnew->prevtrans = ts ) -> nexttrans = tsnew;
              } /* if */
            else
              {
                if ( tsnew->nexttrans = pl->preset )
                  {
                    tsnew->nexttrans->prevtrans = tsnew;
                  } /* if */
                ( pl->preset = tsnew ) -> prevtrans = NULL;
              } /* else */
            if ( ts = tr->postset )
              {
                if ( Kleiner( P_name, ts->place->name ) )
                  {
                    ts = NULL;
                  } /* if */
                else
                  {
                    while ( ts->nextplace
                            && Kleiner( ts->nextplace->place->name, P_name ) )
                      {
                        ts = ts->nextplace;
                      } /* while */
                  } /* else */
              } /* if */
            if ( ts )
              {
                if ( tsnew->nextplace = ts->nextplace )
                  {
                    tsnew->nextplace->prevplace = tsnew;
                  } /* if */
                ( tsnew->prevplace = ts ) -> nextplace = tsnew;
              } /* if */
            else
              {
                if ( tsnew->nextplace = tr->postset )
                  {
                    tsnew->nextplace->prevplace = tsnew;
                  } /* if */
                ( tr->postset = tsnew ) -> prevplace = NULL;
              } /* else */
            tsnew->weight = tsweight;
            tsnew->place = pl;
            tsnew->trans = tr;
          } /* if */
        else
          {
            fprintf( stderr,
"WARNING: NN.insert_TP_con(): place '%s' or transition '%s' does not exist.\n",
                     P_name, T_name );
            return ( NULL );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "insert_TP_con" );
        return ( NULL );
      } /* else */

    return ( tsnew );

  } /* insert_TP_con() */


/*****************************************************************************/
/********************************* insert_PT_con *****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

STEdge insert_PT_con( LMPNet net, char * P_name, char * T_name,
                      Cardinal32 stweight )

#else

STEdge insert_PT_con( net, P_name, T_name, stweight )
  /* setzt einen Verbindungsknoten mit dem Gewicht stweight in die Matrix ein; */
  /* falls ein solcher schon vorhanden, wird das Gewicht entsprechend        */
  /* geaendert.                                                              */

  LMPNet     net;
  char       * P_name,
             * T_name;
  Cardinal32 stweight;                    /* Anzahl der Verbindungen von einer */
                                        /* Stelle zu einer Transition        */

#endif

  {

    register Place      pl;
    register Transition tr;
    register STEdge     st,
                        stnew;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( ! stweight )
          {
            delete_P_to_T_con_node( net,
                                   find_PT_connection( net, P_name, T_name ) );
            return ( NULL );
          } /* if */
        if ( ( pl = find_P( net, P_name ) ) && ( tr = find_T( net, T_name ) ) )
          {
            /* Position des Knotens suchen */
            if ( st = pl->postset )
              {
                if ( Kleiner( T_name, st->trans->name ) )
                  {
                    st = NULL;
                  } /* if */
                else
                  {
                    while ( st->nexttrans
                            && Kleiner( st->nexttrans->trans->name, T_name ) )
                      {
                        st = st->nexttrans;
                      } /* while */
                    if ( ( stnew = st->nexttrans )
                         && Gleich( stnew->trans->name, T_name ) )
                      {
                        stnew->weight = stweight;
                        return ( stnew );
                      } /* if */
                  } /* else */
              } /* if */
            stnew = (STEdge) Malloc( sizeof( STElem ) );
            if ( st )
              {
                if ( stnew->nexttrans = st->nexttrans )
                  {
                    stnew->nexttrans->prevtrans = stnew;
                  } /* if */
                ( stnew->prevtrans = st ) -> nexttrans = stnew;
              } /* if */
            else
              {
                if ( stnew->nexttrans = pl->postset )
                  {
                    stnew->nexttrans->prevtrans = stnew;
                  } /* if */
                ( pl->postset = stnew ) -> prevtrans = NULL;
              } /* else */
            if ( st = tr->preset )
              {
                if ( Kleiner( P_name, st->place->name ) )
                  {
                    st = NULL;
                  } /* if */
                else
                  {
                    while ( st->nextplace
                            && Kleiner( st->nextplace->place->name, P_name ) )
                      {
                        st = st->nextplace;
                      } /* while */
                  } /* else */
              } /* if */
            if ( st )
              {
                if ( stnew->nextplace = st->nextplace )
                  {
                    stnew->nextplace->prevplace = stnew;
                  } /* if */
                ( stnew->prevplace = st ) -> nextplace = stnew;
              } /* if */
            else
              {
                if ( stnew->nextplace = tr->preset )
                  {
                    stnew->nextplace->prevplace = stnew;
                  } /* if */
                ( tr->preset = stnew ) -> prevplace = NULL;
              } /* else */
            stnew->weight = stweight;
            stnew->place = pl;
            stnew->trans = tr;
          } /* if */
        else
          {
            fprintf( stderr,
"WARNING: NN.insert_PT_con(): place '%s' or transition '%s' does not exist.\n",
                     P_name, T_name );
            return ( NULL );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "insert_PT_con" );
        return ( NULL );
      } /* else */

    return ( stnew );

  } /* insert_PT_con() */


/*****************************************************************************/
/*                              Reduktionsfunktionen                         */
/*****************************************************************************/

/*****************************************************************************/
/***************************** delete_P_to_T_con_node ************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void delete_P_to_T_con_node( LMPNet net, STEdge st )

#else

void delete_P_to_T_con_node( net, st )
  /* loescht in Matrix 'net' den Verbindungsknoten, auf den st zeigt, */
  /* und aendert die Verzeigerung der Matrix entsprechend.                   */

  LMPNet net;
  STEdge st;

#endif

  {

    if ( net && st )
      {
        if ( st->nextplace )
          { st->nextplace->prevplace = st->prevplace; }
        if ( st->prevplace )
          { st->prevplace->nextplace = st->nextplace; }
        else
          { st->trans->preset = st->nextplace; }
        if ( st->nexttrans )
          { st->nexttrans->prevtrans = st->prevtrans; }
        if ( st->prevtrans )
          { st->prevtrans->nexttrans = st->nexttrans; }
        else
          { st->place->postset = st->nexttrans; }
      } /* if */
    else
      {
        fprintf( stderr,
            "WARNING: NN.delete_P_to_T_con_node(): not a net or no node.\n" );
      }

  } /* delete_P_to_T_con_node() */


/*****************************************************************************/
/***************************** delete_T_to_P_con_node ************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

void delete_T_to_P_con_node( LMPNet net, TSEdge ts )

#else

void delete_T_to_P_con_node( net, ts )
  /* loescht in Matrix 'net' den Verbindungsknoten, auf den ts zeigt, */
  /* und aendert die Verzeigerung der Matrix entsprechend.                   */

  LMPNet net;
  TSEdge ts;

#endif

  {

    if ( net && ts )
      {
        if ( ts->nextplace )
          { ts->nextplace->prevplace = ts->prevplace; }
        if ( ts->prevplace )
          { ts->prevplace->nextplace = ts->nextplace; }
        else
          { ts->trans->postset = ts->nextplace; }
        if ( ts->nexttrans )
          { ts->nexttrans->prevtrans = ts->prevtrans; }
        if ( ts->prevtrans )
          { ts->prevtrans->nexttrans = ts->nexttrans; }
        else
          { ts->place->preset = ts->nexttrans; }
      } /* if */
    else
      {
        WARNING_NO_NET( "delete_T_to_P_con_node" );
      }

  } /* delete_T_to_P_con_node() */


/*****************************************************************************/
/*************************** delete_T_to_P_connection ************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool delete_T_to_P_connection( LMPNet net, char * P_name, char * T_name )

#else

bool delete_T_to_P_connection( net, P_name, T_name )
  /* loescht eine Verbindung von T_name zu P_name, dekrementier 'weight' im  */
  /* zugehoerigen Knoten.                                                    */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    TSEdge ts;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        /* suche passenden Verbindungsknoten in der Matrix */
        if ( ts = find_TP_connection( net, P_name, T_name ) )
          {
            /* Verbindungsknoten vorhanden */
            if ( ! --ts->weight )
              {
                delete_T_to_P_con_node( net, ts );
              } /* if */
          } /* if */
      } /* if */
    else
      {
        fprintf( stderr,
            "WARNING: NN.delete_P_to_T_connection(): not a net or no P/T.\n" );
        return ( False );
      } /* else */

    return ( True );

  } /* delete_T_to_P_connection() */


/*****************************************************************************/
/************************** delete_P_to_T_connection *************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool delete_P_to_T_connection( LMPNet net, char * P_name, char * T_name )

#else

bool delete_P_to_T_connection( net, P_name, T_name )
  /* loescht Verbindung von P_name zu T_name, reduziert weight im zugehoe-   */
  /* rigen Knoten.                                                           */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    /* Deklarationen */
    STEdge st;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        /* suche passenden Verbindungsknoten in der Matrix */
        if ( st = find_PT_connection( net, P_name, T_name ) )
          {
            /* Verbindungsknoten vorhanden */
            if ( ! --st->weight )
              {
                delete_P_to_T_con_node( net, st );
              } /* if */
          } /* if */
      } /* if */
    else
      {
        fprintf( stderr,
            "WARNING: NN.delete_T_to_P_connection(): not a net or no P/T.\n" );
        return ( False );
      } /* else */

    return ( True );

  } /* delete_T_to_P_connection() */


/*****************************************************************************/
/************************* delete_place_direct *******************************/
/*****************************************************************************/
/* NEU: 17.10.1994 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Place delete_place_direct( LMPNet net, Place place )

#else

Place delete_place_direct( net, place )
  /* loescht die Zeile fuer die Stelle `pl` aus der Matrix                   */
  /* 'net'. Alle Verbindungsknoten werden geloescht und die Spalten entspre- */
  /* chend neu verzeigert.                                                   */
  /* Rueckgabe ist ein Zeiger auf die der geloeschten Stelle folgende Stelle.*/

  LMPNet net;
  Place  place;

#endif

  {

    register Place  pl;
    register STEdge st,
                    stdel;
    register TSEdge ts,
                    tsdel;

    place = ( pl = place )->nextplace;
    ts = pl->preset;
    while ( ts )
      {
        if ( ts->prevplace )
          {
            ts->prevplace->nextplace = ts->nextplace;
          } /* if */
        else
          {
            ts->trans->postset = ts->nextplace;
          } /* else */
        if ( ts->nextplace )
          {
            ts->nextplace->prevplace = ts->prevplace;
          } /* if */
        ts = ( tsdel = ts ) -> nexttrans;
        Free( (char *) tsdel );
      } /* while */
    st = pl->postset;
    while ( st )
      {
        if ( st->prevplace )
          {
            st->prevplace->nextplace = st->nextplace;
          } /* if */
        else
          {
            st->trans->preset = st->nextplace;
          } /* else */
        if ( st->nextplace )
          {
            st->nextplace->prevplace = st->prevplace;
          } /* if */
        st = ( stdel = st ) -> nexttrans;
        Free( (char *) stdel );
      } /* while */
    if ( pl->prevplace )
      {
        pl->prevplace->nextplace = pl->nextplace;
      } /* if */
    else
      {
        net->Places = pl->nextplace;
      } /* else */
    if ( pl->nextplace )
      {
        pl->nextplace->prevplace = pl->prevplace;
      } /* if */
    else
      {
        net->LastPlace = pl->prevplace;
      } /* else */
    Free( pl->name );
    if ( pl->inscript )
      {
        Free( pl->inscript );
      } /* if */
    if ( ! ( net->Flags & 1 ) )
      {
        Free( pl );
      } /* if */
    net->NumPlaces--;

    return ( place );

  } /* delete_place_direct() */


/*****************************************************************************/
/************************** delete_transition_direct *************************/
/*****************************************************************************/
/* NEU: 17.10.1994 Stefan Roemer                                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Transition delete_transition_direct( LMPNet net, Transition trans )

#else

Transition delete_transition_direct( net, trans )
  /* loescht die Spalte fuer die Transition `tr` aus der Matrix 'net'.       */
  /* Alle Verbindungknoten werden geloescht und die Zeilen entsprechend neu  */
  /* verzeigert.                                                             */
  /* Rueckgabe ist ein Zeiger auf die naechste Transition.                   */

  LMPNet     net;
  Transition trans;

#endif

  {

    register Transition tr;
    register STEdge     st,
                        stdel;
    register TSEdge     ts,
                        tsdel;

    trans = ( tr = trans )->nexttrans;
    st = tr->preset;
    while ( st )
      {
        if ( st->prevtrans )
          { st->prevtrans->nexttrans = st->nexttrans; }
        else
          { st->place->postset = st->nexttrans; }
        if ( st->nexttrans )
          { st->nexttrans->prevtrans = st->prevtrans; }
        st = ( stdel = st ) -> nextplace;
        Free( (char *) stdel );
      } /* while */
    ts = tr->postset;
    while ( ts )
      {
        if ( ts->prevtrans )
          { ts->prevtrans->nexttrans = ts->nexttrans; }
        else
          { ts->place->preset = ts->nexttrans; }
        if ( ts->nexttrans )
          { ts->nexttrans->prevtrans = ts->prevtrans; }
        ts = ( tsdel = ts ) -> nextplace;
        Free( (char *) tsdel );
      } /* while */
    if ( tr->prevtrans )
      { tr->prevtrans->nexttrans = tr->nexttrans; }
    else
      { net->Transitions = tr->nexttrans; }
    if ( tr->nexttrans )
      { tr->nexttrans->prevtrans = tr->prevtrans; }
    else
      { net->LastTransition = tr->prevtrans; }
    Free( tr->name );
    if ( tr->inscript )
      {
        Free( tr->inscript );
      } /* if */
    if ( ! ( net->Flags & 1 ) )
      {
        Free( tr );
      } /* if */
    net->NumTransitions--;

    return ( trans );

  } /* delete_transition_direct() */


/*****************************************************************************/
/****************************** delete_place *********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool delete_place( LMPNet net, char * P_name )

#else

bool delete_place( net, P_name )
  /* loescht die Zeile fuer die Stelle mit Namen 'P_name' aus der Matrix     */
  /* 'net'. Alle Verbindungsknoten werden geloescht und die Spalten entspre- */
  /* chend neu verzeigert.                                                   */

  LMPNet net;
  char   * P_name;

#endif

  {

    Place pl;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            (void) delete_place_direct( net, pl );
          } /* if */
        else
          {
            WARNING_P( "delete_place", P_name );
          } /* else */
      }
    else
      {
        WARNING_NO_NET_OR_P( "delete_place" );
        return ( False );
      } /* else */

    return ( True );

  } /* delete_place() */


/*****************************************************************************/
/****************************** delete_transition ****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool delete_transition( LMPNet net, char * T_name )

#else

bool delete_transition( net, T_name )
  /* loescht die Spalte fuer die Transition 'T_name' aus der Matrix 'net'.   */
  /* Alle Verbindungknoten werden geloescht und die Zeilen entsprechend neu  */
  /* verzeigert.                                                             */

  LMPNet net;
  char   * T_name;

#endif

  {

    Transition tr;

    if ( net && T_name && *T_name )
      {
        if ( tr = find_T( net, T_name ) )
          {
            (void) delete_transition_direct( net, tr );
          } /* if */
        else
          {
            WARNING_T( "delete_transition", T_name );
          } /* else */
      }
    else
      {
        WARNING_NO_NET_OR_T( "delete_transition" );
        return ( False );
      } /* else */

    return ( True );

  } /* delete_transition() */


/*****************************************************************************/
/************************** place_successor_number ***************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 place_successor_number( LMPNet net, char * P_name )

#else

Cardinal32 place_successor_number( net, P_name )
  /* zaehlt die Kanten im Nachbereich von 'P_name' */

  LMPNet net;
  char   * P_name;

#endif

  {

    /* Deklarationen */
    Place      pl;
    STEdge     st;
    Cardinal32 number = 0;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            st = pl->postset;
            while ( st )
              {
                number += st->weight;
                st = st->nexttrans;
              } /* while */
          } /* if */
        else
          {
            WARNING_P( "place_successor_number", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "place_successor_number" );
      } /* else */

    return( number );

  } /* place_successor_number() */


/*****************************************************************************/
/************************** trans_successor_number ***************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 trans_successor_number( LMPNet net, char * T_name )

#else

Cardinal32 trans_successor_number( net, T_name )
  /* zaehlt die Kanten im Nachbereich von 'T_name' */

  LMPNet net;
  char   * T_name;

#endif

  {

    /* Deklarationen */
    Transition tr;
    TSEdge     ts;
    Cardinal32 number = 0;

    if ( net && T_name && *T_name )
      {
        if ( tr = find_T( net, T_name ) )
          {
            ts = tr->postset;
            while ( ts )
              {
                number += ts->weight;
                ts = ts->nextplace;
              } /* while */
          } /* if */
        else
          {
            WARNING_T( "trans_successor_number", T_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_T( "trans_successor_number" );
      } /* else */

    return( number );

  } /* trans_successor_number() */


/*****************************************************************************/
/************************** place_predecessor_number *************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 place_predecessor_number( LMPNet net, char * P_name )

#else

Cardinal32 place_predecessor_number( net, P_name )
  /* zaehlt die Kanten im Vorbereich von 'P_name' */

  LMPNet net;
  char   * P_name;

#endif

  {

    /* Deklarationen */
    Place      pl;
    TSEdge     ts;
    Cardinal32 number = 0;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            ts = pl->preset;
            while ( ts )
              {
                number += ts->weight;
                ts = ts->nexttrans;
              } /* while */
          } /* if */
        else
          {
            WARNING_P( "place_predecessor_number", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "place_predecessor_number" );
      } /* else */

    return( number );

  } /* place_predecessor_number() */


/*****************************************************************************/
/************************** trans_predecessor_number *************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 trans_predecessor_number( LMPNet net, char * T_name )

#else

Cardinal32 trans_predecessor_number( net, T_name )
  /* zaehlt die Kanten im Vorbereich von 'T_name' */

  LMPNet net;
  char   * T_name;

#endif

  {

    /* Deklarationen */
    Transition tr;
    STEdge     st;
    Cardinal32 number = 0;

    if ( net && T_name && *T_name )
      {
        if ( tr = find_T( net, T_name ) )
          {
            st = tr->preset;
            while ( st )
              {
                number += st->weight;
                st = st->nextplace;
              } /* while */
          } /* if */
        else
          {
            WARNING_T( "trans_predecessor_number", T_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_T( "trans_predecessor_number" );
      } /* else */

    return( number );

  } /* trans_predecessor_number() */


/*****************************************************************************/
/************************ place_successor_list *******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/* Diese Funktion sollte nur selten eingesetzt werden, da die vorliegende    */
/* Datenstruktur einen direkten Zugriff auf den Nachbereich einer Stelle     */
/* erlaubt. ACHTUNG: DIE LISTE, DIE MIT DIESER FUNKTION ANGEFORDERT WIRD,    */
/* SOLLTE NACH IHRER "BENUTZUNG" AUCH WIEDER AUS DEM SPEICHER GELOESCHT      */
/* WERDEN ( durch Aufruf der Funktion 'remove_list()' ) !                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LIST_POINTER place_successor_list( LMPNet net, char * P_name )

#else

LIST_POINTER place_successor_list( net, P_name )
  /* liefert einen Pointer auf eine Liste mit den Namen der Elemente im      */
  /* Nachbereich der Stelle 'P_name'. Ist der Nachbereich von 'P_name' leer, */
  /* so wird ein nil-Zeiger zurueckgegeben. Die Liste enthaelt die Namen     */
  /* der Nachfolgetransitionen entsprechend oft der Staerke der Verbindung,  */
  /* die das jeweilige Kantengewicht angibt.                                 */

  LMPNet net;
  char   * P_name;

#endif

  {

    /* Deklarationen */
    Place        pl;
    STEdge       st;
    LIST_POINTER list = NULL,
                 listptr;
    Cardinal32   counter;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            if ( st = pl->postset )
              {
                listptr = list = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                Strcpy( list->name, st->trans->name );
                counter = st->weight;
                while ( --counter )
                  {
                    listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                    Strcpy( listptr->name, st->trans->name );
                  } /* while */
                while ( st = st->nexttrans )
                  {
                    counter = st->weight;
                    while ( counter-- )
                      {
                        listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                        Strcpy( listptr->name, st->trans->name );
                      } /* while */
                  } /* while */
                listptr->next = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_P( "place_successor_list", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "place_successor_list" );
      } /* else */

    return ( list );

  } /* place_successor_list() */


/*****************************************************************************/
/************************ trans_successor_list *******************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/* Diese Funktion sollte nur selten eingesetzt werden, da die vorliegende    */
/* Datenstruktur einen direkten Zugriff auf den Nachbereich einer Transition */
/* erlaubt. ACHTUNG: DIE LISTE, DIE MIT DIESER FUNKTION ANGEFORDERT WIRD,    */
/* SOLLTE NACH IHRER "BENUTZUNG" AUCH WIEDER AUS DEM SPEICHER GELOESCHT      */
/* WERDEN ( durch Aufruf der Funktion 'remove_list()' ) !                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LIST_POINTER trans_successor_list( LMPNet net, char * T_name )

#else

LIST_POINTER trans_successor_list( net, T_name )
  /* liefert einen Pointer auf eine Liste mit den Namen der Elemente im      */
  /* Nachbereich der Transition 'P_name'. Ist der Nachbereich von 'T_name'   */
  /* leer, so wird ein nil-Zeiger zurueckgegeben. Die Liste enthaelt die     */
  /* Namen der Nachfolgestellen entsprechend oft der Staerke der Verbindung, */
  /* die das jeweilige Kantengewicht angibt.                                 */

  LMPNet net;
  char   * T_name;

#endif

  {

    /* Deklarationen */
    Transition   tr;
    TSEdge       ts;
    LIST_POINTER list = NULL,
                 listptr;
    Cardinal32   counter;

    if ( net && T_name && *T_name )
      {
        if ( tr = find_T( net, T_name ) )
          {
            if ( ts = tr->postset )
              {
                listptr = list = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                Strcpy( list->name, ts->place->name );
                counter = ts->weight;
                while ( --counter )
                  {
                    listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                    Strcpy( listptr->name, ts->place->name );
                  } /* while */
                while ( ts = ts->nextplace )
                  {
                    counter = ts->weight;
                    while ( counter-- )
                      {
                        listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                        Strcpy( listptr->name, ts->place->name );
                      } /* while */
                  } /* while */
                listptr->next = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_T( "trans_successor_list", T_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_T( "trans_successor_list" );
      } /* else */

    return ( list );

  } /* trans_successor_list() */


/*****************************************************************************/
/************************ place_predecessor_list *****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/* Diese Funktion sollte nur selten eingesetzt werden, da die vorliegende    */
/* Datenstruktur einen direkten Zugriff auf den Vorbereich einer Stelle      */
/* erlaubt. ACHTUNG: DIE LISTE, DIE MIT DIESER FUNKTION ANGEFORDERT WIRD,    */
/* SOLLTE NACH IHRER "BENUTZUNG" AUCH WIEDER AUS DEM SPEICHER GELOESCHT      */
/* WERDEN ( durch Aufruf der Funktion 'remove_list()' ) !                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LIST_POINTER place_predecessor_list( LMPNet net, char * P_name )

#else

LIST_POINTER place_predecessor_list( net, P_name )
  /* liefert einen Pointer auf eine Liste mit den Namen der Elemente im      */
  /* Vorbereich der Stelle 'P_name'. Ist der Vorbereich von 'P_name' leer,   */
  /* so wird ein nil-Zeiger zurueckgegeben. Die Liste enthaelt die Namen     */
  /* der Vorgaengertransitionen entsprechend oft der Staerke der Verbindung, */
  /* die das jeweilige Kantengewicht angibt.                                 */

  LMPNet net;
  char   * P_name;

#endif

  {

    register Place        pl;
    register TSEdge       ts;
    register LIST_POINTER list = NULL,
                          listptr;
    register Cardinal32   counter;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            if ( ts = pl->preset )
              {
                listptr = list = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                Strcpy( list->name, ts->trans->name );
                counter = ts->weight;
                while ( --counter )
                  {
                    listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                    Strcpy( listptr->name, ts->trans->name );
                  } /* while */
                while ( ts = ts->nexttrans )
                  {
                    counter = ts->weight;
                    while ( counter-- )
                      {
                        listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                        Strcpy( listptr->name, ts->trans->name );
                      } /* while */
                  } /* while */
                listptr->next = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_P( "place_predecessor_list", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "place_predecessor_list" );
      } /* else */

    return ( list );

  } /* place_predecessor_list() */


/*****************************************************************************/
/************************ trans_predecessor_list *****************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/* Diese Funktion sollte nur selten eingesetzt werden, da die vorliegende    */
/* Datenstruktur einen direkten Zugriff auf den Vorbereich einer Transition  */
/* erlaubt. ACHTUNG: DIE LISTE, DIE MIT DIESER FUNKTION ANGEFORDERT WIRD,    */
/* SOLLTE NACH IHRER "BENUTZUNG" AUCH WIEDER AUS DEM SPEICHER GELOESCHT      */
/* WERDEN ( durch Aufruf der Funktion 'remove_list()' ) !                    */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

LIST_POINTER trans_predecessor_list( LMPNet net, char * T_name )

#else

LIST_POINTER trans_predecessor_list( net, T_name )
  /* liefert einen Pointer auf eine Liste mit den Namen der Elemente im      */
  /* Vorbereich der Transition 'P_name'. Ist der Vorbereich von 'T_name'     */
  /* leer, so wird ein nil-Zeiger zurueckgegeben. Die Liste enthaelt die     */
  /* Namen der Vorgaengerstellen entsprechend oft der Staerke der Verbindung,*/
  /* die das jeweilige Kantengewicht angibt.                                 */

  LMPNet net;
  char   * T_name;

#endif

  {

    register Transition   tr;
    register STEdge       st;
    register LIST_POINTER list = NULL,
                          listptr;
    register Cardinal32   counter;

    if ( net && T_name && *T_name )
      {
        if ( tr = find_T( net, T_name ) )
          {
            if ( st = tr->preset )
              {
                listptr = list = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                Strcpy( list->name, st->place->name );
                counter = st->weight;
                while ( --counter )
                  {
                    listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                    Strcpy( listptr->name, st->place->name );
                  } /* while */
                while ( st = st->nextplace )
                  {
                    counter = st->weight;
                    while ( counter-- )
                      {
                        listptr = listptr->next
                               = (LIST_POINTER) Malloc( sizeof(LIST_ELEMENT) );
                        Strcpy( listptr->name, st->place->name );
                      } /* while */
                  } /* while */
                listptr->next = NULL;
              } /* if */
          } /* if */
        else
          {
            WARNING_T( "trans_predecessor_list", T_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_T( "trans_predecessor_list" );
      } /* else */

    return ( list );

  } /* trans_predecessor_list() */


/*****************************************************************************/
/*************************** TP_connection ***********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 TP_connection( LMPNet net, char * P_name, char * T_name )

#else

Cardinal32 TP_connection( net, P_name, T_name )
  /* liefert die Staerke der Verbindung zwischen Transition T_name und       */
  /* Stelle P_name.                                                          */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    register TSEdge ts;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( ts = find_TP_connection( net, P_name, T_name ) )
          {
            return( ts->weight );
          } /* if */
      } /* if */
    else
      {
        fprintf( stderr,
                 "WARNING: NN.TP_connection(): not a net or no P/T name.\n" );
      } /* else */

    return( 0 );

  } /* end of function TP_connection */


/*****************************************************************************/
/*************************** PT_connection ***********************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 PT_connection( LMPNet net, char * P_name, char * T_name )

#else

Cardinal32 PT_connection( net, P_name, T_name )
  /* liefert die Staerke der Verbindung zwischen Stelle P_name und Transi-   */
  /* tion T_name.                                                            */

  LMPNet net;
  char   * P_name,
         * T_name;

#endif

  {

    register STEdge st;

    if ( net && P_name && T_name && *P_name && *T_name )
      {
        if ( st = find_PT_connection( net, P_name, T_name ) )
          {
            return ( st->weight );
          } /* if */
      } /* if */
    else
      {
        fprintf( stderr,
                 "WARNING: NN.PT_connection(): not a net or no P/T name.\n" );
      } /* else */

    return( 0 );

  } /* PT_connection() */


/*****************************************************************************/
/*************************** name_free ***************************************/
/*****************************************************************************/
/* KONVERSION: 06.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool name_free( LMPNet net, char * name )

#else

bool name_free( net, name )
/* liefert True zurueck, falls 'name' noch nicht vergeben; False sonst.      */

  LMPNet net;
  char   * name;

#endif

  {

    return ( (bool) ! ( find_P( net, name ) || find_T( net, name ) ) );

  } /* name_free */


/*****************************************************************************/
/*                           Schnittstellenfunktionen                        */
/*****************************************************************************/

/*****************************************************************************/
/*                                 Markierungen                              */
/*****************************************************************************/

/*****************************************************************************/
/******************************** marking ************************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 marking( LMPNet net, char * P_name )

#else

Cardinal32 marking( net, P_name )
/* liefert die Startmarkierung der Stelle mit Namen 'P_name'                 */

  LMPNet net;
  char   * P_name;

#endif

  {

    /* Deklaraitionen */
    Place pl;

    if ( net && P_name && *P_name )
      {
        /* Netz vorhanden */
        if ( pl = find_P( net, P_name ) )
          {
            return ( pl->marks );
          } /* if */
        else
          {
            WARNING_P( "marking", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "marking" );
      } /* else */

    return ( 0 );

  } /* marking() */


/*****************************************************************************/
/***************************** size_of_marking *******************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

Cardinal32 size_of_marking( MARKING marking )

#else

Cardinal32 size_of_marking( marking )
  /* liefert die Anzahl der Stellen in der Markierungsliste                  */

  MARKING marking;

#endif

  {

    Cardinal32 places = 0;

    while ( marking )
      {
        places++;
        marking = marking->next;
      } /* while */

    return( places );

  } /* size_of_marking() */


/*****************************************************************************/
/***************************** change_place_marking **************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool change_place_marking( LMPNet net, char * P_name, Cardinal32 value )

#else

bool change_place_marking( net, P_name, value )
  /* aendert die Markierung der einzelnen Stelle mit Namen 'P_name' in der   */
  /* Matrix 'net' auf den Wert 'value'.                                      */

  LMPNet     net;
  char       * P_name;
  Cardinal32 value;

#endif

  {

    Place pl;

    if ( net && P_name && *P_name )
      {
        if ( pl = find_P( net, P_name ) )
          {
            pl->marks = value;
            return ( True );
          } /* if */
        else
          {
            WARNING_P( "change_place_marking", P_name );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET_OR_P( "change_place_marking" );
      } /* else */

   return ( False );

  } /* change_place_marking() */


/*****************************************************************************/
/***************************** copy_start_marking ****************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/* ACHTUNG: DIE LISTE MUSS NACH IHRER BENUTZUNG AUS DEM SPEICHER ENTFERNT    */
/* WERDEN ( mit 'remove_marklist()' ) !                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

MARKING copy_start_marking( LMPNet net )

#else

MARKING copy_start_marking( net )
  /* liefert eine Liste mit der Startmarkierung der Matrix 'net'.            */

  LMPNet net;       /* Matrix */

#endif

  {

    /* Deklarationen */
    Place   pl;
    MARKING list,
            mhelp = NULL;

    if ( net )
      {
        if ( pl = net->Places )
          {
            mhelp = list = (MARKING) Malloc( sizeof( MARK_LIST ) );
	    Strcpy( list->name, pl->name );
	    list->value = pl->marks;
            /* baue Liste mit Hilfe von 'list' auf, merke die gesamte    */
            /* Liste unter 'mhelp'                                       */
            while ( pl = pl->nextplace )
	      {
                ( list = list->next = (MARKING) Malloc( sizeof( MARK_LIST ) ) )
                                                          -> value = pl->marks;
	        Strcpy( list->name, pl->name );
              } /* while */
	    list->next = NULL;
          } /* end of if */
        else
          {
            WARNING_NO_P( "copy_start_marking" );
          } /* end of else */

      } /* if */
    else
      {
        WARNING_NO_NET( "copy_start_marking" );
      } /* else */

    return( mhelp );

  } /* copy_start_marking() */


/*****************************************************************************/
/*************************** change_start_marking ****************************/
/*****************************************************************************/
/* KONVERSION: 10.05.1993 Stefan Roemer                                      */
/* ACHTUNG: DIE LISTE MUSS NACH IHRER BENUTZUNG AUS DEM SPEICHER ENTFERNT    */
/* WERDEN ( mit 'remove_marklist()' ) !                                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

bool change_start_marking( LMPNet net, MARKING marking )

#else

bool change_start_marking( net, marking )
  /* setzt 'marking' als neue Startmarkierung in die Matrix 'net' ein */

  LMPNet  net;
  MARKING marking;

#endif

  {

    register bool success = False;

    if ( net )
      {
        if ( net->NumPlaces == size_of_marking( marking ) )
          {
            if ( net->NumPlaces )
              {
                success = True;
                while ( marking )
	          {
	            if ( ! change_place_marking( net, marking->name,
                                                            marking->value ) )
		      {
                    fprintf( stderr, "WARNING: NN.change_start_marking(): " );
      fprintf( stderr, "number of marks of place '%s' has not been changed.\n",
                                 marking->name);
                        success = False;
                      } /* if */
	            marking = marking->next;
                  } /* end of while */
              } /* end of if */
          } /* if */
        else
          {
            /* Fehler: Markierungsliste zu kurz oder zu lang */
            fprintf( stderr,
                "WARNING: NN.change_start_marking(): marking list does not " );
            fprintf( stderr, "fit to the number of places.\n" );
          } /* else */
      } /* if */
    else
      {
        WARNING_NO_NET( "change_start_marking" );
      } /* else */

    return( success );

  } /* change_start_marking() */


/*****************************************************************************/
/* END OF FILE neunetze.c                                                    */
/*****************************************************************************/

