/*****************************************************************************/
/* neunetze.c   - VERSION m-1.8.1 - 08.02.1998 - (c) Stefan Roemer 1993-1998 */
/*****************************************************************************/
/* DATENSTRUKTUREN FUER BESCHRIFTETE UND MARKIERTE PETRI-NETZE               */
/*                                                 Stefan Roemer, 17.10.1994 */
/*                                                 Stefan Roemer, 08.02.1998 */
/*****************************************************************************/
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/* ACHTUNG: Diese Datei enthaelt elementare Datenstrukturen und Routinen,    */
/* die von vielen verschiedenen Programmen benoetigt werden.                 */
/* Jede beabsichtigte Aenderung am Programmcode sowie das Kopieren von       */
/* Programmteilen deshalb bitte vorher mit mir absprechen. Danke.            */
/* E-mail: roemer@informatik.tu-muenchen.de                                  */
/*                                                                           */
/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
/*                                                                           */
/*****************************************************************************/


#ifndef __NEUNETZE_H__
#define __NEUNETZE_H__


/*****************************************************************************/


#include <stddef.h>
#include <stdio.h>


/*****************************************************************************/

#define skip


/*****************************************************************************/
/* Allgemeine Typvereinbarungen.                                             */
/*****************************************************************************/

typedef unsigned short int Cardinal;  /* SUN: 16-Bit-Wert !!!!!!! */
typedef unsigned int Cardinal32;      /* SUN: 32-Bit-Wert !!!!!!! */

typedef enum { False = 0, True } bool;


/*****************************************************************************/
/* Typvereinbarungen fuer Aktionen und Multimengen ueber Aktionen.           */
/*****************************************************************************/

typedef unsigned short int Action, Var;  /* SUN: 16-Bit-Wert !!!!!!! */
typedef Action * MultiSet;
typedef MultiSet Function;


/*****************************************************************************/
/* Beschriftung von Stellen/Bedingungen.                                     */
/*****************************************************************************/

typedef char PLabel;

#define EMPTY     0
#define ENTRY     1
#define EXIT      2
#define INTERRUPT 3
#define DELETE    4


/*****************************************************************************/
/* Beschriftete, markierte Netze                                             */
/*****************************************************************************/

                                  /* Stelle bzw. Bedingung                   */
typedef struct PlElem_            /* ---------------------                   */

  {

    struct PlElem_ * prevplace,   /* Zeiger auf die in der Stellenliste vor  */
                                  /* dieser liegende Stelle.                 */
                   * nextplace;   /* Zeiger auf die nachfolgende Stelle in   */
                                  /* der Stellenliste.                       */
    struct TSElem_ * preset;      /* Zeiger auf die Liste der in diese       */
                                  /* Stelle eingehenden Kanten.              */
    struct STElem_ * postset;     /* Zeiger auf die Liste der aus dieser     */
                                  /* Stelle fuehrenden Kanten.               */
    char           * name,        /* Stellenname; Zeiger (!) auf String.     */
                   * inscript;    /* Bei Bedingungen: Zeiger auf String, der */
                                  /* den Namen der entsprechenden Stelle im  */
                                  /* Originalnetz enthaelt.                  */
    Cardinal32     numname;       /* Name ( eindeutige Zahl ) fuer diese     */
                                  /* Stelle bei der Ausgabe.                 */
    PLabel         label;         /* Beschriftung dieser Stelle:             */
                                  /* ENTRY/EMPTY/EXIT/...                    */
    Cardinal32     component,     /* Angabe, in welcher Komponente sich die  */
                                  /* Stelle befindet.                        */
                   marks;         /* Anzahl der Marken auf dieser Stelle.    */

  } PlElem,                       /* Verbundtyp 'Stelle'.                    */
    * Place;                      /* Zeiger auf 'Stelle'.                    */


                                  /* Transition bzw. Ereignis                */
typedef struct TrElem_            /* ------------------------                */

  {

    struct TrElem_ * prevtrans,   /* Zeiger auf vorherige Transition in der  */
                                  /* Transitionenliste.                      */
                   * nexttrans;   /* Zeiger auf die nachfolgende Transition. */
    struct STElem_ * preset;      /* Zeiger auf die Liste der in diese       */
                                  /* Transition fuehrenden Kanten.           */
    struct TSElem_ * postset;     /* Zeiger auf die Liste aus dieser Transi- */
                                  /* tion fuehrender Kanten.                 */
    char           * name,        /* Transitionenname: Zeiger (!) auf String.*/
                   * inscript;    /* Bei Ereignissen: Zeiger auf String, der */
                                  /* den Namen der entsprechenden Transition */
                                  /* im Ursprungsnetz enthaelt.              */
    Cardinal32     numname;       /* Nur fuer Ausgabezwecke: Eindeutiger     */
                                  /* numerischer Code als Name fuer diese    */
                                  /* Transition.                             */
    MultiSet       label;         /* Beschriftung dieser Transition:         */
                                  /* Zeiger auf eine Multimenge ueber den    */
                                  /* Aktionsnamen.                           */
    /* Eventuell kann auch fuer die Transitionen ein Komponenteneintrag */
    /* vorgesehen werden !?!                                            */

  } TrElem,                       /* Verbundtyp 'Transition'.                */
    * Transition;                 /* Zeiger auf 'Transition'.                */


                                  /* Kante Stelle-->Transition               */
typedef struct STElem_            /* -------------------------               */

  {

    struct STElem_ * prevplace,   /* Zeiger auf davorliegende S/T-Kante in   */
                                  /* bezug auf die Stellenliste.             */
                   * nextplace,   /* Zeiger auf die nachfolgende S/T-Kante   */
                                  /* bezueglich der Stellenliste.            */
                   * prevtrans,   /* Zeiger auf die vorangehende S/T-Kante   */
                                  /* bzgl. der Transitionenliste.            */
                   * nexttrans;   /* Zeiger auf die naechste S/T-Kante in    */
                                  /* Richtung der Transitionenliste.         */
    Place          place;         /* Zeiger auf jene Stelle, von der diese   */
                                  /* S/T-Kante ausgeht.                      */
    Transition     trans;         /* Zeiger auf diejenige Transition, in die */
                                  /* diese Kante fuehrt.                     */
    Cardinal32     weight;        /* Kantengewicht S->T, >= 0.               */

  } STElem,                       /* Verbundtyp 'Stellen-Transitionen-Kante'.*/
    * STEdge;                     /* Zeiger auf 'Stellen-Transitionen-Kante'.*/


                                  /* Kante Transition-->Stelle               */
typedef struct TSElem_            /* -------------------------               */

  {

    struct TSElem_ * prevplace,   /* Zeiger auf die dieser Kante vorausge-   */
                                  /* hende T/S-Kante bzgl. der Stellenliste. */
                   * nextplace,   /* Zeiger auf die nachfolgende T/S-Kante   */
                                  /* bzgl. der Stellenliste.                 */
                   * prevtrans,   /* Zeiger auf die dieser Kante vorausge-   */
                                  /* hende T/S-Kante bzgl. der Trans.-Liste. */
                   * nexttrans;   /* Zeiger auf die nachfolgende T/S-Kante   */
                                  /* bzgl. der Transitionenliste.            */
    Place          place;         /* Zeiger auf diejenige Stelle, in die     */
                                  /* diese Kante fuehrt.                     */
    Transition     trans;         /* Zeiger auf die Transition, aus der      */
                                  /* diese Kante kommt.                      */
    Cardinal32     weight;        /* Kantengewicht T->S, >= 0.               */

  } TSElem,                       /* Verbundtyp 'Transitionen-Stellen-Kante'.*/
    * TSEdge;                     /* Zeiger auf 'Transitionen-Stellen-Kante'.*/


                                  /* Kopf eines Netzes                       */
typedef struct                    /* -----------------                       */

  {

    Place      Places,            /* Liste der Stellen (Bedingungen):        */
                                  /* Zeiger auf die erste Stelle der Liste.  */
               LastPlace;         /* Zeiger auf die letzte Stelle der Liste. */
    Transition Transitions,       /* Liste der Transitionen (Ereignisse):    */
                                  /* Zeiger auf die erste Transition.        */
               LastTransition;    /* Zeiger auf die letzte Transition.       */
    Cardinal32 NumPlaces,         /* Anzahl der Stellen im Netz.             */
               NumTransitions,    /* Anzahl der Transitionen im Netz.        */
               NumComponents;     /* Anzahl der zusammenhaengenden Komponen- */
                                  /* ten des Netzes.                         */
    Cardinal32 Flags;             /* Zusatzinformationen zum Netz.           */
          

  } LMPNetHead,                   /* Verbundtyp 'Kopf des Netzes'.           */
    * LMPNet;                     /* Zeiger auf 'Kopf eines Netzes'.         */
                                  /* LMPNet steht fuer                       */
                                  /* Labelled and Marked Petri Net           */


/*****************************************************************************/
/*        Header Datei Praktikum Werkzeuge zur Petri-Netz Analyse            */
/*****************************************************************************/
/* Konversion neue Datenstruktur: Stefan Roemer, Di., 04.05.1993             */
/*****************************************************************************/

#define NAMELENGTH 200
#define LINELENGTH 4096


/* Struktur fuer Listenelemente (let = ListenElementTyp), Vor-- und Nachbereichlisten */
struct let
     {
       struct let *next;      /* Listennachfolger, anderes Element im Vor-/Nachbereich) */
       char   name[NAMELENGTH];
     };

typedef struct let LIST_ELEMENT;
typedef LIST_ELEMENT *LIST_POINTER;

/* Struktur fuer Markierung */
struct mark
     {
       struct mark *next;       /* Listennachfolger */
       char   name[NAMELENGTH]; /* Name der Stelle */
       int    value;            /* Markenanzahl */
     };

typedef struct mark MARK_LIST;
typedef MARK_LIST *MARKING;


/*****************************************************************************/
/*******************************   Variablen   *******************************/
/*****************************************************************************/

extern int NN_ambigious_nodes; /* 0 = ok, 1 = error, 2 = warning */


/*****************************************************************************/
/*******************************   Funktionen   ******************************/
/*****************************************************************************/

/*****************************************************************************/
/*                    String-Vergleichsfunktionen                            */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern bool (* Kleiner)( char *, char * );
extern bool (* Gleich)( char *, char * );
extern void strcmp_lexgraph( void );
extern void strcmp_strcmp( void );
#else
extern bool (* Kleiner)();
extern bool (* Gleich)();
extern void strcmp_lexgraph();
extern void strcmp_strcmp();
#endif
 

/*****************************************************************************/
/*                          Allokationsfunktionen                            */
/*****************************************************************************/

#ifdef Free
#undef Free
#endif
#define Free( _X_ ) ( (void) free( (char *) (_X_) ) )
/*
#define Free( _X_ ) ( (void) free( (char *) (_X_) ) )
*/

#if defined(__STDC__) || defined(__cplusplus)
extern void * Malloc( size_t );
extern void * Calloc( size_t, size_t );
extern void * Realloc( void *, size_t );
extern char * Strdup( char * );
#else
extern void * Malloc();
extern void * Calloc();
extern void * Realloc();
extern char * Strdup();
#endif


/*****************************************************************************/
/*                              Loeschfunktionen                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void remove_list( LIST_POINTER * );
#else
extern void remove_list();
#endif
        /* Argument: LIST_POINTER * list                                     */
        /* Enfernen einer Liste aus dem Speicher.                            */

#if defined(__STDC__) || defined(__cplusplus)
extern void remove_marklist( MARKING * );
#else
extern void remove_marklist();
#endif
        /* Argument: MARKING * marklist                                      */
        /* Enfernen einer Markierungsliste aus dem Speicher.                 */

#if defined(__STDC__) || defined(__cplusplus)
extern void remove_net( LMPNet * );  
#else
extern void remove_net();  
#endif
        /* Argument: LMPNet * net                                            */
        /* Enfernen eines Netzes aus dem Speicher.                           */


/*****************************************************************************/
/*                          Initialisierungsfunktionen                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern LMPNet empty_net( void );
#else
extern LMPNet empty_net();
#endif
        /* Argumente: keine                                                  */
        /* Einrichten eines neuen, leeren Netzes; liefert einen Zeiger auf   */
        /* dieses neue Netz.                                                 */

#if defined(__STDC__) || defined(__cplusplus)
extern LMPNet gspn_ini( char * );
#else
extern LMPNet gspn_ini();
#endif
	/* Argumente: char * filename */
        /* initialisiert die Matrix 'net' mit dem im GreatSPN - Format in    */
        /* der Datei 'filename' abgelegten Netz.                             */

#if defined(__STDC__) || defined(__cplusplus)
extern LMPNet praktikum_ini( char * );
#else
extern LMPNet praktikum_ini();
#endif
	/* Argumente: char * filename */
        /* initialisiert die Matrix 'net' mit dem im Praktikums - Format     */
        /*  abgelegten Netz mit Namen 'netname'.                             */


/*****************************************************************************/
/*                        Funktionen auf Textdateien                         */
/*****************************************************************************/

extern char * NNinput_file;
extern int    NNinput_line;

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void NNError( char * );
#else
extern void NNError();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern char ReadWhiteSpace( FILE * );
#else
extern char ReadWhiteSpace();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern char ReadCharComment( FILE * );
#else
extern char ReadCharComment();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void ReadNewline( FILE *, unsigned char );
#else
extern void ReadNewline();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern bool TestSymbol( FILE *, char );
#else
extern bool TestSymbol();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void ReadSymbol( FILE *, char );
#else
extern void ReadSymbol();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern unsigned char ReadName( FILE *, char *, int );
#else
extern unsigned char ReadName();
#endif

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern int ReadNumber( FILE * );
#else
extern int ReadNumber();
#endif

/*****************************************************************************/
/*                             Ausgabefunktionen                             */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void write_matrix( LMPNet, char * );
#else
extern void write_matrix();
#endif
        /* Argumente: LMPNet net, char * filename                            */
        /* schreibt die Inzidenzmatrix des Netzes 'net' in die Datei         */
        /* 'filename'.                                                       */

#if defined(__STDC__) || defined(__cplusplus)
extern void write_netmarking( LMPNet, char * );
#else
extern void write_netmarking();
#endif
	/* Argumente: LMPNet net, char * filename                            */
        /* haengt die Anfangsmarkierung des Netzes 'net' an die Datei        */
        /* 'filename' an.                                                    */

#if defined(__STDC__) || defined(__cplusplus)
extern void show_matrix( LMPNet );
#else
extern void show_matrix();
#endif
        /* Argumente: LMPNet net                                             */
        /* Gibt die Inzidenzmatrix des Netzes 'net' auf dem Bildschirm aus.  */

#if defined(__STDC__) || defined(__cplusplus)
extern void show_multiset( LIST_POINTER );
#else
extern void show_multiset();
#endif
        /* Argumente: LIST_POINTER list */
        /* Gibt die Liste 'list' als Menge auf dem Bildschirm aus */
	/* Dient zur Anzeige der Vor- und Nachbereichsmengen */
	

/*****************************************************************************/
/*                  Funktionen zur Expansion des Netzes                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern Place fast_find_P_num( LMPNet, Cardinal32, Place * );
#else
extern Place fast_find_P_num();
#endif

#if defined(__STDC__) || defined(__cplusplus)
extern Transition fast_find_T_num( LMPNet, Cardinal32, Transition * );
#else
extern Transition fast_find_T_num();
#endif

#if defined(__STDC__) || defined(__cplusplus)
extern Place fast_insert_place( LMPNet, char *, Cardinal32, Place * );
#else
extern Place fast_insert_place();
#endif
       /* Argumente: LMPNet net, char *place_name,Cardinal32 startmarkierung*/
       /*            Place      * pl_last                                   */
       /* setzt einen Knoten fuer eine Stelle in die Matrix ein. */
       /* Die Stelle erhaelt den Namen 'place_name' und die Markierung in der*/
       /* Anfangsmarkierung von 'startmarkierung'. */
       /* Der Knoten wird in lexikographischer Ordnung in die Stellenliste */
       /* eingefuegt. */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition fast_insert_transition( LMPNet, char *, Transition * );
#else
extern Transition fast_insert_transition();
#endif
       /* Argumente: LMPNet net, char * transition_name */
       /*            Transition * tr_last                                   */
       /* Setzt einen Knoten fuer eine Transition in die Matrix 'net' ein.  */
       /* Die Transition erhaelt den Namen 'transition_name' */
       /* Der Knoten wird wird in lexikographischer Ordnung in die Transitionsliste eingefuegt */

#if defined(__STDC__) || defined(__cplusplus)
extern void fast_insert_con( LMPNet, char *, char *, Cardinal32, Cardinal32,
                             Place *, Transition *, STEdge *, TSEdge * );
#else
extern void fast_insert_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 transition_to_place_weight,                 */
       /*            Cardinal32 place_to_transition_weight                  */
       /*            Place      * pl_last                                   */
       /*            Transition * tr_last                                   */
       /*            STEdge     * st_last                                   */
       /*            STEdge     * ts_last                                   */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der   */
       /* Stelle mit Namen 'place_name' und der Transition mit dem Namen    */
       /* 'transition_name' in die Matrix 'net' ein. Falls ein solcher      */
       /* Knoten schon vorhanden ist, so werden die Gewichte entsprechend   */
       /* geaendert.                                                        */

#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge fast_insert_TP_num( LMPNet, Cardinal32, Cardinal32, Cardinal32,
                                  Place *, Transition *, TSEdge * );
#else
extern TSEdge fast_insert_TP_num();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 transition_to_place_weight                  */
       /*            Place      * pl_last                                   */
       /*            Transition * tr_last                                   */
       /*            STEdge     * ts_last                                   */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der   */
       /* Transition mit dem Namen 'transition_name' und der Stelle namens  */
       /* 'place_name' in die Matrix 'net' ein. Das Gewicht, die Staerke    */
       /* der Verbindung ist transition_to_place_weight. Falls der Knoten   */
       /* schon vorhanden ist, so wird das Gewicht entsprechend geaendert.  */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge fast_insert_PT_num( LMPNet, Cardinal32, Cardinal32, Cardinal32,
                                  Place *, Transition *, STEdge * );
#else
extern STEdge fast_insert_PT_num();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 place_to_transition_weight                  */
       /*            Place      * pl_last                                   */
       /*            Transition * tr_last                                   */
       /*            STEdge     * st_last                                   */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung                */
       /* zwischen der Stelle mit Namen 'place_name' und der                */
       /* Transition mit Namen 'transition_name' in die Matrix              */
       /* 'net' ein. Das Gewicht, die Staerke der Verbindung ist            */
       /* place_to_transition_weight Falls ein Knoten schon vorhanden       */
       /* ist, so werden die Gewichte entsprechend geaendert                */


/*****************************************************************************/
/*                  Funktionen zur Expansion des Netzes                      */
#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge fast_insert_TP_con( LMPNet, char *, char *, Cardinal32,
                                  Place *, Transition *, TSEdge * );
#else
extern TSEdge fast_insert_TP_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 transition_to_place_weight                  */
       /*            Place      * pl_last                                   */
       /*            Transition * tr_last                                   */
       /*            STEdge     * ts_last                                   */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der   */
       /* Transition mit dem Namen 'transition_name' und der Stelle namens  */
       /* 'place_name' in die Matrix 'net' ein. Das Gewicht, die Staerke    */
       /* der Verbindung ist transition_to_place_weight. Falls der Knoten   */
       /* schon vorhanden ist, so wird das Gewicht entsprechend geaendert.  */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge fast_insert_PT_con( LMPNet, char *, char *, Cardinal32,
                                  Place *, Transition *, STEdge * );
#else
extern STEdge fast_insert_PT_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 place_to_transition_weight                  */
       /*            Place      * pl_last                                   */
       /*            Transition * tr_last                                   */
       /*            STEdge     * st_last                                   */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung                */
       /* zwischen der Stelle mit Namen 'place_name' und der                */
       /* Transition mit Namen 'transition_name' in die Matrix              */
       /* 'net' ein. Das Gewicht, die Staerke der Verbindung ist            */
       /* place_to_transition_weight Falls ein Knoten schon vorhanden       */
       /* ist, so werden die Gewichte entsprechend geaendert                */


/*****************************************************************************/
/*                  Funktionen zur Expansion des Netzes                      */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern Place insert_place( LMPNet, char *, Cardinal32 );
#else
extern Place insert_place();
#endif
       /* Argumente: LMPNet net, char *place_name, Cardinal32 startmarkierung*/
       /* setzt einen Knoten fuer eine Stelle in die Matrix ein. */
       /* Die Stelle erhaelt den Namen 'place_name' und die Markierung in der*/
       /* Anfangsmarkierung von 'startmarkierung'. */
       /* Der Knoten wird in lexikographischer Ordnung in die Stellenliste eingefuegt. */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition insert_transition( LMPNet, char * );
#else
extern Transition insert_transition();
#endif
       /* Argumente: LMPNet net, char * transition_name */
       /* Setzt einen Knoten fuer eine Transition in die Matrix 'net' ein. */
       /* Die Transition erhaelt den Namen 'transition_name' */
       /* Der Knoten wird wird in lexikographischer Ordnung in die Transitionsliste eingefuegt */

#if defined(__STDC__) || defined(__cplusplus)
extern void insert_con( LMPNet, char *, char *, Cardinal32, Cardinal32 );
#else
extern void insert_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 transition_to_place_weight,                 */
       /*            Cardinal32 place_to_transition_weight                  */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der   */
       /* Stelle mit Namen 'place_name' und der Transition mit dem Namen    */
       /* 'transition_name' in die Matrix 'net' ein. Falls ein solcher      */
       /* Knoten schon vorhanden ist, so werden die Gewichte entsprechend   */
       /* geaendert.                                                        */

#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge insert_TP_con( LMPNet, char *, char *, Cardinal32 );
#else
extern TSEdge insert_TP_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 transition_to_place_weight                  */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der   */
       /* Transition mit dem Namen 'transition_name' und der Stelle namens  */
       /* 'place_name' in die Matrix 'net' ein. Das Gewicht, die Staerke    */
       /* der Verbindung ist transition_to_place_weight. Falls der Knoten   */
       /* schon vorhanden ist, so wird das Gewicht entsprechend geaendert.  */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge insert_PT_con( LMPNet, char *, char *, Cardinal32 );
#else
extern STEdge insert_PT_con();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name, */
       /*            Cardinal32 place_to_transition_weight */
       /* Setzt einen Verbindungsknoten fuer eine Verbindung zwischen der Stelle mit */
       /* Namen 'place_name' und der Transition mit Namen 'transition_name' in die Matrix 'net' ein. */
       /* Das Gewicht, die Staerke der Verbindung ist place_to_transition_weight */
       /* Falls ein Knoten schon vorhanden ist, so werden die Gewichte entsprechend geaendert */


/*****************************************************************************/
/*                 Funktionen zur Verkleinerung des Netzes                   */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern bool delete_T_to_P_connection( LMPNet, char *, char * );
#else
extern bool delete_T_to_P_connection();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name */
       /* Loescht eine Verbindung von der Transition mit Namen */
       /* 'transition_name' zur Stelle */
       /* mit Namen 'place_name' aus der Matrix 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern bool delete_P_to_T_connection( LMPNet, char *, char * );
#else
extern bool delete_P_to_T_connection();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name */
       /* Loescht eine Verbindung von der Stelle mit Namen 'place_name' */
       /* zur Transition */
       /* mit Namen 'transition_name' aus der Matrix 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern Place delete_place_direct( LMPNet, Place );
#else
extern Place delete_place_direct();
#endif
       /* Argumente: LMPNet net, Place place */
       /* Loescht die Zeile fuer die Stelle 'place' */
       /* Rueckgabe: Zeiger auf naechste Stelle im Netz */

#if defined(__STDC__) || defined(__cplusplus)
extern bool delete_place( LMPNet, char * );
#else
extern bool delete_place();
#endif
       /* Argumente: LMPNet net, char * place_name */
       /* Loescht die Zeile fuer die Stelle mit Namen 'place_name' */
       /* aus der Matrix 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition delete_transition_direct( LMPNet, Transition );
#else
extern Transition delete_transition_direct();
#endif
       /* Argumente: LMPNet net, Transition trans */
       /* Loescht die Spalte fuer die Transition 'trans' */
       /* Rueckgabe: Zeiger auf naechste Transition im Netz */

#if defined(__STDC__) || defined(__cplusplus)
extern bool delete_transition( LMPNet, char * );
#else
extern bool delete_transition();
#endif
       /* Argumente: LMPNet net, char * transition_name */
       /* Loescht die Spalte fuer die Transition mit Namen 'transition_name' */
       /* aus der  Matrix 'net' */ 


/*****************************************************************************/
/*                      Netz-Informations-Funktionen                         */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 place_number( LMPNet );
#else
extern Cardinal32 place_number();
#endif
	/* Argumente: LMPNet net */
        /* zaehlt die Stellen im Netz 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 trans_number( LMPNet );
#else
extern Cardinal32 trans_number();
#endif
	/* Argumente: LMPNet net */
        /* zaehlt die Transitionen im Netz 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern bool name_free( LMPNet, char * );
#else
extern bool name_free();
#endif
	/* Argumente: LMPNet net, char * name */
	/* liefert True, falls der Stellen- bzw. Transitionsname noch nicht */
        /* benutzt ist, sonst False */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal max_placename_length( LMPNet );
#else
extern Cardinal max_placename_length();
#endif
	/* Argumente: LMPNet net */
        /* liefert die Laenge des laengsten Stellennamen in 'net' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal max_transitionname_length( LMPNet );
#else
extern Cardinal max_transitionname_length();
#endif
	/* Argumente: LMPNet net */
	/* liefert die Laenge des laengsten Transitionsnamens */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 TP_connection( LMPNet, char *, char * );
#else
extern Cardinal32 TP_connection();
#endif
        /* Argumente: LMPNet net, char * place_name, * transition_name */
	/* liefert die Staerke der Verbindung von 'transition_name' zu 'place_name' */
	/* Leseoperation auf Feld 'weight1' der Verbindungsknoten */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 PT_connection( LMPNet, char *, char * );
#else
extern Cardinal32 PT_connection();
#endif
        /* Argumente: LMPNet net, char * place_name, * transition_name */
	/* Liefert die Staerke der Verbindung von 'place_name' zu 'transition_name' */
	/* Leseoperation auf Feld 'weight2' der Verbindungsknoten */
         
#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 row_index( LMPNet, char * );
#else
extern Cardinal32 row_index();
#endif
       /* Argumente: LMPNet net, char * place_name */
       /* Liefert die Zeilennummer der Zeile 'place_name'. */
       /* Ist 'place_name' ein Transitionsname oder ist die Stelle nicht */
       /* im Netz enthalten, liefert die Funktion 0 */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 column_index( LMPNet, char * );
#else
extern Cardinal32 column_index();
#endif
       /* Argumente: LMPNet net, char * transition_name */
       /* Liefert die Spaltennummer der Spalte 'transition_name'. */
       /* Ist 'transition_name' ein Stellenname, liefert die Funktion 0 */

#if defined(__STDC__) || defined(__cplusplus)
extern bool is_place_successor( LMPNet, char *, char * );
#else
extern bool is_place_successor();
#endif
        /* Argumente: LMPNet net, char * place_name, char * transition_name */
	/* liefert True, falls die Transition mit dem Namen transition_name */
        /* Nachfolger der Stelle namens place_name ist, sonst False.        */

#if defined(__STDC__) || defined(__cplusplus)
extern bool is_trans_successor( LMPNet, char *, char * );
#else
extern bool is_trans_successor();
#endif
        /* Argumente: LMPNet net, char * place_name, char * transition_name */
	/* liefert True, falls die Stelle mit dem Namen place_name          */
        /* Nachfolger der Transition namens trans_name ist, sonst False.    */


/*****************************************************************************/
/*               Schnittstellenfunktionen zu Markierungslisten               */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern MARKING copy_start_marking( LMPNet );
#else
extern MARKING copy_start_marking();
#endif
         /* Argumente: LMPNet net */
	 /* liefert einen Zeiger auf eine Liste vom Typ MARKING */
	 /* Jeder Knoten enthaelt einen Stellennamen und dessen Markierung */

#if defined(__STDC__) || defined(__cplusplus)
extern bool change_start_marking( LMPNet, MARKING );
#else
extern bool change_start_marking();
#endif
         /* Argumente: LMPNet net, MARKING marking */
	 /* Kopiert die Markierung 'marking' in die Matrix 'net' */
	 /* Die Felder 'weight2' der Stellenknoten der Matrix werden mit den Werten der Knoten in 'marking' */
	 /* belegt fuer die Knoten, die den gleichen Nmane tragen */
	 /* Nach erfolgreicher Ersetzung liefert die Funktion True zurueck */

#if defined(__STDC__) || defined(__cplusplus)
extern bool change_place_marking( LMPNet, char *, Cardinal32 );
#else
extern bool change_place_marking();
#endif
         /* Argumente: LMPNet net, char * place_name, Cardinal32 value */
	 /* Aendert den Startmarkierungswert der Stelle 'place_name' in 'net' zu 'value' */
	 /* Schreiboperation auf das Feld 'weight2' der Stellenknoten */
	 /* Nach erfolgreicher Ersetzung liefert die Funktion True zurueck */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 marking( LMPNet, char * );
#else
extern Cardinal32 marking();
#endif
         /* Argumente: LMPNet net, char * place_name */
	 /* liefert die Startmarkierung der Stelle mit Namen 'place_name' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 size_of_marking( MARKING );
#else
extern Cardinal32 size_of_marking();
#endif
	/* Argumente: MARKING marking */
        /* liefert die Laenge der Liste der Markierung, d.h. die Anzahl von Stellen, */
	/* fuer die die Markierung reicht */


/*****************************************************************************/
/*                               Markierungen                                */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 place_predecessor_number( LMPNet, char * ); 
#else
extern Cardinal32 place_predecessor_number(); 
#endif
	/* Argumente: LMPNet net, char * P_name */
	/* Liefert die Anzahl der Elemente im Vorbereich von 'P_name' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 trans_predecessor_number( LMPNet, char * ); 
#else
extern Cardinal32 trans_predecessor_number(); 
#endif
	/* Argumente: LMPNet net, char * T_name */
	/* Liefert die Anzahl der Elemente im Vorbereich von 'T_name' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 place_successor_number( LMPNet, char * );
#else
extern Cardinal32 place_successor_number();
#endif
	/* Argumente: LMPNet net, char * P_name */
	/* Liefert die Anzahl der Elemente im Nachbereich von 'P_name' */

#if defined(__STDC__) || defined(__cplusplus)
extern Cardinal32 trans_successor_number( LMPNet, char * );
#else
extern Cardinal32 trans_successor_number();
#endif
	/* Argumente: LMPNet net, char * T_name */
	/* Liefert die Anzahl der Elemente im Nachbereich von 'T_name' */

#if defined(__STDC__) || defined(__cplusplus)
extern LIST_POINTER place_successor_list( LMPNet, char * );
#else
extern LIST_POINTER place_successor_list();
#endif
        /* Argumente: LMPNet net, char * P_name */
	/* Liefert einen Zeiger auf eine Liste (vom Typ LIST_POINTER), */
        /* die die Namen der Transitionen ( bzw. Zeiger auf die Namen )*/
        /* im Nachbereich von 'P_name' enthaelt.                       */

#if defined(__STDC__) || defined(__cplusplus)
extern LIST_POINTER trans_successor_list( LMPNet, char * );
#else
extern LIST_POINTER trans_successor_list();
#endif
        /* Argumente: LMPNet net, char * T_name */
	/* Liefert einen Zeiger auf eine Liste (vom Typ LIST_POINTER), */
        /* die die Namen der Stellen ( bzw. Zeiger auf die Namen )     */
        /* im Nachbereich von 'T_name' enthaelt.                       */

#if defined(__STDC__) || defined(__cplusplus)
extern LIST_POINTER place_predecessor_list( LMPNet, char * );
#else
extern LIST_POINTER place_predecessor_list();
#endif
        /* Argumente: LMPNet net, char * P_name */
	/* Liefert einen Zeiger auf eine Liste (vom Typ LIST_POINTER), */
        /* die die Namen der Transitionen ( bzw. Zeiger auf die Namen )*/
        /* im Vorbereich von 'P_name' enthaelt.                        */

#if defined(__STDC__) || defined(__cplusplus)
extern LIST_POINTER trans_successor_list( LMPNet, char * );
#else
extern LIST_POINTER trans_successor_list();
#endif
        /* Argumente: LMPNet net, char * T_name */
	/* Liefert einen Zeiger auf eine Liste (vom Typ LIST_POINTER), */
        /* die die Namen der Stellen ( bzw. Zeiger auf die Namen )     */
        /* im Vorbereich von 'T_name' enthaelt.                        */


/*****************************************************************************/
/*                            Matrixinterne Funktionen                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern void num_net_nodes( LMPNet );
#else
extern void num_net_nodes();
#endif
       /* Argumente: LMPNet net                                            */
       /* Diese Funktion numeriert die Stellenknoten des uebergebenen      */
       /* Netzes in der 'numname'-Komponente von 1..net->NumPlaces durch.  */
       /* Analog werden die 'numname'-Eintraege der Transitionsknoten des  */
       /* Netzes durchnumeriert mit 1..net->NumTransitions.                */
       /* ACHTUNG!!! Diese Funktion muss aufgerufen worden sein, bevor     */
       /* eine der folgenden vier Funktionen benutzt werden kann!          */

#if defined(__STDC__) || defined(__cplusplus)
extern Place find_P_num( LMPNet, Cardinal32 );
#else
extern Place find_P_num();
#endif
       /* Argumente: LMPNet net, Cardinal32 place_numname                  */
       /* Liefert einen Pointer auf den Stellenknoten mit Namen            */
       /* 'place_numname' zurueck.                                         */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition find_T_num( LMPNet, Cardinal32 );
#else
extern Transition find_T_num();
#endif
       /* Argumente: LMPNet net, Cardinal32 transition_numname             */
       /* Liefert Pointer auf Transitionsknoten mit Namen                  */
       /* 'transition_numname' zurueck.                                    */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge find_PT_connection_num( LMPNet, Cardinal32, Cardinal32 );
#else
extern STEdge find_PT_connection_num();
#endif
       /* Argumente: LMPNet net, Cardinal32 place_numname,                 */
       /*                        Cardinal32 transition_numname             */
       /* Liefert Pointer auf den Knoten, der die Kante von Stelle         */
       /* place_numname zur Transition transition_numname repraesentiert.  */
       /* Ist ein solcher Knoten nicht vorhanden, liefert die Funktion     */
       /* einen NULL-Pointer.                                              */

#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge find_TP_connection_num( LMPNet, Cardinal32, Cardinal32 );
#else
extern TSEdge find_TP_connection_num();
#endif
       /* Argumente: LMPNet net, Cardinal32 place_numname,                 */
       /*                        Cardinal32 transition_numname             */
       /* Liefert Pointer auf den Knoten, der die Kante von Transition     */
       /* transition_numname zur Stelle place_numname repraesentiert.      */
       /* Ist ein solcher Knoten nicht vorhanden, liefert die Funktion     */
       /* einen NULL-Pointer.                                              */

/*****************************************************************************/
/*                            Matrixinterne Funktionen                       */
/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)
extern Place find_P( LMPNet, char * );
#else
extern Place find_P();
#endif
       /* Argumente: LMPNet net, char * place_name */
       /* Liefert einen Pointer auf den Stellenknoten mit Namen        */
       /* 'place_name' zurueck */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition find_T( LMPNet, char * );
#else
extern Transition find_T();
#endif
       /* Argumente: LMPNet net, char * transition_name */
       /* Liefert Pointer auf Transitionsknoten mit Namen              */
       /* 'transition_name' zurueck */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge find_PT_connection( LMPNet, char *, char * );
#else
extern STEdge find_PT_connection();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name */
       /* Liefert Pointer auf den Knoten, der die Kante von Stelle         */
       /* place_name zur Transition transition_name repraesentiert.        */
       /* Ist ein solcher Knoten nicht vorhanden, liefert die Funktion     */
       /* einen NULL-Pointer.                                              */

#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge find_TP_connection( LMPNet, char *, char * );
#else
extern TSEdge find_TP_connection();
#endif
       /* Argumente: LMPNet net, char * place_name, char * transition_name */
       /* Liefert Pointer auf den Knoten, der die Kante von Transition     */
       /* transition_name zur Stelle place_name repraesentiert.            */
       /* Ist ein solcher Knoten nicht vorhanden, liefert die Funktion     */
       /* einen NULL-Pointer.                                              */

#if defined(__STDC__) || defined(__cplusplus)
extern Place next_place( LMPNet, Place );
#else
extern Place next_place();
#endif
       /* Argumente: LMPNet net, Place place_pointer */
       /* Falls place_pointer ein NULL-Zeiger ist, liefert die Funktion    */
       /* einen Pointer auf die erste Stelle der Matrix . */
       /* Andernfalls gibt die Funktion einen Zeiger auf die naechste      */
       /* Stelle zurueck. */

#if defined(__STDC__) || defined(__cplusplus)
extern Transition next_transition( LMPNet, Transition );
#else
extern Transition next_transition();
#endif
       /* Argumente: LMPNet net, Transition transition_pointer */
       /* Falls transition_pointer ein NULL-Zeiger ist liefert die       */
       /* Funktion einen Pointer auf */
       /* die erste Transition der Matrix. */
       /* Andernfalls gibt die Funktion einen Zeiger auf die naechste    */
       /* Transition zurueck. */

#if defined(__STDC__) || defined(__cplusplus)
extern STEdge next_PT_connection( LMPNet, STEdge );
#else
extern STEdge next_PT_connection();
#endif
       /* Argumente: LMPNet net, STEdge connection_pointer */

#if defined(__STDC__) || defined(__cplusplus)
extern TSEdge next_TP_connection( LMPNet, TSEdge );
#else
extern TSEdge next_TP_connection();
#endif
       /* Argumente: LMPNet net, TSEdge connection_pointer */

/*****************************************************************************/


#endif /* __NEUNETZE_H__ */


/*****************************************************************************/
/* END OF FILE neunetze.h                                                    */
/*****************************************************************************/

