/*****************************************************************************/
/*  parser.h                                                                 */
/*  C-Headerdatei: Hilfsroutinen fuer den HL-Expression-Parser               */
/*  Mo.,13.02.1995 - Mo.,13.02.1995                           Stefan Roemer  */
/*****************************************************************************/


#ifndef __PARSER_H__
#define __PARSER_H__


#include "main.h"


/*****************************************************************************/

typedef struct btree_
  {
    int           token;
    struct btree_ * left,
                  * right;
  } btree_elem,
    * btree;

extern btree STree;

/*****************************************************************************/

#if defined(__STDC__) || defined(__cplusplus)

extern btree new_node( int, btree, btree );

#else

extern btree new_node();

#endif

/*****************************************************************************/
/* Initialisierungsfunktion, die die vom Parser benutzten globalen Variablen */
/* ( insbesondere die Symboltabellen ) einrichtet. Diese Funktion ist unbe-  */
/* dingt zu Beginn des Hauptprogramms ( vor "Benutzung" des Parsers ) aufzu- */
/* rufen!                                                                    */
/*****************************************************************************/

extern void InitAllSymbolTables(void);
  /* GLOBAL SymTab   SymbolTabelle, (schreibend)
                     VarTable;      (schreibend)
            Cardinal MaxActions,    (schreibend)
                     MaxVars,       (schreibend)
                     ActCounter,    (schreibend)
                     VarCounter,    (schreibend)


/*****************************************************************************/
/* Funktion zur Umwandlung eines Aktionsnamens `Name` (String) in den ent-   */
/* sprechenden Aktionscode (Cardinal bzw. Action), der als Ergebnis zurueck- */
/* geliefert wird. Falls der Name noch nicht in der Aktions-Symboltabelle    */
/* vorhanden ist, wird dieser hinzugefuegt und ein entsprechender Code fuer  */
/* ihn berechnet.                                                            */
/*****************************************************************************/

extern Action StrToAction(char *Name);
  /* GLOBAL: SymTab   SymbolTable; (schreibend)
             Cardinal ActCounter;  (schreibend)
             MultiSet mmDummy,     (schreibend)
                      mmLeaf       (schreibend) */


/*****************************************************************************/
/* Funktion zur Umwandlung eines Aktionscodes `akt` in den entsprechenden    */
/* Namen, der in der bereitgestellten Zeichenkette `str` abgelegt wird.      */
/* `str` sollte also ein Zeiger auf einen zur Aufnahme des Aktionsnamens ge- */
/* nuegend gross initialisierten Speicherbereich sein; `str` wird schliess-  */
/* lich als Funktionsergebnis geliefert.                                     */
/*****************************************************************************/

extern char *ActionToStr(Action akt,char *str);
  /* GLOBAL: SymTab SymbolTable; (lesend)  */


/*****************************************************************************/
/* Funktion zum Umwandeln eines Variablennamens `Name` (String) in den ent-  */
/* sprechenden Variablencode (Var bzw. Cardinal), der als Funktionsergebnis  */
/* geliefert wird. Falls der Variablenname bisher noch nicht existiert, so   */
/* wird ein neuer Code fuer diesen ermittelt und der Name in die Variablen-  */
/* Symboltabelle aufgenommen.                                                */
/*****************************************************************************/

extern Var StrToVariable(char *Name);
  /* GLOBAL: SymTab   VarTable;   (schreibend)
             Cardinal VarCounter; (schreibend)  */


/*****************************************************************************/
/* Funktion zum Umwandeln des Variablencodes `var` in den entsprechenden     */
/* Variablennamen `str`, der in den dafuer zu initialisierenden Speicher-    */
/* bereich kopiert wird. Rueckgabe ist der Zeiger `str`.                     */
/*****************************************************************************/

extern char *VariableToStr(Var var,char *str);
  /* GLOBAL: SymTab VarTable; (lesend)  */


/*****************************************************************************/
/* Funktion zur Freigabe des von der Symboltabelle `table` belegten Haupt-   */
/* speicherplatzes; `table` ist hiernach undefiniert!                        */
/*****************************************************************************/

extern void RemoveSymbolTable(SymTab *table);


/*****************************************************************************/
/* Prozedur zur Darstellung der Symboltabelle `table` auf dem Bildschirm.    */
/* Nur zu Analysezwecken eingerichtet!                                       */
/*****************************************************************************/

extern void PrintSymbolTable(SymTab table);


/*****************************************************************************/


#endif


/*****************************************************************************/
/* END OF FILE parser.h                                                      */
/*****************************************************************************/

