/***********************************/
/* qq_defs.h                       */
/* Frank Wallner  Sep. 98, Apr 99  */
/***********************************/

#ifndef MAXLINE
#define MAXLINE         10000
#endif

#ifndef MAXCHAR
#define MAXCHAR         250
#endif


/******************************************************/
/*  -------- long OR char as BitArray-Base ---------  */
/*                                                    */
/*  Have ONE of the following two lines un-commented  */
/*                                                    */
#define __LONG__BASE__ 1 
/* #define __CHAR__BASE__ 1 */
/*                                                    */
/******************************************************/

#ifdef __LONG__BASE__
#define BA_TYPE 		unsigned long    	
#define BA_BASE			31		
#define BA_START	   	1073741824	
#endif

#ifdef __CHAR__BASE__
#define BA_TYPE 		unsigned char 
#define BA_BASE			7
#define BA_START	   	64	
#endif

typedef BA_TYPE			BitArray;
#define BA_ALL			(BitArray) ( -1 )
#define BA_SZ			sizeof(BitArray)
#define BA_NULL			(BitArray *) NULL
#define EmptyS			(BitArray *) NULL
#define ba_idx(x)		( ( ((x)/BA_BASE) + 1) * BA_BASE )

/***************************************************************************/

#define IL_NULL			(Int_List *) NULL
#define TEL_NULL		(TE_List *) NULL
#define UC_NULL		 	(UC_List *) NULL
#define _ZZ_ \
   "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz"

#define _LINE_			")\n- - - - - - - - - - - - - - - - - -\n"



#define TMP_FILE1		"xxx.yyy"
#define TMP_FILE2		"xxx.yyz"
#define __MY_POS 		100
#define SCHED_X_POS		250
#define SCHED_F_Y_POS		300
#define SCHED_S_Y_POS		330
#define INFTY			33333333
#define MAX2_COUNTEREXAMPLES   	25000
#define MAXTRANSLIST		500
#define MAXMEANING		1000
#define MAXNAME			1000
#define NEW_TPL     		(trans_PL *) MyMalloc(sizeof(trans_PL))
#define NEW_INTLIST             (Int_List *) MyMalloc(sizeof(Int_List))
#define NEW_TELIST              (TE_List *) MyMalloc(sizeof(TE_List))
#define NEW_UCLIST              (UC_List *) MyMalloc(sizeof(UC_List))
#define NEW_CHAR(x)             (char *) MyMalloc((x)*sizeof(char))
#define MIN(x,y)        	((x)<(y) ? (x) : (y))
#define MAX(x,y)        	((x)>(y) ? (x) : (y))

/***************************************************************************/
/* Defines the propositional logic operators */
/* as ->op entry in the trans_PL-Nodes       */
/*********************************************/
#define TPL_NOT		1
#define TPL_TRUE 	2
#define TPL_FALSE 	3
#define TPL_AND 	4
#define TPL_OR		5
#define TPL_OP(x)	( ((int) ((x)->op)) < 6 )
#define TPL_PROP(x)	( ((int) ((x)->op)) > 5 )
#define TPL_AND_OP(x)	( TPL_OP((x)) && ( (x)->op == (char *) TPL_AND) )
#define TPL_OR_OP(x)	( TPL_OP((x)) && ( (x)->op == (char *) TPL_OR) )
#define TPL_NEG_OP(x)	( TPL_OP((x)) && ( (x)->op == (char *) TPL_NOT) )
#define TPL_TRUE_OP(x)	( TPL_OP((x)) && ( (x)->op == (char *) TPL_TRUE) )
#define TPL_FALSE_OP(x)	( TPL_OP((x)) && ( (x)->op == (char *) TPL_FALSE) )
#define BT_TRUE		"}1"
#define BT_FALSE        "}0"

/***************************************************************************/

/*** the labels of the Edges of the graph (edge->final) ***/
#define F_00			6 	/* loc_conf, no Buechi Trans   */
#define F_01			22	/* loc_conf, accepting Trans   */
#define F_02			2	/* loc_conf, NonAccBuechi Trans*/
#define F_10			3	/* ext_conf, no Buechi Trans   */
#define F_11			11	/* ext_conf, accepting Trans   */
#define F_12			1	/* ext_conf, NonAccBuechi Trans*/
#define F__1(x)			( (x) > F_00 )
#define F__0(x)			( ( (x) % 3 ) == 0 )
#define F__2(x)			( (x) < F_10 )
#define F_1_(x)			( ( (x) % 2 ) != 0 )
#define F_0_(x)			( ( (x) % 2 ) == 0 )

/************************************************************************/

/*** info-bits of Events ***/
/* IMMEDIATE_STOP: the local config of x marks an 'accept_all'- state 
   ONSTACK: used for representative cutoffs and 'simple' events 
	    (cf below) for Tarjan's depth-first-search
   CHCKBC:  is set if corr(cutoff) \in [cutoff], but there is no
            accepting event in between, and for cutoff it is already
	    checked if the maxBuchi of cutoff indicates an accepting
	    cycle of the Buchi-automaton 
   VISIBLE, CUTOFF, CORR: is what it is called
   BUECHI:  x is a transition of the automaton
   FINAL:   ... and marks an accepting place
   VISITED: used for Tarjan's dfs-algor. */
#define IMMEDIATE_STOP		255
#define VISIBLE			64
#define ONSTACK			32
#define CHCKBC			32
#define CUTOFF                  16
#define CORR                    8
#define BUECHI                  4
#define FINAL                   2
#define VISITED                 1

/*** info-bits of ENodes ***/
/* LOCAL:   this ENode is shifted back to a local config
   SHIFTED: this ENode is already shifted 
   CHECKED: the adjacents of this ENode are alreday computed */
#define LOCAL                   64
/*      ONSTACK */
#define SHIFTED			16
#define CHECKED			8
#define INITIALIZED             4
/*      VISITED */


/*** info-bits of Buechi-Nodes ***/
/* NOTE CAREFULLY: *********************************************
   the Buechi-Nodes are organized as events !, i.e., the global
   variable 'ep' has to be set to 'buechi_nodes' (and then
   immediately back to 'events' ) whenever Tarjan's algorithm 
   is performed for the automaton 
****************************************************************/
/* Buechi-Nodes are the states of the Buchi-automaton;
   ACCEPTALL: this node is an 'accept_all' state
   SELFLOOP:  this node is self-looped */
/*      ONSTACK	*/
#define ACCEPTALL		8
#define SELFLOOP		4
/*      FINAL   */
/*      VISITED */



/*** extra information for correspondings and cutoffs, stored in ***/
/*** the int-Array ->ip ********************************************/

#define ip_mem(x)		((ep+(x))->ip)

/*** ->ip[0..(X-1)], where X = ....  ***********************************/
#define REPR_CUTOFF_IP		9
#define ALL_CUTOFF_IP		5
#define CORR_IP			9
#define ENCORR_IP		8
#define EN_IP			5

/**** for correspondings and for all cutoff events ****/
	/* maxBuchi, maxAcpt are the maximal buchi, accepting events in */
   	/* the local configuration of x *********************************/
#define maxBuchi(x)		*(((ep+(x))->ip)+0)
#define maxAcpt(x)		*(((ep+(x))->ip)+1)

/**** for corresponding events ****/
	/* CUT(x) is an Int_List of conditions **************************/
	/* enodes(x) is a List of ENodes ********************************/
	/* __lastEdge(x) points to the last Edge included into x->adjs **/
	/* nextcorrs(x) is an UC_List of the ->info-pointers of all
	   direct causal successors of x that are corresponding events **/
	/* confls(x) is a BitArray of size x, containing cutoffs in
	   conflict with e_x ********************************************/
	/* lastENode(x) points to the last ENode included into enodes(x)*/
	/* cutoff_quantity(x) is the number of cutoffs with correspond x*/
	/* cutoff_num(x,y) is the y'th cutoff for corr e_x **************/
#define CUT(x)			*(((ep+(x))->ip)+2)
#define enodes(x)		*(((ep+(x))->ip)+3)
#define __lastEdge(x)		*(((ep+(x))->ip)+4)
#define nextcorrs(x)		*(((ep+(x))->ip)+5)
#define confls(x)		*(((ep+(x))->ip)+6)
#define lastENode(x)		*(((ep+(x))->ip)+7)
#define cutoff_quantity(x)	*(((ep+(x))->ip)+8)
#define cutoff_num(x,y)		*(((ep+(x))->ip)+8+(y))

/**** for all cutoff events ****/
	/* correspond(x) is the corresponding event of cutoff x *********/
	/* ENCorr_co_cands(x) is a list of ENCorrs with x in co *********/
	/* maxCorr(x) is a maximal corresponding event in config(x) *****/
#define correspond(x)        	*(((ep+(x))->ip)+2)
#define ENCorr_co_cands(x)	*(((ep+(x))->ip)+3)
#define maxCorr(x) 	        *(((ep+(x))->ip)+4)
#define lastEdge(x)		__lastEdge( correspond( (x) ) )
#define representative(x)	cutoff_num(correspond( (x) ), 1 )

/**** only for Representative cutoffs ****/
	/* dfn, lowlink are used for the Tarjan- search-Algorithm (scc) */
	/* scc(x) is the scc-num that contains x ************************/
	/* scc_idx(x)=n means that x is the nth element in scc(x) *******/
	/* acc_succ(x)=y(!=0) means that y is an adjacent event of x, 
	   and the edge x->y is accepting *******************************/
#define dfn(x)             	*(((ep+(x))->ip)+5)
#define lowlink(x)              *(((ep+(x))->ip)+6)
#define scc(x) 	        	*(((ep+(x))->ip)+7)
#define scc_idx(x) 	        *(((ep+(x))->ip)+8)
#define acc_succ(x)	 	*(((ep+(x))->ip)+8)


/*** for correspondings of ENodes ***/
	/* these are events that are neither cutoffs nor
	   correspondings events, but correspondings of an ENode,
	   i.e., the (global) config [cutoff] v [corr] is shifted
	   to [x] ((sometimes called 'simple' events *********************/
/* ip[0]: maxBuchi
   ip[1]: maxAcpt
   ip[2]: CUT
   ip[3]: enodes
   ip[4]: maxCorr
   ip[5]: dfn     ( & lastEdge )
   ip[6]: lowlink ( & S_lastENode )
   ip[7]: scc_idx ( & lastENode ) 
   ip[7]: acc_succ */
#define S_lastEdge(x)		*(((ep+(x))->ip)+5)
#define S_lastENode(x)		*(((ep+(x))->ip)+6)

/*** for Enodes ***/
#define EN_lowlink(x)		*(((x)->ip)+0)
#define EN_dfn(x)		*(((x)->ip)+1)
#define EN_scc(x)		*(((x)->ip)+2)
#define EN_scc_idx(x)		*(((x)->ip)+3)
#define EN_acc_succ(x)		*(((x)->ip)+3)
#define EN_maxBuchi(x)		*(((x)->ip)+4)
#define EN_enodes(x)		( ( EN_IS(LOCAL, (x) ) ? \
				  ( (ENode *) (enodes((x)->corr)) ) : \
				  ( (x)->corr->co_en ) \
                                )
#define search_arg(x)		( - ( (int) (x) ) )
#define search_enod(x)		( (ENode *) ( - (x) ) )

/*** the global corrs of ENodes are distributed over a hashtable
     of size E_fin/__HT__ ************************************************/
#define __HT__			4
#define ENC_table_idx(x)	( (x) % (E_Fin/__HT__) )

#define IS(x,y)                 ( (x) == ( ((ep+(y))->info) & (x) ) )
#define IS_ON_STACK(x)		IS(ONSTACK,(x))
#define EN_IS(x,y)              ( (x) == (( (y)->info ) & (x) ) )

/**** ep is a global array of Events ****/
/* - preE(x) are the predecessing events of e_x 
   - preB(x)/postB(x) the input/output conditions of e_x
   - config(x) is the local configuration of e_x
   - T_NAME(x) is the transition-Name of e_x
   - pre_/post_size(x) is the number of pre-/post-conditions of e_x
   - postplaces(x) (ONLY for visible/Buechi events !!) are the 
   		OBSERVED places marked by firing T_NAME(x)
   - preplaces(x) (ONLY for visible events !!) are the
		OBSERVED places unmarked by firing T_NAME(x)
   - max_post is the maximal post condition of e_x
		(if e_x is a cutoff, cutoff_max_post is the maximal
		 postcondition of the (max.) direct predecessor of e_x)
   - adjacents(x) are the adjacent-events of cutoff e_x
		( ! organized at correspond(x) ! )
*/

#define preE(x)			((ep+(x))->pre)
#define preB(x)			((ep+(x))->rem)
#define postB(x)		((ep+(x))->add)
#define config(x)		((ep+(x))->config)
#define T_NAME(x)		((ep+(x))->tname)
#define pre_size(x)		((ep+(x))->pre_post[0])
#define post_size(x)		((ep+(x))->pre_post[1])
#define postplaces(x)		((ep+(x))->post)
#define preplaces(x)		((ep+(x))->preplaces)
#define max_post(x)		postB( (x) )[ ( post_size( (x) ) - 1 ) ]
#define cutoff_max_post(x)	max_post( preE((x))[(pre_size((x))-1)] )
#define INFO(x)                 (ep+(x))->info
#define adjacents(x)		(ep+(correspond((x))))->adjs
#define __adjacents(x)		(ep+(x))->adjs
#define B_adjacents(x)		(ep+(x))->adjs
#define B_dfn(x)           	*(((ep+(x))->ip))
#define B_lowlink(x)            *(((ep+(x))->ip)+1)


/*** the stack for Tarjan's dfs ***/
#define push(x)         	stack[stackpointer++]=(x)
#define pop             	((stackpointer>0)?stack[--stackpointer]:-1)
#define b_push(x)         	b_stack[b_stackpointer++]=(x)
#define b_pop                  ((b_stackpointer>0)?b_stack[--b_stackpointer]:-1)

/*** macros for manipulating the configurations and cuts ***/
#define elem_of(x,y)		( (x)[(y)/BA_BASE] & bitmask[(y)%BA_BASE] )
#define del_entry(x,y)		(x)[(y)/BA_BASE] &= ~(bitmask[(y)%BA_BASE]) 
#define add_entry(x,y)	  	(x)[(y)/BA_BASE] |=  (bitmask[(y)%BA_BASE]) 

/**************************************************************************/

#define InitConfig(x) \
	if ( config( (x) ) == BA_NULL ) { \
	     config( (x) ) = (ConfS) MyMalloc((2+((x)/BA_BASE))*BA_SZ);\
	     Configuration( config( (x) ), (x) ); }

#define InitMaxBuchi(x) \
	if (maxBuchi( (x) ) == (-1) ) { \
	    if ( IS(FINAL, (x) ) ) \
		 maxAcpt( (x) ) = maxBuchi( (x) ) = (x); \
	    else Find_AcptBue(config( (x) ), (x) ); }

/**************************************************************************/
/* Error/Warning-Messages      */
/*******************************/

#define WARN_RENAME( x )	fprintf(stderr, \
	"\nWARNING: Transition \"@%s\" is renamed --> \"_AT_%s\" ", \
	( x ), ( x ) )

#define WARN_SEV_PL_MEAN( x ) 	fprintf(stderr, \
	"\nWARNING: Several Places with the same meaning \"%s\".\n", \
	( x ) )

#define WARN_SEV_PL_NAME( x ) 	fprintf(stderr, \
	"\nWARNING: Several Places with the same name \"%s\".\n", \
	( x ) )

#define WARN_CONFLICT_PL( x , y , z ) 	fprintf(stderr, \
 "\nWARNING: ambiguous name/meaning : \nplace \"%s\" vs. place \"%s\"\
 (with meaning \"%s\").\n(continuing with place \"%s\"...)", \
 ( x ) , ( y ), ( z ) , ( y ))

#define ERR_NO_PL_MEAN_NAME( x ) 	fprintf(stderr, \
	"\nERROR: There is no place with meaning/name \"%s\".\n", \
	( x ) )

#define MAYBE_ENTRY 	fprintf(stderr, \
	"\n(Maybe an entry/exit place, discarded by initialization?)\n\n")

#define ERR_NO_TRANS( x ) 	fprintf(stderr, \
	"\nERROR: synchronize(): No Transition \"%s\".\n", \
	( x ) )

#define ERR_PROD( x )		fprintf(stderr, \
	"\nPROD - Net ERROR: No place with name \"%s\"\n", \
	( x ) )

#define DNF_ERR			fprintf(stderr, \
	"Found illegal operator during DNF conversion\n")



#define EXIT_NO_OUTPUT		printf("\n");\
		printf("Synchronisation: %f\n", sync_time_f);\
		printf("Unfolding:       %f\n", unf_time_f);\
		printf("Elapsed Time:    %f\n\n", all_time_f);\
		fclose(F_OUT); fdopen(stdout_fd, "a")

/**************************************************************************/
/* End of File qq_defs.h        */
/********************************/
