/***********************************/
/* qq_types.h                      */
/* Frank Wallner  Sep. 98, Apr 99  */
/***********************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "files.h"
#include "bits.h"
#include "neunetze.h"

/**************************************************************************/
/* Types                        */
/********************************/

/***** Simple List of (Unsigned Char *)'s ********/
typedef struct _clist {
	unsigned char  *c;
        struct _clist  *next;
} UC_List;

/***** Simple List of Integers ********/
typedef struct _ilist {
	          int  i;
        struct _ilist  *next;
} Int_List;

/***** Simple List of Strings *********/
typedef struct _strlist {
		   char  *s;
        struct _strlist  *next;
} Str_List;

typedef struct _telist {
		  int  t;
             Int_List  *e;
       struct _telist  *next;
} TE_List;


typedef char* PNames;
typedef char* TNames;
typedef BitArray* ConfS;
typedef BitArray* CutS;
typedef BitArray* SetS;
typedef BitArray  ElemS;

/***************************************************************************/
/* annotation of the Buechi-Transitions: expressions of  */
/*                           Propositional               */
/*                           Logic                       */ 
/*********************************************************/

typedef struct _trans_PL {
   	          char  *op;	/* o=OR,a=AND,n=NOT,t,f,'prop'*/
      struct _trans_PL  *lop;	/* left operand */
      struct _trans_PL  *rop;	/* right operand */
} trans_PL;

/***************************************************************************/
/* transitions of the Buechi Automaton:
    - c = string of conjuncted AtomicPropositions "p1,p2,p3"
    - il = List of observed Places (placenumbers)
    - from = state of the automaton, enabling the transition
    - trans
    - to = state, reached after switching "trans" 
    - for "from" and "to" : negative if accepting state
    - tpl = pointer to corresponding expression of Propositional Logic
	    that must hold to enable the transition 
	    (for parsing only; deleted after parsing)                     */

typedef struct _b_trans_list {
                  char  *c; 		/* list of conjuncted props       */
              Int_List  *il;		/* list of observed places        */
                                        /* (index in propositions[])      */
	           int  from;		/* from place ...                 */
	           int  trans;		/* ... via transition ...         */
	           int  to;		/* ...to place 
				           (accepting places are negative)*/
              trans_PL  *tpl;		/* ptr to PL-expression           */
  struct _b_trans_list  *next;
} BuechiTrans;

/**************************************************************************/

typedef struct _buechi_place_list {
	                  int  place_num;
                         char  name[MAXCHAR];
    struct _buechi_place_list  *next;
} buechi_place_list;


typedef struct _edge {
	       int  adj;        /* representative of adjacent node */
    unsigned short  final;      /* label of edge:
				   F__x (see qq_defs.h: #definition ) */
	       int  n;          /* cutoff/corresponding event reachable
				   from current node */
      struct _edge  *next;
} Edge;

typedef struct _ENCorr {
         Int_List  *conf;	/* configuration of this node */
         Int_List  *max;	/* maximal events in conf */
         Int_List  *cut;	/* cut of conf */
    struct _ENode  *enode;	/* (1st) ENode with thisnode as corr */
             Edge  *adjs;	/* Adjacent-List of thisnode */
    struct _ENode  *co_en;	/* ENodes in co w. thisnode  */
	      int  last_corr;	/* max. Corresponding Event in conf */
				/* last_corr < 0 : l_c \notin conf  */
				/* last_corr > 0 : l_c \in conf     */
} ENCorr;

typedef struct _ENode {
     unsigned char  info;       /* cf. qq_defs.h */
    unsigned short  final;	/* cf. Edge */
	       int  idx;        /* index (identifier) of this ENode       */
	       int  src;        /* src>0: (ENode *) src is ENode with 
					  this_node->cutoff in co 
                                   src<0: E_(-src) is the corresponding event
					 in co with this_node->cutoff     */
	       int  cutoff;     /* cutoff in co with (this_node)->src     */
	       int  rep;        /* representative(thisnode->cutoff)       */
            ENCorr  *corr;	/* corresponding configuration            */
				/* resp. E_((int) corr) is local corr     */
               int  *ip;        /* for the scc-search-algorithm:
				   ip[0]: lowlink
				   ip[1]: dfn (Depth-first-Number)
				   ip[2]: scc (component-number)
				   ip[3]: scc_idx
				   ip[4]: maxBuchi
                                 */
     struct _ENode  *next;
} ENode;

typedef struct _event {
	     ConfS  config;      /*(local) configuration                 */
     unsigned char  info;         /*cutoff-corr-buechi-final-visited      */
    unsigned short  *pre_post;    /*number of pre-/post- conditions       */
          Int_List  *post;     	  /*postplaces of Buechi-/visible events  */
          Int_List  *preplaces;   /*preplaces of visible events:          */
                      		  /*   the indexes of 'propositions'      */
               int  *pre;         /*predecessor events in flow relation F */
               int  *rem; 	  /*preconditions                         */
               int  *add;         /*postconditions                        */
	       int  tname;        /*numname of transition                 */
               int  *ip;          /*for corresponding events:
				       ip[0]: Last_Buechi
				       ip[1]: Last_Accepting
			    	       ip[2]: cut (Int_List*)
			    	       ip[3]: cutoffs in co --> ENode
				       ip[4]: last_included edge 
				       ip[5]: next corresponding
				       ip[6]: conflicting events:
	                                      e' <= e :  e' # e  
				       ip[7]: last-encountered ENode
				       ip[8]: no_of_cutoffs
				       ip[9..]: cutoffs 
                                    for cutoff-events:	
				       ip[0]: Last_Buechi
				       ip[1]: Last_Accepting
				       ip[2]: corresponding event
				       ip[3]: last corr event in [e]
				       ip[4]: ENCorrs in co
                                    for representative cutoffs:
				       ip[5]: DFNumber
				       ip[6]: LowLink
				       ip[7]: component (scc)
				       ip[8]: scc_idx                     
				    for corr-events of ENodes:
				       ip[0]: Last_Buechi
				       ip[1]: Last_Accepting
				       ip[2]: cut (Int_List *)
				       ip[3]: cutoffs in co -> ENode
				       ip[4]: last_corr in [e] 
				       ip[5]: DFNumber
				       ip[6]: LowLink
				       ip[7]: component
				       ip[8]: scc_idx
				       ip[9]: ENCorrs with e in max
				       */
              Edge  *adjs;   	/* adjacent events for cutoffs,
				   organized at correspond(e)         */
} Event, Node;


/*************************************************************************/
/* etime_t								 */
/* Wird zur exakten Zeitmessung verwendet, siehe nc_startclock und 	 */
/* nc_stopclock.							 */

typedef struct
{
	clock_t clock;
	time_t time;
} *etime_t, etime_instance;

/**************************************************************************/
/* End of File qq_types.h       */
/********************************/

