%{

unsigned int file_changed = 0;
int yylineno = 0;
OBS observer = (OBS) NULL;

%}


STRING [A-Za-z_][A-Za-z0-9_]*
DELIMITER [ \t\n]+
NUMBER ([1-9][0-9]*)

%START OUT_COMMENT IN_COMMENT

%%

	BEGIN OUT_COMMENT;

<OUT_COMMENT>"/*" {BEGIN IN_COMMENT;
		  }
<IN_COMMENT>"*/"  {BEGIN OUT_COMMENT;
		  }
<IN_COMMENT>.     ;
<IN_COMMENT>\n    { yylineno++; }

<OUT_COMMENT>{DELIMITER};
<OUT_COMMENT>"@"	  {return (F_EOF);}
<OUT_COMMENT>"U"	  {return (F_UNTIL);}
<OUT_COMMENT>"F"	  {return (F_SOME);}
<OUT_COMMENT>"G"  	  {return (F_ALWAYS);}
<OUT_COMMENT>"-"	  {return (F_NOT);}
<OUT_COMMENT>"+"	  {return (F_OR);}
<OUT_COMMENT>"*"	  {return (F_AND);}
<OUT_COMMENT>"=>"	  {return (F_IMP);}
<OUT_COMMENT>"="	  {return (F_EQ);}
<OUT_COMMENT>"<"	  {return (F_LESS);}
<OUT_COMMENT>">"	  {return (F_GREATER);}
<OUT_COMMENT>"<="	  {return (F_LESS_EQ);}
<OUT_COMMENT>">="	  {return (F_GREATER_EQ);}


<OUT_COMMENT>"'"          {return (PRIME);}
<OUT_COMMENT>","          {return (COMMA);}

<OUT_COMMENT>"("          {return (LEFT_PAREN);}
<OUT_COMMENT>")"          {return (RIGHT_PAREN);}
<OUT_COMMENT>"{"          {return (LEFT_BRACE);}
<OUT_COMMENT>"}"          {return (RIGHT_BRACE);}
<OUT_COMMENT>spin_never	  {return (SPINNEVER);}
<OUT_COMMENT>never        {return (NEVER);}
<OUT_COMMENT>if           {return (IF);}
<OUT_COMMENT>fi           {return (FI);}
<OUT_COMMENT>";"          {return (SEQ);}
<OUT_COMMENT>":"          {return (COLON);}
<OUT_COMMENT>"&&"         {return (AND);}
<OUT_COMMENT>"||"         {return (OR);}
<OUT_COMMENT>"!"          {return (NOT);}
<OUT_COMMENT>"->"         {return (IMP);}
<OUT_COMMENT>goto         {return (GOTO);}
<OUT_COMMENT>skip         {return (SKIP);}

<OUT_COMMENT>prop{NUMBER} {strcpy(yylval.string, yytext); 
			   return (PROP); }
<OUT_COMMENT>"0"          {yylval.number = 0; return (CONST); }
<OUT_COMMENT>{NUMBER}	  {yylval.number = atoi(yytext); 
			   return (CONST); }
<OUT_COMMENT>{STRING}     {strcpy (yylval.string, yytext); 
			   return (LABEL); }
<OUT_COMMENT>.            ;
<OUT_COMMENT>\n           { yylineno++; }
%%

int yywrap ()
{
  if ( file_changed ) return (1);
  yy_did_buffer_switch_on_eof = 1;
  unput ('@');
  file_changed = 1;
  return (0);
}
