/*********************************/
/* unfolding.c                   */
/* Frank Wallner       14.8.96   */
/*********************************/

#include "qq_header.h"

/*************************************************************************/
void  ConfAdd(ConfS conf, int elem, int *max_elem)
/* add elem to Config/Cut conf, and set max_elem appropriate */
{
	register int i, j;

	/***** 
	the following condition have to be checked before calling ConfAdd 
	if ( elem < (*max_elem) ) { add_entry(conf, elem); return; }
        *****/
        j = elem/BA_BASE; 
	i = (*max_elem)/BA_BASE;
	while ( i <= j ) conf[i++] = 0;
	conf[i] = BA_ALL;
	conf[j] |= bitmask[elem%BA_BASE];
	(*max_elem) = i * BA_BASE;
}
/*************************************************************************/
int SetCopy(SetS src, SetS cpy)
/* copy src into cpy, and return maximal element of src */
{
	register int i=0;

	while ( (cpy[i]=src[i]) != BA_ALL ) i++ ; cpy[i] = BA_ALL;
	return( ( i * BA_BASE ) );
}
/*************************************************************************/
Int_List  *array2list(int *ta)
/* returns int-array ta (delimited with 0) as an Int_List (in same order) */
{
        int i=0;
	Int_List *tl, *tn;

	tl = IL_NULL; while (ta[i]) { i++; } i--;
	while ( i >= 0 ) { tn = NEW_INTLIST; tn->i = ta[i--];
	                   tn->next = tl; tl = tn; }
	return(tl); 
}
/*************************************************************************/
Int_List  *array2listTA(int *ta)
/* returns int-array ta (delimited with 0) as an Int_List (in same order) */
/* the List Entries are T_NAME(x) where x is in ta */
{
        int i=0;
	Int_List *tl, *tn;

	tl = IL_NULL; while (ta[i]) { i++; } i--;
	while ( i >= 0 ) { tn = NEW_INTLIST; tn->i = T_NAME(ta[i--]);
	                   tn->next = tl; tl = tn; }
	return(tl); 
}
/*************************************************************************/
Int_List  *array2listTA_minus_c(int *ta, int c)
/* returns int-array ta (delimited with 0) as an Int_List (in same order) */
/* (the T_NAMES again), but without those events contained in Config(c) */
{
        int i=0;
	Int_List *tl, *tn;

	tl = IL_NULL; while ( ta[i] ) { i++; } i--;
	while ( i >= 0 ) { if ( (ta[i] <= c) && 
				elem_of(config(c), ta[i]) ) { 
				i--; continue; }
			   tn = NEW_INTLIST; tn->i = T_NAME(ta[i--]);
	                   tn->next = tl; tl = tn; }
	return(tl); 
}
/*************************************************************************/
Int_List *RevBA2IL(BitArray *ba)
/* returns BitArray ba as an Int_List (in reverse order: i > i.next) */
{
        register int bit, w, i=0;
        register BitArray c, *x;
	Int_List *tl, *tn;

	tl = tn = IL_NULL; x = ba;
        while ( (c=(*(x++))) != BA_ALL) {
            if (c == 0) { i++; continue; }
            w = BA_BASE*(i++);
            for ( bit=0; bit<BA_BASE; bit++, w++ ) {
                if (c < bitmask[bit]) continue;
                c -= bitmask[bit];
		tn = NEW_INTLIST; tn->i = w; tn->next = tl; tl = tn; 
                if ( c == 0 ) break; }
            }
	return(tl); 
}
/*************************************************************************/
void  IL2CUT(Int_List *il, CutS ba, int maxelem)
/* create set (Cut) of il-elements; maxelem is the maximal element of il */
{
        int w;
	Int_List *ilp;

	/*** initialize the set ba with Zero-Entries ***/
	w = (maxelem/BA_BASE) + 1; ba[w--] = BA_ALL;
	while ( w >= 0 ) ba[w--] = 0;

	/*** add the il-elements to ba ***/
	ilp=il; while (ilp) { add_entry(ba, ilp->i); 
			      condition[place_num[ilp->i]] = ilp->i;
			      ilp=ilp->next; }
}
/*************************************************************************/
void  IL2CONF(Int_List *il, ConfS ba, int maxelem)
/* create set (Config) of il-elements; maxelem is the maximal element of il */
{
        int w;
	Int_List *ilp;

	/*** initialize the set ba with Zero-Entries ***/
	w = (maxelem/BA_BASE) + 1; ba[w--] = BA_ALL;
	while ( w >= 0 ) ba[w--] = 0;

	/*** add the il-elements to ba ***/
	ilp=il; while (ilp) { add_entry(ba, ilp->i); ilp=ilp->next; }
}
/*************************************************************************/
int In_Conflict_1gt2(int e1, int e2)
/* e1 = cutoff; e2 = corresponding; e1 > e2 */
{
        register int    border;
        ConfS           x, y;

	if (e2 > maxCorr(e1)) 
	     {
	       if ( (y=(ConfS)confls(e2)) == EmptyS) return(0); 
	       x = config(e1);
	     }
        else { 
	       if ( (y=(ConfS)confls(maxCorr(e1))) == EmptyS) return(0);
	       x = config(e2);
             }
	border = (e2/BA_BASE) + 1; 
        while ( border-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
int In_Conflict_2gt1(int e1, int e2)
/* e1 = cutoff; e2 = corresponding;  e2 > e1 > maxCorr(e1)*/
{
        register int    border;
        ConfS           x, y;

	if ( (y=(ConfS)confls(e2)) == EmptyS) return(0);
	x = config(e1);
	border = ( e1/BA_BASE ) + 1; 
        while ( border-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
int  CIn_Conflict(ConfS config, ConfS conflict, int border)
{
        register int    b = border;
        ConfS           x=config, y=conflict;

	if ( y == EmptyS) return(0);
        while ( b-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
int NC_In_Conflict_1gt2(int e1, int e2)
/* e1 = cutoff; e2 = simple; e2 < e1 */
{
        register int    border;
        ConfS           x, y;

	if (maxCorr(e1) > e2) {
	    if ( (y=(ConfS)confls(maxCorr(e1))) == EmptyS) return(0);
	    x = config(e2);
	    border = (e2/BA_BASE) + 1; }
        else { if (maxCorr(e1) > maxCorr(e2) ) {
	           if ( (y=(ConfS)confls(maxCorr(e1))) == EmptyS) 
			 return(0);
                   x = config(e2);
                   border = (maxCorr(e1)/BA_BASE) + 1; }
               else { if ( (y=(ConfS)confls(maxCorr(e2))) == EmptyS )
			    return(0);
                      x = config(e1);
		      border = maxCorr(e2); }
             }
        while ( border-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
int NC_In_Conflict_2gt1(int e1, int e2)
/* e1 = cutoff; e2 = simple;  e2 > e1 > maxCorr(e1)*/
{
        register int    border;
        ConfS           x, y;

	if (maxCorr(e1) > maxCorr(e2) ) {
	    if ( (y=(ConfS)confls(maxCorr(e1))) == EmptyS) return(0);
	    border = (maxCorr(e1)/BA_BASE) + 1;
	    x = config(e2); }
        else {
	    if ( (y=(ConfS)confls(maxCorr(e2))) == EmptyS) return(0);
	    x = config(e1);
	    border = (maxCorr(e2)/BA_BASE) + 1; }
        while ( border-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
int NC_In_Conflict(int e1, int e2)
/* e1 = maxCorr(cutoff); e2 = any event;*/
{
        register int    border;
        ConfS           x, y;

	if ( (y=(ConfS)confls(e1)) == EmptyS) return(0);
	x = config(e2); 
	border = ( MIN(e1, e2) / BA_BASE ) + 1;
        while ( border-- ) { if ( *(x++) & *(y++) ) return(1); }
	return(0);
}
/*************************************************************************/
void Setminus(SetS c1, SetS c2, SetS c3)
/* c3:= c1\c2 */
{
	register int i=0;
	ElemS c1el, c2el;

	while ( ( (c2el=c2[i]) != BA_ALL) && ( (c1el=c1[i]) != BA_ALL) ) 
	      { c3[i++] = c1el & (~c2el); }
        if ( c1el == BA_ALL ) { c3[i]=BA_ALL; return; }
        /* else: (c2el == BA_ALL) */ 
	while ( (c3[i]=c1[i++]) != BA_ALL ) ;
}
/*************************************************************************/
Int_List* TransList(int e1, int e2) 
/* returns the transitions of [e1]\[e2]  */
{
        register   int i=0, bit, w;
        int 	   border;
	Int_List   *tl, *tlp, *tlp2;
        ElemS      c1el, c2el, x;
        ConfS      lc1, lc2;

        lc1 = config(e1); lc2 = config(e2);
	tl = tlp = tlp2 = IL_NULL;

	/*** set bits in conf_diff ***/
	if (e1 > e2) border = (e2/BA_BASE)+1; else border = (e1/BA_BASE)+1;
        while ( i < border ) {
	      if ( (c1el=lc1[i]) == 0 ) { i++; continue; }
              if ((x=(c1el&(~(lc2[i]))))==0) { i++; continue; }
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if (x < bitmask[bit]) continue;
		  x -= bitmask[bit];
		  /*** append T_NAME(w) to tl-List ***/
		  tlp = NEW_INTLIST; tlp->i = T_NAME(w); tlp->next = IL_NULL;
		  if (tl) { tlp2->next = tlp; tlp2 = tlp; }
		  else { tl = tlp2 = tlp; }
		  if (x == 0) break; }
              }

        if ( lc1[i] == BA_ALL) { return(tl); }
        while ( (c1el=lc1[i]) != BA_ALL ) {
              if ( (x = c1el) == 0 ) { i++; continue;}
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if (x < bitmask[bit]) continue;
		  x -= bitmask[bit];
		  /*** append T_NAME(w) to tl-List ***/
		  tlp = NEW_INTLIST; tlp->i = T_NAME(w); tlp->next = IL_NULL;
		  if (tl) { tlp2->next = tlp; tlp2 = tlp; }
		  else { tl = tlp2 = tlp; }
		  if (x == 0) break; }
              }
        return(tl); 
}
/*************************************************************************/
char Concurrent(int e1, int e2, int *ta, CutS cut) 
/* returns 1 if "[e1] co [e2]"                           */
/* i.e. if PreCut([e1]\[e2]) \subseteq Cut([e2])(=cut)   */
/* ta := [e1]\[e2]                                       */
/* cut3 is used for \post{ [e1]\[e2] }                   */
{
        register   int i=0, bit, w;
        int 	   *r, rs, border, j=0;
        ElemS      c1el, c2el, x; 
	ConfS      lc1, lc2;

        lc1 = config(e1); lc2 = config(e2);

	/*** Initialization of cut3 ***/
	if ( IS(CUTOFF, e1) )
	     w = ( cutoff_max_post(e1)/BA_BASE ) + 1;
	else w = ( max_post(e1)/BA_BASE ) + 1;
	for (bit=0; bit<w; ) { cut3[bit++] = 0; } cut3[w] = BA_ALL;
      
	/*** add transitions into ta ***/
	border = ( (MIN(e1,e2))/BA_BASE ) + 1;
        while ( i < border ) {
	      if ( (c1el=lc1[i]) == 0 ) { i++; continue; }
              if ( (x=(c1el&(~(lc2[i])))) == 0 ) { i++; continue; }
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if ( x < bitmask[bit] ) continue;
		  x -= bitmask[bit];
		  rs = pre_size(w); r = preB(w);
		  while (rs--) { if ( elem_of(cut3,(*r)) ||
                                      elem_of(cut, (*r))   ) r++; 
			         else return(0); }
		  ta[j++] = (w); 
		  rs=post_size(w); r = postB(w);
		  while (rs--) { add_entry(cut3, (*r)); r++; }
		  if (x == 0) break; }
              }

        if ( lc1[i] == BA_ALL) { ta[j] = 0; return(1); }
        while ( (x=lc1[i]) != BA_ALL ) {
              if ( x == 0 ) { i++; continue;}
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if ( x < bitmask[bit] ) continue;
		  x -= bitmask[bit];
		  rs = pre_size(w); r = preB(w);
		  while (rs--) { if ( elem_of(cut3,(*r)) ||
                                      elem_of(cut, (*r)) ) r++;
			         else return(0); }
		  ta[j++] = w; 
		  rs=post_size(w); r = postB(w);
		  while (rs--) { add_entry(cut3, (*r)); r++; }
		  if (x == 0) break; }
              }
        ta[j]=0; return(1); 
}
/*************************************************************************/
char GConcurrent(ConfS conf, int e1, int e2, int *ta, CutS cut) 
/* returns 1 if "conf co [e2]"                           */
/* i.e. if PreCut(conf\[e2]) \subseteq Cut([e2])(=cut)   */
/* ta := conf\[e2]  |  e1 is the max event in conf       */
/* cut3 is used for \post{ conf\[e2] }                   */
{
        register   int i=0, bit, w;
        int 	   *r, rs, border, j=0;
        ElemS      c1el, c2el, x; 
	ConfS      lc1, lc2;

        lc1 = conf; lc2 = config(e2);

	/*** Initialization of cut3 ***/
	w = ( max_post(e1)/BA_BASE ) + 1;
	for (bit=0; bit<w; ) { cut3[bit++] = 0; } cut3[w] = BA_ALL;
 
	/*** add transitions into ta ***/
	border = ( (MIN(e1,e2))/BA_BASE ) + 1;
        while ( i < border ) {
	      if ( (c1el=lc1[i]) == 0 ) { i++; continue; }
              if ( (x=(c1el&(~(lc2[i])))) == 0 ) { i++; continue; }
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if ( x < bitmask[bit] ) continue;
		  x -= bitmask[bit];
		  rs = pre_size(w); r = preB(w);
		  while (rs--) { if ( elem_of(cut3,(*r)) ||
                                      elem_of(cut, (*r))   ) r++; 
			         else return(0); }
		  ta[j++] = (w); 
		  rs=post_size(w); r = postB(w);
		  while (rs--) { add_entry(cut3, (*r)); r++; }
		  if (x == 0) break; }
              }

        if ( lc1[i] == BA_ALL) { ta[j] = 0; return(1); }
        while ( (x=lc1[i]) != BA_ALL ) {
              if ( x == 0 ) { i++; continue;}
              w = BA_BASE * (i++);
	      for ( bit=0; bit<BA_BASE; bit++, w++ ) {
		  if ( x < bitmask[bit] ) continue;
		  x -= bitmask[bit];
		  rs = pre_size(w); r = preB(w);
		  while (rs--) { if ( elem_of(cut3,(*r)) ||
                                      elem_of(cut, (*r)) ) r++;
			         else return(0); }
		  ta[j++] = w; 
		  rs=post_size(w); r = postB(w);
		  while (rs--) { add_entry(cut3, (*r)); r++; }
		  if (x == 0) break; }
              }
        ta[j]=0; return(1); 
}
/*************************************************************************/
void Configuration(ConfS conf, int e)
{
	register int   i, j, preds, k, *h, *r;
	int            manage_conflicts=0;
        ElemS          c;
	Int_List       *il;

        /*** Initialization ***/
	if ( conf == config(e) ) {
	     j = (e/BA_BASE) + 1;
	     if ( (IS(CORR, e)) && (! IS(IMMEDIATE_STOP,e)) ) 
		  manage_conflicts = 1; 
	     }
	else j = (E_Fin/BA_BASE) + 1; /* manage_conflicts = 0; */
	k = e/BA_BASE; 
   	for (i=0; i<=k; i++) conf[i] = 0;
        for (i=k+1; i<=j; i++) conf[i] = BA_ALL;
        
       	/*** set bits ***/
        conf[k] |= bitmask[e%BA_BASE]; i=e; 
	if ( manage_conflicts ) {
           while ( k >= 0 )  {
              if ( (c=conf[k]) == 0 ) { k--; i -= BA_BASE; continue; }
              j = i%BA_BASE; 
              while ( j >= 0 ) {
 	           if ( (bitmask[j--] & c) ) {
                       h = preE(i); preds = pre_size(i); r = preB(i);
                       while ( preds-- ) { 
			       add_entry( conf, (*h) );
			       if (b_post[(*r)]) { 
				   il = NEW_INTLIST; il->i = (*r); 
				   il->next = (Int_List *) confls(e);
				   confls(e) = (int) il; }
			       r++; h++; }
		       if ( (c=conf[k]) == 0 ) { i -= j; break; }
		       }
                   i--; 
                   }
              k--;
              } /* while (k >= 0) */
           } /* if (manage_conflict) */
	
	else /* not (manage_conflict) */
	  { while ( k >= 0 )  {
              if ( (c=conf[k]) == 0 ) { k--; i -= BA_BASE; continue; }
              j = i%BA_BASE; 
              while ( j >= 0 ) {
 	           if ( (bitmask[j--] & c) ) {
                       h = preE(i); preds = pre_size(i);
                       while ( preds-- ) { add_entry( conf, (*h) ); h++; }
		       if ( (c=conf[k]) == 0 ) { i -= j; break; }
		       }
                   i--; 
                   }
              k--;
              } /* while (k >= 0) */
	   } /* not (manage_conflict) */
}
/*************************************************************************/
void Cut(CutS cut, ConfS conf, int max)
/* computes Cut(conf), where max is maximal in conf */
{
	register int    bit, w, i=0;
	         int    *a, as, border;
               ElemS    x;
	    Int_List    *il;


	if ( ( ! IS(CUTOFF,max)) && (CUT(max) != 0) ) {
	     il = (Int_List *) CUT(max);
	     IL2CUT( il, cut, il->i);
	     return; }

 	/*** Initialization ***/
	if (IS(CUTOFF,max)) w = ( cutoff_max_post(max)/BA_BASE ) + 1;
	else                w = (    max_post(max)/BA_BASE ) + 1;
	for (i=0; i<w;) { cut[i++]=0; } cut[w]=BA_ALL;

	/*** set bits ***/
	i=0; border = (max/BA_BASE) + 1;
	while ( i < border) {
	    if ( (x=conf[i]) == 0 ) { i++; continue; }
	    w = ( BA_BASE * (i++) );
	    for ( bit=0; bit<BA_BASE; bit++, w++ ) { 
		if ( x < bitmask[bit] ) { continue; }
		x -= bitmask[bit];
		as = post_size(w); a = postB(w); 
		while (as--) { add_entry(cut, (*a)); 
			       condition[place_num[(*a)]] = (*a);
			       a++; }
		as = pre_size(w); a = preB(w);
		while (as--) { del_entry(cut, (*a)); a++;}
		if ( x == 0 ) break;
                }
	    }
	if ( ! IS(CUTOFF,max)) CUT(max) = (int) RevBA2IL(cut);
}
/*************************************************************************/
void Evaluate(ConfS conf)
{
	ElemS    act_element, j;
	Int_List *il;
	register int  x, i, k, eb;

        for (i=0; i<observed_places; i++) current_marking[i] = 0;
	for (i=0; (act_element=conf[i])!=BA_ALL; i++) {
             act_element &= vis_conf[i];
	     if ( act_element == 0 ) continue;
	     k = BA_BASE; j = BA_START; eb = (BA_BASE*i) + BA_BASE;
	     while ( k ) {
		   if (act_element == 0) break;
		   if (act_element >= j) { x = eb-k;
                        il= (Int_List *) postplaces(x);
                        while (il) { current_marking[il->i]++; il=il->next; }
                        il= (Int_List *) preplaces(x);
                        while (il) { current_marking[il->i]-- ; il=il->next; }
		        act_element -= j;
			}
		   j >>= 1; k--;
		   }
	     }
}
/*************************************************************************/
void Initialize_Buechi_Nodes()
{
	        int  i;
	BuechiTrans  *tlptr;
	Int_List     *il;

	for (i=0; i<_PBuechi_; i++) { 
	    /* initialize dfn,lowlink */
	    ip_mem(i) = (int *) MyMalloc(2*sizeof(int)); 
	    B_dfn(i) = B_lowlink(i) = 0; }

	tlptr = BTrans; 
	while ( tlptr ) { 
	      if ( tlptr->from < 0 ) INFO( -1 - tlptr->from) |= FINAL;
	      if ( tlptr->to   < 0 ) INFO( -1 - tlptr->to)   |= FINAL;
	      tlptr = tlptr->next; }

	while ( Accept_all_List ) { il = Accept_all_List;
				    Accept_all_List = il->next;
				    INFO( (il->i) - 1) |= ACCEPTALL;
				    free(il); }
}
/*************************************************************************/
char is_marked(Int_List *il)
{
 	Int_List *ilp;

        ilp = il; 
	while (ilp) {if (current_marking[ilp->i]==0) return(0); ilp=ilp->next;}
        return(1);
}
/*************************************************************************/
void InstantiateAutomaton()
{
	register int i;
	int          from, to;
	Edge  	     *new, *fedge, *prev;
	BuechiTrans  *tlp;

	/*** delete old edges in Buechi-Graph ***/
	for (i=0; i<_PBuechi_; i++) {
	     INFO(i) &= ~VISITED;
	     INFO(i) &= ~SELFLOOP;
	     INFO(i) &= ~ONSTACK;
	     B_dfn(i) = B_lowlink(i) = 0;
	     fedge = B_adjacents(i);
	     while ( fedge ) { 
		     B_adjacents(i)=fedge->next; 
		     free(fedge); 
		     fedge=B_adjacents(i); }
	     B_adjacents(i)= (Edge *) NULL; 
	     if ( IS(ACCEPTALL,i) ) INFO(i) |= SELFLOOP; }

        /*** create new Buechi-Graph ***/
	tlp = BTrans; 
	while ( tlp ) {
             if ( is_marked(tlp->il) ) {
		  if ( tlp->from < 0 ) 
		       from = (- tlp->from) - 1; 
		  else from = tlp->from - 1;
		  if ( tlp->to < 0 )  
		       to = (- tlp->to) - 1; 
		  else to = tlp->to - 1;
		  if ( from == to ) { INFO(from) |= SELFLOOP;
				      tlp = tlp->next; continue; }

                  /*** include 'to' to the adj-List of 'from' ***/
		  i = 0; 
                  fedge = B_adjacents(from);
		  prev = (Edge *) NULL;
		  while ( fedge ) { 
			if ( fedge->adj < to ) {
			     prev = fedge; fedge = fedge->next; continue; }
                        if ( fedge->adj == to ) { i = 1; break; }
			if ( fedge->adj > to ) break;
			}
                  if ( i ) { tlp = tlp->next; continue; }

		  new = (Edge *) MyMalloc(sizeof(Edge));
		  new->adj = to;
		  new->final = ( tlp->to < 0 );
		  if ( fedge == (Edge *) NULL ) {
		       if ( prev ) { new->next = (Edge *) NULL;
				     prev->next = new; }
                       else { new->next = B_adjacents(from);
			      B_adjacents(from) = new; }
                       }
                  else { new->next = fedge;
			 if ( prev ) prev->next = new;
			 else B_adjacents(from) = new; }
		  }
	     tlp=tlp->next;
	     }
}
/*************************************************************************/
void  InsertPost(int cond, int tnum, int event)
{
	Int_List *il, *dil;
	TE_List *tel;

	tel = b_post[cond];
	while ( tel && (tel->t != tnum) ) { tel = tel->next; }
	if ( ! tel ) { tel = NEW_TELIST;
	               tel->t = tnum; tel->e = IL_NULL;
	               tel->next = b_post[cond];
	               b_post[cond] = tel; }
        if ( ! tel->e ) { tel->e = il = NEW_INTLIST;
			  il->i = event; il->next = IL_NULL; return; }
        il = tel->e;
	while ( il ) { dil = il; il=il->next; }
	dil->next = il = NEW_INTLIST;
	il->i = event; il->next = IL_NULL;
}
/*************************************************************************/
char GetEvents(FILE *unf, Event **glob_e, int* B, int* E)
{
	int  		*h, pre_sz, add_sz, corr_ctr=0;
	int             prop_idx, p_sched_f=0;
	int 		*corr_array, last_cutoff;
	int  		e_no, p_no, b_no, next_e;
	char 		line[MAXLINE], str[MAXLINE], name[MAXLINE];
	Int_List        *ilist, *pil, *qil;
	register char 	c;
	register int  	pos, i, j, idx;
	register Event 	*ep;
	Transition	t, midtrans;
	Place		p;
	STEdge		st;
	TSEdge		ts;
   	BitArray        *sc_ba;
	FILE            *start_of_conditions;

	while (fgetline(unf,line,MAXLINE)) {
	      if ((idx=myindex(line,"conditions."))>=0 ||
		  (idx=myindex(line,"condition."))>=0     ) {
		 paste(0, idx, line, str); mystr2int(str, B);
		 real_B_Fin = (*B);
		 /*printf("%6d conditions\n", *B);*/
 		 place_num = (int *) MyMalloc( sizeof(int) * ((*B)+1) );
 		 b_post = (TE_List **) \
			       MyMalloc( sizeof(TE_List *) * ((*B)+1) );
		 }
	      if ((idx=myindex(line,"events."))>=0 ||
		  (idx=myindex(line,"event."))>=0    ) {
		 paste(0, idx, line, str); mystr2int(str, E);
		 /*printf("%6d events\n", (*E));*/
		 *glob_e = ep = (Event *) calloc(((*E)+1),sizeof(Event));
		 if (!ep) {printf("\ncalloc() ep failed!\n");exit(1);}
		 break;
		 }
              }

	p = sysnet->LastPlace;
	while (p) {
	       if ( p->name[0] != '_' ) { p=p->prevplace; continue; }
	       if ( GET_PL_MEANING(p) && 
		    myindex(GET_PL_MEANING(p), "step_formula") >= 0 ) {
		    p_sched_f = p->numname; break; }
	       p = p->prevplace; }

        while (fgetline(unf,line,MAXLINE))
	      if (myindex(line,"numname")>=0) break;

	/*** Find medial Transition ***/
	midtrans = sysnet->Transitions;
	i = (sysnet->LastTransition->numname)/2;
	while (i--) midtrans = midtrans->nexttrans;

        /* Predecessors and marking-info for "E0" */
        preE(0) = (int *) NULL; preB(0) = (int *) NULL;
	add_sz = pre_sz = 0; 
	p = sysnet->Places;
	while (p) {if (p->marks == 1) add_sz++; p = p->nextplace;}
	(ep)->pre_post = (unsigned short *) MyMalloc(2*sizeof(unsigned short));
	pre_size(0) = 0; post_size(0) = 0;
	postB(0) = (int *) MyMalloc( add_sz * sizeof(int) );
	preplaces(0) = IL_NULL;
        INFO(0) |= VISIBLE;
        T_NAME(0) = _T_sync + 1;

        /*** Predecessors and marking-info for E1, E2, ... ***/
	i=1;
        while ( fgetline(unf, line, MAXLINE) ) {
	      if ( (idx=myindex(line,"E")) < 0 ) break;
	      j = p_no = 0;  pos = ++idx;
	      while( line[pos] >= '0' && line[pos] <= '9' ) pos++;
	      while( line[pos] == ' ' ) pos++;
	      while( line[pos] != ' ' ) name[j++] = line[pos++];
	      name[j] = '\0';
                /* ... now, name is the transition-name of E_i */
	      if (name[0] == '@') { INFO(i) |= (BUECHI | FINAL); }
	      else { if ( (name[0] == '_') &&
		          (name[1] == 'c') &&
		          (name[2] == '_') &&
		          (name[3] == 'T')    ) { INFO(i) |= BUECHI; }
                   }
              if (IS(BUECHI, i)) { 
		     t = sysnet->LastTransition;
		     while (strcmp(t->name, name) != 0) t = t->prevtrans;
		     }
              else { if ( Kleiner(name, midtrans->name) ) 
			  t = sysnet->Transitions; 
                     else t = midtrans;
	             while (strcmp(t->name, name) != 0) t = t->nexttrans;
                   }
                /*** ... now t is the transition of E_i ***/
	      T_NAME( i ) = (int) t; 

	      /** allocating memory for preds and removes and adds**/
	      pre_sz = add_sz = 0;
	      st = t->preset;  while (st) { pre_sz++; st=st->nextplace;}
	      ts = t->postset; while (ts) { add_sz++; ts=ts->nextplace;}
	      preE(i) = (int *) MyMalloc( pre_sz * sizeof(int) );
	      preB(i)  = (int *) MyMalloc( pre_sz * sizeof(int) );
	      postB(i)  = (int *) MyMalloc( add_sz * sizeof(int) );
	      (ep+i)->pre_post = (unsigned short *) \
				 MyMalloc( 2 * sizeof(unsigned short) );
	      post_size(i) = 0; pre_size(i) = 0;

	      i++; /*next event*/ }

	while ( fgetline(unf,line,MAXLINE) ) 
              if ( myindex(line,"preset") >= 0 ) break;

	while ( fgetline(unf,line,MAXLINE) ) {
	      if ( (idx=myindex(line, "cutoff")) >= 0 ) {
	            paste(0, idx, line, str);
		    mystr2int(str, &e_no);
		    E_off = e_no;
		    /*******************************************
		    printf("%6d cutoff event", E_off);
		    if (E_off == 1) printf("\n"); else printf("s\n");
		    ********************************************/
		    break; }
	      /* else */
	      if ( myindex(line, "B") < 0 ) continue;
              prop_idx = -1;
	      if ( ((idx=myindex(line,"-->")) >= 0) && (idx < 20) ){
		     paste(0,idx,line,str); mystr2int(str, &e_no); }
              else e_no=0;
	      /*** ... now e_no is the pre-event of condition Bxxx ***/
	      pos = 0; j = 0;
	      while (line[pos++] != 'B');
	      while (line[pos] != ' ') str[j++]=line[pos++];
	      str[j]='\0'; mystr2int(str, &b_no); j = 0;
	         /*** ... now b_no is the number of the actual condition ***/
	      while (line[pos] == ' ') pos++;
	      while (line[pos] != ' ') name[j++] = line[pos++];
	      name[j] = '\0'; 
	         /*** ... now 'name' is the name of the corresponding place ***/
              if ( e_no == 0 ) { if ( name[0] == '_' && name[1] == 'c' &&
		                      name[2] == '_' && name[3] == 'P'   ) {
		                        p = sysnet->LastPlace;
		                        while (strcmp(p->name, name) != 0) 
					       p = p->prevplace;
		                      }
                                 else { p = sysnet->Places;
	                                while (strcmp(p->name, name) != 0) 
					       p = p->nextplace;
			              }
                   }
              else {
		   t = (Transition) T_NAME(e_no);
		   ts = t->postset;
		   while (ts) {
			  if (strcmp(name, ts->place->name) == 0 ) {
			      p = ts->place; break; }
                          ts = ts->nextplace;
			  }
                   }
	         /*** ... now p is the corresponding Place ***/
              place_num[b_no] = p->numname;
              h = postB(e_no); h[post_size(e_no)] = b_no; 
              post_size(e_no) += 1;

              /*** if p is a state of the automaton, and e_no is a ***/
              /*** Buechi-Transition, then set postplaces(e_no)    ***/
	      if ( (GET_PL_MEANING(p)) &&
		   (myindex(GET_PL_MEANING(p), "#_buechi_") >= 0 ) &&
                   (IS(BUECHI,e_no)) ) 
                   { postplaces(e_no) = (Int_List *) c_num(p->name); }
              /*** else check if it is an observed place; ***/
              /*** if so add it to the pre/post sets      ***/
              else { for (j = 0; j < observed_places; j++) {
		          sprintf(name, "OBS(%s)", propositions[j]);
		          if ( (GET_PL_MEANING(p)) &&
			       (myindex(GET_PL_MEANING(p),name) >= 0) ) {
                               prop_idx = j;
                               if ( ! IS(BUECHI,e_no) ) {
		                    ilist = NEW_INTLIST;
		                    ilist->i = j;
		                    ilist->next = postplaces(e_no);
		                    postplaces(e_no) = ilist;
                                    break; }
                               }
                          } /*for */
                   } /*else*/

              j = 0; 
	      while (line[pos] != '>' && line[pos] != '\n') pos++;
	      if (line[pos] == '>') { /* there are some successors */
		 pos++; 
	         while (line[pos] != ',' && line[pos] != '\n') 
		        str[j++] = line[pos++];
	         str[j] = '\0'; j = 0;
	         mystr2int(str, &next_e);
                       /*** ... now 'next_e' is the first successor of e_no ***/
                 h = preE(next_e); h[pre_size(next_e)] = e_no;
                 h =  preB(next_e); h[pre_size(next_e)] = b_no;
                 pre_size(next_e) += 1;
	         if ( (prop_idx != -1) && ( ! IS(BUECHI,next_e)) ) {
		       ilist = NEW_INTLIST;
		       ilist->i = prop_idx;
		       ilist->next = preplaces(next_e);
		       preplaces(next_e) = ilist; }
                 if (line[pos++] == '\n') { /* only one successor event */
	             b_post[b_no] = (TE_List *) 0; (*B) = b_no; }
	         else { /* several successors */
		     b_post[b_no] = (TE_List *) 1; (*B) = b_no;
	             while (line[pos] != '\n') { 
		         while (line[pos] != ',' && line[pos] != '\n') 
			              str[j++] = line[pos++];
		         str[j] = '\0'; j = 0;
		         mystr2int(str, &next_e);
                            /*** ... next_e is the next successor of e_no ***/
                         h = preE(next_e); h[pre_size(next_e)] = e_no;
                         h =  preB(next_e); h[pre_size(next_e)] = b_no;
                         pre_size(next_e) += 1;
		         if ( (prop_idx != -1) && ( ! IS(BUECHI,next_e)) ) {
		               ilist = NEW_INTLIST;
		               ilist->i = prop_idx;
		               ilist->next = preplaces(next_e);
		               preplaces(next_e) = ilist; }
		         if (line[pos] != '\n') pos++;
		         }
                     } /* several successors */
                 } /* some successor */
	      /*** check if e_no is an accepting Buechi state ***/
	      /*** and check for "accept_all" state           ***/
	      if (e_no == 0) {
		    if ( (GET_PL_MEANING(p)) &&
			 (myindex(GET_PL_MEANING(p), "accept_all") >= 0) ) {
			 printf("\n\"accept_all\" state '%s'", p->name);
			 printf(" is initially marked!\n");
			 exit(1); }
		    }/*fi*/
	      else { if ( (GET_PL_MEANING(p)) &&
			  (myindex(GET_PL_MEANING(p), "accept_all") >= 0)) {
			  INFO(e_no) |= IMMEDIATE_STOP; corr_ctr=1; 
			  /*if (! all_counterexamples) return(0);*/ }
                     if ( p->numname == p_sched_f ) INFO(e_no) |= VISIBLE;
                   } /*else*/
              }

	if (corr_ctr) { /*** Accept_ALL State is reachable ***/ 
	    /*** save t->numname as T_NAME(e) ***/
	    for (i=1; i<=(*E); i++) {
		t = (Transition) T_NAME(i); T_NAME(i) = t->numname; }
	    return(0); }


	/*** scan Cutoff and Corresponding Events ***/
	pos = 0;
	while ( (c=getc(unf)) != '.') {
	      if (isdigit(c)) {str[pos++]=c;
			       continue; }
	      if (c=='E')     {continue;}
	      if (c=='(')     {str[pos++]='\0';
			       mystr2int(str, &e_no); /**cutoff found **/
			       last_cutoff = e_no;
			       INFO(e_no) |= CUTOFF;
			       /* remove the postconditions */
	                       post_size(e_no) = 0; free( postB(e_no) ); 
			       pos=0;
			       continue;}
	      if (c==')')     {str[pos++]='\0';
			       mystr2int(str, &e_no); /**corr found*/
			       corr_ctr++;
                               if (e_no > maxcorr) maxcorr = e_no;
			       if ( ! IS(CORR,e_no) ) { 
				    E_corr++;
			            INFO(e_no) |= CORR;
				    add_sz = REPR_CUTOFF_IP * sizeof(int); }
                               else add_sz = ALL_CUTOFF_IP * sizeof(int);
			       ip_mem(last_cutoff) = (int *) MyMalloc(add_sz);
			       maxBuchi(last_cutoff) = (-1);
			       maxAcpt(last_cutoff) = (-1);
			       correspond(last_cutoff) = e_no;
			       ENCorr_co_cands(last_cutoff) = (int) 0;
			       ilist = NEW_INTLIST;
			       ilist->next = (Int_List *) ip_mem(e_no);
			       ilist->i = last_cutoff;
			       ip_mem(e_no) = (int *) ilist;
			       pos=0;
			       continue;}
	      if (c==',')     {continue;}
	      if (c=='s')     {str[pos++]='0';
			       while (getc(unf)!= 'g');
			       continue;}
              }

	/*** store the cutoffs of the corresponding events in an array ***/
	corr_array =  (int *) \
	    MyMalloc( (E_off+(CORR_IP*E_corr))*sizeof(int));
       	j = 0;
	for (e_no=0; e_no<=maxcorr; e_no++) {
	    if ( ! IS(CORR, e_no) ) continue;
	    corr_ctr = 0;
	    pil = qil = (Int_List *) ip_mem(e_no);
	    while (qil) {corr_ctr++; qil=qil->next;}
	      /*** ... now corr_ctr is the number of cutoffs of e_no ***/
	    ip_mem(e_no) = corr_array + j;
	    maxBuchi(e_no) = maxAcpt(e_no) = (-1) ;
	    __lastEdge(e_no) = nextcorrs(e_no) = 0;
	    enodes(e_no) = CUT(e_no) = 0;
            cutoff_quantity(e_no) = corr_ctr;
            h = corr_array + corr_ctr + j + (CORR_IP - 1);
	    while (pil) { qil = pil->next; 
			  *h = pil->i; 
			  h--; free(pil); 
			  pil = qil; }
            j += (corr_ctr + CORR_IP);
	    }

        if (IS(CORR,0)) { 
	    sc_ba = (BitArray *) MyMalloc(BA_SZ);
	    *sc_ba = 0;
	    confls(0) = (int) sc_ba; }

	/*** save t->numname as T_NAME(e) ***/
        for (i=1; i<=(*E); i++) {
	    t = (Transition) T_NAME(i); T_NAME(i) = t->numname; }

	/*
	for (i=0; i<=E_Fin; i++) {
	    printf("E%d :", i); fflush(stdout);
	    j = pre_size(i); h = preB(i);
	    while (j--) { printf("B%d ", (*h)); fflush(stdout); h++; }
	    printf(" --> E%d -->", i); fflush(stdout);
	    j = post_size(i); h = postB(i);
	    while (j--) { printf("B%d ", (*h)); fflush(stdout); h++; }
	    printf(" preds: "); fflush(stdout);
	    j = pre_size(i); h = preE(i);
	    while (j--) { printf("E%d ", (*h)); fflush(stdout); h++; }
	    printf("\n"); }
	*/

	return(1);
}
/*************************************************************************/
void ReadConflicts(FILE *unf)
{
	int  		b_no, next_e;
	char 		line[MAXLINE], str[MAXLINE];
	register int  	pos, j;

	while (fgetline(unf,line,MAXLINE)) 
	      if (myindex(line, "preset")>=0) break;

	while ( fgetline(unf,line,MAXLINE) ) {
	      if ( (myindex(line, "cutoff")) >= 0 ) { return; }
	      if ( myindex(line, "B") < 0 ) continue;
	      pos = 0; j = 0;
	      while (line[pos++] != 'B');
	      while (line[pos] != ' ') str[j++]=line[pos++];
	      str[j]='\0'; mystr2int(str, &b_no); j = 0;
	         /*** ... now b_no is the number of the actual condition ***/
	      while (line[pos] != '>' && line[pos] != '\n') pos++;
	      if (line[pos] == '>') pos++; else continue;
	      while (line[pos] != ',' && line[pos] != '\n') 
		     str[j++] = line[pos++];
	      str[j] = '\0'; j = 0;
	      mystr2int(str, &next_e);
                    /*** ... now 'next_e' is the first successor of e_no ***/
              if (line[pos++] == '\n') {
		  /*** only one successor... ***/
		  b_post[b_no] = NEW_TELIST;
		  b_post[b_no]->t = T_NAME(next_e);
		  b_post[b_no]->next = TEL_NULL;
		  b_post[b_no]->e = NEW_INTLIST;
		  b_post[b_no]->e->i = next_e;
		  b_post[b_no]->e->next = IL_NULL;
		  }
              else { /***  several successors.... ***/
	          b_post[b_no] = NEW_TELIST; 
		  b_post[b_no]->t = T_NAME(next_e);
		  b_post[b_no]->next = TEL_NULL;
		  b_post[b_no]->e = NEW_INTLIST;
		  b_post[b_no]->e->i = next_e;
		  b_post[b_no]->e->next = IL_NULL;
	          while (line[pos] != '\n') { 
		        while (line[pos] != ',' && line[pos] != '\n') 
			       str[j++] = line[pos++];
		        str[j] = '\0'; j = 0;
		        mystr2int(str, &next_e);
                        /*** ... now 'next_e' is the next successor of e_no ***/
			InsertPost(b_no, T_NAME(next_e), next_e);
		        if (line[pos] != '\n') pos++;
		        }
		  } /* .... several successors */
              }
}
/**************************************************************************/
trans_PL  *Copy_TPL(trans_PL *t)
{
	trans_PL *t1;

	if ( t == (trans_PL *) NULL ) return( t );
	t1 = (trans_PL *) MyMalloc(sizeof(trans_PL));
	if (TPL_OP(t)) t1->op = t->op;
	else t1->op = Strdup(t->op);
	t1->lop = Copy_TPL(t->lop);
	t1->rop = Copy_TPL(t->rop);
	return ( t1 );
}
/*************************************************************************/
BuechiTrans *BTransList(trans_PL *t)
{
	BuechiTrans *trli1, *trli2, *trli;
        trans_PL *t1, *t2;
        char *neu;

        if (t == (trans_PL *) NULL) return((BuechiTrans *) NULL);

	/* t->op != TPL_OR    and   != TPL_AND */
        if ( ( ! TPL_OR_OP(t) ) && ( ! TPL_AND_OP(t) ) ) {
   	   trli = (BuechiTrans *) MyMalloc(sizeof(BuechiTrans));
           trli->from = trli->trans = trli->to = 0;
           trli->next = (BuechiTrans *) NULL;
           trli->il = IL_NULL;
           trli->tpl = t;
           if ( TPL_PROP(t) ) trli->c = Strdup(t->op);
	   else { if ( TPL_TRUE_OP(t) ) trli->c = Strdup(BT_TRUE);
		  else trli->c = Strdup(BT_FALSE);
                }
           return(trli); }
          
        /* t->op == TPL_OR */
        if (TPL_OR_OP(t) ) {
           trli1 = BTransList(t->lop);
           trli2 = trli1;
           while (trli2->next) trli2 = trli2->next;
           trli2->next = BTransList(t->rop);
           return(trli1); }

        /* t->op == TPL_AND */
        trli1 = BTransList(t->lop);
        trli2 = BTransList(t->rop);
        trli1->tpl = t;
        neu = NEW_CHAR(2+(strlen(trli1->c)+strlen(trli2->c)) );
        sprintf(neu, "%s,%s" , trli1->c, trli2->c);
        free(trli1->c);
        free(trli2->c);
        trli1->c = trli2->c = (char *) NULL;
        free(trli2);
        trli2 = (BuechiTrans *) NULL;
        trli1->c = neu;
        return(trli1);
}
/*************************************************************************/
char *TransName(int i) { return( tnames[(ep+i)->tname - 1]  ); }
/*************************************************************************/
char *PlaceName(int i) { return( pnames[i-1] ); }
/*************************************************************************/
int  c_num(char *name)
{ 	int i = 0;
	mystr2int(&(name[4]), &i); return(i);
}
/*************************************************************************/
int  PlaceCount(LMPNet n)
{
	Place p=n->Places;
        int i=0;

        while(p) {i++; p = p->nextplace;} return(i);
}
/*************************************************************************/
int  TransCount(LMPNet n)
{
	Transition t=n->Transitions;
        int i=0;

        while(t) {i++; t = t->nexttrans;} return(i);
}
/*************************************************************************/
void Free_trans_PL(trans_PL* tplp)
{
	if (tplp->lop) Free_trans_PL(tplp->lop);
	if (tplp->rop) Free_trans_PL(tplp->rop);
	if ( TPL_PROP(tplp) ) free(tplp->op);
	free(tplp);
        /*tplp = (trans_PL *) NULL;*/
}
/*************************************************************************/
void Free_BuechiTrans(BuechiTrans* bt)
{
	if (bt->c) { free(bt->c); }
        if (bt->il) { Free_Int_List(bt->il); }
        if (bt->tpl) { Free_trans_PL(bt->tpl); }
        free(bt);
}
/*************************************************************************/
void Free_Int_List(Int_List *ilist)
{
	Int_List *delil;

        while(ilist) {delil = ilist->next; free(ilist); ilist=delil;}
}
/*************************************************************************/
void Free_UC_List(UC_List *ilist)
{
	UC_List *delil;

        while(ilist) {delil = ilist->next; free(ilist); ilist=delil;}
}
/*************************************************************************/
void  Make_Conflict(int e)
{	
	Int_List *bl, *dbl, *el;
	TE_List *tel;
	BitArray *c, *cc;

	bl = (Int_List *) confls(e);
	cc = (BitArray *) MyMalloc( ((e/BA_BASE)+1) * BA_SZ);
	confls(e) = (int) cc;
	c = config(e);

	while ( bl ) {
	      tel = b_post[bl->i];
	      while ( tel ) {
		    el = tel->e;
		    while ( el && (el->i <= e) ) {
	                  if ( ! elem_of(c,el->i) ) add_entry(cc, el->i);
                          el = el->next; }
	            while (el && ( ((el->i)/BA_BASE) == e/BA_BASE ) ) {
	                  if ( ! elem_of(c,el->i) ) add_entry(cc, el->i);
                          el = el->next; }
                    tel = tel->next;
		    }
	      dbl = bl->next; free(bl); bl = dbl; }
}
/*************************************************************************/
int  Find_Extension(CutS cut, int trans, int *max)
/* returns an event mapped to "trans" which extends cut, 
   and update cut, min, max appropriately */
{
	int 		p, cond, rs, *r, valid;
	Int_List 	*el;
	TE_List 	*tel;

	cond = condition[pre_trans[trans]];
	if ( ! elem_of(cut, cond) ) return(0);
	tel = b_post[cond];
        while ( tel && ( tel->t != trans ) ) tel = tel->next;
	if ( ! tel ) return(0);
	el = tel->e;
	while ( el ) {
	        valid = 1;
	        rs = pre_size(el->i); r = preB(el->i); 
		while (rs--) {
		       if ( (*r) >= (*max) ) {valid=0; break;}
		       if ( ! elem_of(cut, (*r)) ) {valid=0; break;}
		       r++; }
                if ( ! valid ) { el = el->next; continue; }
	        r = preB(el->i); rs = pre_size(el->i);
	        while (rs--) { del_entry(cut, (*r)); r++; }
	        r = postB(el->i); rs = post_size(el->i);
	        while (rs--) {
		      if ( (*r) < (*max) ) add_entry(cut, (*r));
		      else ConfAdd(cut, (*r), max);
		      condition[place_num[(*r)]] = (*r);
		      r++; }
                return( el->i ); }
	return(0);
}
/*************************************************************************/
Int_List  *ReverseCopyIntList(Int_List * il)
{
	Int_List *ilp, *rl, *rln;

	rl = IL_NULL; ilp = il; 
	while (ilp) { rln=NEW_INTLIST; rln->i=ilp->i; rln->next=rl; rl=rln;
		      ilp = ilp->next; }
	return( rl );
}
/*************************************************************************/
Int_List *CopyIntList(Int_List *il)
{
	Int_List *ilp, *ril, *elem, *elem2;

	ilp=il; ril=elem=NEW_INTLIST; elem2=IL_NULL;
	while (ilp) {
           elem->i=ilp->i; elem->next=NEW_INTLIST; elem2=elem; elem=elem->next;
	   ilp = ilp->next; }
        free(elem);
	if (elem2) { elem2->next=IL_NULL; return(ril); } else return(IL_NULL);
}
/*************************************************************************/
Int_List  *CoTRList(Int_List *cl, ConfS conf, int max)
/* returns list of Transitions contained in cl\conf */
{
	Int_List *rl, *clp, *new;

	clp = cl; rl = new = IL_NULL;
	while ( clp && (clp->i > max) ) {
	      new=NEW_INTLIST; new->i=T_NAME(clp->i); new->next=rl; rl=new;
	      clp = clp->next; }
	while ( clp ) {
	      if ( elem_of(conf, clp->i) ) { clp=clp->next; continue; }
	      new=NEW_INTLIST; new->i=T_NAME(clp->i); new->next=rl; rl=new;
	      clp = clp->next; }
        return( rl );
}
/*************************************************************************/
void  Shift(int v, Int_List **tl, ENode **w0ci)
/* shift the configuration [w0ci->conf] \cup [v] into the prefix */
/* v is a cutoff event  |   tl == list of transitions in [w0ci->conf]\[v] */
{
	int      max_c;   /* max. condition in cut3 */
	int      max_e;   /* max. event in conf3    */
	int      v0, rep, next, simpleshift;
	char     inherit=0;
	Int_List *tlp, *cotl, *maxtl, *maxtlp;
	ENode    *source, *newnode;
	ENCorr   *encorr;
	Edge     *nedge, *lastedge;
	Int_List *ctr;

	/*** initializing the pointers ***/
	source = newnode = (ENode *) NULL; 
	encorr = (ENCorr *) NULL;
	maxtl = cotl = IL_NULL; 
	lastedge = nedge = (Edge *) NULL;
	/*** set cutoff of *w0ci ***/
	(*w0ci)->cutoff = v; 
	(*w0ci)->info |= SHIFTED;
	(*w0ci)->ip = (int *) MyMalloc(sizeof(int) * EN_IP );
	/*** Initialize v0, cut3, conf3 ***/
	v0 = correspond(v);
	Cut(cut3, config(v0), v0);
	max_c = ba_idx( max_post(v0) ); 
	max_e = SetCopy(config(v0), conf3);
	tlp = (*tl); 
	while ( tlp ) {
	      next = Find_Extension(cut3, tlp->i, &max_c);
              /*** next is a valid extension ***/
	      /*if ( ! next ) { printf("No next!\n"); exit(1); }*/
	      if ( ! next ) { while (1) source=source->next;  }
              if ( next < max_e ) 
		   add_entry(conf3, next);
	      else ConfAdd(conf3, next, &max_e);
	      tlp->i = ( - next ) ; 
	      if ( ! IS(CUTOFF, next) ) { tlp=tlp->next; continue; }
	      /*** next is a cutoff event ***/
	      rep = representative(next);
	      /* forall e \in tl, e \notin Conf(next):  e|->Trans(e) ***/
	      /* forall e \in tl, e \in Conf(next):  remove e        ***/
	      updateTL(tl, next);

	      simpleshift = ( (! source) && (v0 == correspond(next)));
	      if (! simpleshift ) { simpleshift = 1;
	         if ( encorr ) { maxtlp = encorr->max;
	           while (maxtlp) { 
		         if ( ( maxtlp->i > next ) ||
		              ( ! elem_of(config(next), maxtlp->i)) ) {
			       simpleshift = 0; break; }
                         maxtlp = maxtlp->next; }
                   }
                 else { if ( ( v0 > next ) ||
			     ( ! elem_of(config(next), v0) ) )
			      simpleshift = 0; }
                 }

              if ( simpleshift ) {
		  v0 = correspond(next);
		  Cut(cut3, config(v0), v0);
		  max_c = ba_idx( max_post(v0) );
		  max_e = SetCopy(config(v0), conf3);
		  maxtl = IL_NULL;
		  source = (ENode *) NULL;
		  encorr = (ENCorr *) NULL;
		  tlp = (*tl); continue; }
              /* else :  not simpleshift; add next to the ENode-List of v0 */
	      if ( source && ( ! EN_IS(LOCAL, source)) ) {
		   newnode = ExtGrab_ENode( source, rep);
		   if ( ! EN_IS(SHIFTED, newnode) ) {
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = CoTRList(encorr->conf, config(next), next);
                        Shift( next, &cotl, &newnode );
		        if (cotl) Free_Int_List(cotl); 
			IL2CONF(encorr->conf, conf3, ba_idx(encorr->max->i));
			newnode->final = 2 * edge_label_lg(next, conf3); 
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) = 
			        MAX ( maxBuchi(next),
				      Find_Buechi(conf3, encorr->max->i)
                                    ) ;
                        }
                   }
	      else { newnode = Grab_ENode(v0, rep);
	             if ( ! EN_IS(SHIFTED, newnode) ) {
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = TransList( v0, next );
		        Shift(next, &cotl, &newnode); 
		        if (cotl) Free_Int_List(cotl);
			newnode->final = 2 * edge_label(next, v0);
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) =
				MAX( maxBuchi(next), maxBuchi(v0));
			}
                   }
	      if ( ! (*tl) ) { inherit = 1; break; }
	      if ( EN_IS(LOCAL, newnode) ) {
		     v0 = (int) newnode->corr;
		     max_c = ba_idx( max_post(v0) );
		     Cut(cut3, config(v0), v0);
	             max_e = SetCopy(config(v0), conf3); 
		     encorr = (ENCorr *) NULL; }
              else { encorr = newnode->corr;
                     max_c = ba_idx( (encorr->cut->i) );
	             IL2CUT(encorr->cut, cut3, max_c); 
	             max_e = ba_idx( (encorr->max->i) );
	             IL2CONF(encorr->conf, conf3, max_e ); }
	      source = newnode; tlp = (*tl); 
	      }
        /*** the entire list of transitions is added ***/
	if ( inherit ) {
	     (*w0ci)->corr = newnode->corr;
	     if (EN_IS(LOCAL, newnode)) (*w0ci)->info |= LOCAL; 
	     return; }
	maxtl = Max_Events( conf3 , max_e );
	/*** shifted to a global configuration ?? ***/
	if ( maxtl->next ) {
	     encorr = Grab_ENCorr( maxtl, (*w0ci) );
	     (*w0ci)->corr = encorr;
	     if ( encorr->conf ) { Free_Int_List(maxtl); return; }
	     globEN++ ;
	     encorr->conf = RevBA2IL(conf3);
	     encorr->cut = RevBA2IL(cut3);
	     ADJ_EN_glob(encorr, conf3);
	     /*if ( maxtl ) Free_Int_List(maxtl);*/
	     return; }
        /*** shifted to a local configuration !! ***/
	v0 = maxtl->i; free(maxtl);
	(*w0ci)->info |= LOCAL; 
	(*w0ci)->corr = (ENCorr *) v0;
	if ( ! ip_mem(v0) ) {
	     ip_mem(v0) = (int *) MyMalloc( ENCORR_IP * sizeof(int) );
	     maxBuchi(v0) = ( - 1 ); InitMaxBuchi(v0); 
	     if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	     /* update the edges of v0 */ ADJ_EN_loc(v0); 
	     locEN++; }
	if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	source = (ENode *) NULL;
	/* add ->rep<- to the Adjacent-List of w0ci->src */
	if ( IS( CORR, v0) ) rep = cutoff_num(v0, 1);
	else rep = v0;
	if ( (*w0ci)->src > 0 ) {
	     source = (ENode *) (*w0ci)->src;
	     if ( EN_IS(LOCAL, source) ) {
		  max_e = (int) source->corr;
		  if ( IS(CORR, max_e) ) {
		         max_c = cutoff_num(max_e, 1);
			 lastedge = (Edge *) __lastEdge(max_e);
		         nedge = Grab_Edge(max_c, rep); 
			 __lastEdge(max_e) = (int) lastedge;}
                  else { lastedge = __adjacents(max_e);
			 nedge = S_Grab_Edge(max_e, rep, &lastedge); }
                  }
             else { lastedge = source->corr->adjs;
		    nedge = EN_Grab_Edge( (source->corr) , rep, &lastedge); }
	     }
        else { source = (ENode *) NULL;
	       next = ( - ((*w0ci)->src) );
	       if ( IS( CORR, next) ) 
	              nedge = Grab_Edge(cutoff_num(next, 1), rep);
 	       else { lastedge = __adjacents(next);
		      nedge = S_Grab_Edge(next, rep, &lastedge); }
             }
       
        if ( F__1(nedge->final) ) return;

	if ( source ) {
	       if ( ! EN_IS(LOCAL, source)) {
	              v0 = ba_idx( (source->corr->max->i) );
	              IL2CONF(source->corr->conf, conf2, v0);
	              simpleshift = edge_label_lg(v, conf2); }
               else { v0 = (int) source->corr;
		      simpleshift = edge_label(v, v0); }
	       }
	else { simpleshift = edge_label(v, next); }
        if ( simpleshift > nedge->final ) {
	      nedge->final = (2 *simpleshift); nedge->n = v; }
}
/*************************************************************************/
void  Extend_GC(ENCorr *enco, int v, Int_List **tl, ENode **w0ci)
/* add the Trans-List tl to Config(enco) */
{
	int max_c;   /* max. condition in cut3 */
	int max_e;   /* max. event in conf3    */
	int v0, rep, next, simpleshift;
	char inherit=0;
	Int_List *tlp, *cotl, *maxtl, *maxtlp;
	ENode *source, *newnode;
	ENCorr *encorr;
	Edge *nedge, *lastedge;
	Int_List *ctr;

	/*** initializing the pointers ***/
	newnode = (ENode *) NULL; 
	source = enco->enode;
	encorr = enco;
	maxtl = cotl = IL_NULL; 
	lastedge = nedge = (Edge *) NULL;
	/*v0 = correspond(v);*/
	/*** set cutoff of *w0ci ***/
	(*w0ci)->cutoff = v; 
	(*w0ci)->info |= SHIFTED;
	(*w0ci)->ip = (int *) MyMalloc(sizeof(int) * EN_IP );
	/*** Initialize cut3, conf3 ***/
	max_c = ba_idx( (encorr->cut->i) ); 
	IL2CUT(encorr->cut, cut3, max_c);
	max_e = ba_idx( (encorr->max->i) );
	IL2CONF(encorr->conf, conf3, max_e);
	tlp = (*tl); 
	while ( tlp ) {
	      next = Find_Extension(cut3, tlp->i, &max_c);
              /*** next is a valid extension ***/
	      if ( ! next ) { while(1) source=source->next; }
              if ( next < max_e ) 
		   add_entry(conf3, next);
	      else ConfAdd(conf3, next, &max_e);
	      tlp->i = ( - next ) ; 
	      if ( ! IS(CUTOFF, next) ) { tlp=tlp->next; continue; }
	      /*** next is a cutoff event ***/
	      rep = representative(next);
	      /* forall e \in tl, e \notin Conf(next):  e|->Trans(e) ***/
	      /* forall e \in tl, e \in Conf(next):  remove e        ***/
	      updateTL(tl, next);

	      simpleshift = ( (! source) && (v0 == correspond(next)));
	      if (! simpleshift ) { simpleshift = 1;
	         if ( encorr ) { maxtlp = encorr->max;
	           while (maxtlp) { 
		         if ( ( maxtlp->i > next ) ||
		              ( ! elem_of(config(next), maxtlp->i)) ) {
			       simpleshift = 0; break; }
                         maxtlp = maxtlp->next; }
                   }
                 else { if ( ( v0 > next ) ||
			     ( ! elem_of(config(next), v0) ) )
			      simpleshift = 0; }
                 }

              if ( simpleshift ) {
		  v0 = correspond(next);
		  Cut(cut3, config(v0), v0);
		  max_c = ba_idx( max_post(v0) );
		  max_e = SetCopy(config(v0), conf3);
		  maxtl = IL_NULL;
		  source = (ENode *) NULL;
		  encorr = (ENCorr *) NULL;
		  tlp = (*tl); continue; }
              /* else :  not simpleshift; add next to the ENode-List of v0 */
	      if ( source && ( ! EN_IS(LOCAL, source)) ) {
		   newnode = ExtGrab_ENode( source, rep);
		   if ( ! EN_IS(SHIFTED, newnode) ) {
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = CoTRList(encorr->conf, config(next), next);
                        Shift( next, &cotl, &newnode );
		        if (cotl) Free_Int_List(cotl); 
			IL2CONF(encorr->conf, conf3, ba_idx(encorr->max->i));
			newnode->final = 2 * edge_label_lg(next, conf3); 
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) = 
			        MAX ( maxBuchi(next),
				      Find_Buechi(conf3, encorr->max->i)
                                    ) ;
                        }
                   }
	      else { newnode = Grab_ENode(v0, rep);
	             if ( ! EN_IS(SHIFTED, newnode) ) {
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = TransList( v0, next );
		        Shift(next, &cotl, &newnode); 
		        if (cotl) Free_Int_List(cotl);
			newnode->final = 2 * edge_label(next, v0);
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) =
				MAX( maxBuchi(next), maxBuchi(v0));
			}
                   }
	      if ( ! (*tl) ) { inherit = 1; break; }
	      if ( EN_IS(LOCAL, newnode) ) {
		     v0 = (int) newnode->corr;
		     max_c = ba_idx( max_post(v0) );
		     Cut(cut3, config(v0), v0);
	             max_e = SetCopy(config(v0), conf3); 
		     encorr = (ENCorr *) NULL; }
              else { encorr = newnode->corr;
                     max_c = ba_idx( (encorr->cut->i) );
	             IL2CUT(encorr->cut, cut3, max_c); 
	             max_e = ba_idx( (encorr->max->i) );
	             IL2CONF(encorr->conf, conf3, max_e ); }
	      source = newnode; tlp = (*tl); 
	      }
        /*** the entire list of transitions is added ***/
	if ( inherit ) {
	     (*w0ci)->corr = newnode->corr;
	     if (EN_IS(LOCAL, newnode)) (*w0ci)->info |= LOCAL; 
	     return; }
	maxtl = Max_Events( conf3 , max_e );
	/*** shifted to a global configuration ?? ***/
	if ( maxtl->next ) {
	     encorr = Grab_ENCorr( maxtl, (*w0ci) );
	     (*w0ci)->corr = encorr;
	     if ( encorr->conf ) { Free_Int_List(maxtl); return; }
	     globEN++ ;
	     encorr->conf = RevBA2IL(conf3);
	     encorr->cut = RevBA2IL(cut3);
	     ADJ_EN_glob(encorr, conf3);
	     /*Free_Int_List(maxtl);*/
	     return; }
        /*** shifted to a local configuration !! ***/
	v0 = maxtl->i; free(maxtl);
	(*w0ci)->info |= LOCAL; 
	(*w0ci)->corr = (ENCorr *) v0;
	if ( ! ip_mem(v0) ) {
	     ip_mem(v0) = (int *) MyMalloc( ENCORR_IP * sizeof(int) );
	     maxBuchi(v0) = ( - 1 ); InitMaxBuchi(v0); 
	     if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	     /* update the edges of v0 */ ADJ_EN_loc(v0); 
	     locEN++; }
	if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	source = (ENode *) NULL;
	/* add ->rep<- to the Adjacent-List of w0ci->src */
	if ( IS( CORR, v0) ) rep = cutoff_num(v0, 1);
	else rep = v0;
	if ( (*w0ci)->src > 0 ) {
	     source = (ENode *) (*w0ci)->src;
	     if ( EN_IS(LOCAL, source) ) {
		  max_e = (int) source->corr;
		  if ( IS(CORR, max_e) ) {
		         max_c = cutoff_num(max_e, 1);
			 lastedge = (Edge *) __lastEdge(max_e);
		         nedge = Grab_Edge(max_c, rep); 
			 __lastEdge(max_e) = (int) lastedge;}
                  else { lastedge = __adjacents(max_e);
			 nedge = S_Grab_Edge(max_e, rep, &lastedge); }
                  }
             else { lastedge = source->corr->adjs;
		    nedge = EN_Grab_Edge( (source->corr) , rep, &lastedge); }
	     }
        else { source = (ENode *) NULL;
	       next = ( - ((*w0ci)->src) );
	       if ( IS( CORR, next) ) 
	              nedge = Grab_Edge(cutoff_num(next, 1), rep);
 	       else { lastedge = __adjacents(next);
		      nedge = S_Grab_Edge(next, rep, &lastedge); }
             }
       
        if ( F__1(nedge->final) ) return;

	if ( source ) {
	       if ( ! EN_IS(LOCAL, source)) {
	              v0 = ba_idx( (source->corr->max->i) );
	              IL2CONF(source->corr->conf, conf2, v0);
	              simpleshift = edge_label_lg(v, conf2); }
               else { v0 = (int) source->corr;
		      simpleshift = edge_label(v, v0); }
	       }
	else { simpleshift = edge_label(v, next); }
        if ( simpleshift > nedge->final ) {
	      nedge->final = (2 *simpleshift); nedge->n = v; }
}
/*************************************************************************/
void  Extend_LC(int enco, int v, Int_List **tl, ENode **w0ci)
/* add the Trans-List tl to [enco] */
{
	int max_c;   /* max. condition in cut3 */
	int max_e;   /* max. event in conf3    */
	int v0, rep, next, simpleshift;
	char inherit=0;
	Int_List *tlp, *cotl, *maxtl, *maxtlp;
	ENode *source, *newnode;
	ENCorr *encorr;
	Edge *nedge, *lastedge;
	Int_List *ctr;

	/*** initializing the pointers ***/
	source = newnode = (ENode *) NULL; 
	encorr = (ENCorr *) NULL;
	maxtl = cotl = IL_NULL; 
	lastedge = nedge = (Edge *) NULL;
	/*** set cutoff of *w0ci ***/
	(*w0ci)->cutoff = v; 
	(*w0ci)->info |= SHIFTED;
	(*w0ci)->ip = (int *) MyMalloc(sizeof(int) * EN_IP );
	/*** Initialize v0, cut3, conf3 ***/
	/** !!! **/ v0 = enco;
	max_c = ba_idx( max_post(enco) ); 
	IL2CUT( (Int_List *) CUT(enco), cut3, max_c);
	max_e = SetCopy(config(enco), conf3);
	tlp = (*tl); 
	while ( tlp ) {
	      next = Find_Extension(cut3, tlp->i, &max_c);
              /*** next is a valid extension ***/
	      if ( ! next ) { while(1) source=source->next; }
              if ( next < max_e ) 
		   add_entry(conf3, next);
	      else ConfAdd(conf3, next, &max_e);
	      tlp->i = ( - next ) ; 
	      if ( ! IS(CUTOFF, next) ) { tlp=tlp->next; continue; }
	      /*** next is a cutoff event ***/
	      rep = representative(next);
	      /* forall e \in tl, e \notin Conf(next):  e|->Trans(e) ***/
	      /* forall e \in tl, e \in Conf(next):  remove e        ***/
	      updateTL(tl, next);

	      simpleshift = ( (! source) && (v0 == correspond(next)));
	      if (! simpleshift ) { simpleshift = 1;
	         if ( encorr ) { maxtlp = encorr->max;
	           while (maxtlp) { 
		         if ( ( maxtlp->i > next ) ||
		              ( ! elem_of(config(next), maxtlp->i)) ) {
			       simpleshift = 0; break; }
                         maxtlp = maxtlp->next; }
                   }
                 else { if ( ( v0 > next ) ||
			     ( ! elem_of(config(next), v0) ) )
			      simpleshift = 0; }
                 }

              if ( simpleshift ) {
		  v0 = correspond(next);
		  Cut(cut3, config(v0), v0);
		  max_c = ba_idx( max_post(v0) );
		  max_e = SetCopy(config(v0), conf3);
		  maxtl = IL_NULL;
		  source = (ENode *) NULL;
		  encorr = (ENCorr *) NULL;
		  tlp = (*tl); continue; }
              /* else :  not simpleshift; add next to the ENode-List of v0 */
	      if ( source && ( ! EN_IS(LOCAL, source)) ) {
		   newnode = ExtGrab_ENode( source, rep);
		   if ( ! EN_IS(SHIFTED, newnode) ) {
			encorr = source->corr;
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = CoTRList(encorr->conf, config(next), next);
                        Shift( next, &cotl, &newnode );
		        if (cotl) Free_Int_List(cotl); 
			IL2CONF(encorr->conf, conf3, ba_idx(encorr->max->i));
			newnode->final = 2 * edge_label_lg(next, conf3); 
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) = 
			        MAX ( maxBuchi(next),
				      Find_Buechi(conf3, encorr->max->i)
                                    ) ;
                        }
                   }
	      else { newnode = Grab_ENode(v0, rep);
	             if ( ! EN_IS(SHIFTED, newnode) ) {
		        if (newnode->idx != (EN_idx - 1)) {tlp=(*tl); continue;}
			cotl = TransList( v0, next );
		        Shift(next, &cotl, &newnode); 
		        if (cotl) Free_Int_List(cotl);
			newnode->final = 2 * edge_label(next, v0);
			if ( F__0(newnode->final) )
			     EN_maxBuchi(newnode) =
				MAX( maxBuchi(next), maxBuchi(v0));
			}
                   }
	      if ( ! (*tl) ) { inherit = 1; break; }
	      if ( EN_IS(LOCAL, newnode) ) {
		     v0 = (int) newnode->corr;
		     max_c = ba_idx( max_post(v0) );
		     Cut(cut3, config(v0), v0);
	             max_e = SetCopy(config(v0), conf3); 
		     encorr = (ENCorr *) NULL; }
              else { encorr = newnode->corr;
                     max_c = ba_idx( (encorr->cut->i) );
	             IL2CUT(encorr->cut, cut3, max_c); 
	             max_e = ba_idx( (encorr->max->i) );
	             IL2CONF(encorr->conf, conf3, max_e ); }
	      source = newnode; tlp = (*tl); 
	      }
        /*** the entire list of transitions is added ***/
	if ( inherit ) {
	     (*w0ci)->corr = newnode->corr;
	     if (EN_IS(LOCAL, newnode)) (*w0ci)->info |= LOCAL; 
	     return; }
	maxtl = Max_Events( conf3 , max_e );
	/*** shifted to a global configuration ?? ***/
	if ( maxtl->next ) {
	     encorr = Grab_ENCorr( maxtl, (*w0ci) );
	     (*w0ci)->corr = encorr;
	     if ( encorr->conf ) { Free_Int_List(maxtl); return; }
	     globEN++ ;
	     encorr->conf = RevBA2IL(conf3);
	     encorr->cut = RevBA2IL(cut3);
	     ADJ_EN_glob(encorr, conf3);
	     /*Free_Int_List(maxtl);*/
	     return; }
        /*** shifted to a local configuration !! ***/
	v0 = maxtl->i; free(maxtl);
	(*w0ci)->info |= LOCAL; 
	(*w0ci)->corr = (ENCorr *) v0;
	if ( ! ip_mem(v0) ) {
	     ip_mem(v0) = (int *) MyMalloc( ENCORR_IP * sizeof(int) );
	     maxBuchi(v0) = ( - 1 ); InitMaxBuchi(v0); 
	     if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	     /* update the edges of v0 */ ADJ_EN_loc(v0); 
	     locEN++; }
	if ( ! CUT(v0) ) Cut(cut3, config(v0), v0);
	source = (ENode *) NULL;
	/* add ->rep<- to the Adjacent-List of w0ci->src */
	if ( IS( CORR, v0) ) rep = cutoff_num(v0, 1);
	else rep = v0;
	if ( (*w0ci)->src > 0 ) {
	     source = (ENode *) (*w0ci)->src;
	     if ( EN_IS(LOCAL, source) ) {
		  max_e = (int) source->corr;
		  if ( IS(CORR, max_e) ) {
		         max_c = cutoff_num(max_e, 1);
			 lastedge = (Edge *) __lastEdge(max_e);
		         nedge = Grab_Edge(max_c, rep); 
			 __lastEdge(max_e) = (int) lastedge;}
                  else { lastedge = __adjacents(max_e);
			 nedge = S_Grab_Edge(max_e, rep, &lastedge); }
                  }
             else { lastedge = source->corr->adjs;
		    nedge = EN_Grab_Edge( (source->corr) , rep, &lastedge); }
	     }
        else { source = (ENode *) NULL;
	       next = ( - ((*w0ci)->src) );
	       if ( IS( CORR, next) ) 
	              nedge = Grab_Edge(cutoff_num(next, 1), rep);
 	       else { lastedge = __adjacents(next);
		      nedge = S_Grab_Edge(next, rep, &lastedge); }
             }
       
        if ( F__1(nedge->final) ) return;

	if ( source ) {
	       if ( ! EN_IS(LOCAL, source)) {
	              v0 = ba_idx( (source->corr->max->i) );
	              IL2CONF(source->corr->conf, conf2, v0);
	              simpleshift = edge_label_lg(v, conf2); }
               else { v0 = (int) source->corr;
		      simpleshift = edge_label(v, v0); }
	       }
	else { simpleshift = edge_label(v, next); }
        if ( simpleshift > nedge->final ) {
	      nedge->final = (2 * simpleshift); nedge->n = v; }
}
/*************************************************************************/
Int_List  *Max_Events(ConfS c, int maxe)
/* returns a list of the causally maximal events in Conf. c */
/* maxe is the upper border of c ! */
{
	Int_List *ntl, *ntlp, *dtlp;
	int      bit, w, i, valid, maxevent;
	ElemS    ce;

	ntl = ntlp = dtlp = IL_NULL;
	i = (maxe/BA_BASE) - 1; 
	w = maxe;
	ce = c[i];

	/* find maximal event in c */
        for (bit = (BA_BASE-1); bit >= 0; bit-- ) {
	    w--;
	    if ( bitmask[bit] != (bitmask[bit] & ce) ) continue;
            del_entry(c, w); maxevent = w;
	    ntl = NEW_INTLIST; ntl->i = w; ntl->next = IL_NULL;
	    InitConfig(ntl->i); 
	    dtlp = ntl;
	    break; }

	/*** skip all non-maximal elements of c ***/
	while ( i >= 0 ) {
	    if ( (ce=c[i]) == 0 ) {i--; continue; }
	    w = BA_BASE * (i--) + BA_BASE;
            for (bit = (BA_BASE-1); bit >= 0; bit-- ) {
	        w--;
	        if ( bitmask[bit] != (bitmask[bit] & ce) ) continue;
	        ce -= bitmask[bit];
	        ntlp = ntl; valid = 1;
	        while (ntlp) { 
		   if ( elem_of(config(ntlp->i), w) ) { valid=0; break; }
		   ntlp=ntlp->next; }
                if ( ! valid ) { if (ce == 0) break; else continue; }
	        dtlp->next = NEW_INTLIST; dtlp = dtlp->next;
	        dtlp->next = IL_NULL; dtlp->i = w;
	        InitConfig(w);  
	        if ( ce==0 ) break; }
            }
        add_entry(c, maxevent);
	return(ntl);
}
/*************************************************************************/
void  *updateTL(Int_List **tl, int cutoff)
/* remove those events from tl that belong to  Config(cutoff); 
   the remaining events are mapped (back) to their Transitions */
{
	Int_List *dtlp, *ptlp;
	int event;

	dtlp = ptlp = (*tl); 
	while (ptlp) {
	       event = - ptlp->i;
	       if ( event < 0 ) { /* keep ptlp */
		      if ((*tl) != ptlp ) dtlp = ptlp;
		      ptlp = ptlp->next; continue; }
	       /* event > 0 */
	       if ( (event <= cutoff) &&
		     elem_of(config(cutoff), event) )  { /*delete ptlp*/
		     if ((*tl) == ptlp) { dtlp = (*tl) = ptlp->next;
					  free(ptlp); ptlp = dtlp; }
		     else { dtlp->next = ptlp->next;
			    free(ptlp); ptlp = ptlp->next; }
		     if ( event == cutoff ) return;
		     }
	       else { /* keep ptlp and map ptlp->i back to T_NAME */
		     if ((*tl) != ptlp ) dtlp = ptlp;
		     ptlp->i = T_NAME(event);
		     ptlp = ptlp->next; }
	       }
}
/*************************************************************************/
char  IL_equal(Int_List *il1, Int_List *il2)
{
	while ( il1 && il2 ) {
		if ( il1->i != il2->i ) return(0);
		il1=il1->next; il2=il2->next; }
        if ( il1 || il2 ) return(0);
	else return(1);
}
/*************************************************************************/
char  IL_less(Int_List *il1, Int_List *il2)
{
	while ( il1 && il2 ) {
		if ( il1->i > il2->i ) return(0);
		if ( il1->i < il2->i ) return(1);
		il1=il1->next; il2=il2->next; }
        if ( ( ! il1 ) && il2 ) return(1);
        if ( ( ! il1 ) && ( ! il2) ) return(2);
	return(0);
}
/*************************************************************************/
trans_PL *create_tree(int op, trans_PL *lop, trans_PL *rop)
/* Legt einen neuen Knoten mit Unterbaeumen an.				 */
{
	trans_PL *tree;
	
	tree = NEW_TPL;
	tree->op = (char *) op;
	tree->lop = lop;
	tree->rop = rop;
	return tree;
}
/*************************************************************************/
trans_PL *create_leaf(char *prop)
/* Legt einen Blattknoten an. Dieser kann mit einer Beschriftung und einem*/
/* beliebigen Inhalt versehen sein, welcher in lop "versteckt" wird. Um	den*/
/* Inhalt wieder zurueckzugewinnen, ist Typecasting erforderlich.	   */
{
	trans_PL *tree;

	tree = NEW_TPL;
        if ( ( prop == ( char *) TPL_TRUE  ) || 
	     ( prop == ( char *) TPL_FALSE )   )
	     tree->op = prop;
	else tree->op = Strdup(prop);
	tree->lop = (trans_PL *) NULL;
	tree->rop = (trans_PL *) NULL;
	return tree;
}
/*************************************************************************/
void free_tree(trans_PL *tree)
/* Durchlaeuft den Binaerbaum und gibt allen Speicher frei.		 */
{
	if (tree->lop) {
		free_tree(tree->lop);
		free_tree(tree->rop);
	        }
	else if (TPL_PROP(tree)) free(tree->op);
	free(tree);
}
/*************************************************************************/
trans_PL *copy_tree(trans_PL *tree)
{
	if (tree->lop)
		return create_tree((int) tree->op,
			tree->lop? copy_tree(tree->lop) : NULL,
			tree->rop? copy_tree(tree->rop) : NULL);
	else return create_leaf(tree->op);
}


/*************************************************************************/
trans_PL *DNF_convert(trans_PL *tree)
/* Bringt einen Formelbaum, der frei von temporalen Operatoren ist, in	 */
/* disjunktive Normalform.						 */
{
	trans_PL *ntree;
	int tree_op;

	if ( ! tree->lop) return tree;

	tree_op = (int) tree->op;
	switch (tree_op)
	{
	    case TPL_NOT:
		return DNF_negate(tree);
	    case TPL_AND:
		tree->lop = DNF_convert(tree->lop);
		tree->rop = DNF_convert(tree->rop);
		ntree = DNF_multiply_left(tree->lop,tree->rop);
		free_tree(tree);
		return ntree;
	    case TPL_OR:
		tree->lop = DNF_convert(tree->lop);
		tree->rop = DNF_convert(tree->rop);
		return tree;
	    default:	/* can't happen */
		DNF_ERR;
		return NULL;
	}
}
/*************************************************************************/
trans_PL *DNF_multiply_left(trans_PL *left, trans_PL *right)
/* Berechne DNF(left AND right), wenn left und right DNF sind.		 */
{
	if ( TPL_OR_OP(left) )
		return create_tree(TPL_OR,
					DNF_multiply_left(left->lop,right),
					DNF_multiply_left(left->rop,right));
	else
		return DNF_multiply_right(left,right);
}
/*************************************************************************/
trans_PL *DNF_multiply_right(trans_PL *left, trans_PL *right)
{
	if ( TPL_OR_OP(right) )
		return create_tree(TPL_OR,
					DNF_multiply_right(left,right->lop),
					DNF_multiply_right(left,right->rop));
	else
		return create_tree(TPL_AND,
					copy_tree(left),
					copy_tree(right));
}
/*************************************************************************/
trans_PL *DNF_negate(trans_PL *tree)
/* Berechne DNF(tree), wenn tree->op == FORM_NOT ist.			 */
{
	trans_PL *ctree;
	int ctree_op;
	char *prop;

	ctree = tree->lop;

	if ( TPL_PROP(ctree) ) {
	     if ( ctree->op[0] != '!' ) {
		    prop = NEW_CHAR(strlen(ctree->op) + 2);
		    sprintf(prop, "!%s", ctree->op);
		  }
             else { prop = NEW_CHAR(strlen(ctree->op));
		    strcpy(prop, &(ctree->op[1]));
		  }
             free(ctree->op);
	     free(ctree);
	     tree->op = prop;
	     tree->lop = NULL;
	     return tree;
	     }

        if (TPL_TRUE_OP(ctree)) {  tree->op = (char *) TPL_FALSE; 
				   tree->lop = NULL;
				   free(ctree);
				   return tree; } 
        if (TPL_FALSE_OP(ctree)) { tree->op = (char *) TPL_TRUE; 
				   tree->lop = NULL;
				   free(ctree);
				   return tree; } 
	ctree_op = (int) ctree->op;
	switch (ctree_op)
	{
	    case TPL_NOT:
		ctree = ctree->lop;
		free(tree);
		free(tree->lop);
		return DNF_convert(ctree);
	    case TPL_AND:
		tree->lop = ctree->rop;
		ctree->op = (char *) TPL_NOT;
		ctree->rop = NULL;
		return create_tree(TPL_OR,DNF_negate(ctree),DNF_negate(tree));
	    case TPL_OR:
		tree->lop = ctree->rop;
		ctree->op = (char *) TPL_NOT;
		ctree->rop = NULL;
		return DNF_convert(create_tree(TPL_AND,ctree,tree));
	    default:	/* can't happen */
		DNF_ERR;
		return NULL;
	}
}
/*************************************************************************/
void  ENode_Config(ConfS conf, ENode *enod)
{
	int    i, border, srclocal;
	ENode  *enod_src;
	ENCorr *encorr;
	ConfS  src;
	Int_List *il;

	if (EN_IS(LOCAL, enod)) border = MAX(enod->cutoff, (int) enod->corr);
	else { encorr = enod->corr;
	       border = MAX(enod->cutoff, encorr->max->i); }
	i = 0; border = border/BA_BASE;
	while ( i <= border ) conf[i++] = 0;
	conf[i] = BA_ALL;

        /* enod->src is an event */
	if ( enod->src < 0 ) {
	       srclocal = ( - enod->src );
	       src = config(srclocal); i = 0;
               while ( src[i] != BA_ALL ) conf[i] = src[i++]; 
	       src = config(enod->cutoff); i = 0; 
	       while ( src[i] != BA_ALL ) conf[i] |= src[i++];
	       return; }
        /* enod->src is a global configuration */
	enod_src = (ENode *) enod->src;
	if ( EN_IS(LOCAL, enod_src) ) {
	       src = config( (int) enod_src->corr); i=0;
	       while ( src[i] != BA_ALL ) conf[i] = src[i++]; }
	else { encorr = enod_src->corr; il = encorr->conf;
	       while (il) { add_entry(conf, il->i); il=il->next; }
	     }
	src = config(enod->cutoff); i = 0;
	while ( src[i] != BA_ALL ) conf[i] |= src[i++];
}
/*************************************************************************/
void  Init_Extended_Search()
{
	int i;
	ENode *en;

	/* Re-Initialize the stack */ 
	free(stack);
	stack = (int *) MyMalloc(sizeof(int) * (E_corr + EN_idx));
      	/* Initialize the Nodes and the stack-info*/
	for (i=0; i<=E_Fin; i++) {
	    if ( ! ip_mem(i) ) continue; 
	    INFO( i ) &= ~VISITED; INFO( i ) &= ~ONSTACK; 
	    if ( IS(CUTOFF, i) && (i != representative(i)) ) { 
		 INFO(i) |= VISITED; continue; } 
	    if ( ! IS( CORR, i) ) { 
		 dfn(i) = lowlink(i) = scc(i) = scc_idx(i) = 0; }
	    if ( IS(CUTOFF, i) ) continue; 
	    en = (ENode *) enodes(i); 
	    while ( en) { Init_ENodes( en ); en = en->next; } 
	    } 
	b_stackpointer = stackpointer = accepting_component = acc_ctr = 0; 
	count = component = comp_ctr = 1;
}
/*************************************************************************/
/* End of File unfolding.c      */
/********************************/

